/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef CALCULATEPOLAR_SRC_BWB_BWBDEFAULTDATA_H_
#define CALCULATEPOLAR_SRC_BWB_BWBDEFAULTDATA_H_

#include <atmosphere/atmosphere.h>
#include <standardFiles/typedefs.h>
#include <aixml/endnode.h>

#include <moduleBasics/runtimeIO.h>

#include <string>
#include <unordered_map>
#include <vector>
#include <map>

/** \brief Class contains all properties of calculated aircraft objects
*/
class bwbDefaultData {
 public:
    /* Typedefs */
    /** \brief Constructor of aircraft class
     *  \param &mySettings Reference to calculatePolarSettings object
     */
    explicit bwbDefaultData(const std::shared_ptr<RuntimeIO>& rtIO);
    /** \brief Virtual destructor of aircraft class
     */
    virtual ~bwbDefaultData() {}
    std::string engineType;         /**< Information about the propulsor type (e.g. fan) [-] */
    double delta_isa;               /**< temperature delta to ISA Standard Atmosphere [K] */
    double initialCruiseMachNumber; /**< Mach number flown in initial cruise phase at initial cruise altitude [-] */
    /* !!! WARNING: All EndnodeReadOnly Values are converted automatically to SI-Units/RADIAN when read in */
    Endnode<double> bestCruiseCL;
    Endnode<double>designCoGX;
    Endnode<double>designCoGY;
    Endnode<double>designCoGZ;
    Endnode<double>forwardCoGX;
    Endnode<double>forwardCoGY;
    Endnode<double>forwardCoGZ;
    Endnode<double>aftCoGX;
    Endnode<double>aftCoGY;
    Endnode<double>aftCoGZ;
    Point designCoG;              /**< Center of gravity coordinates */
    Point forwardCoG;       /**< Forward position of center of gravity */
    Point aftCoG;           /**< Average flight time position of center of gravity */
    Point cruiseTrimRefCoG; /**< Cruise reference center of gravity position */
    std::unordered_map <std::string, Point> CoGPositions;    /**< Map connecting CoG names with points */
    size_t numberDevices;
    struct wingDevice {
        std::string type;
        double maxPositiveDeflection;
        double maxNegativeDeflection;
        double innerRelSpanPosition;
        double innerRelFwdChordPosition;
        double innerRelAftChordPosition;
        double outerRelSpanPosition;
        double outerRelFwdChordPosition;
        double outerRelAftChordPosition;
    };
    std::vector<wingDevice> wingDevices;
 private:
    const std::shared_ptr<RuntimeIO>& rtIO_;                    /**< Pointer to the RuntimeIO */
};

#endif  // CALCULATEPOLAR_SRC_BWB_BWBDEFAULTDATA_H_
