/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "energy_based.h"
#include <algorithm>
#include <filesystem>
#include <format>
#include <list>
#include <ranges>
#include <cmath>
#include <vector>
#include <numbers>
#include <iostream>
#include <string>
#include <moduleBasics/module.h>
#include <engine/engine.h>
#include "../../include/constraint_analysis/ca_functions.h"
#include "../../include/constraint_analysis/ca_parser.h"
#include "../constraint_analysis.h"
#include "../../include/constraint_analysis/ca_plotting.h"
#include "../io/aircraft_xml.h"
#include "../../include/constraint_analysis/ca_minfinder.h"


void EnergyBased::initialize()
{
    if (!this->aircraft_xml()) {
        throw std::runtime_error("[EnergyBased::initialize()] aircraft_xml is null!");
    };
    if (!this->configuration()) {
        throw std::runtime_error("[EnergyBased::initialize()] config_xml is null!");
    };
    if (!this->polar()) {
        throw std::runtime_error("[EnergyBased::initialize()] polar_xml is null!");
    };
}

void EnergyBased::operator()()
{

    const std::string mode_selection(this->configuration()->at("module_configuration_file/program_settings/mode_selection/value"));

    int WS_start = 0;
    int WS_end = 0;
    size_t WS_size = 1;

    if (mode_selection == "mode_0")
    {
        /*Define relevant Wing_loading range. Selected values should be valid for all possible configurations*/
        WS_start = 50;
        WS_end = 820;
        WS_size = WS_end - WS_start + 1; // Range from 50 to 820, inclusive
    }
    else if (mode_selection == "mode_1")
    {
        WS_start = this->aircraft_xml()->at("aircraft_exchange_file/sizing_point/wing_loading/value");
        WS_size = 1; // Constant W/S
    }

    std::vector<double> W_over_S_data(WS_size);

    /*Transform from kg/m^2 into N/m^2*/
    for (int i = 0; i < WS_size; ++i) {
        W_over_S_data[i] = (WS_start + i) * 9.81;
    }

    this->lower_boundaries.push_back(50.0 * 9.81);
    this->upper_boundaries.push_back(820.0 * 9.81);

    Engine engine = Engine(this->engine_directory());

    Mattingly::constraint_analysis constraint_analysis_tool;

    atmosphere atm;

    // Set atmosphere according to the delta ISA requirement for the sizing process
    double temperature_ISA = ISA_TEMPERATURE + this->aircraft_xml()->at("aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/design_mission/delta_ISA/value");
    atm.setAtmosphere(0.0, temperature_ISA, ISA_PRESSURE);

    constraint_analysis_tool.atm = atm;

    std::filesystem::path missionCSV = this->mission_directory();

    for (const auto& entry : fs::directory_iterator(this->mission_directory())) {
        std::string filename = entry.path().filename().string();

        if (filename.find("design_mission") != std::string::npos && filename.find("_out.csv") != std::string::npos) {
            missionCSV /= filename;
            break;
        }
    }

    Mattingly::constraint_analysis constraint_analysis_plotting = constraint_analysis_tool;

    readMission mission_data(missionCSV);

    assess_constraints(W_over_S_data, constraint_analysis_tool, mission_data, engine);

    /* Define Simple_Analysis*/
    Simple_Analysis min_finder = Simple_Analysis(this->constraint_list, this->aircraft_xml(), this->lower_boundaries, this->upper_boundaries);

    /* Find dominant curve of the design chart*/
    min_finder.find_dominant_curve();

    double safety_factor = this->configuration()->at("module_configuration_file/program_settings/safety_factor/value");

    const double buffer_factor(this->configuration()->at("module_configuration_file/program_settings/buffer_factor/value"));

    /*Get the previous design point*/
    double previous_WS = this->aircraft_xml()->at("aircraft_exchange_file/sizing_point/wing_loading/value");
    double previous_TW = this->aircraft_xml()->at("aircraft_exchange_file/sizing_point/thrust_to_weight/value");

    /*Find design point with safety factor*/
    min_finder.find_converged_design_point(previous_TW, previous_WS * 9.81, safety_factor, buffer_factor);

    /*Update design point in ACxml*/
    min_finder.update_design_point();

    std::string plot_dir_string = this->plot_dir().string();

    // Add output of design point in logfile

    myRuntimeInfo->out << "New wing loading: " << min_finder.design_wing_loading * 9.81 << " N/m^2. New thrust to weight ratio: " << min_finder.design_thrust_to_weight << std::endl;

    min_finder.write_design_point(plot_dir_string + "/csv_files/design_point.csv");

    /* Plot the complete design chart with constraints, boundaries and design point*/
    std::string plotting = this->configuration()->at("module_configuration_file/control_settings/plot_output/enable/value");
    if (plotting == "true") {

        WS_start = 50;
        WS_end = 820;
        WS_size = WS_end - WS_start + 1; // Range from 50 to 820, inclusive

        std::vector<double> W_over_S_data(WS_size);

        for (int i = 0; i < WS_size; ++i) {
            W_over_S_data[i] = (WS_start + i) * 9.81;
        }

        this->constraint_list = {};
        this->lower_boundaries = { 50. * 9.81 };
        this->upper_boundaries = { 820. * 9.81 };

        assess_constraints(W_over_S_data, constraint_analysis_plotting, mission_data, engine);

        Simple_Analysis min_finder_2 = Simple_Analysis(this->constraint_list, this->aircraft_xml(), this->lower_boundaries, this->upper_boundaries);

        min_finder_2.find_dominant_curve();

        ConstraintAnalysis_Plot ca_plot;

        std::vector<double> boundaries = { min_finder_2.lower_wing_loading_boundary , min_finder_2.upper_wing_loading_boundary };

        ca_plot.fill_infeasible_area(W_over_S_data, min_finder_2.dominant_thrust_to_weight, boundaries);

        for (auto ca_case : this->constraint_list)
        {
            ca_plot.add_curve(ca_case.wing_loading, ca_case.thrust_to_weight, ca_case.constraint_case);
        }
        for (auto boundary : boundaries)
        {
            ca_plot.add_curve(boundary, "W/S limit");
        }
        ca_plot.add_curve({ min_finder.design_wing_loading * 9.81 }, { min_finder.design_thrust_to_weight }, "Design Point");

        ca_plot.save_figure(plot_dir_string + "/constraint_plot.svg");
    };
}

void  EnergyBased::assess_constraints(std::vector<double>& W_over_S_data, Mattingly::constraint_analysis& constraint_analysis_tool, readMission& mission_data, Engine& engine)
{
    auto calculate_thrust_lapse = [&constraint_analysis_tool, &engine](double M, double altitude) {
        const double thrust_lapse = engine.get_thrust_lapse("cruise", constraint_analysis_tool.atm, M, altitude);
        return thrust_lapse;
        };

    auto get_num_engines = [&]() {
        double num_engines = 0;

        // Get the node vector for all defined propulsion systems
        auto propulsors = this->aircraft_xml()->getVector(
            "aircraft_exchange_file/requirements_and_specifications/design_specification/propulsion/propulsor");

        // Increment num_engines for each propulsor
        for (const auto& propulsor : propulsors) {
            num_engines += 1;
        }
        return num_engines;
        };

    auto std_constraints = this->configuration()->at(
        "module_configuration_file/program_settings/constraint_selection/standard_set").getChildren();

    for (auto constraint : std_constraints)
    {
        std::string activated = constraint->at("active/value");
        if (activated == "true")
        {
            double CD = 0.0;
            double CL = 0.0;
            std::vector<double> CD_vector = {};
            std::vector<double> CL_vector = {};

            if (constraint->name == "takeoff_ground_roll")
            {
                double CL_max_TO = this->aircraft_xml()->at("aircraft_exchange_file/analysis/aerodynamics/lift_coefficients/C_LmaxT-O/value");
                double s_G_TO = this->aircraft_xml()->at("aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/takeoff_distance/value");
                double M = constraint->at("Mach_TO/value");
                double thrust_lapse = calculate_thrust_lapse(M, 0.0);
                double weight_fraction = mission_data.get_beta(" takeoff", 1.0); //from mission analysis
                const auto constraint_vector = constraint_analysis_tool.takeoff_ground_roll(
                    W_over_S_data,
                    weight_fraction,
                    thrust_lapse,
                    M,
                    0.0,
                    s_G_TO,
                    CL_max_TO);
                Constraint Case(W_over_S_data, constraint_vector, constraint->name);
                this->constraint_list.push_back(Case);
            }
            else if (constraint->name == "takeoff_climb")
            {
                double M = constraint->at("Mach_TO/value");
                double takeoff_climb_angle = constraint->at("takeoff_climb_angle/value");
                double thrust_lapse = calculate_thrust_lapse(M, 0.0);
                double weight_fraction = mission_data.get_beta(" takeoff", 1.0); //from mission analysis
                double CL_max_TO = this->aircraft_xml()->at("aircraft_exchange_file/analysis/aerodynamics/lift_coefficients/C_LmaxT-O/value");
                readPolar takeoff_config(this->polar(), "takeoff", "linear", M);
                CD = takeoff_config.interpCD(CL_max_TO);
                for (int i = 0; i < W_over_S_data.size(); ++i) {
                    CD_vector.push_back(CD);
                };
                const auto constraint_vector = constraint_analysis_tool.takeoff_climb_angle(
                    W_over_S_data,
                    CD_vector,
                    weight_fraction,
                    thrust_lapse,
                    M,
                    0.0,
                    CL_max_TO,
                    takeoff_climb_angle);
                Constraint Case(W_over_S_data, constraint_vector, constraint->name);
                this->constraint_list.push_back(Case);
            }
            else if (constraint->name == "OEI_climb")
            {
                /* OEI Climb */
                double climb_speed = this->aircraft_xml()->at(
                    "aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/design_mission/climb_speed_schedule/climb_speed_below_FL100/value");
                double climb_gradient_OEI = constraint->at("climb_gradient_OEI/value");
                double M = climb_speed / constraint_analysis_tool.atm.getSpeedOfSound(0.0);
                double weight_fraction = mission_data.get_beta(" takeoff"); //from mission analysis
                double thrust_lapse = calculate_thrust_lapse(M, 0.0);
                double num_engines = get_num_engines();

                readPolar climb_config(this->polar(), "takeoff", "linear", M);
                for (int i = 0; i < W_over_S_data.size(); ++i) {
                    CD = climb_config.getCD(W_over_S_data[i], 1.0, weight_fraction, (climb_speed / constraint_analysis_tool.atm.getSpeedOfSound(0.0)), 0.0);
                    CD_vector.push_back(CD);
                };

                double thrust_factor_OEI = thrust_lapse * ((num_engines - 1) / num_engines);
                double climb_rate_OEI = climb_gradient_OEI * climb_speed;
                /*minimal climb rate of 2.4% defined in CS25 for second segment climb*/
                const auto constraint_vector = constraint_analysis_tool.constant_speed_climb(
                    W_over_S_data,
                    CD_vector,
                    weight_fraction,
                    thrust_factor_OEI,
                    M,
                    0.0,
                    climb_rate_OEI); /* DEFINE CLIMB RATE CONSTRAINT */
                Constraint Case(W_over_S_data, constraint_vector, constraint->name);
                this->constraint_list.push_back(Case);
            }
            else if (constraint->name == "top_of_climb")
            {

            }
            else if (constraint->name == "cruise")
            {

            }
            else if (constraint->name == "SEP")
            {
                /* Service Ceiling */
                double altitude = this->aircraft_xml()->at(
                    "aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/flight_envelope/maximum_operating_altitude/value");
                double M = this->aircraft_xml()->at(
                    "aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/study_mission/initial_cruise_mach_number/value");

                double weight_fraction = mission_data.get_beta(" climb_to_cruise");
                readPolar clean_config_M_cruise(this->polar(), "clean", "linear", M);
                double thrust_lapse = calculate_thrust_lapse(M, altitude);
                CD_vector = {};
                CL_vector = {};

                for (int i = 0; i < W_over_S_data.size(); ++i) {
                    CD = clean_config_M_cruise.getCD(W_over_S_data[i], 1.0, weight_fraction, M, altitude);
                    CD_vector.push_back(CD);
                }

                double minimum_climb_rate = constraint->at("minimum_climb_rate/value");

                /* Service ceiling with the minimum climb rate requirement */
                const auto constraint_vector = constraint_analysis_tool.constant_speed_climb(
                    W_over_S_data,
                    CD_vector,
                    weight_fraction,
                    thrust_lapse,
                    M,
                    altitude,
                    minimum_climb_rate);
                Constraint Case(W_over_S_data, constraint_vector, constraint->name);
                this->constraint_list.push_back(Case);
            }
            else if (constraint->name == "landing_field_length")
            {
                double k_TD = constraint->at("k_TD/value");
                double h_obs = constraint->at("obstacle_height/value");
                double M = constraint->at("Mach/value");
                double t_FR = constraint->at("t_FR/value");
                double V_app = this->aircraft_xml()->at(
                    "aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/flight_envelope/maximum_approach_speed/value");
                double CL_max_L = this->aircraft_xml()->at("aircraft_exchange_file/analysis/aerodynamics/lift_coefficients/C_LmaxLanding/value");
                double s_A = h_obs / std::tan(3. * std::numbers::pi / 180.);
                readPolar landing_config(this->polar(), "landing", "linear", M);
                double CD_max_L = landing_config.interpCD(CL_max_L);
                double max_landing_weight = this->aircraft_xml()->at("aircraft_exchange_file/analysis/masses_cg_inertia/maximum_landing_mass/mass_properties/mass/value");
                double MTOM = this->aircraft_xml()->at("aircraft_exchange_file/analysis/masses_cg_inertia/maximum_takeoff_mass/mass_properties/mass/value");
                double weight_fraction = max_landing_weight / MTOM;
                const auto approach = constraint_analysis_tool.approach_speed(V_app, CL_max_L, weight_fraction, s_A, CD_max_L, h_obs, k_TD, 0.0);
                this->upper_boundaries.push_back(approach);
                double V_stall = constraint_analysis_tool.get_V_stall(V_app, k_TD, CL_max_L, CD_max_L, h_obs, s_A);
                double s_FR = V_stall * k_TD * t_FR;

                /* Landing */
                double thrust_lapse = calculate_thrust_lapse(M, 0.0);
                double my_B = this->aircraft_xml()->at("aircraft_exchange_file/requirements_and_specifications/requirements/additional_requirements/landing_gear/braking_coefficient/value");
                double s_G_L = 0.6 * this->aircraft_xml()->at("aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/landing_field_length/value") - s_A - s_FR;
                const auto constraint_border = constraint_analysis_tool.braking_roll(CD_max_L, CL_max_L, weight_fraction, thrust_lapse, M, 0.0, my_B, s_G_L);
                this->upper_boundaries.push_back(constraint_border);
            }
            else if (constraint->name == "gust")
            {
                /* Gust */
                double M = constraint->at("Mach/value");
                readPolar landing_config(this->polar(), "landing", "linear", M);
                double weight_fraction = mission_data.get_beta(" takeoff", 1.0); //from mission analysis
                double CL_max_L = this->aircraft_xml()->at("aircraft_exchange_file/analysis/aerodynamics/lift_coefficients/C_LmaxLanding/value");
                double AoA_1 = landing_config.interpalpha(CL_max_L);
                double AoA_2 = landing_config.interpalpha(CL_max_L - CL_max_L * 0.9);
                double CL_alpha = (CL_max_L - (CL_max_L - CL_max_L * 0.9)) / ((AoA_1 - AoA_2) * PI / 180.0);
                double V_TO_L = this->aircraft_xml()->at("aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/flight_envelope/maximum_approach_speed/value");
                double w_g = this->configuration()->at("module_configuration_file/program_settings/constraint_selection/standard_set/gust/gust_speed/value");
                double dn_G = this->configuration()->at("module_configuration_file/program_settings/constraint_selection/standard_set/gust/gust_load_factor/value");
                const auto constraint_border = constraint_analysis_tool.gust(CL_alpha, 0.0, V_TO_L, w_g, dn_G, weight_fraction);
                this->lower_boundaries.push_back(constraint_border);
            }
        }
    };
};