# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

def calculate_capital_costs(dict_ac_data, dict_operating_costs, runtime_output):
    """Calculate capital costs.

    The annuity formula, which is based on a modified mortgage equation, addresses both yearly depreciation and
    interest. The reason for the annuity method modification is to include the residual aircraft value at the end of
    the depreciation period into the capital costs. That does assume that an operator is buying an aircraft at a
    constant price per kg and spends the corresponding capital cost constantly per year all over the depreciation
    period. The capital costs are a major element of the route independent cost. They can be assumed as a linear
    function of the operating empty weight if the aircraft market influence is considered negligible.

    This function adds the 'capital_costs' dict to the 'dict_operating_costs' that contains the following parameters:
        - float capital_costs: Capital costs of the aircraft in EUR

    :param dict dict_ac_data: Dict containing parameters and values from aircraft exchange and module config file
    :param dict dict_operating_costs: Dict containing parameters from cost estimation calculation
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict dict_operating_costs: Dict containing parameters from cost estimation calculation
    """

    # Extract necessary variables from 'dict_ac_data'.
    depreciation_period = dict_ac_data['depreciation_period']
    m_operating_empty = dict_ac_data['m_operating_empty']
    price_per_operating_empty_mass = dict_ac_data['price_per_operating_empty_mass']
    rate_insurance = dict_ac_data['rate_insurance']
    rate_interest = dict_ac_data['rate_interest']
    exchange_rate = dict_ac_data['exchange_rate']

    # Determine residual value factor.
    depreciation_periods = [5, 10, 15, 20]
    residual_value_factors = [0.7, 0.5, 0.3, 0.1]
    residual_value_factor = 0
    for i in range(len(depreciation_periods)):
        if depreciation_period <= depreciation_periods[i]:
            residual_value_factor = residual_value_factors[i]
            break
    else:
        residual_value_factor = residual_value_factors[-1]

    # Calculate annuity factor.
    annuity_factor = rate_interest * ((1 - residual_value_factor * (1 / (1 + rate_interest)) **
                                       depreciation_period) / (1 - (1 / (1 + rate_interest)) **
                                                               depreciation_period))
    # Calculate capital costs.
    capital_costs = ((price_per_operating_empty_mass * m_operating_empty * (annuity_factor + rate_insurance))
                     *exchange_rate)

    # Print.
    runtime_output.print('Capital costs: ' + f"{int(capital_costs):,}" + ' USD')

    # Prepare output data.
    dict_operating_costs['capital_costs'] = {}
    dict_operating_costs['capital_costs']['capital_costs'] = capital_costs

    return dict_operating_costs
