# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

"""Module providing report functionalities for current calculation method."""
from yattag import Doc
from bs4 import BeautifulSoup
import os


def method_html_report(paths_and_names, routing_dict, data_dict, method_specific_output_dict, runtime_output):
    """HTML report function.

    This function is responsible for creating reports.
    [Add further information here...]

    :param dict paths_and_names: Dictionary containing system paths and ElementTrees
    :param dict routing_dict: Dictionary containing routing parameter
    :param dict data_dict: Dictionary containing results of module execution
    :return: None
    """
    runtime_output.print(f'HTML report is generated and saved...')

    """ Preparation."""
    # Extract tool and project name.
    tool_name = paths_and_names['tool_name']
    tool_version = paths_and_names['tool_version']
    report_title = (tool_name.replace('_', ' ')).capitalize()
    project_name = paths_and_names['name_of_project']
    root_of_module_config_tree = paths_and_names['root_of_module_config_tree']
    plot_switch = (eval(root_of_module_config_tree.find('.//plot_output/enable/value').text.capitalize()))
    study_exists = ("study" in data_dict.keys())

    # Extract data.
    # Note: Please ensure to format the data here (e.g., round floats)!
    # Cost overview data.
    cost_overview = [
        ("Route independent costs",
        f"{round(data_dict['design']['route_independent_costs']['route_independent_costs_annual']):,.0f} USD"),
        ("Route dependent costs",
        f"{round(data_dict['design']['route_dependent_costs']['route_dependent_costs_annual_design_point']):,.0f} USD"),
        ("Direct operating costs",
        f"{round(data_dict['design']['direct_operating_costs']['direct_operating_costs_annual_design_point']):,.0f} USD"),
    ]
    cost_overview_title = "Cost overview"
    cost_overview_column_titles = ["Cost share", "Design mission"]
    cost_overview_alignment = ["l", "r"]
    if study_exists:
        study_values = [
            f"{round(data_dict['study']['route_independent_costs']['route_independent_costs_annual']):,.0f} USD",
            f"{round(data_dict['study']['route_dependent_costs']['route_dependent_costs_annual_design_point']):,.0f} USD",
            f"{round(data_dict['study']['direct_operating_costs']['direct_operating_costs_annual_design_point']):,.0f} USD"
        ]
        # Add study values to each row in route_dependent_data.
        cost_overview = [
            (design_data[0], design_data[1], study_value)
            for design_data, study_value in zip(cost_overview, study_values)
        ]
        cost_overview_column_titles.append("Study mission")
        cost_overview_alignment.append("r")

    # Route independent data.
    route_independent_data = [
        ("Capital cost",
         f"{round(data_dict['design']['capital_costs']['capital_costs']):,.0f} USD"),
        ("Crew cost",
         f"{round(data_dict['design']['crew_costs']['crew_costs_annual']):,.0f} USD"),
        ("Route independent costs",
        f"{round(data_dict['design']['route_independent_costs']['route_independent_costs_annual']):,.0f} USD"),
    ]
    route_independent_data_title = "Route independent cost"
    route_independent_data_column_titles = ["Cost share", "Design mission"]
    route_independent_data_alignment = ["l", "r"]
    if study_exists:
        study_values = [
            f"{round(data_dict['study']['capital_costs']['capital_costs']):,.0f} USD",
            f"{round(data_dict['study']['crew_costs']['crew_costs_annual']):,.0f} USD",
            f"{round(data_dict['study']['route_independent_costs']['route_independent_costs_annual']):,.0f} USD"
        ]
        # Add study values to each row in route_dependent_data.
        route_independent_data = [
            (design_data[0], design_data[1], study_value)
            for design_data, study_value in zip(route_independent_data, study_values)
        ]
        route_independent_data_column_titles.append("Study mission")
        route_independent_data_alignment.append("r")

    # Route dependent data.
    route_dependent_data = [
        ("Fuel cost",
         f"{round(data_dict['design']['fuel_costs']['fuel_costs_design_point']):,.0f} USD"),
        ("Handling cost",
         f"{round(data_dict['design']['handling_costs']['handling_costs']):,.0f} USD"),
        ("Landing cost",
         f"{round(data_dict['design']['landing_costs']['landing_costs']):,.0f} USD"),
        ("Air traffic control cost",
         f"{round(data_dict['design']['air_traffic_control_costs']['air_traffic_control_costs_design_point']):,.0f} USD"),
        ("Maintenance cost",
         f"{round(data_dict['design']['maintenance_costs']['maintenance_costs']):,.0f} USD"),
        ("Route dependent costs",
         f"{round(data_dict['design']['route_dependent_costs']['route_dependent_costs_annual_design_point']):,.0f} USD"),
    ]
    route_dependent_data_title = "Route dependent cost"
    route_dependent_data_column_titles = ["Cost share", "Design mission"]
    route_dependent_data_alignment = ["l", "r"]
    if study_exists:
        study_values = [
            f"{round(data_dict['study']['fuel_costs']['fuel_costs_design_point']):,.0f} USD",
            f"{round(data_dict['study']['handling_costs']['handling_costs']):,.0f} USD",
            f"{round(data_dict['study']['landing_costs']['landing_costs']):,.0f} USD",
            f"{round(data_dict['study']['air_traffic_control_costs']['air_traffic_control_costs_design_point']):,.0f} USD",
            f"{round(data_dict['study']['maintenance_costs']['maintenance_costs']):,.0f} USD",
            f"{round(data_dict['study']['route_dependent_costs']['route_dependent_costs_annual_design_point']):,.0f} USD"
        ]
        # Add study values to each row in route_dependent_data.
        route_dependent_data = [
            (design_data[0], design_data[1], study_value)
            for design_data, study_value in zip(route_dependent_data, study_values)
        ]
        route_dependent_data_column_titles.append("Study mission")
        route_dependent_data_alignment.append("r")

    """Generate report."""
    # Start Yattag document.
    doc, tag, text = Doc().tagtext()

    # Generate header (without creating a new <html>).
    with tag('html', lang="en"):
        # Generate report header (adds <head>).
        generate_report_header(doc, report_title)

        # Generate body.
        with tag('body'):
            with tag('div', klass="logo"):
                text("")
            with tag('div', klass="content"):
                with tag('h1'):
                    text(f"Report - {report_title} of {project_name}")
                    with tag('br'):  # Line break for formatting
                        pass
                    with tag('span'):
                        with tag('font', size="2"):
                            text(f"All outputs of the program '{tool_name}' were created with version {tool_version}")
                    
                with tag('div', klass="container"):
                    # Data on the left side.
                    with tag('div', klass="box data"):
                        # Add small description.
                        with tag('p', style="font-size: 0.7rem; margin-left: 0.5rem; margin-top: 1rem; text-align: left;"):
                            text("Note: All data shown are annual costs.")
                        # Add title.
                        with tag('h2'):
                            text("Data")
                        # Add table for cost overview.
                        generate_table(doc, tag, text, cost_overview, cost_overview_title,
                                       cost_overview_column_titles, cost_overview_alignment)
                        # Add table for route independent costs.
                        generate_table(doc, tag, text, route_independent_data, route_independent_data_title,
                                       route_independent_data_column_titles, route_independent_data_alignment)
                        # Add table for route dependent costs.
                        generate_table(doc, tag, text, route_dependent_data, route_dependent_data_title,
                                       route_dependent_data_column_titles, route_dependent_data_alignment)

                    # Plots on the right side.
                    # Ensure that plots are enabled before displaying them on report.
                    if plot_switch:
                        with tag('div', klass="box plot"):
                            with tag('h2'):
                                text("Plots")
                                # PRD and DOC/TKO.
                                with tag('h3'):
                                    plot_name = "Payload range diagram and SKC for design mission"
                                    src_name = "../plots/cost_estimation_" + plot_name + ".svg"
                                with tag('img',src=src_name, klass="image-plot", style="margin-bottom: 20px;"):
                                    pass
                                # DOC per TKO.
                                with tag('h3'):
                                    plot_name = "DOC per ton kilometers offered (TKO) and TKO for design mission"
                                    src_name = "../plots/cost_estimation_" + plot_name + ".svg"
                                with tag('img',src=src_name, klass="image-plot", style="margin-bottom: 20px;"):
                                    pass
                                # SKC and TKO.
                                with tag('h3'):
                                    plot_name = "SKC and TKO for design mission"
                                    src_name = "../plots/cost_estimation_" + plot_name + ".svg"
                                with tag('img',src=src_name, klass="image-plot", style="margin-bottom: 20px;"):
                                    pass

    # Convert the document to HTML string.
    html_output = doc.getvalue()

    # Pretty HTML output with BeautifulSoup.
    soup = BeautifulSoup(html_output, 'html.parser')
    pretty_html = soup.prettify()

    """ Save report to current aircraft project directory. """
    # Save the report.
    project_directory = paths_and_names["project_directory"]
    report_directory = os.path.join(project_directory, 'reporting', 'report_html')
    os.makedirs(report_directory, exist_ok=True)
    output_path = os.path.join(report_directory, f"{tool_name}_report.html")

    # Write the HTML to the file.
    with open(output_path, "w", encoding="utf-8") as report:
        report.write(pretty_html)
    
    return


def generate_table(doc, tag, text, data, caption, column_titles, alignment=None):
    """Generate an HTML table.

    This function creates an HTML table from the provided data.
    Note: Options for the alignment types are "l" for left, "r" for right, and "m" for center. If not provided, columns
    are left-aligned by default.

    :param doc: Yattag Doc instance
    :param tag: Yattag tag function
    :param text: Yattag text function
    :param str caption: Table caption
    :param list column_titles: Titles of table columns
    :param list alignment: List of alignment types for ech column (defaults to None)
    :param list data: List of tuples containing table data
    """
    with tag('table', klass="content-table"):
        with tag('caption'):
            text(caption)
        with tag('thead'):
            with tag('tr'):
                for col in column_titles:
                    with tag('th'):
                        text(f"{col}")
        with tag('tbody'):
            row_number = 0
            for row in data:
                is_last_row = (row_number == len(data) - 1)
                with tag('tr', style="font-weight: normal;" if is_last_row else ""):
                    column_number = 0
                    for col in row:
                        if col is None:
                            with tag('td'):
                                text('N.A.')
                        else:
                            if alignment is not None:
                                if alignment[column_number] == "m":
                                    with tag('td', style="text-align: center;"):
                                        text(col)
                                elif alignment[column_number] == "r":
                                    with tag('td', style="text-align: right;"):
                                        text(col)
                                elif alignment[column_number] == "l":
                                    with tag('td', style="text-align: left;"):
                                        text(col)
                                else:
                                    with tag('td'):
                                        text(col)
                            else:
                                with tag('td'):
                                    text(col)
                        column_number += 1
                row_number += 1
    
    return


def generate_report_header(doc, tool_name):
    """Generate HTML header.
    
    This function generates an HTML header with meta information, style sheets, and title.

    :param doc: Yattag Doc instance
    :param str tool_name: Name of the tool
    """
    with doc.tag('head'):
        doc.stag('meta', name="viewport", content="width=device-width, initial-scale=1.0")
        doc.stag('meta', charset="UTF-8")
        with doc.tag('title'):
            doc.text(tool_name)
        doc.stag('link', rel="stylesheet", href="style.css")
    
    return
