/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "missionIOData.h"
#include <unordered_map>
#include <aixml/node.h>
#include <aixml/endnode.h>
#include <runtimeInfo/runtimeInfo.h>
#include <moduleBasics/runtimeIO.h>
#include <standardFiles/functions.h>

using std::endl;
using std::ostream;
using std::string;
using std::vector;

MissionIOData::MissionIOData(std::shared_ptr<RuntimeIO> rtIO)
    :
    config(MissionConfig()),
    myEngine(rtIO->getEngineDataDir() + "/" +
            EndnodeReadOnly<std::string>("component_design/propulsion/specific/propulsion/engine/model").read(rtIO->acxml).value(),
            EndnodeReadOnly<double>("component_design/propulsion/specific/propulsion/engine/scale_factor").read(rtIO->acxml).value()),
    flights_per_year(EndnodeReadOnly<int>("assessment/cost_estimation/operating_cost/direct_operating_cost/flights_per_year_study_mission").read(rtIO->acxml).value()),
    duration_operation(EndnodeReadOnly<int>("requirements_and_specifications/assessment_scenario/duration_operation").read(rtIO->acxml).value_xml()),
    number_of_engines(0) {
        config.read_all_config_data(rtIO->moduleConfig);
        number_of_engines = rtIO->acxml.getVector("/aircraft_exchange_file/component_design/propulsion/specific/propulsion").size();
        myEngine.set_same_engine_type_quantity(number_of_engines);
    }


/* Output data methods*/
void MissionIOData::update_ecological_assessment_xml(std::shared_ptr<node> ecologicalAssessmentXML, const Mission& thisMission) {
    myRuntimeInfo->out << "  Update ecological_assessment_results.xml" << std::endl;
    std::string subPath;
    subPath = "ecological_data_file/mission_emissions/" + thisMission.mission_type +"_mission/emissions/LTO_cycle";
    (Endnode<double>("CO2", "LTO CO2 emissions of one " + thisMission.mission_type +" mission",
                        Rounding(thisMission.LTO.CO2, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    (Endnode<double>("H2O", "LTO H2O emissions of one " + thisMission.mission_type +" mission",
                         Rounding(thisMission.LTO.H2O, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    (Endnode<double>("SO2", "LTO SO2 emissions of one " + thisMission.mission_type +" mission",
                        Rounding(thisMission.LTO.SO2, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    (Endnode<double>("SO4", "LTO SO4 emissions of one " + thisMission.mission_type +" mission",
                        Rounding(thisMission.LTO.SO4, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    (Endnode<double>("HC", "LTO HC emissions of one " + thisMission.mission_type +" mission",
                        Rounding(thisMission.LTO.HC, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    (Endnode<double>("CH4", "LTO CH4 emissions of one " + thisMission.mission_type +" mission",
                        Rounding(thisMission.LTO.CH4, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    (Endnode<double>("CO", "LTO CO emissions of one " + thisMission.mission_type +" mission",
                         Rounding(thisMission.LTO.CO, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    (Endnode<double>("NOx", "LTO NOx emissions of one " + thisMission.mission_type +" mission",
                        Rounding(thisMission.LTO.NOx, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    (Endnode<double>("soot", "LTO Soot emissions of one " + thisMission.mission_type +" mission",
                        Rounding(thisMission.LTO.soot, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    (Endnode<double>("c_soot_LTO_max", "Maximum soot concentration during " + thisMission.mission_type +" mission LTO",
                        Rounding(thisMission.c_soot_LTO_max, 3)*pow(10, -6), "1")).update((*ecologicalAssessmentXML)[subPath]);
    subPath = "ecological_data_file/mission_emissions/" + thisMission.mission_type +"_mission/emissions/cruise";
    (Endnode<double>("CO2", "Cruise CO2 emissions of one " + thisMission.mission_type +" mission",
                        Rounding(thisMission.cruise.CO2, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    (Endnode<double>("H2O", "Cruise H2O emissions of one " + thisMission.mission_type +" mission",
                        Rounding(thisMission.cruise.H2O, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    (Endnode<double>("SO2", "Cruise SO2 emissions of one " + thisMission.mission_type +" mission",
                        Rounding(thisMission.cruise.SO2, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    (Endnode<double>("SO4", "Cruise SO4 emissions of one " + thisMission.mission_type +" mission",
                        Rounding(thisMission.cruise.SO4, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    (Endnode<double>("HC", "Cruise HC emissions of one " + thisMission.mission_type +" mission",
                        Rounding(thisMission.cruise.HC, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    (Endnode<double>("CH4", "Cruise CH4 emissions of one " + thisMission.mission_type +" mission",
                        Rounding(thisMission.cruise.CH4, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    (Endnode<double>("CO", "Cruise CO emissions of one " + thisMission.mission_type +" mission",
                        Rounding(thisMission.cruise.CO, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    (Endnode<double>("NOx", "Cruise NOx emissions of one " + thisMission.mission_type +" mission",
                        Rounding(thisMission.cruise.NOx, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    (Endnode<double>("soot", "Cruise Soot emissions of one " + thisMission.mission_type +" mission",
                        Rounding(thisMission.cruise.soot, 3), "kg")).update((*ecologicalAssessmentXML)[subPath]);
    /* Add descriptions */
    ecologicalAssessmentXML->at("ecological_data_file/mission_emissions").addAttrib("description", "Results of mission emission calculation");
    ecologicalAssessmentXML->at("ecological_data_file/mission_emissions/" + thisMission.mission_type +"_mission").addAttrib("description",
        "Results of " + thisMission.mission_type +" mission calculation");
    ecologicalAssessmentXML->at("ecological_data_file/mission_emissions/" + thisMission.mission_type +"_mission/emissions").addAttrib("description",
        "Emissions of " + thisMission.mission_type +" mission");
    ecologicalAssessmentXML->at("ecological_data_file/mission_emissions/" + thisMission.mission_type +"_mission/emissions/LTO_cycle").addAttrib("description",
        "Emissions of landing and takeoff cycle");
    ecologicalAssessmentXML->at("ecological_data_file/mission_emissions/" + thisMission.mission_type +"_mission/emissions/cruise").addAttrib("description",
        "Emissions of cruise phase");
    ecologicalAssessmentXML->at("ecological_data_file/mission_emissions/design_mission").addAttrib("description", "Results of design mission calculation");
    ecologicalAssessmentXML->at("ecological_data_file/mission_emissions/design_mission/emissions").addAttrib("description", "Emissions of design mission ");
    ecologicalAssessmentXML->at("ecological_data_file/mission_emissions/design_mission/emissions/LTO_cycle").addAttrib("description",
        "Emissions of landing and takeoff cycle");
    ecologicalAssessmentXML->at("ecological_data_file/mission_emissions/design_mission/emissions/cruise").addAttrib("description", "Emissions of cruise phase");
}
