/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include <iostream>
#include <memory>
#include <algorithm>
#include <string>
#include <moduleBasics/runtimeIO.h>
#include <aixml/node.h>
#include "AQI_schaefer.h"
#include "AQISchaeferIOData.h"

AQISchaefer::AQISchaefer() : tolerance_CO(0.), tolerance_HC(0.), tolerance_NOx(0.), tolerance_soot(0.), AQI(0.) {}

void AQISchaefer::run(const std::shared_ptr<RuntimeIO> &rtIO, std::shared_ptr<node> ecoAssXML) {
    myRuntimeInfo->out << "  Calculation of air quality index" << std::endl;
    AQIschaeferIOData = std::make_shared<AQISchaeferIOData>(rtIO, ecoAssXML);
    calculateAQISchaefer(rtIO->aircraft_energy_carrier_type());
}

void AQISchaefer::calculateAQISchaefer(const std::string &fuelType) {
    AQIschaeferIOData->myEngine.set_same_engine_type_quantity(AQIschaeferIOData->number_of_engines);
    /*ICAO Emission Regulatory (subsonic AC), s. Annex 16 S.III-2-1*/
    /*for CO*/
    double DpF00_CO = 118.; //max. permitted output of CO [g] per kN thrust (per engine)
    /*for HC*/
    double DpF00_HC = 19.6; //max. permissible output at HC [g] per kN thrust (per engine)
    /*for NOx*/
    double DpF00_NOx;
    AQIschaeferIOData->myEngine.calculate_N1_with_penalties(0, 0., myAtmosphere, 1., "takeoff", 0., 0.);
    double thrust_takeoff = AQIschaeferIOData->myEngine.get_thrust_aircraft();
    double P3t_TO = AQIschaeferIOData->myEngine.get_physical_properties_stage(StageProperties::TotalPressure, EngineStage::St3);
    double P2t_TO = AQIschaeferIOData->myEngine.get_physical_properties_stage(StageProperties::TotalPressure, EngineStage::St2);
    double pressure_ratio = P3t_TO / P2t_TO; //s. Def. ICAO Annex 16 I-1-1 (Referencen Pressure ratio)
    double takeoff_thrust_per_engine = convertUnit(NEWTON, KILO, NEWTON, thrust_takeoff) / AQIschaeferIOData->number_of_engines; // [kN]
    //note: Regulatory NOx for Engines with first production model after 31 Dec. 2007 (III-2-4 d); for preliminary design based on most recent TW
    if (pressure_ratio <= 30.) {
        if (takeoff_thrust_per_engine > 89.) {
            DpF00_NOx = 16.72 + 1.4080 * pressure_ratio;
        } else if (takeoff_thrust_per_engine > 26.7 && takeoff_thrust_per_engine <= 89.) {
            DpF00_NOx = 38.5486 + 1.6823 * pressure_ratio - 0.2453 * takeoff_thrust_per_engine - 0.00308 * pressure_ratio * takeoff_thrust_per_engine;
        } else {
            // NOTE(franz#1#): Mantis ID: 0000318 if the worst comes to the worst: just set a value!
            myRuntimeInfo->err << "Thrust category of " << takeoff_thrust_per_engine << " kN not in ICAO regulations" << std::endl;
            exit(1);
        }
    } else if (pressure_ratio > 30. && pressure_ratio < 82.6) {
        if (takeoff_thrust_per_engine > 89.) {
            DpF00_NOx = -1.04 + 2.0 * pressure_ratio;
        } else if (takeoff_thrust_per_engine > 26.7 && takeoff_thrust_per_engine <= 89.) {
            DpF00_NOx = 46.16 + 1.4286 * pressure_ratio - 0.5303 * takeoff_thrust_per_engine + 0.00642 * pressure_ratio * takeoff_thrust_per_engine;
        } else {
            // NOTE(franz#1#): Mantis ID: 0000319 if the worst comes to the worst: just set a value!
            myRuntimeInfo->err << "Thrust category of " << takeoff_thrust_per_engine << " kN not in ICAO regulations" << std::endl;
            exit(1);
        }
    } else {
        DpF00_NOx = 32.0 + 1.6 * pressure_ratio;
    }
    /*for soot*/
    double c_soot_ICAO(0.);
    if (fuelType == "kerosene") {
        double SN_1 = 83.6 * pow(takeoff_thrust_per_engine, -0.274);
        double SN_2 = 50.;
        double SN_ICAO = std::min(SN_1, SN_2);
        c_soot_ICAO = 0.0000026156 * pow(SN_ICAO, 4.) - 0.00010998 * pow(SN_ICAO, 3.) + 0.0022367 * pow(SN_ICAO, 3.) + 0.10955 * SN_ICAO + 0.0012842; //mg/kg
    } else if (fuelType == "liquid_hydrogen") {
    } else {
        myRuntimeInfo->warn << "No smoke number defined for chosen fuel type!" << std::endl;
    }

    /*AQI Calculation*/
    // TODO(franz#1#): T139 add weighting factors
    tolerance_CO = AQIschaeferIOData->CO_LTO / AQIschaeferIOData->number_of_engines * 1000. / (DpF00_CO * takeoff_thrust_per_engine);
    tolerance_HC = AQIschaeferIOData->HC_LTO / AQIschaeferIOData->number_of_engines * 1000. / (DpF00_HC * takeoff_thrust_per_engine);
    tolerance_NOx = AQIschaeferIOData->NOx_LTO / AQIschaeferIOData->number_of_engines * 1000. / (DpF00_NOx * takeoff_thrust_per_engine);
    if (fabs(c_soot_ICAO) > ACCURACY_HIGH) {
     tolerance_soot = AQIschaeferIOData->c_soot_LTO_max / c_soot_ICAO;
    } else {
        tolerance_soot = 0;
    }

    AQI = (tolerance_CO + tolerance_HC + tolerance_NOx + tolerance_soot) / 4.;
    myRuntimeInfo->info << "Tolerance Factor CO: " << tolerance_CO << std::endl;
    myRuntimeInfo->info << "Tolerance Factor HC: " << tolerance_HC << std::endl;
    myRuntimeInfo->info << "Tolerance Factor NOx: " << tolerance_NOx << std::endl;
    myRuntimeInfo->info << "Tolerance Factor Soot: " << tolerance_soot << std::endl;
    myRuntimeInfo->info << "Air Quality Index: " << AQI << std::endl;
}

