/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "volumeCoefficientMethod.h"

#include <cmath>
#include <functional>

#include "lib/general_methods/newtonAlgorithm.h"

/**
 * \brief Compute Geometry by volume coefficient
 *
 * \param number_of_tails
 * \param volume_coefficient
 * \param sweep
 * \param taper_ratio
 * \param aspect_ratio
 * \param reference_area
 * \param reference_length
 * \param fuselage_length
 * \param reference_neutral_point
 * \param offset_position
 * \param aligned
 */
void VolumeCoefficientMethod::compute_geometry(int number_of_tails, double volume_coefficient, double sweep,
                                               double taper_ratio, double aspect_ratio, double reference_area,
                                               double reference_length, double fuselage_length,
                                               geom2::Point_3& reference_neutral_point, geom2::Point_3& offset_position,
                                               geom2::Direction_3 orientation, bool symmetric) {
  this->span_multiplier = symmetric ? 2 : 1;

  auto find_root_chord = [=,*this](double root_chord) {
    return aspect_ratio / 4 * (1 + taper_ratio) * (1 + taper_ratio) * root_chord * root_chord *
               (fuselage_length - offset_position.x() - reference_neutral_point.x()) +
           aspect_ratio / 4 * (1 + taper_ratio) * (1 + taper_ratio) * root_chord * root_chord * root_chord *
               (aspect_ratio / (3 * this->span_multiplier) * (1 + 2 * taper_ratio) * tan(sweep) +
                1. / 6. * (1 + taper_ratio + taper_ratio * taper_ratio) / (1 + taper_ratio) - 1) -
           volume_coefficient / number_of_tails * reference_length * reference_area;
  };
  double initial_root_chord = 0.1;
  int root_counter = 0;
  do {
    this->root_chord = newton_algorithm(find_root_chord, initial_root_chord);
    initial_root_chord *= 2;
  } while((this->root_chord < 0 || this->root_chord > 15) && root_counter++ < 5 );
  this->tip_chord = taper_ratio * this->root_chord;
  this->span = aspect_ratio * this->root_chord * (1 + taper_ratio) * 0.5;
  this->origin =
      geom2::Point_3(fuselage_length - offset_position.x() - root_chord, offset_position.y(), offset_position.z());
  this->orientation = orientation;
  this->sweep = sweep;
}

/**
 * \brief Compute by volume coefficient and fixed root value
 *
 * \param number_of_tails
 * \param volume_coefficient
 * \param sweep
 * \param taper_ratio
 * \param root_chord
 * \param wing_area
 * \param wing_span
 * \param wing_neutral_point_position
 * \param initial_position
 * \param orientation
 */
void VolumeCoefficientMethod::compute_geometry(int number_of_tails, double volume_coefficient, double sweep,
                                               double taper_ratio, double root_chord, double wing_area,
                                               double wing_span, geom2::Point_3& wing_neutral_point_position,
                                               geom2::Point_3& reference_position, bool symmetric) {
  this->span_multiplier = symmetric ? 2 : 1;
  this->root_chord = root_chord;
  this->tip_chord = root_chord * taper_ratio;
  this->sweep = sweep;
  double M = reference_position.x() +
             1.0 / 6.0 * root_chord * (1 + taper_ratio + taper_ratio * taper_ratio) / (1 + taper_ratio) -
             wing_neutral_point_position.x();
  double G = 1.0 / 3.0 * (1 + 2.0 * taper_ratio) / (1 + taper_ratio) * std::tan(sweep);
  double K = number_of_tails * (root_chord * (1 + taper_ratio) * 0.5);

  // pq formula
  double p = M / G;
  double q = -volume_coefficient * wing_span * wing_area / (K * G);

  // Only positive solution of pq formula
  this->span = -0.5 * p + std::sqrt(0.25 * p * p - q);
}

geom2::MultisectionSurface<geom2::AirfoilSection> VolumeCoefficientMethod::get_design(
    geom2::AirfoilSection root_profile, geom2::AirfoilSection tip_profile, const double dihedral) {
  geom2::AirfoilSection root(root_profile);
  geom2::AirfoilSection tip(tip_profile);
  root.origin = geom2::Point_3(0., 0., 0.);
  root.set_chord_length(root_chord);

  tip.origin = geom2::Point_3(this->span / this->span_multiplier * tan(this->sweep),
                              tan(dihedral) * this->span / this->span_multiplier, -this->span / this->span_multiplier);
  tip.set_chord_length(tip_chord);

  this->surface.sections.push_back(root);
  this->surface.sections.push_back(tip);
  this->surface.origin = this->origin;
  this->surface.normal = this->orientation;
  return this->surface;
}
