# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

def estimate_positions_and_cog(dict_ac_exchange, dict_mod_config, fuselage_description, runtime_output):
    """ Function to estimate global and local positions of fuselage and the cog coordinates.

    :param dict dict_ac_exchange: Dict containing parameter and according values from aircraft exchange file
    :param dict dict_mod_config: Dict containing parameter and according values from module configuration file
    :param dict fuselage_description: Dict containing parameter and values of current fuselage design
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict fuselage_description: Dict containing parameter and values of current fuselage design
    """
    ''' initialize local parameter '''
    convert_inch_to_meter = 0.0254
    fuselage_description['positions'] = {'total_position': {'x': float(),
                                                            'y': float(),
                                                            'z': float()},
                                         'fuselage_position': {'x': float(),
                                                               'y': float(),
                                                               'z': float()},
                                         'fuselage_accommodation_position': {'x': float(),
                                                                             'y': float(),
                                                                             'z': float()},
                                         'payload_deck_position': {}
                                         }

    fuselage_description['center_of_gravities'] = {'total_cog': {'x': float(),
                                                                 'y': float(),
                                                                 'z': float()},
                                                   'fuselage_cog': {'x': float(),
                                                                    'y': float(),
                                                                    'z': float()},
                                                   'fuselage_accommodation_cog': {'x': float(),
                                                                                  'y': float(),
                                                                                  'z': float()},
                                                   'payload_tube_cog': {},
                                                   'payload_deck_cog': {}
                                                   }

    ''' estimate total fuselage position '''
    # total x position of fuselage design group
    fuselage_description['positions']['total_position']['x'] = \
        -1 * dict_ac_exchange['global_reference_point_in_x_direction']
    # total y position of fuselage design group
    fuselage_description['positions']['total_position']['y'] = \
        -1 * dict_ac_exchange['global_reference_point_in_y_direction']
    # total z position of fuselage design group
    fuselage_description['positions']['total_position']['z'] = \
        -1 * dict_ac_exchange['global_reference_point_in_z_direction']

    ''' estimate fuselage position '''
    # x position of fuselage
    fuselage_description['positions']['fuselage_position']['x'] = 0.0
    # y position of fuselage
    fuselage_description['positions']['fuselage_position']['y'] = 0.0
    # z position of fuselage
    fuselage_description['positions']['fuselage_position']['z'] = 0.0

    ''' estimate fuselage accommodation position '''
    # x position of fuselage accommodation group
    fuselage_description['positions']['fuselage_accommodation_position']['x'] = \
        fuselage_description['positions']['fuselage_position']['x'] + dict_mod_config['length_of_flight_deck']
    # y position of fuselage accommodation group
    fuselage_description['positions']['fuselage_accommodation_position']['y'] = \
        fuselage_description['positions']['fuselage_position']['y'] + 0
    # z position of fuselage accommodation group
    fuselage_description['positions']['fuselage_accommodation_position']['z'] = \
        fuselage_description['positions']['fuselage_position']['z'] + 0

    ''' estimate payload deck position '''
    # Check if a multi passenger deck configuration is selected
    #  -> if true: -> estimate positions of each passenger deck.
    if dict_ac_exchange['multi_passenger_deck_layout']:
        for i in range(1, dict_ac_exchange['specific_number_of_passenger_decks']+1):
            fuselage_description['positions']['payload_deck_position']['passenger_deck_' + str(i)] = {}
            # x position of passenger deck
            fuselage_description['positions']['payload_deck_position']['passenger_deck_' + str(i)]['x'] = \
                ((i - 1) * dict_mod_config['length_of_flight_deck'])
            # y position of passenger deck
            fuselage_description['positions']['payload_deck_position']['passenger_deck_' + str(i)]['y'] = 0.0
            # z position of passenger deck
            fuselage_description['positions']['payload_deck_position']['passenger_deck_' + str(i)]['z'] = \
                ((dict_mod_config['passenger_aisle_standing_height']
                 + fuselage_description['geometry_description']['cabin_floor_thickness'])
                 * convert_inch_to_meter) * (i - 1) \
                - (dict_mod_config['cabin_floor_offset'] * convert_inch_to_meter)
    # Else condition: Single passenger deck configuration is selected -> estimate postion for the lowest passenger deck
    else:
        fuselage_description['positions']['payload_deck_position']['passenger_deck_1'] = {}
        # x position of passenger deck
        fuselage_description['positions']['payload_deck_position']['passenger_deck_1']['x'] = 0.0
        # y position of passenger deck
        fuselage_description['positions']['payload_deck_position']['passenger_deck_1']['y'] = 0.0
        # z position of passenger deck
        fuselage_description['positions']['payload_deck_position']['passenger_deck_1']['z'] = 0.0

    # cargo deck position
    fuselage_description['positions']['payload_deck_position']['cargo_deck'] = {}
    # x position of cargo deck
    fuselage_description['positions']['payload_deck_position']['cargo_deck']['x'] = \
        dict_mod_config['length_of_flight_deck']
    # y position of cargo deck
    fuselage_description['positions']['payload_deck_position']['cargo_deck']['y'] = 0.0
    # z position of cargo deck
    fuselage_description['positions']['payload_deck_position']['cargo_deck']['z'] = \
        round((- dict_mod_config['cabin_floor_offset']
               - fuselage_description['geometry_description']['cabin_floor_thickness']
               - fuselage_description['geometry_description']['fuselage_dimension']['cargo_bay_height'])
              * convert_inch_to_meter, 9)

    ''' estimate total center of gravity position '''
    # total x position of center of gravity
    fuselage_description['center_of_gravities']['total_cog']['x'] = \
        (fuselage_description['geometry_description']['fuselage_dimension']['fuselage_length']
         * convert_inch_to_meter * 0.45) - dict_ac_exchange['global_reference_point_in_x_direction']
    # total y position of center of gravity
    fuselage_description['center_of_gravities']['total_cog']['y'] = \
        0.0 - dict_ac_exchange['global_reference_point_in_x_direction']
    # total z position of center of gravity
    fuselage_description['center_of_gravities']['total_cog']['z'] =\
        (- 1/2 * dict_mod_config['cabin_floor_offset']) * convert_inch_to_meter \
        - dict_ac_exchange['global_reference_point_in_x_direction']

    ''' estimate fuselage center of gravity position '''
    # total x position of center of gravity
    fuselage_description['center_of_gravities']['fuselage_cog']['x'] = 0.0
    # total y position of center of gravity
    fuselage_description['center_of_gravities']['fuselage_cog']['y'] = 0.0
    # total z position of center of gravity
    fuselage_description['center_of_gravities']['fuselage_cog']['z'] = 0.0

    ''' estimate fuselage accommodation center of gravity position '''
    # total x position of center of gravity
    fuselage_description['center_of_gravities']['fuselage_accommodation_cog']['x'] = 0.0
    # total y position of center of gravity
    fuselage_description['center_of_gravities']['fuselage_accommodation_cog']['y'] = 0.0
    # total z position of center of gravity
    fuselage_description['center_of_gravities']['fuselage_accommodation_cog']['z'] = 0.0

    ''' estimate payload tube center of gravity position '''
    # total x position of center of gravity
    fuselage_description['center_of_gravities']['payload_tube_cog']['x'] = 0.0
    # total y position of center of gravity
    fuselage_description['center_of_gravities']['payload_tube_cog']['y'] = 0.0
    # total z position of center of gravity
    fuselage_description['center_of_gravities']['payload_tube_cog']['z'] = 0.0

    ''' estimate payload deck cog center of gravity position '''
    # total x position of center of gravity
    fuselage_description['center_of_gravities']['payload_deck_cog']['x'] = 0.0
    # total y position of center of gravity
    fuselage_description['center_of_gravities']['payload_deck_cog']['y'] = 0.0
    # total z position of center of gravity
    fuselage_description['center_of_gravities']['payload_deck_cog']['z'] = 0.0

    return fuselage_description
