# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

# Import standard modules.
import math


def payload_tube_water_volume_calculation(dict_ac_exchange, dict_mod_config, payload_tube_description, runtime_output):
    """ Calculation of water volume of passenger and cargo deck for each passenger tube.
            References and calculation method:
            [1] M.Sc Andreas Gobbin.
                Numerische Modellierung des Auslegungsprozesses für Passagierkabinen von Verkehrsflugzeugen
                unter Berücksichtigung der wichtigsten Auslegungsforderungen in MatLab.
                Fachgebiet Luftfahrzeugbau und Leichtbau, TU Berlin. Berlin. 2015.

    :param dict dict_ac_exchange: Dict containing parameter and according values from aircraft exchange file
    :param dict dict_mod_config: Dict containing parameter and according values from module configuration file
    :param dict payload_tube_description: Dict containing parameter and values of current blended wing body design
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict payload_tube_center_reference_points: Dict containing center reference points of each payload tube
            dict payload_tube_wall_reference_points: Dict containing wall reference points of each payload tube
            dict payload_tube_length: Dict containing the length of each payload tube
    """
    ''' initialize local parameter '''
    passenger_deck_water_volume = {}
    cargo_deck_water_volume = {}
    passenger_deck_height = {}
    cargo_deck_height = {}

    '''' calculation of passenger and cargo deck volume and height '''
    for tube in range(0, dict_ac_exchange['number_of_payload_tubes']):
        # estimate passenger deck height for each payload tube in [inch]
        passenger_deck_height['payload_tube_' + str(tube)] = \
            round(abs(payload_tube_description['payload_tube_center_reference_points'][
                          'payload_tube_center_front_reference_points']['payload_tube_' + str(tube)][
                          'center_reference_point_in_z'])
                  + abs(payload_tube_description['payload_tube_center_reference_points'][
                          'payload_tube_center_front_reference_points']['payload_tube_' + str(tube)][
                          'center_reference_point_upper_z'])
                  + abs(dict_mod_config['cabin_floor_offset'])
                  - payload_tube_description['payload_tube_wall_thickness']['payload_tube_' + str(tube)], 3)

        # estimate cargo deck height for each payload tube in [inch]
        cargo_deck_height['payload_tube_' + str(tube)] = \
            round(abs(payload_tube_description['payload_tube_center_reference_points'][
                          'payload_tube_center_front_reference_points']['payload_tube_' + str(tube)][
                          'distance_to_cargo_bottom'])
                  - abs(dict_mod_config['cabin_floor_offset'])
                  - abs(payload_tube_description['payload_tube_floor_thickness']['payload_tube_' + str(tube)]), 3)
        if cargo_deck_height['payload_tube_' + str(tube)] < 20:
            cargo_deck_height['payload_tube_' + str(tube)] = 0

        # estimate passenger deck water volume for each payload tube in [inch*inch*inch]
        cabin_width = (payload_tube_description['payload_tube_width']['payload_tube_' + str(tube)]
                       - 2 * payload_tube_description['payload_tube_wall_thickness']['payload_tube_' + str(tube)])
        passenger_deck_water_volume['payload_tube_' + str(tube)] = \
            round(payload_tube_description['payload_tube_length']['payload_tube_' + str(tube)]
                  * (pow(cabin_width / 2, 2)
                     * math.acos((cabin_width / 2 - passenger_deck_height['payload_tube_' + str(tube)])
                                 / (cabin_width / 2))
                     - (cabin_width / 2 - passenger_deck_height['payload_tube_' + str(tube)])
                     * pow(2 * cabin_width / 2 * passenger_deck_height['payload_tube_' + str(tube)]
                           - pow(passenger_deck_height['payload_tube_' + str(tube)], 2), 0.5)), 3)

        # estimate cargo deck water volume for each payload tube in [inch*inch*inch]
        cabin_width = (payload_tube_description['payload_tube_width']['payload_tube_' + str(tube)]
                       - 2 * payload_tube_description['payload_tube_wall_thickness']['payload_tube_' + str(tube)])
        cargo_deck_water_volume['payload_tube_' + str(tube)] = \
            round(payload_tube_description['payload_tube_length']['payload_tube_' + str(tube)]
                  * (pow(cabin_width / 2, 2)
                     * math.acos((cabin_width / 2 - cargo_deck_height['payload_tube_' + str(tube)])
                                 / (cabin_width / 2))
                     - (cabin_width / 2 - cargo_deck_height['payload_tube_' + str(tube)])
                     * pow(2 * cabin_width / 2 * cargo_deck_height['payload_tube_' + str(tube)]
                           - pow(cargo_deck_height['payload_tube_' + str(tube)], 2), 0.5)), 3)

    ''' debug print '''
    runtime_output.debug('Debug: The payload_tube_reference_point_calculation function was successfully executed!')

    return passenger_deck_water_volume, cargo_deck_water_volume, passenger_deck_height, cargo_deck_height
