# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

"""Module providing plotting functionalities for current calculation method."""
# Import standard libraries.
import os
import numpy as np
from matplotlib import pyplot as plt


def method_plot(paths_and_names, routing_dict, data_dict, method_specific_output_dict, runtime_output):
    """Plot function.

    This function is responsible for creating plots.
    [Add further information here...]

    :param dict paths_and_names: Dictionary containing system paths and ElementTrees
    :param dict routing_dict: Dictionary containing routing parameters
    :param dict data_dict: Dictionary containing results of module execution
    :param dict method_specific_output_dict: Dictionary containing method specific output data
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return: None
    """
    # Check path to directory for figure outputs
    project_directory = paths_and_names["project_directory"]
    plot_directory = project_directory + '/reporting/plots'
    if not os.path.isdir(plot_directory):
        os.makedirs(plot_directory)

    ''' visualisation of results '''
    # generate acn plot
    # call function to create the plots of acn methode
    # create_acn_plot(paths_and_names, data_dict['acn_estimation']['gross_mass_list'],
    #                 data_dict['acn_estimation']['aircraft_classification_number_output_list'],
    #                 data_dict['acn_estimation']['InputAlpha'], plot_directory)

    # call function to create the top and side view plots of landing gear design
    create_fuselage_3d_plot(paths_and_names, data_dict, plot_directory, runtime_output)

    runtime_output.print('Plots for fuselage design are successfully generated and saved.')

    return


def create_fuselage_3d_plot(paths_and_names, data_dict, plot_directory, runtime_output):
    """ This function create a isometric view of the fuselage of current aircraft project.

    :param dict paths_and_names: Dictionary containing system paths and ElementTrees
    :param dict data_dict: Dictionary containing system paths and ElementTrees
    :param string plot_directory: String contains the absolute system path to save the output figures.
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return: none
    """
    ''' Plot of front view '''
    # Generate figure.
    plot_title = 'Isometric view of fuselage shape'
    fig = plt.figure(figsize=(12, 7.5))
    ax = fig.add_subplot(111, projection='3d')

    # read ellipse data from dictionary
    cockpit_section = data_dict['dict_fuselage_deign_output']["cockpit_section"]
    tail_section = data_dict['dict_fuselage_deign_output']["tail_section"]

    # plot cockpit section
    x = []
    y = []
    z_up = []
    z_down = []
    x_direction_coordinates = cockpit_section["x_direction_coordinates"]
    y_direction_coordinates = cockpit_section["y_direction_coordinates"]
    z_upwards_coordinates = cockpit_section["z_upward_coordinates"]
    z_downwards_coordinates = cockpit_section["z_downward_coordinates"]

    i = 0
    while i < len(x_direction_coordinates):
        x.append(np.array(x_direction_coordinates["ellipse_number_" + str(i + 1)]) * 2.54 / 100)
        y.append(np.array(y_direction_coordinates["ellipse_number_" + str(i + 1)]) * 2.54 / 100)
        z_up.append(np.array(z_upwards_coordinates["ellipse_number_" + str(i + 1)]) * 2.54 / 100)
        z_down.append(np.array(z_downwards_coordinates["ellipse_number_" + str(i + 1)]) * 2.54 / 100)

        i += 1

    # plot tail section
    x_direction_coordinates = tail_section["x_direction_coordinates"]
    y_direction_coordinates = tail_section["y_direction_coordinates"]
    z_upwards_coordinates = tail_section["z_upward_coordinates"]
    z_downwards_coordinates = tail_section["z_downward_coordinates"]

    i = 0
    while i < len(x_direction_coordinates):
        x.append(np.array(x_direction_coordinates["ellipse_number_" + str(i + 1)]) * 2.54 / 100)
        y.append(np.array(y_direction_coordinates["ellipse_number_" + str(i + 1)]) * 2.54 / 100)
        z_up.append(np.array(z_upwards_coordinates["ellipse_number_" + str(i + 1)]) * 2.54 / 100)
        z_down.append(np.array(z_downwards_coordinates["ellipse_number_" + str(i + 1)]) * 2.54 / 100)

        i += 1

    ax.plot_surface(np.array(x), np.array(y), np.array(z_up), alpha=0.3, color='grey')
    ax.plot_surface(np.array(x), np.array(y), np.array(z_down), alpha=0.3, color='grey')

    ax.set_xlabel('fuselage length in meter', labelpad=60)

    ax.set_yticks([])
    ax.set_zticks([])
    ax.set_box_aspect((np.ptp(x), np.ptp(y), np.ptp(z_up)))

    # ax.view_init(30, 240)
    ax.view_init(0, 270)
    #plt.show()

    """ Save figures to current aircraft project directory. """
    # Save as *.svg
    tool_name = paths_and_names['tool_name']
    name_of_fig = tool_name + '_' + plot_title + '.svg'
    plt.savefig(plot_directory + '/' + name_of_fig, transparent=True)
    # Save as *.pdf
    name_of_fig = tool_name + '_' + plot_title + '.pdf'
    plt.savefig(plot_directory + '/' + name_of_fig, transparent=True)

    return
