/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

/* The following statements import necessary libraries */
#include <cstdlib> //system()/exit()                                           // Import c++ standard library for using cstdlib
#include <exception>                                                           // Import c++ standard library for using exception
#include <iostream> //Passing on required additional objects cout/cerr/endl    // Import c++ standard library for using iostream
#include <string>                                                              // Import c++ standard library for using string

/* The following statements import necessary header files */
#include <aixml/node.h>                                  // Import of header file from external aircraft xml library for using node
#include <runtimeInfo/runtimeInfo.h>                     // Import of header file from external runtime library for using runtimeInfo
#include <standardFiles/functions.h>                     // Import of header file from external standard file library for using function
#include "aircraft.h"                                    // Import of header file from src folder for using aircraft
#include "initialSizing.h"                               // Import of header file from src folder for using initialSizing
#include "initialSizingOutput.h"                         // Import of header file from src folder for using initialSizingOutput
#include "initialSizingSettings.h"                       // Import of header file from src folder for using initialSizingSettings
#include "toolinfo.h"                                    // Import of header file from src folder for using toolinfo

/* Using directives for namespaces */
using std::cerr;
using std::endl;
using std::exception;
using std::string;

/** \brief The main function is used to start the initialSizing process
*/
int main(int argc, const char *argv[]) {
    try {  /* Checking the different prerequisites and output parameter */
        /* Open initialSizing_conf. xml */
        std::unordered_map<string, string> argumentLookup;
        for (int argument(2); argument < argc; argument+=2) {
            argumentLookup[argv[argument-1]] = argv[argument];
        }
        if (argumentLookup["-c"].empty()) {
            argumentLookup["-c"] = getSettingsFilename(TOOL_NAME);
        }
        if (!fileExists(argumentLookup["-c"])) {
            cerr << "ERROR: Setting file " << argumentLookup["-c"] << " not found. Program termination!" << endl;
            system("PAUSE"); //Waiting for input (to display error, since not written into log)
            exit(1);
        }
        node& configXML = aixml::openDocument(argumentLookup["-c"]);
        /* Initialize runtimeInfo */
        myRuntimeInfo = new runtimeInfo(checkBoundaries(configXML.at("ConsoleOutputOn"), 0, true, 3, true),
                                        checkBoundaries(configXML.at("LogfileOutputOn"), 0, true, 3, true),
                                        configXML.at("LogFile"), TOOL_NAME);
        /* Initialize initialSizingSettings with config */
        const initialSizingSettings mySettings(configXML);
        myRuntimeInfo->out << "Use file: " << mySettings.getIoDir() + mySettings.getIoFilename() << endl;

        /* Initialize aircraft */
        aircraft myAcft(mySettings);

        /* Generating the initialSizing object for the calculations */
        initialSizing myInitialSizing(&myAcft, mySettings);
        myRuntimeInfo->out << "Start calculation ..." << endl;
        //myInitialSizing.initialize();
        myInitialSizing.run();
        myRuntimeInfo->out << "... calculation finished!" << endl;

        /* Creating output object */
        initialSizingOutput out(myInitialSizing, mySettings);
        out.initialize();

        /* Writing Output */
        myRuntimeInfo->out << "Write output ..." << endl;
        out.writeOutput();
        delete myRuntimeInfo;
        return 0;

    /* Catch: Trigger: if a C++ code specific error occurs */
    } catch (const char * text) {
        if (myRuntimeInfo == nullptr) {
            cerr << text << endl;
        } else {
            myRuntimeInfo->err << text << endl;
            delete myRuntimeInfo;
        }
        exit(1);

    /* Catch: Trigger: if a C++ code specific error occurs */
    } catch (const string& text) {
        if (myRuntimeInfo == nullptr) {
            cerr << text << endl;
        } else {
            myRuntimeInfo->err << text << endl;
            delete myRuntimeInfo;
        }
        exit(1);

    /* Catch: Trigger: if an error in c++ standard libraries occurs -> C++ standard error will occurred */
    } catch (const exception& e) {
        if (myRuntimeInfo == nullptr) {
            cerr << "The following C++ standard error has occurred: " << e.what() << endl;
        } else {
            myRuntimeInfo->err << "The following C++ standard error has occurred: " << e.what() << endl;
            delete myRuntimeInfo;
        }
        exit(1);

    /* Catch: Trigger: if any other error occurs */
    } catch (...) {
         if (myRuntimeInfo == nullptr) {
            cerr << "An uncaught error has occurred - Exit." << endl;
        } else {
            myRuntimeInfo->err << "An uncaught error has occurred - Exit." << endl;
            delete myRuntimeInfo;
        }
        exit(1);
    }
}
