/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include <iostream>
#include <cmath>
#include <memory>
#include <string>
#include "TawAnalytical.h"
#include "../TawInitialSizingIOData.h"
#include "../TawInitialSizingData.h"
#include "../../initialSizingDefines.h"
#include "atmosphere/atmosphere.h"


TawAnalytical::TawAnalytical(std::shared_ptr<RuntimeIO> rtIO) : rtIO_(rtIO), report_(rtIO) {}

void TawAnalytical::initialize() {
    const node& config = rtIO_->moduleConfig;
    /* Initializing IOData */
    data_ = std::make_shared<TawInitialSizingIOData>(rtIO_, config);
    /* Initializing values with data from the aircraft xml */
    data_->readMissionInfo(rtIO_->acxml);
    data_->readAccomodation(rtIO_->acxml);
    data_->readEnergyCarrier(rtIO_->acxml);
}

void TawAnalytical::run() {
    myRuntimeInfo->out << "Run initial sizing analytical strategy" << std::endl;
    // Check the design case configuration
    if (data_->configuration.designCase) {
        myRuntimeInfo->out << "Run - SizingMode" << std::endl;
        size();
    } else {
        myRuntimeInfo->out << "Run - StudyMode" << std::endl;
        study();
    }
}

void TawAnalytical::update() {
    data_->setGlobalReferencePoint();
    data_->updateTW(designTW_);
    data_->updateWS(designWS_);
}

void TawAnalytical::report() {
    myRuntimeInfo->debug << "write reports" << std::endl;
    setHtmlBody();
    setTexBody();
    if (rtIO_->plotOn) {
        generatePlotData();
        myRuntimeInfo->out << "Plots generated!" << std::endl;
    }
    report_.generateReports();
}

void TawAnalytical::save() {
     // not used on this level
}

/* METHODS */

void TawAnalytical::size() {
        myRuntimeInfo->out << "Start calculation ..." << std::endl;
   
        myRuntimeInfo->out << "Calculating payload ..." << std::endl;
        calculatePayloadAndCrew();

        myRuntimeInfo->out << "Calculating fuel fraction ..." << std::endl;
        calculateFuelFraction();

        myRuntimeInfo->out << "Calculating takeoff and empty mass ..." << std::endl;
        calculateMassEstimation();

        calculateCDO();

        calculateTakeoff();

        calculateClimb();

        calculateCruise();

        calculateLanding();
        
        findIntersectionPoints_LTO();

        findIntersectionPoints_CL();

        findIntersectionPoints_CR();

        findIntersectionPoints_CT();
        
        findIntersectionPoints_TC();

       getoptiCL();

       findIntersectionPoints_optiCL_Cruise();

       findIntersectionPoints_optiCL_Climb();

       findIntersectionPoints_optiCL_Takeoff();

        myRuntimeInfo->out << "Finding the design point..." << std::endl;
        findDesignPoint();

        myRuntimeInfo->out << "... calculation finished!" << std::endl;

}


void TawAnalytical::study() {
    // not implemented yet
}


void TawAnalytical::calculatePayloadAndCrew() {
    // Calculate passenger weight
    passengerWeight = data_->data.Accommodation.numberPassengers * (data_->data.Accommodation.massPassenger + data_->data.Accommodation.massLuggage);

    // Calculate cargo weight
    cargoWeight = data_->data.Accommodation.massCargo;

    // The ratios (0/0/0/0.1/0.9) represent the relative passenger distribution for the classes
    // The ratios (10/14/28/35/40) represent how many crew members are needed for every 10, 14, 28, 35, 40 passengers in each class. 
    
    numPresidentCrew = std::ceil(data_->data.Accommodation.numberPassengers * 0 / 10.0); 
    numFirstClassCrew = std::ceil(data_->data.Accommodation.numberPassengers * 0 / 14.0);
    numBusinessClassCrew = std::ceil(data_->data.Accommodation.numberPassengers * 0 / 28.0);
    numPremiumEconomyCrew = std::ceil(data_->data.Accommodation.numberPassengers * 0.1 / 35.0);
    numEconomyCrew = std::ceil(data_->data.Accommodation.numberPassengers * 0.9 / 40.0);

    numCrew = numPresidentCrew + numFirstClassCrew + numBusinessClassCrew + numPremiumEconomyCrew + numEconomyCrew + data_->configuration.numberOfPilots;

    // Calculate crew weight
    crewWeight = (data_->data.Accommodation.massPassenger + data_->data.Accommodation.massLuggage) * numCrew;

    // Calculate total payload
    Payload = passengerWeight + cargoWeight;
    
    std::cout << "Passenger Weight: " << passengerWeight << std::endl;
    std::cout << "Crew Number: " << numCrew << std::endl;
    std::cout << "Crew Weight: " << crewWeight << std::endl;
    std::cout << "cargoWeight: " << cargoWeight << std::endl;
    std::cout << "Payload: " << Payload << std::endl;
    
}


void TawAnalytical::calculateFuelFraction() {
    // Access necessary data from TawInitialSizingIOData
    range = data_->data.MissionInfo.range;
    contingency = data_->data.MissionInfo.contingency;
    machNumber = data_->data.MissionInfo.machNumber;
    SFCKerosene = data_->configuration.SFCKerosene;
    LDInitialCruise = data_->configuration.LDInitialCruise;
	LDInitialLoiter = data_->configuration.LDInitialLoiter;
    energyCarrier = data_->data.EnergyCarrier.energyCarrier;


    ff_warmup = data_->configuration.ff_warmup;
    ff_taxi = data_->configuration.ff_taxi;
    ff_takeoff = data_->configuration.ff_takeoff;
    ff_climb = data_->configuration.ff_climb;
    ff_descent = data_->configuration.ff_descent;
    ff_land = data_->configuration.ff_land;
    ff_missedandclimb = data_->configuration.ff_missedandclimb;
    altitude = data_->data.MissionInfo.altitude;

    // Create an instance of the atmosphere class
    atmosphere atm;

    // Set the atmosphere parameters using the reference values
    atm.setAtmosphere(0, ISA_TEMPERATURE, ISA_PRESSURE);

    // Get the speed of sound at the given altitude
    speedOfSound = atm.getSpeedOfSound(altitude);
    vCruise = speedOfSound * machNumber;
    std::cout << "Cruise Speed: " << vCruise << std::endl;

    // Calculate fuel fractions of different mission segments
    CruiseFuel = exp((-range * SFCKerosene) / (vCruise * LDInitialCruise));
    distAlter = data_->data.MissionInfo.distalt;
    std::cout << "Alternate Distance: " << distAlter << std::endl;
	AltFuel = exp((-distAlter * SFCKerosene) / (vCruise * LDInitialCruise));
    timeLoiter = data_->data.MissionInfo.loitertime;
    std::cout << "Loiter time: " << timeLoiter << std::endl;

	LoiterFuel = exp((-timeLoiter * (0.85 * SFCKerosene)) / (LDInitialLoiter));
   
    // Calculate fuel fraction based on fuel type (0: kerosene, 1: liquid hydrogen)

  if (energyCarrier == "kerosene") {
    // Calculate FuelFraction for kerosene
    FuelFraction = 1.02 * (1 - ((ff_warmup * ff_taxi * ff_takeoff *
         ff_climb  * ff_descent * CruiseFuel * (contingency + 1)) * ff_missedandclimb * AltFuel * LoiterFuel * ff_land));
    std::cout << "Fuel Fraction (Kerosene): " << FuelFraction << std::endl;
  } else if (energyCarrier == "hydrogen") {
    // Calculate FuelFraction for liquid hydrogen
    FuelFraction = 1.05 * (1 / 2.8) * (1 - ((ff_warmup * ff_taxi * ff_takeoff *
         ff_climb  * ff_descent * CruiseFuel * (contingency + 1)) * ff_missedandclimb * AltFuel  * LoiterFuel * ff_land));
    std::cout << "Fuel Fraction (Liquid Hydrogen): " << FuelFraction << std::endl;
  } else {
    // Abort fuel fraction calculation for unsupported fuel types
    std::cout << "Unsupported fuel type. Fuel fraction calculation aborted." << std::endl;
    return;  // Abort the function execution
  }
}


void TawAnalytical::calculateMassEstimation() {
    // Constants
    mo_iteration = 5.5 * Payload;
    c_kerosine = 0.55;
    c_LH2 = 0.44;
    mo_iteration = mo_initial;

    // Select the appropriate coefficient based on the fuel type
    c_fuel = (energyCarrier == "kerosene") ? c_kerosine : c_LH2;


    // Perform the iterative calculations
    do {
        mo_iteration_previous = mo_iteration;

        // Calculate me/mo using mo_iteration
        me_mo = c_fuel * cos(0.0000012 * mo_iteration);

        // Calculate new mo_iteration
        mo_iteration = (Payload + crewWeight) / (1 - (FuelFraction) - (me_mo));

        // Debug new mo_iteration
    } while (!checkConvergence(mo_iteration, mo_iteration_previous));  // Loop until convergence is achieved

    // Calculate final mass estimation
    final_mass_estimation_ = mo_iteration * me_mo;
    mo_iteration_ = mo_iteration;

    // Print the calculated values
    std::cout << "MTOM: " << mo_iteration_ << std::endl;
    std::cout << "OME: " << final_mass_estimation_ << std::endl;

    // Update the data
    data_->updateMTOM(mo_iteration_);
    data_->updateOME(final_mass_estimation_);
}

bool TawAnalytical::checkConvergence(double mo_iteration, double mo_iteration_previous) {
    // Convergence check here
    double convergence_criteria = 0.0005;  // 0.05%
    return std::abs(mo_iteration - mo_iteration_previous) < convergence_criteria;
}


void TawAnalytical::calculateCDO() {
    // Read fricition coefficient
    double Cfeq = data_->configuration.Cfeq;
    // Swet/Sref constants for kerosine and LH2
    double SwetSref_kerosene = 6;
    double SwetSref_LH2 = 8.5;
    double SwetSref = (energyCarrier == "kerosene") ? SwetSref_kerosene : SwetSref_LH2;
    CD0_clean = Cfeq * (SwetSref);
}

void TawAnalytical::calculateTakeoff() {
    // Read friction coefficient
    CLmaxTO = data_->configuration.CLmaxTO;
    numberOfEngines = data_->configuration.numberOfEngines;
    double STOFL = data_->data.MissionInfo.sTOFL;
    double KTO;
    if (numberOfEngines == 2) {
        KTO = 2.45;
    } else if (numberOfEngines == 3) {
        KTO = 2.26;
    } else if (numberOfEngines == 4) {
        KTO = 2.06;
    } else {
        throw std::runtime_error("Unsupported number of engines");
    }
        takeoffDataWS.clear();
        takeoffDataTW.clear();

    for (double mo_S = 100; mo_S <= 1000; mo_S += 0.5) {
        double T_W_TO = ((KTO)/ (CLmaxTO * STOFL))*(mo_S);
        takeoffDataWS.push_back(mo_S);
        takeoffDataTW.push_back(T_W_TO);
    }

}


void TawAnalytical::calculateClimb() {
    double DeltaCD_HL = data_->configuration.DeltaCD_HL;
    AR = data_->configuration.AR;
    e = data_->configuration.e;
    double Gamma = data_->data.MissionInfo.Gamma;

    double mc_mo = ff_warmup*ff_taxi*ff_takeoff;
    double CD_TO = CD0_clean + DeltaCD_HL + (std::pow(CLmaxTO, 2) / (PI * AR * e));
    double T_W_C = mc_mo * ((numberOfEngines / (numberOfEngines - 1.0)) * (Gamma + (CD_TO / CLmaxTO)));

        climbDataWS.clear();
        climbDataTW.clear();

    for (double mo_S = 100; mo_S <= 1000; mo_S += 0.5) {
         double T_W_C = mc_mo * ((numberOfEngines / (numberOfEngines - 1.0)) * (Gamma + (CD_TO / CLmaxTO)));
        climbDataWS.push_back(mo_S);
        climbDataTW.push_back(T_W_C);
    }
}

void TawAnalytical::calculateCruise() {
    double AR = data_->configuration.AR;
    double e = data_->configuration.e;
    double To_Tcr = 0.95;
    double g = 9.81;
    altitude = data_->data.MissionInfo.altitude;

    // Create an instance of the atmosphere class
    atmosphere atm;

    // Set the atmosphere parameters using the reference values
    atm.setAtmosphere(0, ISA_TEMPERATURE, ISA_PRESSURE);

    mcr_mo = data_->configuration.mcr_mo;
    double densityCruise = atm.getDensity(altitude);
    double q = (densityCruise / 2) * std::pow(vCruise, 2);

        cruiseDataWS.clear();
        cruiseDataTW.clear();

    for (double mo_S = 100; mo_S <= 1000; mo_S += 0.5) {
        double T_W_R = ((To_Tcr) * ((CD0_clean*q/(mo_S)*g)+((mo_S)*g/(q*PI*AR*e))*std::pow(mcr_mo, 2)))/g;
        cruiseDataWS.push_back(mo_S);
        cruiseDataTW.push_back(T_W_R);
    }
}

void TawAnalytical::calculateLanding() {
    double CLmaxL = data_->configuration.CLmaxL;
    double ml_mo = data_->configuration.ml_mo;
    double vApproach = data_->data.MissionInfo.vApproach;
    double g = 9.81;
    double densitySL = 1.225;

    landingDataWS.clear();
    landingDataTW.clear();

    size_t size = cruiseDataTW.size();  

    for (size_t i = 0; i < size; ++i) {
        double moS_landing = ((std::pow(vApproach, 2))*densitySL*CLmaxL)/((std::pow(densitySL, 2))*2*g*(ml_mo));
        landingDataWS.push_back(moS_landing);
        landingDataTW.push_back(cruiseDataTW[i]);
    }
}


void TawAnalytical::findIntersectionPoints_LTO() {

    // Iterate over the landing and takeoff data to find intersections
    for (size_t i = 0; i < landingDataWS.size() - 1; ++i) {
        double landingWS1 = landingDataWS[i];
        double landingTW1 = landingDataTW[i];
        double landingWS2 = landingDataWS[i + 1];
        double landingTW2 = landingDataTW[i + 1];

        for (size_t j = 0; j < takeoffDataWS.size() - 1; ++j) {
            double takeoffWS1 = takeoffDataWS[j];
            double takeoffTW1 = takeoffDataTW[j];
            double takeoffWS2 = takeoffDataWS[j + 1];
            double takeoffTW2 = takeoffDataTW[j + 1];

            double denominator = (takeoffWS2 - takeoffWS1) * (landingTW2 - landingTW1) - (takeoffTW2 - takeoffTW1) * (landingWS2 - landingWS1);

            if (std::abs(denominator) > 1e-6) {
                double ua = ((takeoffWS2 - takeoffWS1) * (landingTW1 - takeoffTW1) - (takeoffTW2 - takeoffTW1) * (landingWS1 - takeoffWS1)) / denominator;
                double ub = ((landingWS2 - landingWS1) * (landingTW1 - takeoffTW1) - (landingTW2 - landingTW1) * (landingWS1 - takeoffWS1)) / denominator;

                if (ua >= 0 && ua <= 1 && ub >= 0 && ub <= 1) {
                    double intersectWS = landingWS1 + ua * (landingWS2 - landingWS1);
                    double intersectTW = landingTW1 + ua * (landingTW2 - landingTW1);
                    intersectionPoints_LTO.emplace_back(intersectWS, intersectTW);
                }
            }
        }
    }
}


void TawAnalytical::findIntersectionPoints_CL() {
    
    // Iterate over the landing and climb data to find intersections
    for (size_t i = 0; i < landingDataWS.size() - 1; ++i) {
        double landingWS1 = landingDataWS[i];
        double landingTW1 = landingDataTW[i];
        double landingWS2 = landingDataWS[i + 1];
        double landingTW2 = landingDataTW[i + 1];

        for (size_t j = 0; j < climbDataWS.size() - 1; ++j) {
            double climbWS1 = climbDataWS[j];
            double climbTW1 = climbDataTW[j];
            double climbWS2 = climbDataWS[j + 1];
            double climbTW2 = climbDataTW[j + 1];

            double denominator = (climbWS2 - climbWS1) * (landingTW2 - landingTW1) - (climbTW2 - climbTW1) * (landingWS2 - landingWS1);

            if (std::abs(denominator) > 1e-6) {
                double ua = ((climbWS2 - climbWS1) * (landingTW1 - climbTW1) - (climbTW2 - climbTW1) * (landingWS1 - climbWS1)) / denominator;
                double ub = ((landingWS2 - landingWS1) * (landingTW1 - climbTW1) - (landingTW2 - landingTW1) * (landingWS1 - climbWS1)) / denominator;

                if (ua >= 0 && ua <= 1 && ub >= 0 && ub <= 1) {
                    double intersectWS = landingWS1 + ua * (landingWS2 - landingWS1);
                    double intersectTW = landingTW1 + ua * (landingTW2 - landingTW1);
                    intersectionPoints_CL.emplace_back(intersectWS, intersectTW);
                }
            }
        }
    }
}


void TawAnalytical::findIntersectionPoints_CR() {
    // Iterate over the landing and cruise data to find intersections
    for (size_t i = 0; i < landingDataWS.size() - 1; ++i) {
        double landingWS1 = landingDataWS[i];
        double landingTW1 = landingDataTW[i];
        double landingWS2 = landingDataWS[i + 1];
        double landingTW2 = landingDataTW[i + 1];

        for (size_t j = 0; j < cruiseDataWS.size() - 1; ++j) {
            double cruiseWS1 = cruiseDataWS[j];
            double cruiseTW1 = cruiseDataTW[j];
            double cruiseWS2 = cruiseDataWS[j + 1];
            double cruiseTW2 = cruiseDataTW[j + 1];

            double denominator = (cruiseWS2 - cruiseWS1) * (landingTW2 - landingTW1) - (cruiseTW2 - cruiseTW1) * (landingWS2 - landingWS1);

            if (std::abs(denominator) > 1e-6) {
                double ua = ((cruiseWS2 - cruiseWS1) * (landingTW1 - cruiseTW1) - (cruiseTW2 - cruiseTW1) * (landingWS1 - cruiseWS1)) / denominator;
                double ub = ((landingWS2 - landingWS1) * (landingTW1 - cruiseTW1) - (landingTW2 - landingTW1) * (landingWS1 - cruiseWS1)) / denominator;

                if (ua >= 0 && ua <= 1 && ub >= 0 && ub <= 1) {
                    double intersectWS = landingWS1 + ua * (landingWS2 - landingWS1);
                    double intersectTW = landingTW1 + ua * (landingTW2 - landingTW1);
                    intersectionPoints_CR.emplace_back(intersectWS, intersectTW);
                }
            }
        }
    }
}


void TawAnalytical::findIntersectionPoints_CT() {

    // Iterate over the climb and takeoff data to find intersections
    for (size_t i = 0; i < climbDataWS.size() - 1; ++i) {
        double climbWS1 = climbDataWS[i];
        double climbTW1 = climbDataTW[i];
        double climbWS2 = climbDataWS[i + 1];
        double climbTW2 = climbDataTW[i + 1];

        for (size_t j = 0; j < takeoffDataWS.size() - 1; ++j) {
            double takeoffWS1 = takeoffDataWS[j];
            double takeoffTW1 = takeoffDataTW[j];
            double takeoffWS2 = takeoffDataWS[j + 1];
            double takeoffTW2 = takeoffDataTW[j + 1];

            double denominator = (takeoffWS2 - takeoffWS1) * (climbTW2 - climbTW1) - (takeoffTW2 - takeoffTW1) * (climbWS2 - climbWS1);

            if (std::abs(denominator) > 1e-6) {
                double ua = ((takeoffWS2 - takeoffWS1) * (climbTW1 - takeoffTW1) - (takeoffTW2 - takeoffTW1) * (climbWS1 - takeoffWS1)) / denominator;
                double ub = ((climbWS2 - climbWS1) * (climbTW1 - takeoffTW1) - (climbTW2 - climbTW1) * (climbWS1 - takeoffWS1)) / denominator;

                if (ua >= 0 && ua <= 1 && ub >= 0 && ub <= 1) {
                    double intersectWS = climbWS1 + ua * (climbWS2 - climbWS1);
                    double intersectTW = climbTW1 + ua * (climbTW2 - climbTW1);
                    intersectionPoints_CT.emplace_back(intersectWS, intersectTW);
                }
            }
        }
    }
}


void TawAnalytical::findIntersectionPoints_TC() {

    // Iterate over the takeoff and cruise data to find intersections
    for (size_t i = 0; i < takeoffDataWS.size() - 1; ++i) {
        double takeoffWS1 = takeoffDataWS[i];
        double takeoffTW1 = takeoffDataTW[i];
        double takeoffWS2 = takeoffDataWS[i + 1];
        double takeoffTW2 = takeoffDataTW[i + 1];

        for (size_t j = 0; j < cruiseDataWS.size() - 1; ++j) {
            double cruiseWS1 = cruiseDataWS[j];
            double cruiseTW1 = cruiseDataTW[j];
            double cruiseWS2 = cruiseDataWS[j + 1];
            double cruiseTW2 = cruiseDataTW[j + 1];

            double denominator = (cruiseWS2 - cruiseWS1) * (takeoffTW2 - takeoffTW1) - (cruiseTW2 - cruiseTW1) * (takeoffWS2 - takeoffWS1);

            if (std::abs(denominator) > 1e-6) {
                double ua = ((cruiseWS2 - cruiseWS1) * (takeoffTW1 - cruiseTW1) - (cruiseTW2 - cruiseTW1) * (takeoffWS1 - cruiseWS1)) / denominator;
                double ub = ((takeoffWS2 - takeoffWS1) * (takeoffTW1 - cruiseTW1) - (takeoffTW2 - takeoffTW1) * (takeoffWS1 - cruiseWS1)) / denominator;

                if (ua >= 0 && ua <= 1 && ub >= 0 && ub <= 1) {
                    double intersectWS = takeoffWS1 + ua * (takeoffWS2 - takeoffWS1);
                    double intersectTW = takeoffTW1 + ua * (takeoffTW2 - takeoffTW1);
                    intersectionPoints_TC.emplace_back(intersectWS, intersectTW);
                }
            }
        }
    }
}


void TawAnalytical::getoptiCL() {
    double optimalCL = data_->configuration.optimalCL;
    double g = 9.81;
    // Create an instance of the atmosphere class
    atmosphere atm;
    atm.setAtmosphere(0, ISA_TEMPERATURE, ISA_PRESSURE);

    double densityCruise = atm.getDensity(altitude);
    double q = (densityCruise / 2) * std::pow(vCruise, 2);

    optiCLDataWS.clear();
    optiCLDataTW.clear();

    size_t size = cruiseDataTW.size();

    for (size_t i = 0; i < size; ++i) {
        double moS_optiCL = (mcr_mo) * (q / g) * (optimalCL);
        optiCLDataWS.push_back(moS_optiCL);
        optiCLDataTW.push_back(cruiseDataTW[i]);
    }
}


void TawAnalytical::findIntersectionPoints_optiCL_Takeoff() {
    // Iterate over the optiCL and takeoff data to find intersections
    for (size_t i = 0; i < optiCLDataWS.size() - 1; ++i) {
        double optiCLWS1 = optiCLDataWS[i];
        double optiCLTW1 = optiCLDataTW[i];
        double optiCLWS2 = optiCLDataWS[i + 1];
        double optiCLTW2 = optiCLDataTW[i + 1];

        for (size_t j = 0; j < takeoffDataWS.size() - 1; ++j) {
            double takeoffWS1 = takeoffDataWS[j];
            double takeoffTW1 = takeoffDataTW[j];
            double takeoffWS2 = takeoffDataWS[j + 1];
            double takeoffTW2 = takeoffDataTW[j + 1];

            double denominator = (takeoffWS2 - takeoffWS1) * (optiCLTW2 - optiCLTW1) - (takeoffTW2 - takeoffTW1) * (optiCLWS2 - optiCLWS1);

            if (std::abs(denominator) > 1e-6) {
                double ua = ((takeoffWS2 - takeoffWS1) * (optiCLTW1 - takeoffTW1) - (takeoffTW2 - takeoffTW1) * (optiCLWS1 - takeoffWS1)) / denominator;
                double ub = ((optiCLWS2 - optiCLWS1) * (optiCLTW1 - takeoffTW1) - (optiCLTW2 - optiCLTW1) * (optiCLWS1 - takeoffWS1)) / denominator;

                if (ua >= 0 && ua <= 1 && ub >= 0 && ub <= 1) {
                    double intersectWS = optiCLWS1 + ua * (optiCLWS2 - optiCLWS1);
                    double intersectTW = optiCLTW1 + ua * (optiCLTW2 - optiCLTW1);
                    intersectionPoints_optiCL_Takeoff.emplace_back(intersectWS, intersectTW);
                }
            }
        }
    }
}


void TawAnalytical::findIntersectionPoints_optiCL_Cruise() {
    // Iterate over the optiCL and cruise data to find intersections
    for (size_t i = 0; i < optiCLDataWS.size() - 1; ++i) {
        double optiCLWS1 = optiCLDataWS[i];
        double optiCLTW1 = optiCLDataTW[i];
        double optiCLWS2 = optiCLDataWS[i + 1];
        double optiCLTW2 = optiCLDataTW[i + 1];

        for (size_t j = 0; j < cruiseDataWS.size() - 1; ++j) {
            double cruiseWS1 = cruiseDataWS[j];
            double cruiseTW1 = cruiseDataTW[j];
            double cruiseWS2 = cruiseDataWS[j + 1];
            double cruiseTW2 = cruiseDataTW[j + 1];

            double denominator = (cruiseWS2 - cruiseWS1) * (optiCLTW2 - optiCLTW1) - (cruiseTW2 - cruiseTW1) * (optiCLWS2 - optiCLWS1);

            if (std::abs(denominator) > 1e-6) {
                double ua = ((cruiseWS2 - cruiseWS1) * (optiCLTW1 - cruiseTW1) - (cruiseTW2 - cruiseTW1) * (optiCLWS1 - cruiseWS1)) / denominator;
                double ub = ((optiCLWS2 - optiCLWS1) * (optiCLTW1 - cruiseTW1) - (optiCLTW2 - optiCLTW1) * (optiCLWS1 - cruiseWS1)) / denominator;

                if (ua >= 0 && ua <= 1 && ub >= 0 && ub <= 1) {
                    double intersectWS = optiCLWS1 + ua * (optiCLWS2 - optiCLWS1);
                    double intersectTW = optiCLTW1 + ua * (optiCLTW2 - optiCLTW1);
                    intersectionPoints_optiCL_Cruise.emplace_back(intersectWS, intersectTW);
                }
            }
        }
    }
}


void TawAnalytical::findIntersectionPoints_optiCL_Climb() {
    // Iterate over the optiCL and climb data to find intersections
    for (size_t i = 0; i < optiCLDataWS.size() - 1; ++i) {
        double optiCLWS1 = optiCLDataWS[i];
        double optiCLTW1 = optiCLDataTW[i];
        double optiCLWS2 = optiCLDataWS[i + 1];
        double optiCLTW2 = optiCLDataTW[i + 1];

        for (size_t j = 0; j < climbDataWS.size() - 1; ++j) {
            double climbWS1 = climbDataWS[j];
            double climbTW1 = climbDataTW[j];
            double climbWS2 = climbDataWS[j + 1];
            double climbTW2 = climbDataTW[j + 1];

            double denominator = (climbWS2 - climbWS1) * (optiCLTW2 - optiCLTW1) - (climbTW2 - climbTW1) * (optiCLWS2 - optiCLWS1);

            if (std::abs(denominator) > 1e-6) {
                double ua = ((climbWS2 - climbWS1) * (optiCLTW1 - climbTW1) - (climbTW2 - climbTW1) * (optiCLWS1 - climbWS1)) / denominator;
                double ub = ((optiCLWS2 - optiCLWS1) * (optiCLTW1 - climbTW1) - (optiCLTW2 - optiCLTW1) * (optiCLWS1 - climbWS1)) / denominator;

                if (ua >= 0 && ua <= 1 && ub >= 0 && ub <= 1) {
                    double intersectWS = optiCLWS1 + ua * (optiCLWS2 - optiCLWS1);
                    double intersectTW = optiCLTW1 + ua * (optiCLTW2 - optiCLTW1);
                    intersectionPoints_optiCL_Climb.emplace_back(intersectWS, intersectTW);
                }
            }
        }
    }
}


void TawAnalytical::findDesignPoint() {
    // Function to find the maximum TW in a vector of intersection points
    auto findTW = [](const std::vector<std::pair<double, double>>& points) -> double {
        double maxTW = 0;
        for (const auto& point : points) {
            if (point.second > maxTW) {
                maxTW = point.second;
            }
        }
        return maxTW;
    };

    // Function to find the corresponding WS for a given TW
    auto findCorrespondingWS = [](const std::vector<std::pair<double, double>>& points, double TW) -> double {
        for (const auto& point : points) {
            if (point.second == TW) {
                return point.first;
            }
        }
        return 0.0;
    };

    // optimal CL data 
    getoptiCL();

    double TW_CL = findTW(intersectionPoints_CL);
    double ws_CL = findCorrespondingWS(intersectionPoints_CL, TW_CL);

    double TW_CR = findTW(intersectionPoints_CR);
    double ws_CR = findCorrespondingWS(intersectionPoints_CR, TW_CR);

    double TW_LTO = findTW(intersectionPoints_LTO);
    double ws_LTO = findCorrespondingWS(intersectionPoints_LTO, TW_LTO);

    // Variable to store the design TW and WS
    std::string designPoint;

    // Determine the highest TW among CL, CR, and LTO
    if (TW_CL > TW_CR && TW_CL > TW_LTO) {
        designTW_ = TW_CL;
        designWS_ = ws_CL;
        designPoint = "Climb and Landing";
    } else if (TW_CR > TW_CL && TW_CR > TW_LTO) {
        designTW_ = TW_CR;
        designWS_ = ws_CR;
        designPoint = "Cruise and Landing";
    } else {
        // the intersectionPoints_LTO "takeoff-landing" has the maximum T/W among the other constraints
        // Find T/W for takeoff-climb and takeoff-cruise intersections

        double TW_CT = findTW(intersectionPoints_CT);
        double ws_CT = findCorrespondingWS(intersectionPoints_CT, TW_CT);

        double TW_TC = findTW(intersectionPoints_TC);
        double ws_TC = findCorrespondingWS(intersectionPoints_TC, TW_TC);

        // Determine the higher T/W between takeoff-climb and takeoff-cruise intersections to be 
        // interpolated with the highest point of T/W in takeoff-landing constraint

        double TW_temp = std::max(TW_CT, TW_TC);
        double WS_temp = (TW_temp == TW_CT) ? ws_CT : ws_TC;

        // Check if optiCL is within the range
        bool optiCLInRange = false;
        double TW_OptiCLTO = findTW(intersectionPoints_optiCL_Takeoff);
        double ws_OptiCLTO = findCorrespondingWS(intersectionPoints_optiCL_Takeoff, TW_OptiCLTO);

        if ((ws_OptiCLTO >= std::min(WS_temp, ws_LTO) && ws_OptiCLTO <= std::max(WS_temp, ws_LTO)) &&
            (TW_OptiCLTO >= std::min(TW_temp, TW_LTO) && TW_OptiCLTO <= std::max(TW_temp, TW_LTO))) {
            optiCLInRange = true;
        }

        if (optiCLInRange) {
            designWS_ = ws_OptiCLTO;
            designTW_ = TW_OptiCLTO;
            designPoint = "Optimal CL";
        } else {
            // Perform linear interpolation between LTO and the higher of CT/TC
            double slope = (TW_LTO - TW_temp) / (ws_LTO - WS_temp);
            designWS_ = (ws_LTO + WS_temp) / 2; // Midpoint as a starting guess
            double interpolatedTW = TW_temp + slope * (designWS_ - WS_temp);

            // Update interpolated TW and WS based on interpolation
            designTW_ = interpolatedTW;

            // Determine the appropriate design point label
            designPoint = (TW_temp == TW_CT) ? "Takeoff and Climb" : "Takeoff and Cruise";

            // Print out of range message
            std::cout << "Optimal CL set in config file is out of range. Interpolation method was used to calculate the design point\n";
        }
    }

    // Print the final design point WS and TW
    std::cout << "W/S: " << designWS_ << ", T/W: " << designTW_ << '\n';
    std::cout << "Design Point Intersection: " << designPoint << '\n';

}
