# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

# Import own modules.
import math
import pyaixml as aixml
import pyaircraftgeometry2 as geom2
import pycoordinatesystemconversion as py11csc


def prepare_external_data(paths_and_names, routing_dict, dict_ac_exchange, dict_mod_config, runtime_output):
    """  Function prepare external data for project specific calculations.

        :param dict paths_and_names: Dictionary containing system paths and ElementTrees
        :param dict routing_dict: Dictionary containing routing parameter
        :param dict dict_ac_exchange: Dict containing parameter and according values from aircraft exchange file
        :param dict dict_mod_config: Dict containing parameter and according values from module configuration file
        :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
        :return dict dict_ac_exchange: Dict containing parameter and according values from aircraft exchange file
    """
    
    ''' set most forward and most aftward center of gravity position for maximum take-off mass condition '''
    # TODO: this have to remove if the correct cog positions for mtom are writen!!! (line 20 to 45)
    dict_ac_exchange['dihedral_angle_wing'] = 4.5  # angle in degree
    if dict_ac_exchange['center_of_gravity_in_x'] is not None \
        and dict_ac_exchange['wing_reference_point_global_in_x'] is not None:
        # Create the factory
        airfoil_data_dir = f'{paths_and_names["project_directory"]}/geometry_data/airfoil_data'
        aircraft_exchange_file = aixml.openDocument(paths_and_names['path_to_aircraft_exchange_file'])
        factory = geom2.factory.WingFactory(aircraft_exchange_file, airfoil_data_dir)
        # Create the surface
        wing = factory.create("wing/specific/geometry/aerodynamic_surface@0")
        mean_aerodynamic_chord_length = geom2.measure.mean_aerodynamic_chord(wing)
        dict_ac_exchange['most_aft_center_of_gravity_in_x_direction'] = round(dict_ac_exchange['center_of_gravity_in_x'] 
                                                                         + mean_aerodynamic_chord_length * 0.1, 9)
        dict_ac_exchange['most_forward_center_of_gravity_in_x_direction'] = round(dict_ac_exchange['center_of_gravity_in_x'] 
                                                                             - mean_aerodynamic_chord_length * 0.1, 9)
        dict_ac_exchange['dihedral_angle_wing'] = round(math.degrees(geom2.measure.dihedral(
            wing, dict_ac_exchange['wing_section_position_in_y']['wing_section_position_in_y_ID0_ID2'])), 9)
    elif dict_ac_exchange['center_of_gravity_in_x'] is not None:
        dict_ac_exchange['most_aft_center_of_gravity_in_x_direction'] = (dict_ac_exchange['center_of_gravity_in_x'] + 0.5)
        dict_ac_exchange['most_forward_center_of_gravity_in_x_direction'] = (dict_ac_exchange['center_of_gravity_in_x'] - 0.5)
    
    if dict_ac_exchange['most_aft_center_of_gravity_in_z_direction'] is not None:
        dict_ac_exchange['most_aft_center_of_gravity_in_z_direction'] = dict_ac_exchange['center_of_gravity_in_z'] - 0.1
        
    if dict_ac_exchange['most_forward_center_of_gravity_in_z_direction'] is not None:
        dict_ac_exchange['most_forward_center_of_gravity_in_z_direction'] = dict_ac_exchange['center_of_gravity_in_z'] - 0.05
        
    ''' transform local position of input parameter into global positions '''
    # transform wing reference points
    if dict_ac_exchange['wing_reference_point_global_in_x'] is None \
            or dict_ac_exchange['aerodynamic_surface_position_in_x'] is None:
        dict_ac_exchange['wing_section_position_in_x'] = {'wing_section_position_in_x_ID0': None}
    else:
        dict_ac_exchange['wing_section_position_in_x'] = \
            {key: value + dict_ac_exchange['wing_reference_point_global_in_x']
                + dict_ac_exchange['aerodynamic_surface_position_in_x']
             for key, value in dict_ac_exchange['wing_section_position_in_x'].items()}

    if dict_ac_exchange['wing_reference_point_global_in_y'] is None \
            or dict_ac_exchange['aerodynamic_surface_position_in_y'] is None:
        dict_ac_exchange['wing_section_position_in_y'] = {'wing_section_position_in_y_ID0': None}
    else:
        dict_ac_exchange['wing_section_position_in_y'] = \
            {key: -1 * (value + dict_ac_exchange['wing_reference_point_global_in_y']
                        + dict_ac_exchange['aerodynamic_surface_position_in_y'])
             for key, value in dict_ac_exchange['wing_section_position_in_y'].items()}

    if dict_ac_exchange['wing_reference_point_global_in_z'] is None \
            or dict_ac_exchange['aerodynamic_surface_position_in_z'] is None:
        dict_ac_exchange['wing_section_position_in_z'] = {'wing_section_position_in_z_ID0': None}
    else:
        dict_ac_exchange['wing_section_position_in_z'] = \
            {key: value + dict_ac_exchange['wing_reference_point_global_in_z']
                + dict_ac_exchange['aerodynamic_surface_position_in_z']
             for key, value in dict_ac_exchange['wing_section_position_in_z'].items()}

    # estimate origin dependent values
    _, wing_chord_length_on_center_line_value = \
        next(iter(dict_ac_exchange['wing_chord_length_on_section'].items()), (None, None))
    _, wing_reference_point_on_center_line_in_x_value = \
        next(iter(dict_ac_exchange['wing_section_position_in_x'].items()), (None, None))
    if wing_chord_length_on_center_line_value is None or wing_reference_point_on_center_line_in_x_value is None:
        dict_ac_exchange['rear_spar_position_center_wing_box_in_x_direction'] = 0
    else:
        dict_ac_exchange['rear_spar_position_center_wing_box_in_x_direction'] = \
            (wing_reference_point_on_center_line_in_x_value + 0.65 * wing_chord_length_on_center_line_value)

    _, wing_reference_point_on_center_line_in_z_value = \
        next(iter(dict_ac_exchange['wing_section_position_in_z'].items()), (None, None))
    if wing_reference_point_on_center_line_in_z_value is None:
        dict_ac_exchange['rear_spar_position_center_wing_box_in_z_direction'] = -0.8
    else:
        dict_ac_exchange['rear_spar_position_center_wing_box_in_z_direction'] = \
            wing_reference_point_on_center_line_in_z_value

    ''' estimate total fuselage section dictionaries '''
    # Convert values of both dictionaries to lists
    section_upper_height_list = list(dict_ac_exchange['fuselage_section_upper_height'].values())
    section_lower_height_list = list(dict_ac_exchange['fuselage_section_lower_height'].values())

    # Add corresponding elements of the lists
    section_total_height_list = [x + y if x is not None and y is not None else None for x, y in
                                 zip(section_upper_height_list, section_lower_height_list)]

    # Create a new dictionary with keys from the first dictionary
    result_dict = {key: value for key, value in zip(dict_ac_exchange['fuselage_section_upper_height'].keys(),
                                                    section_total_height_list)}

    # total fuselage section heights
    dict_ac_exchange['fuselage_section_total_height'] = \
        {key.replace('upper', 'total'): value for key, value in result_dict.items()}

    ''' debug print '''
    runtime_output.debug('Debug: The prepare_external_data function was successfully executed!')
    
    # dict_ac_exchange = coordinate_system_conversion(dict_ac_exchange)

    return dict_ac_exchange


def coordinate_system_conversion(dict_ac_exchange):
    for key in dict_ac_exchange.keys():
        if "x" in key.split("_"):
            if key.split("_")[-1] != "x":
                if dict_ac_exchange[key.split("_x")[0] + "_x" + key.split("_x")[-1]] is not None:
                    if isinstance(dict_ac_exchange[key.split("_x")[0] + "_x" + key.split("_x")[-1]], dict):
                        dict_ac_exchange[key.split("_x")[0] + "_x" + key.split("_x")[-1]], dict_ac_exchange[
                            key.split("_x")[0] + "_y" + key.split("_x")[-1]], dict_ac_exchange[
                            key.split("_x")[0] + "_z" + key.split("_x")[-1]] = coordinate_system_conversion2(
                            dict_ac_exchange[key.split("_x")[0] + "_x" + key.split("_x")[-1]],
                            dict_ac_exchange[key.split("_x")[0] + "_y" + key.split("_x")[-1]],
                            dict_ac_exchange[key.split("_x")[0] + "_z" + key.split("_x")[-1]])
                    else:
                        element_3d = \
                            py11csc.Element3D(dict_ac_exchange[key.split("_x")[0] + "_x" + key.split("_x")[-1]],
                                              dict_ac_exchange[key.split("_x")[0] + "_y" + key.split("_x")[-1]],
                                              dict_ac_exchange[key.split("_x")[0] + "_z" + key.split("_x")[-1]])
                        element_3d.CGAL2AC()
                        dict_ac_exchange[key.split("_x")[0] + "_x" + key.split("_x")[-1]] = element_3d.x()
                        dict_ac_exchange[key.split("_x")[0] + "_y" + key.split("_x")[-1]] = element_3d.y()
                        dict_ac_exchange[key.split("_x")[0] + "_z" + key.split("_x")[-1]] = element_3d.z()
            else:
                if dict_ac_exchange[key.split("_x")[0] + "_x"] is not None:
                    if isinstance(dict_ac_exchange[key.split("_x")[0] + "_x"], dict):
                        dict_ac_exchange[key.split("_x")[0] + "_x"], dict_ac_exchange[key.split("_x")[0] + "_y"], \
                            dict_ac_exchange[key.split("_x")[0] + "_z"] = coordinate_system_conversion2(
                                dict_ac_exchange[key.split("_x")[0] + "_x"],
                                dict_ac_exchange[key.split("_x")[0] + "_y"],
                                dict_ac_exchange[key.split("_x")[0] + "_z"])
                    else:
                        element_3d = py11csc.Element3D(dict_ac_exchange[key.split("_x")[0] + "_x"],
                                                       dict_ac_exchange[key.split("_x")[0] + "_y"],
                                                       dict_ac_exchange[key.split("_x")[0] + "_z"])
                        element_3d.CGAL2AC()
                        dict_ac_exchange[key.split("_x")[0] + "_x"] = element_3d.x()
                        dict_ac_exchange[key.split("_x")[0] + "_y"] = element_3d.y()
                        dict_ac_exchange[key.split("_x")[0] + "_z"] = element_3d.z()

    return dict_ac_exchange


def coordinate_system_conversion2(dict_ac_exchange, dict_ac_exchangey, dict_ac_exchangez):
    for key in dict_ac_exchange.keys():
        if "x" in key.split("_"):
            if key.split("_")[-1] != "x":
                if dict_ac_exchange[key.split("_x")[0] + "_x" + key.split("_x")[-1]] is not None:
                    if isinstance(dict_ac_exchange[key.split("_x")[0] + "_x" + key.split("_x")[-1]], dict):
                        dict_ac_exchange[key.split("_x")[0] + "_x" + key.split("_x")[-1]], dict_ac_exchange[
                            key.split("_x")[0] + "_y" + key.split("_x")[-1]], dict_ac_exchange[
                            key.split("_x")[0] + "_z" + key.split("_x")[-1]] = coordinate_system_conversion2(
                            dict_ac_exchange[key.split("_x")[0] + "_x" + key.split("_x")[-1]],
                            dict_ac_exchange[key.split("_x")[0] + "_y" + key.split("_x")[-1]],
                            dict_ac_exchange[key.split("_x")[0] + "_z" + key.split("_x")[-1]])
                    else:
                        element_3d = \
                            py11csc.Element3D(dict_ac_exchange[key.split("_x")[0] + "_x" + key.split("_x")[-1]],
                                              dict_ac_exchangey[key.split("_x")[0] + "_y" + key.split("_x")[-1]],
                                              dict_ac_exchangez[key.split("_x")[0] + "_z" + key.split("_x")[-1]])
                        element_3d.CGAL2AC()
                        dict_ac_exchange[key.split("_x")[0] + "_x" + key.split("_x")[-1]] = element_3d.x()
                        dict_ac_exchangey[key.split("_x")[0] + "_y" + key.split("_x")[-1]] = element_3d.y()
                        dict_ac_exchangez[key.split("_x")[0] + "_z" + key.split("_x")[-1]] = element_3d.z()
            else:
                if dict_ac_exchange[key.split("_x")[0] + "_x"] is not None:
                    if isinstance(dict_ac_exchange[key.split("_x")[0] + "_x"], dict):
                        dict_ac_exchange[key.split("_x")[0] + "_x"], dict_ac_exchange[key.split("_x")[0] + "_y"], \
                            dict_ac_exchange[key.split("_x")[0] + "_z"] = coordinate_system_conversion2(
                                dict_ac_exchange[key.split("_x")[0] + "_x"],
                                dict_ac_exchange[key.split("_x")[0] + "_y"],
                                dict_ac_exchange[key.split("_x")[0] + "_z"])
                    else:
                        element_3d = py11csc.Element3D(dict_ac_exchange[key.split("_x")[0] + "_x"],
                                                       dict_ac_exchangey[key.split("_x")[0] + "_y"],
                                                       dict_ac_exchangez[key.split("_x")[0] + "_z"])
                        element_3d.CGAL2AC()
                        dict_ac_exchange[key.split("_x")[0] + "_x"] = element_3d.x()
                        dict_ac_exchangey[key.split("_x")[0] + "_y"] = element_3d.y()
                        dict_ac_exchangez[key.split("_x")[0] + "_z"] = element_3d.z()

    return dict_ac_exchange, dict_ac_exchangey, dict_ac_exchangez
