/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 * 
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

/* List of references
    [1] Airbus customer services, Flight Operations Support & Line Assistance: Getting to grips with the cost index, Issue II, May 1998, p.7
*/
#ifndef MISSION_ANALYSIS_SRC_LIBS_CONDITION_METHODS_OPERATING_CONDITIONS_H_
#define MISSION_ANALYSIS_SRC_LIBS_CONDITION_METHODS_OPERATING_CONDITIONS_H_

#include <string>
#include <vector>

#include "../mission_file_data_handler/mission_file_data_handler.h"

/** \class OperatingConditions
 *  \brief Class to gather all parameters characterizing the operation
 */
class OperatingConditions {
 public:
    uint16_t mission_step;                  /**< Index of current mission step [-] */
    std::string rating;                     /**< Rating */
    double derate;                          /**< Derate [-] */
    double CI;                              /**< Cost index [kg/min], scaled 0 to 999 according to Sperry/Honeywell [1] */
    std::string config;                     /**< Configuration */
    std::vector<double> bleed_times;        /**< Vector of bleed times [s] */
    std::vector<double> bleed_values;       /**< Vector of bleed values [kg/s] */
    std::vector<double> shaft_power_times;  /**< Vector of shaft power times [s] */
    std::vector<double> shaft_power_values; /**< Vector of shaft power values [W] */
    double background_bleed;                /**< Background bleed [kg/s] */
    double background_shaft_power;          /**< Background shaft power [W] */

    std::vector<MissionFileDataHandler::MissionDataSet>* mission_data; /**< reference to the mission data from mission XML */

    /* Constructor and destructor */
    /** \brief Constructor for the class OperatingConditions
     * \param mission_data: reference to the mission data set
     */
    explicit OperatingConditions(std::vector<MissionFileDataHandler::MissionDataSet>* mission_data);
    /** \brief Destructor for the class OperatingConditions
     */
    virtual ~OperatingConditions();

    /* Methods */
    /** \brief Function sets segment operating conditions
     * \param i: index parameter of mission step [-]
     */
    void set_segment_operating_conditions(const uint16_t &i);
     /** \brief Function deletes the time step processed values from the bleed- and shaft_power-Offtakes vectors of operationConditions
     * \param time_step constant reference of the overall time step [s]
     */
    void delete_used_bleed_and_shaft_power_offtakes(const double& time_step);
};

#endif // MISSION_ANALYSIS_SRC_LIBS_CONDITION_METHODS_OPERATING_CONDITIONS_H_
