/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 * 
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */
#ifndef PERFORMANCE_ASSESSMENT_SRC_TAW_DEFAULTMETHODS_PERFORMANCE_ASSESSMENT_SETTINGS_H_
#define PERFORMANCE_ASSESSMENT_SRC_TAW_DEFAULTMETHODS_PERFORMANCE_ASSESSMENT_SETTINGS_H_

#include <moduleBasics/module.h>
#include <vector>
#include <string>
#include <memory>

using std::vector;
using std::string;

class performance_assessment_settings {
 public:
    const std::shared_ptr<RuntimeIO>& rtIO;

    /** \brief Constructor performance_assessment_settings
    */
    explicit performance_assessment_settings(const std::shared_ptr<RuntimeIO>& rtIOin);

    /** \brief Destructor performance_assessment_settings
    */
    ~performance_assessment_settings();

    class switches {
     public:
        bool doFuelPlanning = false; /**< Switch for fuel planning */
        bool doPayloadRange = false; /**< Switch for payload range diagram */
        bool doEnginePerformance = false; /**< Switch for engine performance calculation */
        bool doFlightEnvelopePerformance = false; /**< Switch for flight envelope calculation */
        bool doSARCalculation = false;
        bool overwriteInitialValues = false; /**< Switch if initial flight envelope values are overwritten */
        bool doClimbPerformance = false; /**< Switch for climb performance calculation */
        bool doTOPerformance = false; /**< Switch for estimating the take-off distance */
        bool doBLFLPerformance = false; /**< Switch for estimating the balanced field length */
        bool doLandingPerformance = false; /**< Switch for landing distance estimation */
        bool doVnDiagram = false; /**< "Switch for calculation of V-n diagram */
    };
    switches check_sucessfully_executed; /**< Indicated if a performance check has been executed sucessfully  */
    bool overwrite_initial_values; /** Causes recalculation of dive speeds */

    bool designForMTOM; /**< Switch for a new determination of the MTOM */
    bool useStudyMissionForAnalysis; /**< Switch if missionStudy.xml should be used instead of missionDesign.xml */
    bool optimizeMission; /** Switch if the mission profile will be optimized */

    /** \class enginePerfData Includes the engine performance rating
    *
    */
    class enginePerfData {
     public:
        vector<string> Rating; /**< List of the different engine ratings (Cruise, Climb, MaxCont, TO) */
    };
    enginePerfData enginePerformance;

    /** \class fieldData Includes the characteristics of the runway and head wind
    *
    */
    class fieldData {
     public:
        double runwaySlope; /**< Slope of the runway (rad) */
        double headWind; /**< Head wind speed (m/s) */
    };
    fieldData myField;

    /** \class VnDiagramConstants Includes the load factor for the clean configuration and extended flaps
    *
    */
    class VnDiagramConstants {
     public:
        double maxNClean; /**< Maximum load factor in clean configuration (fixed for CS-25 2.5) */
        double minNClean; /**< Minimum load factor in clean configuration (fixed -1.0 for CS-25) */
        double maxNFlaps; /**< Maximum load factor with flaps (fixed for CS-25 2.0) */
        double minNFlaps; /**< Minimum load factor with flaps (fixed at CS-25 0.0) */
    };

    VnDiagramConstants myVnDiagramConstants;

    /** \class fuelPlan Sets the amount of the contingency fuel
    *
    */
    class fuelPlan {
     public:
        double contingencyFuel; /**< Relative amount of contingency fuel in reference to the total trip fuel (-) */

        fuelPlan() {
            contingencyFuel = 0.0;
        }
    };
    fuelPlan myFuelPlan;

    /** \brief Exists program if a value is NAN, infinite or out of range
    * \param Value Value that is checked
    * \param lowerBound Lower boundary of the value
    * \param upperBound Upper boundary of the value
    * \param Text Text that will be displayed in the error message
    */
    void checkForValidity(double Value, double lowerBound, double upperBound, const string &Text);
};

#endif  // PERFORMANCE_ASSESSMENT_SRC_TAW_DEFAULTMETHODS_PERFORMANCE_ASSESSMENT_SETTINGS_H_
