/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
*/

#include "aircraftSystem.h"

#include "systemsIOData.h"
#include <runtimeInfo/runtimeInfo.h>
#include <unitConversion/constants.h>
#include <memory>
#include <unordered_set>

/* constructor for all systems*/
aircraftSystem::aircraftSystem(const std::string& systemName, bool systemIsOperating, std::shared_ptr<systemsIOData> data,
    const RuntimeIO& rtIO, size_t index, std::string type)
    :
    data_(data),
    rtIO_(rtIO),
    Name(systemName),
    ATA(""),
    Operating(systemIsOperating),
    Mass(0.0),
    systemType(type),
    systemIndex(index) {
    if (Name == "virtualSystem") {
        ATA = "ATA-virtual";
        myRuntimeInfo->out << "  Create: " << Name << "  ->  " << ATA << std::endl;
    } else if (Name == "dummySystemName") {
        ATA = "ATA-Dummy";
    }
    refPoint[0] = 0.0;
    refPoint[1] = 0.0;
    refPoint[2] = 0.0;
    bleedAir.designPower = (0.0);
    bleedAir.maxMissionPower = (0.0);
    heatLoad.designPower = 0.0;
    heatLoad.maxMissionPower = 0.0;
    /* Initializing the vectors for the mission-dependent power requirement **/
    initializePowerProfile(systemType, systemIndex);
    // ctor
}

aircraftSystem::~aircraftSystem() {
    //dtor
}

void aircraftSystem::initializePowerProfile(std::string type, size_t index) {
    // find systemDescrition in configuration
    vector<standardConfig::port> inputDesc;
    vector<standardConfig::port> outputDesc;
    if (type == "sink") {
        inputDesc = data_->configuration.Sinks.at(index).inputs;
    } else if (type == "conductor") {
        inputDesc = data_->configuration.Conductors.at(index).inputs;
        outputDesc = data_->configuration.Conductors.at(index).outputs;
    } else if (type == "source") {
        inputDesc = data_->configuration.Sources.at(index).inputs;
        outputDesc = data_->configuration.Sources.at(index).outputs;
    } else if (type == "interface") {
        outputDesc = data_->configuration.Interfaces.at(index).outputs;
    } else if (type == "ecs") {
        inputDesc = data_->configuration.theECS.inputs;
    } else if (type == "virtual") {
        inputDesc = data_->configuration.virtualSystem.at(index).inputs;
    } else {
        myRuntimeInfo->err << "Invalid system type " << type << endl;
        exit(1);
    }
    // loop over port descriptions and initialize respective power profiles
    for (size_t i=0; i < inputDesc.size(); i++) {
        powerProfile newPowerProfile;
        // set information
        newPowerProfile.type = inputDesc.at(i).type;
        newPowerProfile.ID = inputDesc.at(i).ID;
        // set power values to 0
        newPowerProfile.missionPower.clear();
        newPowerProfile.designPower = (0.0);
        newPowerProfile.maxMissionPower = (0.0);
        for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
            newPowerProfile.missionPower.push_back(missionLoads());
        }
        this->inputs.push_back(newPowerProfile);
    }
    for (size_t i=0; i < outputDesc.size(); i++) {
        powerProfile newPowerProfile;
        // set information
        newPowerProfile.type = outputDesc.at(i).type;
        newPowerProfile.ID = outputDesc.at(i).ID;
        // set power values to 0
        newPowerProfile.missionPower.clear();
        newPowerProfile.designPower = (0.0);
        newPowerProfile.maxMissionPower = (0.0);
        for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
            newPowerProfile.missionPower.push_back(missionLoads());
        }
        this->outputs.push_back(newPowerProfile);
    }
    heatLoad.missionPower.clear();
    heatLoad.designPower = 0.0;
    heatLoad.maxMissionPower = 0.0;
    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        heatLoad.missionPower.push_back(missionLoads());
    }
}

int aircraftSystem::findPort(std::string type, std::string uID, std::string inOrOut) const {
    size_t portIndex = 0;
    bool portFound = false;
    std::vector<powerProfile> ports;
    if (inOrOut == "input") {
        ports = inputs;
    } else {
        ports = outputs;
    }
    for (size_t m=0; m < ports.size(); m++) {
        if (ports.at(m).type == type && ports.at(m).ID == uID) {
            portIndex = m;
            portFound = true;
            break;
        }
    }
    if (!portFound) {
        myRuntimeInfo->err << "Could not find " << inOrOut << " port with ID " << uID << " in system " << this->Name << endl;
        exit(1);
    }
    return portIndex;
}

void aircraftSystem::calculateMaxMissionPowerPort(std::string uID) {
    //bleedAir.maxMissionPower = bleedAir.getMaxPower();
    bool found_in_inputs = false;
    bool found_in_outputs = false;
    int index = 0;
    //search in inputs for uID
    for (size_t i(0); i < inputs.size(); i++) {
        //if ID matches uID, the port is an input port at index = i
        if (inputs.at(i).ID == uID) {
            found_in_inputs = true;
            index = i;
        }
    }
    //search in outputs for uID
    for (size_t i(0); i < outputs.size(); i++) {
        //if ID matches uID, the port is an output port at index = i
        if (outputs.at(i).ID == uID) {
            found_in_outputs = true;
            index = i;
        }
    }
    //port with uID not found
    if (!found_in_inputs && !found_in_outputs) {
        myRuntimeInfo->err << "No port with uID " << uID << " found in " << ATA
                           << ". Error in aircraftSystem::calculateMaxMissionPowerPort."
                           << std::endl;
        exit(1);
    }
    //port uID not unique
    if (found_in_inputs && found_in_outputs) {
        myRuntimeInfo->err << "Port with uID " << uID << " found in inputs and outputs of " << ATA
                           << ". Port uID's must be unique!" << std::endl;
        exit(1);
    }
    //input port
    if (found_in_inputs) {
        inputs.at(index).maxMissionPower = inputs.at(index).getMaxPower();
    }
    //output port
    if (found_in_outputs) {
        outputs.at(index).maxMissionPower = outputs.at(index).getMaxPower();
    }
}

void aircraftSystem::updateMass(std::string node_name, double value, std::string ATA) {
    std::string subPath = "aircraft_exchange_file/component_design/systems/specific/geometry/mass_properties";
    if (!(rtIO_.acxml.find(subPath))) {
        rtIO_.acxml[subPath];
    }
    Endnode<double> mass = data_->data.create_endnode(value, "kg", subPath + "/" + node_name, "mass of " + ATA);
    mass.set_boundaries(0, INFINITY);
    mass.update(rtIO_.acxml);
}

void aircraftSystem::checkPorts() {
    checkPortNumber();
    checkPortTypes();
}

void aircraftSystem::checkPortTypes() {
    //check input ports
    // Convert expected types to sets for faster lookup
    std::unordered_set<std::string> expected_input_set(expected_input_types.begin(), expected_input_types.end());
    std::unordered_set<std::string> expected_output_set(expected_output_types.begin(), expected_output_types.end());

    // Check input ports
    for (const auto& input : inputs) {
        if (expected_input_set.find(input.type) == expected_input_set.end()) {
            myRuntimeInfo->err << "Input port of type " << input.type << " defined but not expected for "
                            << ATA << std::endl;
            exit(1); // Consider throwing an exception or using another error handling strategy
        }
    }

    // Check output ports
    for (const auto& output : outputs) {
        if (expected_output_set.find(output.type) == expected_output_set.end()) {
            myRuntimeInfo->err << "Output port of type " << output.type << " defined but not expected for "
                            << ATA << std::endl;
            exit(1); // Consider throwing an exception or using another error handling strategy
        }
    }
}

void aircraftSystem::checkPortNumber() {
    // find systemDescrition in configuration
    vector<standardConfig::port> inputDesc;
    vector<standardConfig::port> outputDesc;
    if (systemType == "sink") {
        inputDesc = data_->configuration.Sinks.at(systemIndex).inputs;
    } else if (systemType == "conductor") {
        inputDesc = data_->configuration.Conductors.at(systemIndex).inputs;
        outputDesc = data_->configuration.Conductors.at(systemIndex).outputs;
    } else if (systemType == "source") {
        inputDesc = data_->configuration.Sources.at(systemIndex).inputs;
        outputDesc = data_->configuration.Sources.at(systemIndex).outputs;
    } else if (systemType == "interface") {
        outputDesc = data_->configuration.Interfaces.at(systemIndex).outputs;
    } else if (systemType == "ecs") {
        inputDesc = data_->configuration.theECS.inputs;
    } else {
        myRuntimeInfo->err << "Invalid system type " << systemType << endl;
        exit(1);
    }
    // check that number of ports are below the limits of the system
    if (inputDesc.size() > allowed_inputs) {
        myRuntimeInfo->err << "The number of defined input ports (" << inputDesc.size() << ") exceeds the number of input ports "
                           << ATA << " can handle (" << allowed_inputs << "). Abort." << std::endl;
        exit(1);
    }
    if (outputDesc.size() > allowed_outputs) {
        myRuntimeInfo->err << "The number of defined output ports (" << outputDesc.size() << ") exceeds the number of output ports "
                           << ATA << " can handle (" << allowed_outputs << "). Abort." << std::endl;
        exit(1);
    }
}
