/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
*/

/* The following statements import necessary header files */
#include "ATAXX.h"                                // Import of header file from src/aircraftSystems folder for ATAXX
#include <runtimeInfo/runtimeInfo.h>                          // Import of header file from external runtime library for using runtimeInfo

/* Constructor for ATAXX */
ATAXX::ATAXX(const std::string& systemName, bool systemIsOperating, std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO, size_t index)
    :
    aircraftSystem(systemName, systemIsOperating, data, rtIO, index, "sink"),
    refPtX_cockpit(0.),
    refPtX_avionicsBay(0.) {
    ATA = "ATA-XX";
    /* Define how many input and output ports the code can handle: */
    allowed_inputs = 3;
    allowed_outputs = 0;
    /* Define port types */
    expected_input_types = {"Electric", "Hydraulic"};
    expected_output_types = {""};
    myRuntimeInfo->out << "  Create: " << Name << "  ->  " << ATA << std::endl;
    //ctor
}

/* Destructor for ATAXX */
ATAXX::~ATAXX() {
    //dtor
}

/* Reading system specific parameters of configuration xml file */
void ATAXX::readParameters(const node& config) {
    std::string remainingConsumerConfig = "/module_configuration_file/program_settings/systems_constants/remaining_consumers";
    avBayLocation = config.at("/module_configuration_file/program_settings/systems_constants/avionics/location_of_aviationics_bay/value");
    data_->configuration.checkForValidity(avBayLocation, 0, 2, "avBayLocation");
    percentageBleedAir   = config.at(remainingConsumerConfig + "/percentage_bleed_air/value");
    percentageElectric   = config.at(remainingConsumerConfig + "/percentage_electric/value");
    percentageHydraulic  = config.at(remainingConsumerConfig + "/percentage_hydraulic/value");
    scalingFactorATAXX =  checkBoundaries(config.at(remainingConsumerConfig + "/mass/scaling_factor/value"), 0, true, 1, true);
    percInstrOfATAXX = checkBoundaries(config.at(remainingConsumerConfig + "/mass/distribution/percentage_instrumentation_of_ATAXX/value"), 0, true, 1, true);
    percAutoFlightOfATAXX = checkBoundaries(config.at(remainingConsumerConfig + "/mass/distribution/percentage_auto_flight_of_ATAXX/value"), 0, true, 1, true);
    percNavigOfATAXX = checkBoundaries(config.at(remainingConsumerConfig + "/mass/distribution/percentage_navigation_of_ATAXX/value"), 0, true, 1, true);
    percCommunicOfATAXX = checkBoundaries(config.at(remainingConsumerConfig + "/mass/distribution/percentage_communication_of_ATAXX/value"), 0, true, 1, true);
    if ((percInstrOfATAXX + percAutoFlightOfATAXX + percNavigOfATAXX + percCommunicOfATAXX) != 1.) {
        myRuntimeInfo->err << "Sum of percentages for remaining consumers is not 1. Use scaling factor of remaining consumers instead. Abort program." << std::endl;
        exit(1);
    }
    scaling_factor = EndnodeReadOnly<double>("/module_configuration_file/program_settings/scaling_factors/remaining_consumers").read(config).value();
}

/* Updating system specific parameters in aircraft xml file */
void ATAXX::updateXML() {
    updateMass("instruments", percInstrOfATAXX * Mass, "ATA31");
    //AutomaticFlightSystem
    updateMass("automatic_flight_system", percAutoFlightOfATAXX * Mass, "ATA22");
    //Navigation
    updateMass("navigation", percNavigOfATAXX * Mass, "ATA34");
    //Communication
    updateMass("communication", percCommunicOfATAXX * Mass, "ATA23");
}

/* Calculation of the system power */
void ATAXX::calculateSystemsPower() {
    getPower();
    //calculate max mission power for input ports
    for (size_t i(0); i < inputs.size(); i++) {
        calculateMaxMissionPowerPort(inputs.at(i).ID);
    }
}

/* Calculation of the skin mass */
void ATAXX::calculateSinkMass() {
    /* Calculate Systems Mass */
    Mass = scaling_factor * getMass();
    refPoint = getRefPoint();
}

/* Determination of the power */
void ATAXX::getPower() {
    /* Calculate Systems Power */
    /* Remaining Consumers including AVIONICS */
    /* New estimation for electric from F. Peter 18.6.14 after [Bue14] p. 80 */
    //(design power and base load required for hydraulic and electric)
    powerProfile previousSystemsHydraulic;
    powerProfile previousSystemsElectric;
    //initialize power profiles
    previousSystemsElectric.designPower = 0.;
    previousSystemsHydraulic.designPower = 0.;
    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        missionLoads tmp_mission;
        tmp_mission.baseLoad = 0.;
        previousSystemsElectric.missionPower.push_back(tmp_mission);
        previousSystemsHydraulic.missionPower.push_back(tmp_mission);
    }
    /*Get power from the previous sink systems*/
    for (size_t i=0; i < data_->Systems.energySink.size(); i++) {
        std::vector<std::string> input_types = data_->Systems.energySink.at(i)->expected_input_types;
        //check if system has electric input port
        auto electric_exists = std::find(input_types.begin(), input_types.end(), "Electric");
        if (electric_exists != input_types.end()) {
            //there is electic input
            //add electric power
            previousSystemsElectric.designPower += data_->Systems.energySink.at(i)->inputs.at(
                data_->Systems.energySink.at(i)->findPort("Electric", "ElectricIn", "input")).designPower;
            for (size_t j=0; j < data_->mission.Steps.size(); j++) {
                previousSystemsElectric.missionPower.at(j).baseLoad +=
                    data_->Systems.energySink.at(i)->inputs.at(
                        data_->Systems.energySink.at(i)->findPort("Electric", "ElectricIn", "input")).missionPower.at(j).baseLoad;
            }
        }
        //check if system has hydraulic input port
        auto hydraulic_exists = std::find(input_types.begin(), input_types.end(), "Hydraulic");
        if (hydraulic_exists != input_types.end()) {
            //there is hydraulic input
            //add hydraulic power
            previousSystemsHydraulic.designPower += data_->Systems.energySink.at(i)->inputs.at(
                data_->Systems.energySink.at(i)->findPort("Hydraulic", "HydraulicIn", "input")).designPower;
            for (size_t j=0; j < data_->mission.Steps.size(); j++) {
                previousSystemsHydraulic.missionPower.at(j).baseLoad +=
                    data_->Systems.energySink.at(i)->inputs.at(
                        data_->Systems.energySink.at(i)->findPort("Hydraulic", "HydraulicIn", "input")).missionPower.at(j).baseLoad;
            }
        }
    }
    /*Loop over electric circuits*/
    inputs.at(findPort("Electric", "ElectricIn", "input")).designPower = percentageElectric * previousSystemsElectric.designPower;
    for (unsigned i = 0; i < data_->mission.Steps.size(); i++) {
        inputs.at(findPort("Electric", "ElectricIn", "input")).missionPower.at(i).baseLoad =
            percentageElectric * previousSystemsElectric.missionPower.at(i).baseLoad;
    }
    inputs.at(findPort("Hydraulic", "HydraulicIn", "input")).designPower = percentageHydraulic * previousSystemsHydraulic.designPower;
    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
            inputs.at(findPort("Hydraulic", "HydraulicIn", "input")).missionPower.at(i).baseLoad =
                percentageHydraulic * previousSystemsHydraulic.missionPower.at(i).baseLoad;
    }
    /*For heatLoad electrical powerProfile is accumulated*/
    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        heatLoad.missionPower.at(i).baseLoad = inputs.at(findPort("Electric", "ElectricIn", "input")).missionPower.at(i).baseLoad;
    }
}

/* Determination of the mass */
double ATAXX::getMass() {
    /* From [Tor13], S. 289 (pdf S. 308) -> method requires the maximum range, but at this point only the DesignRange in MICADO is known */
    double designRange = data_->data.Performance.designRange / 1000; // conversion from m to km
    return  static_cast<double>(scalingFactorATAXX * 0.347 * pow(data_->data.Mass.OME, 0.556) * pow(designRange, 0.25));
}

/* Determination of the reference point */
Vec3 ATAXX::getRefPoint() {
    /* Instruments: Placement at the end of the cockpit segment */
    this->refPtX_cockpit = 0.0;
    this->refPtX_cockpit = data_->data.Accommodation.payload_tube_x_ref;
    /* Auto Flight System, Navigation & Communication: Placement in Avionics Bay */
    if (avBayLocation == 0) {
        this->refPtX_avionicsBay = data_->theWing.at(0).aerodynamicSurface.sections.front().origin[0] -
                                   0.25 * data_->theWing.at(0).aerodynamicSurface.sections.front().get_chord_length(); // Initial method
    } else if (avBayLocation == 1) {
        this->refPtX_avionicsBay = this->refPtX_cockpit * 0.25; // Location of the avionics bay in the nose for business-jet-like AC; approximation: 25% of the cockpit length
    } else if (avBayLocation == 2) {
        this->refPtX_avionicsBay = this->refPtX_cockpit;    // Location of the avionics bay behind the cockpit segment for A320-like AC
    } else {
        myRuntimeInfo->err << "Error: invalid location of the avionics bay. Program exit!" << std::endl;
        exit(1);
    }
    /* Determination of the total CG point in x-direction */
    double refX_Pt_tmp = percInstrOfATAXX * this->refPtX_cockpit + (percAutoFlightOfATAXX + percNavigOfATAXX + percCommunicOfATAXX)* this->refPtX_avionicsBay;
    return Vec3(data_->theFuselage.back().origin[0] + refX_Pt_tmp, data_->theFuselage.back().origin[1], data_->theFuselage.back().origin[2]);
}
