/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
*/

#ifndef SYSTEMS_DESIGN_SRC_AIRCRAFTSYSTEMS_ATAXX_H_
#define SYSTEMS_DESIGN_SRC_AIRCRAFTSYSTEMS_ATAXX_H_

/* The following statements import necessary libraries */
#include <string>                                                            // Import c++ standard library for using string
#include <vector>                                                            // Import c++ standard library for using vector

/* The following statements import necessary header files */
#include "../aircraftSystem.h"                                // Import of header file from src folder for aircraft system
#include "../powerProfile.h"                                  // Import of header file from src folder for powerProfile
#include "../systemsIOData.h"                                 // Import of header file from systemsIOData for accessing data_. Can't be included in
                                                              // aircraftSystem.h because that would lead to circular dependencies
#include <svl/SVL.h>                                          // Import of header file from src folder for SVL
#include <memory>


/** \class ATAXX
 * \brief class that contains all properties of the ATAXX system (avionics / remaining consumers)
 */
class ATAXX : public aircraftSystem {
 public:
    int numberOfPowerSources; /**< numver of power sources*/
    int avBayLocation; /**< Location of the avionics bay*/
    std::vector<PowerSource> powerSources; /**< Vector containing the power sources of ATAXX */

    /*Additional percentages of bleed, electric, and hydraulic power for ATA-XX*/
    double percentageBleedAir; /** < [-] percentage of unrecorded bleed air for ATA-XX */
    double percentageElectric; /** < [-] percentage of unrecorded electric power for ATA-XX */
    double percentageHydraulic; /** < [-] percentage of unrecorded hydraulic power for ATA-XX */

    double scalingFactorATAXX;/**< [-] global scaling factor for ATAXX */
    double percInstrOfATAXX;/**< [-] percentage instrumentation of ATAXX */
    double percAutoFlightOfATAXX;/**< [-] percentage auto flight of ATAXX */
    double percNavigOfATAXX;/**< [-] percentage navigation of ATAXX */
    double percCommunicOfATAXX;/**< [-] percentage communication of ATAXX */

    double refPtX_cockpit; /**< [m] global reference point of the cockpit in x direction */
    double refPtX_avionicsBay; /**< [m] global reference point of the avionics bay in x direction */

    /*Inherited functions from aircraftSystem*/
    /**
     * \brief function for reading the system specific parameters from the config
     * 
     * \param config const node&: systemsDesign configuration file
    */
    void readParameters(const node& config) override;
    /**
     * \brief function for updating the system specific values in the aircraft xml
    */
    void updateXML() override;
    void calculateSystemsPower() override;
    void calculateSinkMass() override;

    void getPower();

    double getMass();
    Vec3 getRefPoint();

    ATAXX(const std::string& systemName, bool systemIsOperating, std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO, size_t index);
    virtual ~ATAXX();
};

#endif // SYSTEMS_DESIGN_SRC_AIRCRAFTSYSTEMS_ATAXX_H_
