/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
*/

/* The following statements import necessary header files */
#include "conventionalATA24.h"                            // Import of header file from src/aircraftSystems folder for conventionalATA24
#include <runtimeInfo/runtimeInfo.h>                      // Import of header file from external runtime library for using runtimeInfo
#include <unitConversion/constants.h>                     // Import of header file from external unitConversion library for constants
#include <standardFiles/functions.h>                      // Import of header file from external standardFiles library for functions
#include "../geometricalFunctions.h"                      // Import of header file from src folder for using geometricalFunctions

/* Constructor for conventionalATA24 */
conventionalATA24::conventionalATA24(const std::string& systemName, bool systemIsOperating, std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO, size_t index)
    :
    aircraftSystem(systemName, systemIsOperating, data, rtIO, index, "conductor"),
    cableMass(0.0),
    otherMass(0.0),
    lengthCable(0.0) {
    ATA = "ATA-24";
    /* Define how many input and output ports the code can handle: */
    allowed_inputs = 3;//shaft power, electric, hydraulic
    allowed_outputs = 3;//arbitrary, is checked again once number of electric circuits is known
    /* Define port types */
    expected_input_types = {"Electric", "ShaftPower"};
    expected_output_types = {"Electric"};
    /* Check ports */
    checkPorts();
    /* Runtime Info */
    myRuntimeInfo->out << "  Create: " << Name << "  ->  " << ATA << std::endl;
}

/* Destructor for conventionalATA24 */
conventionalATA24::~conventionalATA24() {
    //dtor
}
/** Reading system specific parameters of configuration xml file
 * electric circuits are written to standardConfig so that other systems can access these values as well
*/
void conventionalATA24::readParameters(const node& config) {
    scaling_factor = EndnodeReadOnly<double>("/module_configuration_file/program_settings/scaling_factors/electric_system").read(config).value();
    std::string electricSystemConfig = "/module_configuration_file/program_settings/systems_constants/electric_system";
    efficiencyFactorElectricSystem = config.at(electricSystemConfig + "/efficiency_factor_electric_system/value");
    data_->configuration.checkForValidity(efficiencyFactorElectricSystem, 0.0, 1.0, "efficiencyFactorElectricSystem");
    maxRelPower = config.at(electricSystemConfig + "/rel_max_power/value");
    data_->configuration.checkForValidity(maxRelPower, 1.0, 10.0, "maxRelPower");
    specificCableMass = config.at(electricSystemConfig + "/specific_cable_mass/value");
    data_->configuration.checkForValidity(specificCableMass, 0.0, 20.0, "specificCableMass");
    numberOfElectricCircuits = config.at(electricSystemConfig + "/electric_circuits/number_of_circuits/value");
    //check again now that number of circuits are defined that number of output ports matches
    allowed_outputs = numberOfElectricCircuits;
    checkPortNumber();
    data_->configuration.Electric.numberOfElectricCircuits = numberOfElectricCircuits;
    for (int i(0); i < numberOfElectricCircuits; i++) {
        data_->configuration.Electric.circuits.push_back(standardConfig::electricSystem::electricCircuit());
        data_->configuration.Electric.circuits.back().numberOfGenerators =
               config.at(electricSystemConfig + "/electric_circuits/electric_circuit@" + num2Str(i) + "/components/generators/number_of_generators/value");
        for (int j(0); j < data_->configuration.Electric.circuits.back().numberOfGenerators; j++) {
            data_->configuration.Electric.circuits.back().generators.push_back(standardConfig::electricSystem::electricCircuit::generator());
            data_->configuration.Electric.circuits.back().generators.back().name =
                std::string(config.at(electricSystemConfig + "/electric_circuits/electric_circuit@" + num2Str(
                                             i) + "/components/generators/generator@" + num2Str(j) + "/name/value"));
            data_->configuration.Electric.circuits.back().generators.back().type =
                std::string(config.at(electricSystemConfig + "/electric_circuits/electric_circuit@" + num2Str(
                                             i) + "/components/generators/generator@" + num2Str(j) + "/type/value"));
            data_->configuration.Electric.circuits.back().generators.back().powerSource.operationFactor = 1.;
            data_->configuration.Electric.circuits.back().generators.back().efficiency =
                config.at(electricSystemConfig + "/electric_circuits/electric_circuit@" + num2Str(
                                 i) + "/components/generators/generator@" + num2Str(j) + "/generator_efficiency/value");
            data_->configuration.Electric.circuits.back().generators.back().operationFactor =
                config.at(electricSystemConfig + "/electric_circuits/electric_circuit@" + num2Str(
                                 i) + "/components/generators/generator@" + num2Str(j) + "/operation_factor/value");
            data_->configuration.checkForValidity(data_->configuration.Electric.circuits.back().generators.back().efficiency,
                                                  0.0, 1., "circuits.back().generators.back().efficiency");
            data_->configuration.checkForValidity(data_->configuration.Electric.circuits.back().generators.back().operationFactor,
                                                  0.0, 1., "circuits.back().generators.back().operationFactor");
        }
        percentageElectricPower = config.at("/module_configuration_file/program_settings/systems_constants/propulsion_system/percentage_electrical_power/value");
    }
    /*aigner: Check the config settings*/
    for (size_t i(0); i < data_->configuration.Electric.circuits.size(); i++) {
        double genOperationFactor(0.);
        for (size_t j(0); j < data_->configuration.Electric.circuits.at(i).generators.size(); j++) {
            genOperationFactor += Rounding(data_->configuration.Electric.circuits.at(i).generators.at(j).operationFactor, 10);
        }
        if (genOperationFactor < 1.) {
            myRuntimeInfo->err << "The work distribution of the generators in the electrical circuit " << i
                               << " results together in less than 100% (" << genOperationFactor * 100 <<
                               "%). Adjust in config. Abort!" << std::endl;
            exit(1);
        } else if (genOperationFactor > 1.) {
            myRuntimeInfo->err << "The work distribution of the generators in the electrical circuit " << i
                               << " results together in more than 100% (" << genOperationFactor * 100 <<
                               "%). Adjust in config. Abort!" << std::endl;
            exit(1);
        }
    }
}

/* Updating system specific parameters in aircraft xml file */
void conventionalATA24::updateXML() {
    //multiplied by scaling factor since that is only applied to total mass and not sub masses
    /* electrical generation */
    updateMass("electrical_generation", scaling_factor * otherMass, ATA);
    /* electrical distribution */
    updateMass("electrical_distribution", scaling_factor * cableMass, ATA);
}

/* Calculation of conducting power */
void conventionalATA24::calculateConductingPower() {
    getPower();
    //calculate max mission power for input ports
    for (size_t i(0); i < inputs.size(); i++) {
        calculateMaxMissionPowerPort(inputs.at(i).ID);
    }
}

/* Calculation of conductor mass */
void conventionalATA24::calculateConductorMass() {
    /* Calculate Conductor Mass */
    Mass = scaling_factor * getMass();
    refPoint = getRefPoint();
}

/* Determination of the power */
void conventionalATA24::getPower() {
    /* This function clalculates the losses of the electric system which will later be added to the conductor consumption */
    double generatorEfficiency(1.);
    system_design_power = 0.;//reset after each iteration
    // calculate electric circuit design power - each output port represents one circuit
    for (size_t n(0); n < data_->configuration.Electric.circuits.size(); n++) {
        // get corresponding output port - take same index n
        for (size_t k(0); k < data_->configuration.Electric.circuits.at(n).generators.size(); k++) {
            double generatorOperationFactor = data_->configuration.Electric.circuits.at(n).generators.at(k).operationFactor;
            double efficiencyFactor_total = generatorEfficiency * efficiencyFactorElectricSystem;
            double power_tmp = outputs.at(n).designPower * generatorOperationFactor * (1. / efficiencyFactor_total);
            system_design_power += power_tmp;
            //write to input
            for (size_t i(0); i < inputs.size(); i++) {
                inputs.at(i).designPower = power_tmp;
            }
        }
    }
    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        /* Electric baseLoad */
        for (size_t n(0); n < data_->configuration.Electric.circuits.size(); n++) {
            for (size_t k(0); k < data_->configuration.Electric.circuits.at(n).generators.size(); k++) {
                double generatorOperationFactor = data_->configuration.Electric.circuits.at(n).generators.at(k).operationFactor;
                double efficiencyFactor_total = generatorEfficiency * efficiencyFactorElectricSystem;
                // total power including conductor power demand
                double power_tmp = outputs.at(n).missionPower.at(i).baseLoad * generatorOperationFactor * (1. / efficiencyFactor_total);
                double power_tmp_no_gen = outputs.at(n).missionPower.at(i).baseLoad * (1. / efficiencyFactorElectricSystem);//for electric input
                //because electric input from fuel cell does not have a generator: write power_tmp to all inputs, distribution handels percentages
                for (size_t m=0; m < inputs.size(); m++) {
                    if (inputs.at(m).type == "Electric") {
                        power_tmp = power_tmp_no_gen; //no generator efficiency for electric input
                    }
                    inputs.at(m).missionPower.at(i).baseLoad += power_tmp;
                    // get from outputs
                for (size_t j = 0; j < outputs.at(n).missionPower.at(i).peakLoad.size(); j++) {
                    /* Create a new peak, only if not already existing */
                    if (inputs.at(m).missionPower.at(i).peakLoad.size() < j + 1) {
                        inputs.at(m).missionPower.at(i).peakLoad.push_back(0.);
                        inputs.at(m).missionPower.at(i).timePeak.push_back(outputs.at(n).missionPower.at(i).timePeak.at(j));
                    } else {
                        if (!accuracyCheck(inputs.at(m).missionPower.at(i).timePeak.at(j), outputs.at(n).missionPower.at(i).timePeak.at(j), ACCURACY_HIGH)) {
                            inputs.at(m).missionPower.at(i).peakLoad.push_back(0.);
                            inputs.at(m).missionPower.at(i).timePeak.push_back(outputs.at(n).missionPower.at(i).timePeak.at(j));
                        }
                    }
                    for (size_t k(0); k < data_->configuration.Electric.circuits.at(n).generators.size(); k++) {
                        double generatorOperationFactor = data_->configuration.Electric.circuits.at(n).generators.at(k).operationFactor;
                        double efficiencyFactor_total = generatorEfficiency * efficiencyFactorElectricSystem;
                        double power_tmp = 0.;
                        if (inputs.at(m).type == "Electric") {
                            //only efficiency of electric system relevant for electric input
                            power_tmp = outputs.at(n).missionPower.at(i).peakLoad.at(j) / efficiencyFactorElectricSystem;
                        } else {
                            power_tmp = outputs.at(n).missionPower.at(i).peakLoad.at(j) * generatorOperationFactor * (1. / efficiencyFactor_total);
                        }
                        inputs.at(m).missionPower.at(i).peakLoad.at(j) += power_tmp;//Leistung wird von Hydraulic abgezogen
                    }
                }
            }
        }
    }
    /* Electric power is accumulated for the electric circuits*/
    powerProfile myElectricSum_outputs(getPowerProfileSum(outputs));
    /* Power losses are converted into heat */
    heatLoad.missionPower.at(i).baseLoad = inputs.at(findPort("ShaftPower", "ShaftPowerIn", "input")).missionPower.at(i).baseLoad
                                           - myElectricSum_outputs.missionPower.at(i).baseLoad;
    }
    /* Clean up peaks */
    for (size_t n(0); n < inputs.size(); n++) {
        cleanPowerPeaks(&inputs.at(n));
    }
    cleanPowerPeaks(&heatLoad);
}

/* Determination of the mass */
double conventionalATA24::getMass() {
    double Range            = data_->data.Performance.designRange;
    geom2::MultisectionSurface<geom2::PolygonSection> Fuselage = data_->theFuselage.back();
    systemsIOData::Wing Wing     = data_->theWing.back();
    std::vector<geom2::MultisectionSurface<geom2::PolygonSection>> Nacelles = data_->theNacelle;
    double maxPower = maxRelPower * system_design_power / 1000.; //Conversion from max. required power to max. continuous power in kW
    lengthCable = getCableLength(Fuselage, Wing, Nacelles);
    std::vector<double> m_ATA24;
    /* [Ste10] */
    double SF(0.0);  // Factor from [Ste10], S.30, Gl.4.16
    if (Range <= 7408. * 1000.) { // in [m]
        SF = 0.95;   // Short Range
    } else {
        SF = 1.12;   // Long Range
    }
    /*NOTE(aigner): The formula from [Ste10], page 30, eq. 4.16  (tempMass = 16.3 * SF * maxPower * (1. - 0.033 * sqrt(maxPower));) has been corrected.
                    The reason for this is as follows:
                    For large values of maxPower, the curve from [Ste10]'s formula has a maximum and eventually even drops below zero.
                    This does not make sense, since the mass of the system should always increase with required power!
                    Therefore, K. Risse and F. Peter implemented the formula below (root function) as a work-around.
                    The values match the curve from [Ste10] for lower values of maxPower quite well.
                    For large values of maxPower, the mass increases continuously, which is physically feasible.*/
    double tempMass = 115. * SF * sqrt(maxPower);//new formula
    cableMass = specificCableMass * lengthCable;
    /*NOTE(aigner): Factor 0.61 originates from [Koe06] S.101 Figure 4.25,
    subtracting the cable mass by omitting the proportions "gemeinsame Elemente" (29.9%)
    and "nicht erfasste Elemente" (9.3%) -> 100-29.9-9.3=60.7 */
    otherMass = 0.61 * tempMass - ((1. - percentageElectricPower) * 150.);  // Oswald: reduction of generator mass, if an other power source than
                                                                            // the propulsion unit generates all or part of the power
    m_ATA24.push_back(otherMass + cableMass);

    double massATA24(0.);
    for (size_t i = 0; i < m_ATA24.size(); i++) {
        massATA24 += m_ATA24.at(i);
    }
    massATA24 = massATA24 / static_cast<double>(m_ATA24.size());
    return  massATA24;
}

/* Determination of the cable length */
double conventionalATA24::getCableLength(const geom2::MultisectionSurface<geom2::PolygonSection>& Fuselage, const systemsIOData::Wing& Wing,
                                         const std::vector<geom2::MultisectionSurface<geom2::PolygonSection>>& Nacelles) {
    const geometricalFunctions myGeoFunctions;
    double theCableLength(0.);
    /* Determination of the cable length */
    std::vector<geom2::MultisectionSurface<geom2::PolygonSection>> wingNacelle;
    std::vector<geom2::MultisectionSurface<geom2::PolygonSection>> fuselageNacelle;
    /* Determination of the engine positions */
    for (size_t i = 0; i < Nacelles.size(); i++) {
        /* Engines on the wing (symmetry) */
        if (Nacelles.at(i).name == "wing") {
            wingNacelle.push_back(Nacelles.at(i));
        } else if (Nacelles.at(i).name == "fuselage" || Nacelles.at(i).name == "empennage") {
            /* Engines on the fuselage resp. vertical tail (no symmetry) */
            fuselageNacelle.push_back(Nacelles.at(i));
        } else {
            myRuntimeInfo->err << "The engine position \"" << Nacelles.at(i).name << "\" is unknown!" << std::endl;
            myRuntimeInfo->err << "ATA24.cpp, getMass" << std::endl;
            myRuntimeInfo->err << "Abort!" << std::endl;
            exit(1);
        }
    }
    theCableLength = 0.0;
    /* Determination of the conductor lengths */
    /* MainBus: Front to back, 2 bus systems */
    theCableLength = 2.0 * geom2::measure::length(Fuselage);
    /* Connection: Wing engines to AvionicBay */
    if (static_cast<int>(wingNacelle.size()) > 0) {
        /* Wing to center */
        theCableLength += 2.0 * myGeoFunctions.getConductorLengthWingEngines(Wing.aerodynamicSurface, wingNacelle);
        /* Center to Avionic Bay */
        theCableLength += 2.0 * fabs(Wing.aerodynamicSurface.origin[0] - data_->data.Accommodation.payload_tube_x_ref);
    }
    /* Connection: fuselage engines to Avionicbay */
    if (static_cast<int>(wingNacelle.size()) > 0) {
        /* Wing to mid-fuselage */
        theCableLength += myGeoFunctions.getConductorLengthFuselageEngines(Fuselage, fuselageNacelle);
        /* Mid-fuselage to avionics bay */
        for (size_t i = 0; i < fuselageNacelle.size(); i++) {
            theCableLength += fabs(fuselageNacelle.at(i).origin[0] - data_->data.Accommodation.payload_tube_x_ref);
        }
    }
    return theCableLength;
}

/* Determination of the reference point */
Vec3 conventionalATA24::getRefPoint() {
    /* Assumption: center of gravity at 45% fuselage length **/
    double fuselage_length = geom2::measure::length(data_->theFuselage.back());
    return Vec3(data_->theFuselage.back().origin[0] + 0.45 * fuselage_length, data_->theFuselage.back().origin[1], data_->theFuselage.back().origin[2]);
}
