/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
*/

/* The following statements import necessary header files */
#include "conventionalATA25.h"                               // Import of header file from src/aircraftSystems folder for conventionalATA25
#include <runtimeInfo/runtimeInfo.h>                         // Import of header file from external runtime library for using runtimeInfo
/* Constructor for conventionalATA25 */
conventionalATA25::conventionalATA25(const std::string& systemName, bool systemIsOperating, std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO, size_t index)
    :
    aircraftSystem(systemName, systemIsOperating, data, rtIO, index, "sink") {
    ATA = "ATA-25";
    /* Define how many input and output ports the code can handle: */
    allowed_inputs = 1;
    allowed_outputs = 0;
    /* Define port types */
    expected_input_types = {"Electric"};
    expected_output_types = {""};
    /* Check ports */
    checkPorts();
    myRuntimeInfo->out << "  Create: " << Name << "  ->  " << ATA << std::endl;
    //ctor
}

/* Destructor for conventionalATA25 */
conventionalATA25::~conventionalATA25() {
    //dtor
}

/* Reading system specific parameters of configuration xml file */
void conventionalATA25::readParameters(const node& config) {
    std::string furnishingConfig = "/module_configuration_file/program_settings/furnishing";
    galleyLoadFractionTakeOff = config.at(furnishingConfig + "/galley_load_fraction_takeoff/value");
    data_->configuration.checkForValidity(galleyLoadFractionTakeOff, 0.0, 1., "galleyLoadFractionTakeOff");
    galleyLoadFractionCruise = config.at(furnishingConfig + "/galley_load_fraction_cruise/value");
    data_->configuration.checkForValidity(galleyLoadFractionCruise, 0.0, 1., "galleyLoadFractionCruise");
    galleyLoadFractionDescent = config.at(furnishingConfig + "/galley_load_fraction_descent/value");
    data_->configuration.checkForValidity(galleyLoadFractionDescent, 0.0, 1., "galleyLoadFractionDescent");
    nonPersonelIFEPower = config.at(furnishingConfig + "/non_personal_ife_power/value");
    data_->configuration.checkForValidity(nonPersonelIFEPower, 0.0, 500., "nonPersonelIFEPower");
    powerPersonelIFE = config.at(furnishingConfig + "/power_personal_ife/value");
    data_->configuration.checkForValidity(powerPersonelIFE, 0.0, 500., "powerPersonelIFE");
    personelIFELoadFractionClimb = config.at(furnishingConfig + "/personal_ife_load_fraction_climb/value");
    data_->configuration.checkForValidity(personelIFELoadFractionClimb, 0.0, 1., "personalIFELoadFractionClimb");
    personelIFELoadFractionCruise = config.at(furnishingConfig + "/personal_ife_load_fraction_cruise/value");
    data_->configuration.checkForValidity(personelIFELoadFractionCruise, 0.0, 1., "personalIFELoadFractionCruise");
    personelIFELoadFractionDescent = config.at(furnishingConfig + "/personal_ife_load_fraction_descent/value");
    data_->configuration.checkForValidity(personelIFELoadFractionDescent, 0.0, 1., "personalIFELoadFractionDescent");
}
/* Updating system specific parameters in aircraft xml file */
void conventionalATA25::updateXML() {}

/* Calculation of the system power */
void conventionalATA25::calculateSystemsPower() {
    /* Calculate Systems Power */
    if (Operating == true) {
        getPower();
    }
    //calculate max mission power for input ports
    for (size_t i(0); i < inputs.size(); i++) {
        calculateMaxMissionPowerPort(inputs.at(i).ID);
    }
}

/* Calculation of the skin mass */
void conventionalATA25::calculateSinkMass() {
    /* Calculate Systems Mass */
    //Attention: ATA33 (Lighting) is subtracted here, because in [Tor13] formula for ATA-25 (Furnishing) already contained
    Mass = getMass();
    //- 0.02 * 0.92 * pow(data_->data.Mass.MTOM, 0.85);//ata33 Masse not substracted because furnishing mass comes from fuselage design
    refPoint = getRefPoint();
}

/* Determination of the power */
void conventionalATA25::getPower() {
    /* IFE design power estimation */
    double designIFEPower(0.);
    double designGalleyPower(0.);
    /* Estimate designIFEPower */
    designIFEPower = (powerPersonelIFE + nonPersonelIFEPower) * data_->data.Accommodation.numberIFE;
    /* Estimate designGalleyPower */
    designGalleyPower = data_->data.Accommodation.galleyPower;
    //assumption: single electric input port
    inputs.at(0).designPower = designIFEPower + designGalleyPower;
    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        inputs.at(0).missionPower.at(i).baseLoad = 0.0;
        if (data_->mission.Steps.at(i).missionPhase == "Departure") {
            inputs.at(0).missionPower.at(i).baseLoad += personelIFELoadFractionClimb * designIFEPower;
            if (data_->mission.Steps.at(i).modeStep == "Accelerate" ||
                    data_->mission.Steps.at(i).modeStep == "Climb") {
                inputs.at(0).missionPower.at(i).baseLoad += galleyLoadFractionCruise * designGalleyPower;
            } else {
                inputs.at(0).missionPower.at(i).baseLoad += galleyLoadFractionTakeOff * designGalleyPower;
            }
        } else if (data_->mission.Steps.at(i).missionPhase == "Approach" || data_->mission.Steps.at(i).modeStep == "Descend_to_approach") {
            inputs.at(0).missionPower.at(i).baseLoad += personelIFELoadFractionDescent * designIFEPower;
            inputs.at(0).missionPower.at(i).baseLoad += galleyLoadFractionDescent * designGalleyPower;
        } else if (data_->mission.Steps.at(i).missionPhase == "Cruise") {
            if (data_->mission.Steps.at(i).modeStep == "change_speed_to_CAS" ||
                    data_->mission.Steps.at(i).modeStep == "Climb") {
                inputs.at(0).missionPower.at(i).baseLoad += personelIFELoadFractionClimb * designIFEPower;
                inputs.at(0).missionPower.at(i).baseLoad += galleyLoadFractionCruise * designGalleyPower;
            } else {
                inputs.at(0).missionPower.at(i).baseLoad += personelIFELoadFractionCruise * designIFEPower;
                inputs.at(0).missionPower.at(i).baseLoad += galleyLoadFractionCruise * designGalleyPower;
            }
        } else {
            inputs.at(0).missionPower.at(i).baseLoad += designIFEPower;
            inputs.at(0).missionPower.at(i).baseLoad += designGalleyPower;
        }
        // NOTE (aigner#1#): Temporarily uniform distribution of the baseLiad to the electrical circuits
        inputs.at(0).missionPower.at(i).baseLoad *= 1 / data_->configuration.Electric.numberOfElectricCircuits;
        /* Accumulate the electrical power to a total powerProfile*/
        powerProfile myElectricSum(getPowerProfileSum(inputs));
        /* Direct conversion from energy to heat - das bleibt wie es ist */
        heatLoad.missionPower.at(i).baseLoad = 0.2 * myElectricSum.missionPower.at(i).baseLoad;
    }
}

/* Determination of the mass */
double conventionalATA25::getMass() {
    //calculated in fuselageDesign
    double massATA25 = data_->data.Accommodation.furnishingMass;
    return  massATA25;
}

/* Determination of the reference point */
Vec3 conventionalATA25::getRefPoint() {
    double x_cg(0.);
    double fuselage_length = geom2::measure::length(data_->theFuselage.back());
    // 0.45*l_fus (arrangement kri&fab), because cockpit+cabin elements are included (and no passenger items)
    x_cg = data_->theFuselage.back().origin.x() + (0.45 * fuselage_length);
    return Vec3(x_cg, data_->theFuselage.back().origin.y(), data_->theFuselage.back().origin.z());
}
