/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
*/

/* The following statements import necessary header files */
#include "conventionalATA26.h"                                       // Import of header file from src/aircraftSystems folder for conventionalATA26
#include <runtimeInfo/runtimeInfo.h>                                 // Import of header file from external runtime library for using runtimeInfo

/* Constructor for conventionalATA26 */
conventionalATA26::conventionalATA26(const std::string& systemName, bool systemIsOperating, std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO, size_t index)
    :
    aircraftSystem(systemName, systemIsOperating, data, rtIO, index, "sink") {
    ATA = "ATA-26";
    /* Define how many input and output ports the code can handle: */
    allowed_inputs = 0;
    allowed_outputs = 0;
    /* Define port types */
    expected_input_types = {""};
    expected_output_types = {""};
    /* Check ports */
    checkPorts();
    myRuntimeInfo->out << "  Create: " << Name << "  ->  " << ATA << std::endl;
    //ctor
}

/* Destructor for conventionalATA26 */
conventionalATA26::~conventionalATA26() {
    //dtor
}

/* Reading system specific parameters of configuration xml file */
void conventionalATA26::readParameters(const node& config) {
    scaling_factor = EndnodeReadOnly<double>("/module_configuration_file/program_settings/scaling_factors/fire_protection_system").read(config).value();
}

/* Updating system specific parameters in aircraft xml file */
void conventionalATA26::updateXML() {
    updateMass("fire_protection", Mass, ATA);
}

/* Calculation of the system power */
void conventionalATA26::calculateSystemsPower() {
    /* Calculate Systems Power */
    if (Operating == true) {
        getPower();
    }
    //calculate max mission power for input ports
    for (size_t i(0); i < inputs.size(); i++) {
        calculateMaxMissionPowerPort(inputs.at(i).ID);
    }
}

/* Calculation of the skin mass */
void conventionalATA26::calculateSinkMass() {
    /* Calculate Systems Mass */
    Mass = scaling_factor * getMass();
    refPoint = getRefPoint();
}

/* Determination of the power */
void conventionalATA26::getPower() {
    //no power required
}

/* Determination of the mass */
double conventionalATA26::getMass() {
    // From [Tor13] Tab. 8-12 (Furnishings), page 291 (pdf page 310)
    /* Fire Detection and Extinguishing System */
    double fireDetectionMass;
    // method is based on engine type of the first propulsor
    std::string propulsionType = data_->data.Propulsion.Propulsor.at(0).type;
    if (propulsionType == "FAN" || propulsionType == "JET" || propulsionType == "GTF") {
        fireDetectionMass = 0.0012 * data_->data.Mass.MTOM;
    } else {
        fireDetectionMass = 0.0030 * data_->data.Mass.MTOM;
        myRuntimeInfo->info << "Since propulsion type is not FAN, JET, or GTF, ATA26 mass method for turbo prop is used." << std::endl;
    }
    return fireDetectionMass;
}

/* Determination of the reference point */
Vec3 conventionalATA26::getRefPoint() {
    /* Assumption: center of gravity at 45% fuselage length */
    double fuselage_length = geom2::measure::length(data_->theFuselage.back());
    return Vec3(data_->theFuselage.back().origin[0] + 0.45 * fuselage_length, data_->theFuselage.back().origin[1], data_->theFuselage.back().origin[2]);
}
