/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
*/

/* The following statements import necessary header files */
#include "conventionalATA28.h"                                       // Import of header file from src/aircraftSystems folder for conventionalATA28
#include <runtimeInfo/runtimeInfo.h>                                 // Import of header file from external runtime library for using runtimeInfo

/* Constructor for conventionalATA28 */
conventionalATA28::conventionalATA28(const std::string& systemName, bool systemIsOperating, std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO, size_t index)
    :
    aircraftSystem(systemName, systemIsOperating, data, rtIO, index, "sink") {
    ATA = "ATA-28";
    /* Define how many input and output ports the code can handle: */
    allowed_inputs = 1;
    allowed_outputs = 0;
    /* Define port types */
    expected_input_types = {"Electric"};
    expected_output_types = {""};
    /* Check ports */
    checkPorts();
    myRuntimeInfo->out << "  Create: " << Name << "  ->  " << ATA << std::endl;
    //ctor
}

/* Destructor for conventionalATA28 */
conventionalATA28::~conventionalATA28() {
    //dtor
}

/* Reading system specific parameters of configuration xml file */
void conventionalATA28::readParameters(const node& config) {
    scaling_factor = EndnodeReadOnly<double>("/module_configuration_file/program_settings/scaling_factors/fuel_system").read(config).value();
}

/* Updating system specific parameters in aircraft xml file */
void conventionalATA28::updateXML() {
    updateMass("fuel_system", Mass, ATA);
}

/* Calculation of the system power */
void conventionalATA28::calculateSystemsPower() {
    /* Calculate Systems Power */
    if (Operating == true) {
        getPower();
    }
    //calculate max mission power for input ports
    for (size_t i(0); i < inputs.size(); i++) {
        calculateMaxMissionPowerPort(inputs.at(i).ID);
    }
}

/* Calculation of the skin mass */
void conventionalATA28::calculateSinkMass() {
    /* Calculate Systems Mass */
    Mass = scaling_factor * getMass();
    refPoint = getRefPoint();
}

/* Determination of the power */
void conventionalATA28::getPower() {
    /* Factor calibrated according to [Bue14] p.66 (MTOM 77000t at 11.9kW -> 0.154547)*/
    //assumption: only one electric input port
    inputs.at(0).designPower = 0.154547 * data_->data.Mass.MTOM;
    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        heatLoad.missionPower.at(i).baseLoad = 0.0;
        inputs.at(0).missionPower.at(i).baseLoad = inputs.at(0).designPower;
    }
}

/* Determination of the mass */
double conventionalATA28::getMass() {
    std::vector<double> m_ATA28;
    //Method from [Ste10], p. 27 eq. 4-15, with 0.7 as technology factor.
    m_ATA28.push_back(0.7 * 0.9184 * pow((data_->data.Mass.maxUsableFuel / data_->data.fuelDensity), 0.6));
    /* [Ray18]: m_ATA28 = 2.405 * V_useableFuel^0.606 * 0.5 * 1 * n_tanks^0.5 */
    double V_useableFuel = data_->data.Mass.maxUsableFuel / data_->data.fuelDensity * 0.264172037; //kg -> l -> gal
    m_ATA28.push_back(0.5 * 2.405 * pow(V_useableFuel, 0.606) * pow(3., 0.5));
    m_ATA28.back() = m_ATA28.back() * 0.4535924; //lb -> kg
    double massATA28(0.);
    for (size_t i = 0; i < m_ATA28.size(); i++) {
        massATA28 += m_ATA28.at(i);
    }
    massATA28 = massATA28 / static_cast<double>(m_ATA28.size());
    return  massATA28;
}

/* Determination of the reference point */
Vec3 conventionalATA28::getRefPoint() {
    /* Assumption: center of gravity is in the CG of the wing*/
    return Vec3(data_->data.CofG.Wing);
}
