/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
*/

#include "conventionalATA32.h"

#include <unitConversion/constants.h>
#include <standardFiles/functions.h>
#include <runtimeInfo/runtimeInfo.h>

conventionalATA32::conventionalATA32(const std::string& systemName, bool systemIsOperating, std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO, size_t index)
    :
    aircraftSystem(systemName, systemIsOperating, data, rtIO, index, "sink") {
    ATA = "ATA-32";
    /* Define how many input and output ports the code can handle: */
    allowed_inputs = 1;
    allowed_outputs = 0;
    /* Define port types */
    expected_input_types = {"Hydraulic"};
    expected_output_types = {""};
    /* Check ports */
    checkPorts();
    /* Runtime Info */
    myRuntimeInfo->out << "  Create: " << Name << "  ->  " << ATA << std::endl;
    //ctor
}

conventionalATA32::~conventionalATA32() {
    //dtor
}

/* Reading system specific parameters of configuration xml file */
void conventionalATA32::readParameters(const node& config) {
    scaling_factor = EndnodeReadOnly<double>("/module_configuration_file/program_settings/scaling_factors/landing_gear_system").read(config).value();
    efficiencyFactor = config.at("/module_configuration_file/program_settings/systems_constants/gear/efficiency_factor/value");
    retractionTime = config.at("/module_configuration_file/program_settings/systems_constants/gear/retraction_time/value");
    extensionTime = config.at("/module_configuration_file/program_settings/systems_constants/gear/extension_time/value");
    data_->configuration.checkForValidity(efficiencyFactor, 0.0, 1.0, "Gear.efficiencyFactor");
}

/* Updating system specific parameters in aircraft xml file */
void conventionalATA32::updateXML() {
    //not written to acxml since calculated by landingGearDesign
}

void conventionalATA32::calculateSystemsPower() {
    /* Calculate Systems Power */
    if (Operating == true) {
        getPower();
    }
    //calculate max mission power for input ports
    for (size_t i(0); i < inputs.size(); i++) {
        calculateMaxMissionPowerPort(inputs.at(i).ID);
    }
}
void conventionalATA32::calculateSinkMass() {
    /* Calculate Systems Mass */
    if (data_->data.Mass.Gear > 0.0) {
        Mass = data_->data.Mass.Gear;
        refPoint = data_->data.CofG.landingGear;
    } else {
        Mass = scaling_factor * getMass();
        refPoint = getRefPoint();
    }
}
void conventionalATA32::getPower() {
    //Initializing auxiliary variables
    double Efficiency(efficiencyFactor);
    double gearRetractionTime(retractionTime);
    double gearExtensionTime(extensionTime);
    double massGear(data_->data.Mass.Gear);
    double retractionPower(0.);
    double extensionPower(0.);
    double gearStrutLength(0.);//average strut length of the landing gear
    int mainGearCount(0);//counter for the main gears
    /*Calculation of the strut length as average value of all strut lengths*/
    for (size_t i(0); i < data_->theLandingGear.size(); i++) {
        //Calculation strictly speaking only applies to Main Gear, the most massive part of the landing gear
        //--> Strut Length is only included in the calculation for Main Gear
        if (compareStrings(data_->theLandingGear.at(i).name, "main", false) == true) {
            mainGearCount++;
            gearStrutLength += data_->theLandingGear.at(i).strut_length;
        }
    }
    if (mainGearCount == 0) {
        myRuntimeInfo->warn << "There was no main landing gear found. This might lead to errors in the calculation of hydraulic system power." << std::endl;
    }
    gearStrutLength *= 1 / static_cast<double>(mainGearCount);
    retractionPower =        4. * (2. * G_FORCE * massGear * gearStrutLength) / gearRetractionTime;
    extensionPower  = 0.75 * 4. * (2. * G_FORCE * massGear * gearStrutLength) / gearExtensionTime;;
    for (int n(0); n < data_->configuration.Hydraulic.numberOfhydraulicCircuits; n++) {
        double designPower_tmp(0.);
        designPower_tmp = retractionPower / Efficiency;
        inputs.at(0).designPower += designPower_tmp;
    }
    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        //bleedAir.missionPower.at(i).baseLoad     = 0.0;
        inputs.at(0).missionPower.at(i).baseLoad = 0.0;
        if (i > 0 && i < data_->mission.Steps.size() - 1) {
            // mission segment in which the landing gear is deployed
            double peakLoad_tmp;
            if (data_->mission.Steps.at(i - 1).modeStep == "takeoff") { //should always be deployed in 2nd mission segment
                peakLoad_tmp = retractionPower / (Efficiency);
                //Create and calculate peak loads of the hydraulic demand
                inputs.at(0).missionPower.at(i).peakLoad.push_back(peakLoad_tmp);
                inputs.at(0).missionPower.at(i).timePeak.push_back(gearRetractionTime);
            } else if (data_->mission.Steps.at(i - 1).configuration == "approach" && data_->mission.Steps.at(i).configuration == "approach_landing_gear_out") {
                peakLoad_tmp = extensionPower / (Efficiency);
                //Create and calculate peak loads of the hydraulic demand
                inputs.at(0).missionPower.at(i).peakLoad.push_back(peakLoad_tmp);
                inputs.at(0).missionPower.at(i).timePeak.push_back(gearExtensionTime);
            }
        }
        heatLoad.missionPower.at(i).baseLoad     = 0.0;
    }
    cleanPowerPeaks(&heatLoad);
    cleanPowerPeaks(&inputs.at(0));
}
double conventionalATA32::getMass() {
    return  static_cast<double>(0.00271 * pow(data_->data.Mass.MLM, 1.24));
}

Vec3 conventionalATA32::getRefPoint() {
    return Vec3(data_->data.CofG.landingGear);
}
