/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
*/

#include "conventionalATA35.h"

#include <runtimeInfo/runtimeInfo.h>

conventionalATA35::conventionalATA35(const std::string& systemName, bool systemIsOperating, std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO, size_t index)
    :
    aircraftSystem(systemName, systemIsOperating, data, rtIO, index, "sink") {
    ATA = "ATA-35";
    /* Define how many input and output ports the code can handle: */
    allowed_inputs = 0;
    allowed_outputs = 0;
    /* Define port types */
    expected_input_types = {""};
    expected_output_types = {""};
    /* Check ports */
    checkPorts();
    myRuntimeInfo->out << "  Create: " << Name << "  ->  " << ATA << std::endl;
    //ctor
}

conventionalATA35::~conventionalATA35() {
    //dtor
}

/* Reading system specific parameters of configuration xml file */
void conventionalATA35::readParameters(const node& config) {
    scaling_factor = EndnodeReadOnly<double>("/module_configuration_file/program_settings/scaling_factors/oxygen_system").read(config).value();
}

/* Updating system specific parameters in aircraft xml file */
void conventionalATA35::updateXML() {
    updateMass("oxygen_system", Mass, ATA);
}

void conventionalATA35::calculateSystemsPower() {
    /* Calculate Systems Power */
    if (Operating == true) {
        getPower();
    }
    //calculate max mission power for input ports
    for (size_t i(0); i < inputs.size(); i++) {
        calculateMaxMissionPowerPort(inputs.at(i).ID);
    }
}
void conventionalATA35::calculateSinkMass() {
    /* Calculate Systems Mass */
    Mass = scaling_factor * getMass();
    refPoint = getRefPoint();
}
void conventionalATA35::getPower() {
    //no power required
}

double conventionalATA35::getMass() {
    int PAX = data_->data.Accommodation.PAX;
    int flightCrew = data_->data.Accommodation.flightCrew;
    int cabinCrew = data_->data.Accommodation.cabinCrew;
    int nrPAX = PAX + flightCrew + cabinCrew;
    std::vector<double> m_ATA35;
    /* [Ros89] (Toreenbeek Method) **/
    if (data_->data.Performance.designRange < 6000. * 1000.) { //[m]
        m_ATA35.push_back(13.6 + 0.544 * nrPAX);
    } else {
        m_ATA35.push_back(18.1 + 1.09 * nrPAX);
    }
//    /** [Ros89] (GD Method): m_ATA35 = 7 * (PAX+Crew)^0.702 (lb)**/
//    m_ATA35.push_back( 7. * pow(nrPAX, 0.702) );
//    m_ATA35.back() = m_ATA35.back() * 0.4535924; //lb -> kg
    double massATA35(0.);
    for (size_t i = 0; i < m_ATA35.size(); i++) {
        massATA35 += m_ATA35.at(i);
    }
    massATA35 = massATA35 / static_cast<int>(m_ATA35.size());
    return  massATA35;
}

Vec3 conventionalATA35::getRefPoint() {
    /* Assumption: center of gravity at 45% fuselage length **/
    double fuselage_length = geom2::measure::length(data_->theFuselage.back());
    return Vec3(data_->theFuselage.back().origin[0] + 0.45 * fuselage_length, data_->theFuselage.back().origin[1], data_->theFuselage.back().origin[2]);
}
