/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
*/

#include "conventionalATA49.h"

#include <unitConversion/constants.h>
#include <standardFiles/functions.h>
#include <runtimeInfo/runtimeInfo.h>

conventionalATA49::conventionalATA49(const std::string& systemName, bool systemIsOperating,
                                     std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO, size_t index)
    :
    aircraftSystem(systemName, systemIsOperating, data, rtIO, index, "source") {
    ATA = "ATA-49";
    /* Define how many input and output ports the code can handle: */
    allowed_inputs = 1;
    allowed_outputs = 2;
    /* Define port types */
    expected_input_types = {"Kerosene"};
    expected_output_types = {"Bleed", "ShaftPower"};
    /* Check ports */
    checkPorts();
    /* Runtime Info */
    myRuntimeInfo->out << "  Create: " << Name << "  ->  " << ATA << std::endl;
    //ctor
}

conventionalATA49::~conventionalATA49() {
    //dtor
}

/* Reading system specific parameters of configuration xml file */
void conventionalATA49::readParameters(const node& config) {
    scaling_factor = EndnodeReadOnly<double>("/module_configuration_file/program_settings/scaling_factors/auxiliary_power_unit").read(config).value();
    std::string APUConfig = "/module_configuration_file/program_settings/systems_constants/auxiliary_power_unit";
    posXFuselage   = config.at(APUConfig + "/position_of_auxiliary_power_unit_in_fuselage_length/value");
    data_->configuration.checkForValidity(posXFuselage, 0.0, 1.0, "APU.posXFuselage");
    efficiencyFactorBleedAir    = config.at(APUConfig + "/efficiency_factor_bleed_air/value");
    data_->configuration.checkForValidity(efficiencyFactorBleedAir, 0.0, 1.0, "APU.efficiencyFactorBleedAir");
    installationFactor     = config.at(APUConfig + "/installation_factor/value");
    data_->configuration.checkForValidity(installationFactor, 1.0, 3.0, "APU.installationFactor");
}

/* Updating system specific parameters in aircraft xml file */
void conventionalATA49::updateXML() {
    updateMass("auxiliary_power_unit", Mass, ATA);
}

void conventionalATA49::calculateSourcePower() {
    getPower();
    //calculate max mission power for input ports
    for (size_t i(0); i < inputs.size(); i++) {
        calculateMaxMissionPowerPort(inputs.at(i).ID);
    }
}
void conventionalATA49::calculateSourceMass() {
    Mass = scaling_factor * getMass();
    refPoint = getRefPoint();
}

void conventionalATA49::getPower() {
    //check which output port types there are
    bool bleedAirOutput = false;
    bool shaftPowerOutput = false;
    for (size_t n(0); n < outputs.size(); n++) {
        if (outputs.at(n).type == "Bleed") {
            bleedAirOutput = true;
            myRuntimeInfo->debug << "Bleed air output detected." << std::endl;
        }
        if (outputs.at(n).type == "ShaftPower") {
            shaftPowerOutput = true;
            myRuntimeInfo->debug << "Shaft power output detected." << std::endl;
        }
    }
    /* If source provides Bleed Air: */
    if (bleedAirOutput == true) {
        //find the correct output port
        int portIndex = findPort("Bleed", "BleedOut", "output");
        bleedAir.designPower  = (1 / efficiencyFactorBleedAir) * outputs.at(portIndex).designPower;
        for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
            missionLoads tmp_loads;
            bleedAir.missionPower.push_back(tmp_loads);
            bleedAir.missionPower.at(i).baseLoad = (1 / efficiencyFactorBleedAir) * outputs.at(portIndex).missionPower.at(i).baseLoad;
            for (size_t j = 0; j < outputs.at(portIndex).missionPower.at(i).peakLoad.size(); j++) {
                bleedAir.missionPower.at(i).peakLoad.push_back((1 / efficiencyFactorBleedAir) * outputs.at(portIndex).missionPower.at(i).peakLoad.at(j));
                bleedAir.missionPower.at(i).timePeak.push_back(outputs.at(portIndex).missionPower.at(i).timePeak.at(j));
            }
        }
    }
    /* If source provides Shaft Power: */
    if (shaftPowerOutput == true) {
        //find the correct output port
        int portIndex = findPort("ShaftPower", "ShaftPowerOut", "output");
        shaftPower.designPower = outputs.at(portIndex).designPower;
        for (size_t i(0); i < data_->mission.Steps.size(); i++) {
            missionLoads tmp_loads;
            shaftPower.missionPower.push_back(tmp_loads);
            shaftPower.missionPower.at(i).baseLoad = outputs.at(portIndex).missionPower.at(i).baseLoad;
            for (size_t j(0); j < outputs.at(portIndex).missionPower.at(i).peakLoad.size(); j++) {
                shaftPower.missionPower.at(i).peakLoad.push_back(outputs.at(portIndex).missionPower.at(i).peakLoad.at(j));
                shaftPower.missionPower.at(i).timePeak.push_back(outputs.at(portIndex).missionPower.at(i).timePeak.at(j));
            }
        }
    }
    /* Calculate required input (Kerosene)*/
    /*Fuel consumption for design load*/
    double effective_design_power = shaftPower.designPower + bleedAir.designPower * 1005. * 200.;
    double design_fuel_consumption = getKerosene(convertUnit(WATT, KILO, WATT, effective_design_power));
    myRuntimeInfo->debug << "APU fuel consumption with design loads (" << effective_design_power << " W): " << design_fuel_consumption << " kg/h" << std::endl;
    inputs.at(0).designPower = design_fuel_consumption;
    /* Fuel consumption for mission loads */
    for (size_t n(0); n < data_->mission.Steps.size(); n++) {
        double shaftPower_tmp = 0.;
        double bleedPower_tmp = 0.;
        //check which power types are output
        if (bleedAirOutput) {
            bleedPower_tmp = bleedAir.missionPower.at(n).baseLoad * 1005. * 200.;// DA Steinke, p.48, eq. 6.7, with cp = 1005. [kJ/
        }
        if (shaftPowerOutput) {
            shaftPower_tmp = shaftPower.missionPower.at(n).baseLoad;
        }
        double effectivePower = shaftPower_tmp + bleedPower_tmp;
        myRuntimeInfo->debug << "Mission effective power: " << effectivePower;
        inputs.at(0).missionPower.at(n).baseLoad = getKerosene(convertUnit(WATT, KILO, WATT, effectivePower)) / 3600;//in kg/s
    }
}

double conventionalATA49::getKerosene(double Power) {
    double SFC = 0.0;
    if (Power == 0.0) {
        /* If no power is required, the SFC is 0 */
    } else {
        /* specific fuel consumption in [kg/h] */
        SFC = -0.31 * log(abs(Power)) + 2.24; // DA Steinke S.49 basierend auf LTH Daten
    }
    return SFC;
}

double conventionalATA49::getMass() {
// TODO(peter#1#): Mantis ID 0000213: Neue Modellierung der APU-Nutzleistung ...
//- Recherche ob LTH Nutzlastdaten NUR Wellenleistung sind (vgl. reale Datenblaetter)
//- Berechnung der bleedpower nur thermisch (cp * delta T), beruecksichtigung der Kompressionsarbeit fehlt
//- Massenabschaetzung ueberarbeiten
    double bleedPower(0.0); // DA Steinke, p.48, eq. 6.7, with cp = 1005. [kJ/
    bleedPower = bleedAir.designPower * 1005. * 200.;
    double effectivePower(0.0);    // DA Steinke, p.48 in [kW]!!!
    effectivePower = (bleedPower / efficiencyFactorBleedAir
                      +   shaftPower.designPower) / 1000.;
    double APUMass(0.);
    // F. Peter 22.7.4: new regression of the APU power mass [kg/kW] according to LTH data (p. APU-Massenabschaetzung.xls -> sheet LTH_red_2)
    APUMass = installationFactor * (6.8694073751 * pow(effectivePower, -0.4311907202) * effectivePower); //Regression fuer Leistungsgewicht
    // alternatively: linear regression of the mass: APUMass = installationFactor * (0.28198117 * effectivePower + 89.85363477);
    // myRuntimeInfo->debug << std::endl << "APU Design Parameters:" << std::endl;
    // myRuntimeInfo->debug << "theBleed.designPower: " << theBleed.designPower << std::endl;
    myRuntimeInfo->debug << "bleedPower (kg/s): " << bleedAir.designPower << std::endl;
    myRuntimeInfo->debug << "bleedPower (kW): " << bleedPower << std::endl;
    myRuntimeInfo->debug << "shaftPower: " << shaftPower.designPower << std::endl;
    myRuntimeInfo->debug << "APU effectivePower in kW: " << effectivePower << std::endl;
    // myRuntimeInfo->debug << "installationFactor: " << installationFactor << std::endl;
    // myRuntimeInfo->debug << "APUMass: " << APUMass << std::endl;
    // myRuntimeInfo->debug << std::endl;
    // myRuntimeInfo->debug << "APU Power in kW (shaft+bleed): " << effectivePower << std::endl;
    // myRuntimeInfo->debug << "APU Mass in kg: " << APUMass << std::endl;
    return  APUMass;
}

Vec3 conventionalATA49::getRefPoint() {
    Vec3 cgPoint;
    cgPoint[0] = data_->theFuselage.back().origin[0] + static_cast<double>(posXFuselage * geom2::measure::length(data_->theFuselage.back()));
    cgPoint[1] = data_->theFuselage.back().origin[1];
    cgPoint[2] = data_->theFuselage.back().origin[2];
    return cgPoint;
}
