/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
*/

#ifndef SYSTEMS_DESIGN_SRC_AIRCRAFTSYSTEMS_CONVENTIONALATA49_H_
#define SYSTEMS_DESIGN_SRC_AIRCRAFTSYSTEMS_CONVENTIONALATA49_H_

/* The following statements import necessary libraries */
#include <string>                                                   // Import c++ standard library for using string
#include <vector>                                                   // Import c++ standard library for using vector
#include <memory>                                                   // Import c++ standard library for using memory

/* The following statements import necessary header files */
#include <svl/SVL.h>                                                // Import of header file from src folder for SVL
#include "../aircraftSystem.h"                                      // Import of header file from src folder for aircraft system
#include "../powerProfile.h"                                        // Import of header file from src folder for power profile
#include "../systemsIOData.h"                                       // Import of header file from systemsIOData for accessing data_. Can't be included in
                                                                    // aircraftSystem.h because that would lead to circular dependencies

/** \class conventionalATA49
 * \brief class that contains all properties of the auxiliary power unit (ATA49)
 */
class conventionalATA49: public aircraftSystem {
 public:
    double posXFuselage;/**< [-] Relative position of the APU in reference to the total fuselage length **/

    double efficiencyFactorBleedAir; /**< [-] Efficiency of Bleed Air extraction **/

    double installationFactor;/**< [-] Installation factor for attached parts such as fire protection, noise protection, etc. [Tor13] p. 288 2.0-2.5*/

    powerProfile shaftPower;/** Power profile for storing shaft power */
    powerProfile bleedAir;/** power profile for stroring bleed air power */

    /**
     * \brief function for reading the system specific parameters from the config
     * 
     * \param config const node&: systemsDesign configuration file
    */
    void readParameters(const node& config) override;
    /**
     * \brief function for updating the system specific values in the aircraft xml
    */
    void updateXML() override;

    void calculateSourcePower() override;
    void calculateSourceMass() override;

    /** \brief calculates the power of the APU (ATA49)
     * \return void
     */
    void getPower();

    /** \brief Calculates specific fuel consumption of the APU based on required power
     * \param Power [kW] total power APU has to provide
     * \return SFC (double) [kg/kWh] specific fuel consumption 
    */
    double getKerosene(double Power);

    /** \brief calculates the mass of the APU (ATA49)
     * \return double: mass of the APU [kg]
     */
    double getMass();

    /** \brief calculates the reference point of the APU (ATA49)
     * \return Vec3: APU reference point in global x, y, z-coordinates [m]
     */
    Vec3 getRefPoint();

    /** \brief Constructor
     * \param systemName string
     * \param systemIsOperating bool: switch whether the system is operating (1) or not(0)
     * \param data pointer to IOData
     */
    conventionalATA49(const std::string& systemName, bool systemIsOperating, std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO, size_t index);
    virtual ~conventionalATA49();
};

#endif // SYSTEMS_DESIGN_SRC_AIRCRAFTSYSTEMS_CONVENTIONALATA49_H_
