/*  Copyright (C) 2011-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "efficienciesATA70.h"

#include <unitConversion/constants.h>
#include <standardFiles/functions.h>
#include <runtimeInfo/runtimeInfo.h>

efficienciesATA70::efficienciesATA70(const std::string& systemName, bool systemIsOperating,
                                     std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO, size_t index)
    :
    aircraftSystem(systemName, systemIsOperating, data, rtIO, index, "source") {
    /* Define ATA name */
    ATA = "ATA-70";
    /* Define how many input and output ports the code can handle: */
    allowed_inputs = 2;
    allowed_outputs = 2;
    /* Define port types */
    expected_input_types = {"Bleed", "ShaftPower"};
    expected_output_types = {"Bleed", "ShaftPower"};
    /* Check ports */
    checkPorts();
    /* Runtime Info */
    myRuntimeInfo->out << "  Create: " << Name << "  ->  " << ATA << std::endl;
    //ctor
}

efficienciesATA70::~efficienciesATA70() {
    //dtor
}

/* Reading system specific parameters of configuration xml file */
void efficienciesATA70::readParameters(const node& config) {
    percentageBleedAir          = config.at("/module_configuration_file/program_settings/systems_constants/propulsion_system/percentage_bleed_air/value");
    data_->configuration.checkForValidity(percentageBleedAir, 0.0, 1.0, "Propulsion.percentageBleedAir");
    percentageElectricPower     = config.at("/module_configuration_file/program_settings/systems_constants/propulsion_system/percentage_electrical_power/value");
    data_->configuration.checkForValidity(percentageElectricPower, 0.0, 1.0, "Propulsion.percentageElectricPower");
    percentageHydraulicPower    = config.at("/module_configuration_file/program_settings/systems_constants/propulsion_system/percentage_hydraulic_power/value");
    data_->configuration.checkForValidity(percentageHydraulicPower, 0.0, 1.0, "Propulsion.percentageHydraulicPower");
    efficiencyFactorBleedAir    = config.at("/module_configuration_file/program_settings/systems_constants/propulsion_system/efficiency_factor_bleed_air/value");
    data_->configuration.checkForValidity(efficiencyFactorBleedAir, 0.0, 1.0, "Propulsion.efficiencyFactorBleedAir");
//    efficiencyFactorHydraulic   = config.at("/module_configuration_file/program_settings/systems_constants/propulsion_system/efficiencyFactorHydraulic");
//    data_->configuration.checkForValidity(efficiencyFactorHydraulic, 0.0, 1.0, "Propulsion.efficiencyFactorHydraulic" );
}

void efficienciesATA70::calculateSourcePower() {
    addEngineEfficiencies();
    //calculate max mission power for input ports
    for (size_t i(0); i < inputs.size(); i++) {
        calculateMaxMissionPowerPort(inputs.at(i).ID);
    }
}

void efficienciesATA70::addEngineEfficiencies() {
    /* Check for output ports */
    bool bleedAirOutput = false;
    bool shaftPowerOutput = false;
    for (size_t n(0); n < outputs.size(); n++) {
        if (outputs.at(n).type == "Bleed") {
            bleedAirOutput = true;
        }
        if (outputs.at(n).type == "ShaftPower") {
            shaftPowerOutput = true;
        }
    }
    /* If source provides Bleed Air: */
    if (bleedAirOutput == true) {
        //find the correct output port
        int portIndex = findPort("Bleed", "BleedOut", "output");
        bleedAir.designPower  = (percentageBleedAir / efficiencyFactorBleedAir) * outputs.at(portIndex).designPower;
        for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
            missionLoads tmp_loads;
            bleedAir.missionPower.push_back(tmp_loads);
            bleedAir.missionPower.at(i).baseLoad = (percentageBleedAir / efficiencyFactorBleedAir) * outputs.at(portIndex).missionPower.at(i).baseLoad;
            for (size_t j = 0; j < outputs.at(portIndex).missionPower.at(i).peakLoad.size(); j++) {
                bleedAir.missionPower.at(i).peakLoad.push_back((percentageBleedAir / efficiencyFactorBleedAir) * outputs.at(portIndex).missionPower.at(i).peakLoad.at(j));
                bleedAir.missionPower.at(i).timePeak.push_back(outputs.at(portIndex).missionPower.at(i).timePeak.at(j));
            }
        }
        cleanPowerPeaks(&bleedAir);
        /* Write to input ports*/
        int port_ID = findPort("Bleed", "BleedIn", "input");
        inputs.at(port_ID).designPower = bleedAir.designPower;
        for (size_t i(0); i < data_->mission.Steps.size(); i++) {
            inputs.at(port_ID).missionPower.at(i).baseLoad = bleedAir.missionPower.at(i).baseLoad;
            //add peak loads
            for (size_t j(0); j < bleedAir.missionPower.at(i).peakLoad.size(); j++) {
                inputs.at(port_ID).missionPower.at(i).peakLoad.push_back(bleedAir.missionPower.at(i).peakLoad.at(j));
                inputs.at(port_ID).missionPower.at(i).timePeak.push_back(bleedAir.missionPower.at(i).timePeak.at(j));
            }
        }
    }
    /* If source provides Shaft Power: */
    if (shaftPowerOutput == true) {
        //find the correct output port
        int portIndex = findPort("ShaftPower", "ShaftPowerOut", "output");
        shaftPower.designPower = outputs.at(portIndex).designPower;
        for (size_t i(0); i < data_->mission.Steps.size(); i++) {
            missionLoads tmp_loads;
            shaftPower.missionPower.push_back(tmp_loads);
            shaftPower.missionPower.at(i).baseLoad = outputs.at(portIndex).missionPower.at(i).baseLoad;
            for (size_t j(0); j < outputs.at(portIndex).missionPower.at(i).peakLoad.size(); j++) {
                shaftPower.missionPower.at(i).peakLoad.push_back(outputs.at(portIndex).missionPower.at(i).peakLoad.at(j));
                shaftPower.missionPower.at(i).timePeak.push_back(outputs.at(portIndex).missionPower.at(i).timePeak.at(j));
            }
        }
        cleanPowerPeaks(&shaftPower);
        /* Write to input ports*/
        //find shaft power port
        int port_ID = findPort("ShaftPower", "ShaftPowerIn", "input");
        inputs.at(port_ID).designPower = fabs(shaftPower.designPower);
        for (size_t i(0); i < data_->mission.Steps.size(); i++) {
            inputs.at(port_ID).missionPower.at(i).baseLoad = fabs(shaftPower.missionPower.at(i).baseLoad);
            /* Power peaks resulting from the electric power requirement */
            for (size_t j(0); j < shaftPower.missionPower.at(i).peakLoad.size(); j++) {
                inputs.at(port_ID).missionPower.at(i).peakLoad.push_back(fabs(shaftPower.missionPower.at(i).peakLoad.at(j)));
                inputs.at(port_ID).missionPower.at(i).timePeak.push_back(fabs(shaftPower.missionPower.at(i).timePeak.at(j)));
            }
        }
    }
}
