/*  Copyright (C) 2011-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/

/* The following statements import necessary header files */
#include "electricATA21.h"                                           // Import of header file from src/aircraftSystems folder for electricATA21
#include <unitConversion/unitConversion.h>                               // Import of header file from external unitConversion library for unitConversion
#include <memory>

/* Constructor for electricATA21 */
electricATA21::electricATA21(const std::string& systemName, bool systemIsOperating, std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO, size_t index)
    :
    aircraftSystem(systemName, systemIsOperating, data, rtIO, index, "ecs") {
    //ctor
    ATA = "ATA-21";
    /* Define how many input and output ports the code can handle: */
    allowed_inputs = 1;//bleed air and electric
    allowed_outputs = 0;
    /* Define port types */
    expected_input_types = {"Electric"};
    expected_output_types = {""};
    /* Check ports */
    checkPorts();
    /* Runtime Info */
    myRuntimeInfo->out << "  Create: " << Name << "  ->  " << ATA << std::endl;
    /*Initialization*/
    componentsECS.massDucts = 0.0;
    componentsECS.massAirConPack = 0.0;
    componentsECS.massOutlet = 0.0;
    componentsECS.massRamInlet = 0.0;
    componentsECS.massVents = 0.0;
    componentsECS.massMisc = 0.0;
    totalAirflow = 0.0;
}

/* Destructor for electricATA21 */
electricATA21::~electricATA21() {
    //dtor
}

/* Reading system specific parameters of configuration xml file */
void electricATA21::readParameters(const node& config) {
    std::string unit;
    offTakeOff          = config.at("/module_configuration_file/program_settings/systems_constants/environmental_control_system/off_on_take_off/value");
    data_->configuration.checkForValidity(offTakeOff, 0, 1, "offOnTakeOff");
    ecoMode             = config.at("/module_configuration_file/program_settings/systems_constants/environmental_control_system/eco_mode/value");
    data_->configuration.checkForValidity(ecoMode, 0, 1, "ecoMode");
    airflowPerPax       = config.at("/module_configuration_file/program_settings/systems_constants/environmental_control_system/airflow_per_pax/value");
    data_->configuration.checkForValidity(airflowPerPax, 0.0, 0.02, "airflowPerPax");
    Recirculation       = config.at("/module_configuration_file/program_settings/systems_constants/environmental_control_system/recirculation/value");
    data_->configuration.checkForValidity(Recirculation, 0.0, 1.0, "Recirculation");
    heatConvection      = config.at("/module_configuration_file/program_settings/systems_constants/environmental_control_system/heat_convection/value");
    data_->configuration.checkForValidity(heatConvection, -5.0, 0.0, "heatConvection");
    heatSun             = config.at("/module_configuration_file/program_settings/systems_constants/environmental_control_system/heat_sun/value");
    data_->configuration.checkForValidity(heatSun, 0.0, 5000.0, "heatSun");
    heatPerPAX          = config.at("/module_configuration_file/program_settings/systems_constants/environmental_control_system/heat_per_pax/value");
    data_->configuration.checkForValidity(heatPerPAX, 0.0, 500.0, "heatPerPAX");
    heatPerLightLength  = config.at("/module_configuration_file/program_settings/systems_constants/environmental_control_system/heat_per_light_length/value");
    data_->configuration.checkForValidity(heatPerLightLength, 0.0, 500.0, "heatPerLightLength");
    efficiencyFactor    = config.at("/module_configuration_file/program_settings/systems_constants/environmental_control_system/efficiency_factor/value");
    data_->configuration.checkForValidity(efficiencyFactor, 0.0, 1.0, "efficiencyFactor");
    heatCapacityAir     = config.at("/module_configuration_file/program_settings/systems_constants/environmental_control_system/heat_capacity_air/value");
    cabinTemperature    = config.at("/module_configuration_file/program_settings/systems_constants/environmental_control_system/cabin_temperature/value");
    unit = std::string(config.at("/module_configuration_file/program_settings/systems_constants/environmental_control_system/cabin_temperature/unit"));
    if (unit == "K") {
        // No conversion required
    } else if (unit == "C") {
        // Conversion from Celsius to Kelvin
        cabinTemperature += 273.15;
    } else {
        myRuntimeInfo->err << "The unit of the cabin temperature is not known.!" << std::endl;
        myRuntimeInfo->err << "cabin temperature: " << cabinTemperature << " " << unit << std::endl;
        exit(1);
    }
    data_->configuration.checkForValidity(cabinTemperature, 273.15, 310.0, "cabinTemperature");
    temperatureBleedAir         = config.at("/module_configuration_file/program_settings/systems_constants/bleed_air/temperature_bleed_air/value");
    unit = std::string(config.at("/module_configuration_file/program_settings/systems_constants/bleed_air/temperature_bleed_air/unit"));
    if (unit == "K") {
        // No conversion required
    } else if (unit == "C") {
        // conversion from Celsius to Kelvin
        temperatureBleedAir += 273.15;
    } else {
        myRuntimeInfo->err << "The unit of bleed-air temperature is not known.!" << std::endl;
        myRuntimeInfo->err << "Bleed-Air Temperature: " << temperatureBleedAir << " " << unit << std::endl;
        exit(1);
    }
    data_->configuration.checkForValidity(temperatureBleedAir, 0.0, 1000.0, "BleedAir.temperatureBleedAir");
    windowArea          = config.at("/module_configuration_file/program_settings/systems_constants/environmental_control_system/window_area/value");
    unit = std::string(config.at("/module_configuration_file/program_settings/systems_constants/environmental_control_system/window_area/unit"));
    if (unit == "m2") {
        // No conversion required
    } else if (unit == "cm2") {
        // Conversion from cm2 to m2
        windowArea = (1 / pow(100, 2)) * windowArea;
    } else if (unit == "mm2") {
        // Conversion from mm2 to m2
        windowArea = (1 / pow(1000, 2)) * windowArea;
    } else {
        myRuntimeInfo->err << "The unit of the window area is not known.!" << std::endl;
        myRuntimeInfo->err << "window area: " << windowArea << " " << unit << std::endl;
        exit(1);
    }
    data_->configuration.checkForValidity(windowArea, 0.0, 2.0,  "windowArea ");
}

/* Updating system specific parameters in aircraft xml file */
void electricATA21::updateXML() {
    updateMass("air_conditioning", Mass, ATA);
}

/* Calculation of the ECS power */
void electricATA21::calculateECSPower(powerProfile heatSinks, powerProfile heatECS, powerProfile heatConductor) {
    //initialize bleed air power profile (because there is no input port for bleed currently)
    bleedAir.missionPower.clear();
    heatLoad.missionPower.clear();
    bleedAir.designPower = (0.0);
    heatLoad.designPower = (0.0);
    bleedAir.maxMissionPower = (0.0);
    heatLoad.maxMissionPower = (0.0);
    heatLoad.missionPower.clear();
    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        bleedAir.missionPower.push_back(missionLoads());
        heatLoad.missionPower.push_back(missionLoads());
    }
    if (Operating == true) {
        //Accumulate HeatLoads
        std::vector<double> theLoads;
        for (size_t i = 0; i < heatSinks.missionPower.size(); i++) {
            theLoads.push_back(heatSinks.missionPower.at(i).baseLoad
                               + heatECS.missionPower.at(i).baseLoad
                               + heatConductor.missionPower.at(i).baseLoad);
        }
        getPower(theLoads);
        //calculate max mission power for input ports
    for (size_t i(0); i < inputs.size(); i++) {
        calculateMaxMissionPowerPort(inputs.at(i).ID);
    }
    }
}

/* Calculation of the skin mass */
void electricATA21::calculateSinkMass() {
    /* Calculate Systems Mass */
    //has to be adapted since there is no bleed air input
    Mass = getMass(inputs.at(findPort("Bleed", "BleedIn", "input")).designPower);
    refPoint = getRefPoint();
}

/* Determination of the power */
void electricATA21::getPower(std::vector<double> existingHeatLoads) {
    /*Initialization of auxiliary variables from systemsIOData*/
    int numberPAX(data_->data.Accommodation.PAX);
    int numberFlightCrew(data_->data.Accommodation.flightCrew);
    int numberCabinCrew(data_->data.Accommodation.cabinCrew);
    double wettedSurfaceFuselage(0.);
    for (size_t i(0); i < data_->theFuselage.size(); i++) {
        wettedSurfaceFuselage += geom2::measure::area(data_->theFuselage.at(i));
    }
    int nrWindows(data_->data.Accommodation.numberWindows);
    /* Air conidition pack (ACP) */
    double m_ACP = 0.;
    double T_ACP = 0.;
    double P_ACP = 0.;

    /* Design Point */
    double maxSystemHeatLoad = 0.;
    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        if (existingHeatLoads.at(i) > maxSystemHeatLoad) {
            maxSystemHeatLoad = existingHeatLoads.at(i);
        }
    }
    getHeatFluxForACP(&T_ACP, &m_ACP, &P_ACP,
                      0.35 * maxSystemHeatLoad,
                      Recirculation, 24. + 273.15, airflowPerPax,
                      static_cast<int>(1.0 * numberPAX) + numberFlightCrew + numberCabinCrew, numberPAX + numberFlightCrew + numberCabinCrew, heatPerPAX,
                      0., windowArea, nrWindows,
                      heatConvection, wettedSurfaceFuselage, 273.15 + 25.,
                      heatCapacityAir, temperatureBleedAir);
    //bleedAir.designPower  = 1.03 * m_ACP;
    /*Write design loads to ports*/
    //inputs.at(findPort("Bleed", "BleedIn", "input")).designPower = 1.03 * m_ACP;
    bleedAir.designPower = 1.03 * m_ACP;
    inputs.at(findPort("Electric", "ElectricIn", "input")).designPower = getElectricConvECSConsumption();

    m_ACP = 0.;
    T_ACP = 0.;
    P_ACP = 0.;
    /* Determination of mission-dependent powers -> Heat balance of the air conditioning system */
    /* Check if vector of the entered system-side heat flows has the correct length. */
    if (static_cast<int>(data_->mission.Steps.size()) != static_cast<int>(existingHeatLoads.size())) {
        myRuntimeInfo->err << "Vector lenghts of the entered system-side heat flows do not match. Abort!" << std::endl;
        exit(1);
    }

    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        getHeatFluxForACP(&T_ACP, &m_ACP, &P_ACP,
                          0.25 * existingHeatLoads.at(i),
                          Recirculation, cabinTemperature, airflowPerPax,
                          data_->mission.missionUtilization + numberFlightCrew + numberCabinCrew, numberPAX + numberFlightCrew + numberCabinCrew, heatPerPAX,
                          heatSun, windowArea, nrWindows,
                          heatConvection, wettedSurfaceFuselage, data_->mission.Steps.at(i).Point.Temperature,
                          heatCapacityAir, temperatureBleedAir);
        bleedAir.missionPower.at(i).baseLoad = 1.03 * m_ACP;
        //inputs.at(findPort("Bleed", "BleedIn", "input")).missionPower.at(i).baseLoad = 1.03 * m_ACP;
        /* Electrical consumption of fans  */
        inputs.at(findPort("Electric", "ElectricIn", "input")).missionPower.at(i).baseLoad = getElectricConvECSConsumption();

        /* Heat Load of the ECS */
        //accumulate the electrical power to an overall powerProfile
        heatLoad.missionPower.at(i).baseLoad = (1. - efficiencyFactor) * (P_ACP
                                               + 0.1 * inputs.at(findPort("Electric", "ElectricIn", "input")).missionPower.at(i).baseLoad);
        //Switching off the air condition during takeoff
        if (offTakeOff == true && data_->mission.Steps.at(i).rating == "TO" && data_->mission.Steps.at(i).configuration == "TakeOff") {
            //inputs.at(findPort("Bleed", "BleedIn", "input")).missionPower.at(i).baseLoad = 0.0;
            bleedAir.missionPower.at(i).baseLoad = 0.0;
            heatLoad.missionPower.at(i).baseLoad = 0.0;
        }
        /* ECO-Mode: Reduction of Bleed-Air by 25% (Airbus Getting Grips On Fuel Savings) */
        if (ecoMode == true) {
            inputs.at(findPort("Electric", "ElectricIn", "input")).missionPower.at(i).baseLoad *= 0.75;
            //inputs.at(findPort("Bleed", "BleedIn", "input")).missionPower.at(i).baseLoad *= 0.75;
            bleedAir.missionPower.at(i).baseLoad *= 0.75;
            heatLoad.missionPower.at(i).baseLoad *= 0.75;
        }
    }
}

/* Determination of the mass */
double electricATA21::getMass(double massFlowECS) {
    int numberOfEngines     = data_->data.Propulsion.numberEngines;
    int numberPAX           = data_->data.Accommodation.PAX;
    int numberFlightCrew    = data_->data.Accommodation.flightCrew;
    int numberCabinCrew     = data_->data.Accommodation.cabinCrew;
    std::vector<double> massECS;
    double occupants = static_cast<double>(numberCabinCrew + numberFlightCrew + numberPAX);  // Abhaengigkeit der Systemmasse von Passagieren
    occupants += massFlowECS / airflowPerPax / 1.70; // Abhaengigkeit der Systemmasse vom benoetigtem Bleed Air Strom
    occupants = 0.5 * occupants;

    double totalLengthOfDeck(0.0);
    for (size_t j = 0; j < data_->data.Accommodation.lengthDeck.size(); j++) {
        totalLengthOfDeck += data_->data.Accommodation.lengthDeck.at(j);
    }
    massECS.push_back(11.45 * pow((totalLengthOfDeck * pow(static_cast<double>(numberOfEngines), 0.3)), 1.188));
    massECS.push_back(4. * occupants + 120.);
    myRuntimeInfo->debug << "Deck length = " << totalLengthOfDeck << std::endl;
    myRuntimeInfo->debug << "Occupants = " << occupants << std::endl;
    /* NOTE(aigner): Checked Raymer formula (Mantis ID 0000202, UNICADO task T71 https://unicado.ilr.rwth-aachen.de/T71).
                     Raymer masses are given in lbs, therefore a conversion to kg is needed. However, the mass is still much higher as LTH and Howe.
                     Therefore, I would suggest to leave it out for now.*/
    double mECS(0.0);
    for (size_t i = 0; i < massECS.size(); i++) {
        mECS += massECS.at(i);
    }
    mECS = mECS / static_cast<double>(massECS.size());
    /* Percentage distribution of the component masses according to Koeppen */
    componentsECS.massDucts = 0.249 * mECS;
    componentsECS.massAirConPack = 0.230 * mECS;
    componentsECS.massOutlet = 0.076 * mECS;
    componentsECS.massRamInlet = 0.072 * mECS;
    componentsECS.massVents = 0.041 * mECS;
    componentsECS.massMisc = 0.332 * mECS;
    return mECS;
}

/* Determination of the reference point */
Vec3 electricATA21::getRefPoint() {
    /* Assumption: ECS is located in the belly fairing */
    return Vec3(data_->theWing.back().aerodynamicSurface.origin[0] +
                0.5 * data_->theWing.back().aerodynamicSurface.sections.at(0).get_chord_length(),
                data_->theWing.back().aerodynamicSurface.origin[1], data_->theWing.back().aerodynamicSurface.origin[2]);
}

/* Determination of the heat flux for ACP */
void electricATA21::getHeatFluxForACP(double *theACP_temp, double *theACP_mass, double *theACP_power,
        double Q_systems,
        double airRecirculate, double tempCabin, double airflowPAX,
        int Utilization, int maxUtilization, double heatPAX,
        double sunHeat, double windowArea, int numberWindows,
        double convectionHeat, double Surface, double tempSkin,
        double heatCapAir, double tempATA36) {
    /* Define unused function parameters to avoid linter warnings */
    UNUSED(tempATA36);
    /* Mission-independent Heat Fluxes */
    /* Heat flux from sun radiation through windows */
    double Q_sun = sunHeat * windowArea * static_cast<double>(numberWindows);
    /* Heat flux from human body */
    double Q_bio = heatPAX * static_cast<double>(Utilization);
    /* Mission-dependent Heat Fluxes */
    /* Heat flux from convection over aircraft skin */
    double Q_convection = (-1.) * convectionHeat * Surface * (tempSkin - tempCabin);
    /* Parasite heat loads from aircraft systems */
    /* s. function transfer parameters! */
    /* Total Loads */
    double Q_loads = Q_sun + Q_bio + Q_convection + Q_systems;
    /* Minimum fresh air stream */
    double theAirFlow_min = airflowPAX * static_cast<double>(Utilization);
    if (theAirFlow_min < 0.15 * airflowPAX * static_cast<double>(maxUtilization)) {
        theAirFlow_min = 0.15 * airflowPAX * static_cast<double>(maxUtilization);
    }
    /* Required temperature for inflow */
    double theInletTemperature = tempCabin - Q_loads / (heatCapAir * theAirFlow_min);
    /* Required Mass Flow */
    // acc. to SAE ARP 85 minimum temperature Inlet flow into cabin 10 deg Ceslius, maximum temperature +50 deg Celsius.
    if (theInletTemperature < 273.15 + 10.) {
        theInletTemperature = 273.15 + 10.;
    } else if (theInletTemperature > 273.15 + 50.) {
        theInletTemperature = 273.15 + 50.;
    }
    *theACP_mass = Q_loads / (heatCapAir * (tempCabin - theInletTemperature));
    *theACP_temp = (theInletTemperature - airRecirculate * tempCabin) / (1. - airRecirculate);
    *theACP_power = heatCapAir * *theACP_mass * fabs(tempCabin - theInletTemperature);
}

/* Determination of electric conv. of ECS consumption */
double electricATA21::getElectricConvECSConsumption() {
    /*  Method implemented by F. Peter 18.6.2014 no reference, calibrated after A320 ELA's (s. DA Buente p. 58).
        The cabin volume is used since ECS is part of certification and PAX number might change due to seat layout */
    return (data_->data.Accommodation.cabinVolumeTotal * 136.296);
}
