/*  Copyright (C) 2011-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef SYSTEMS_DESIGN_SRC_AIRCRAFTSYSTEMS_ELECTRICATA21_H_
#define SYSTEMS_DESIGN_SRC_AIRCRAFTSYSTEMS_ELECTRICATA21_H_

/* The following statements import necessary libraries */
#include <string>                                                       // Import c++ standard library for using string
#include <vector>                                                       // Import c++ standard library for using vector

/* The following statements import necessary header files */
#include <svl/SVL.h>                              // Import of header file form src folder for SVL
#include <runtimeInfo/runtimeInfo.h>              // Import of header file from src folder for runtimeInfo
#include <memory>
#include "../aircraftSystem.h"                           // Import of header file from src folder for aircraftSystem
#include "../powerProfile.h"                             // Import of header file from src folder for powerProfile
#include "../systemsIOData.h"                            // Import of header file from systemsIOData for accessing data_. Can't be included in
                                                         // aircraftSystem.h because that would lead to circular dependencies
#include <aircraftGeometry2/fuselage.h>
#include <aircraftGeometry2/processing/measure.h>

/** \class electricATA21
 * \brief Contains all properties of the environmental control system class (ATA21)
 */
class electricATA21: public aircraftSystem {
 public:
    class systemComponentsECS {
     public:
        double massDucts; /**< [kg] mass of ducting of the ECS */
        double massAirConPack;/**< [kg] mass of the air conditioning pack */
        double massOutlet; /**< [kg] mass of the outlet */
        double massRamInlet; /**< [kg] mass of the ram inlet */
        double massVents; /** < [kg] mass of the vents */
        double massMisc; /** < [kg] miscellaneous mass */
    };
    systemComponentsECS componentsECS; /**< Contains masses of the ECS components */
    double airflowPerPax;       /**< [lbs/min] airflow per PAX **/
    double Recirculation;       /**< [---] percentage of air being recirculated **/
    double heatConvection;      /**< [W/(m^2*K)] heat convection **/
    double cabinTemperature;    /**< [K] cabin temperature **/
    double heatSun;             /**< [W/m^2] heat from the sun **/
    double windowArea;          /**< [m^2] window area **/
    double heatPerPAX;          /**< [W] heat from each PAX **/
    double heatPerLightLength;  /**< [W/m] heat per meter of lighting **/
    double efficiencyFactor;    /**< [---] efficiency factor of the ECS **/
    double heatCapacityAir;     /**< [J/(kg*K)] heat capacity of the air **/
    bool offTakeOff;            /**< switch if ECS is off during takeoff */
    bool ecoMode;               /**< switch for eco mode of ECS */
    double temperatureBleedAir;    /** [K] temperature of the bleed air **/

    /**
     * \brief function for reading the system specific parameters from the config
     * 
     * \param config const node&: systemsDesign configuration file
    */
    void readParameters(const node& config) override;
    /**
     * \brief function for updating the system specific values in the aircraft xml
    */
    void updateXML() override;
    void calculateECSPower(powerProfile heatSinks, powerProfile heatECS, powerProfile heatConductor) override;
    void calculateSinkMass() override;

    void getPower(std::vector<double> existingHeatLoads);

    double getMass(double massFlowECS);
    Vec3 getRefPoint();

    electricATA21(const std::string& systemName, bool systemIsOperating, std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO, size_t index);
    virtual ~electricATA21();

 private:
    double totalAirflow;/**< [kg/s] total air flow **/

    void getHeatFluxForACP(double *theACP_temp, double *theACP_mass, double *theACP_power,
                           double Q_systems,
                           double airRecirculate, double tempCabin, double airflowPAX,
                           int Utilization, int maxUtilization, double heatPAX,
                           double sunHeat, double windowArea, int numberWindows,
                           double convectionHeat, double Surface, double tempSkin,
                           double heatCapAir, double tempATA36);
    double getElectricConvECSConsumption();
};

#endif // SYSTEMS_DESIGN_SRC_AIRCRAFTSYSTEMS_ELECTRICATA21_H_
