/*  Copyright (C) 2011-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <memory>
#include <string>
#include <algorithm>

#include "fuelCellATA49.h"

#include <unitConversion/constants.h>
#include <standardFiles/functions.h>
#include <runtimeInfo/runtimeInfo.h>

fuelCellATA49::fuelCellATA49(const std::string& systemName, bool systemIsOperating, std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO, size_t index)
    :
    aircraftSystem(systemName, systemIsOperating, data, rtIO, index, "source") {
    ATA = "ATA-49";
    /* Define how many input and output ports the code can handle: */
    allowed_inputs = 1;
    allowed_outputs = 1;
    /* Define port types */
    expected_input_types = {"Hydrogen"};
    expected_output_types = {"Electric"};
    /* Check ports */
    checkPorts();
    myRuntimeInfo->out << "  Create: " << Name << "  -> fuel cell " << ATA << std::endl;
    //ctor
}

fuelCellATA49::~fuelCellATA49() {
    //dtor
}

/* Reading system specific parameters of configuration xml file */
void fuelCellATA49::readParameters(const node& config) {
    /*Component Specific Variables*/
    std::string fuelCellConfig = "/module_configuration_file/program_settings/systems_constants/fuel_cell";
    U_stack_nominal = config.at(fuelCellConfig + "/nominal_voltage/value");
    T_operating = config.at(fuelCellConfig + "/operating_temperature/value");
    p_operating = config.at(fuelCellConfig + "/operating_pressure/value");
    a_CT = config.at(fuelCellConfig + "/charge_transfer_coefficient/value");
    r_spec = config.at(fuelCellConfig + "/area_specific_resistance/value");
    alpha = config.at(fuelCellConfig + "/alpha/value");
    beta = config.at(fuelCellConfig + "/beta/value");
    i_0 = config.at(fuelCellConfig + "/exchange_current_density/value");
    H2_fuelEfficiency = config.at(fuelCellConfig + "/H2_fuel_efficiency/value");
    i_rel_design = config.at(fuelCellConfig + "/relative_current_density_design/value");
    P_spec_grav = config.at(fuelCellConfig + "/grav_power_density/value");
    P_spec_vol = config.at(fuelCellConfig + "/vol_power_density/value");
    //calculated for design point but config value used for mission:
    theDesignVariables.numberOfCellsInStack = config.at(fuelCellConfig + "/number_of_cells_in_stack/value");
    theDesignVariables.P_electric_max = config.at(fuelCellConfig + "/P_electric_max/value");
}

/* Updating system specific parameters in aircraft xml file */
void fuelCellATA49::updateXML() {
    updateMass("fuel_cell", Mass, ATA);
}

void fuelCellATA49::calculateSourcePower() {
    // initializePowerProfile();
    getPower();
    //calculate max mission power for input ports
    for (size_t i(0); i < inputs.size(); i++) {
        calculateMaxMissionPowerPort(inputs.at(i).ID);
    }
}
void fuelCellATA49::calculateSourceMass() {
    Mass = getMass();
    refPoint = getRefPoint();
}

void fuelCellATA49::getPower() {
    U_heatingValue = (T_operating < 100) ? 1.48 : 1.25;
    eta = 0.;
    P_electric = 0.;
    P_thermal = 0.;
    double lower_heating_value_H2 = 118429000;
    /*DESIGN POWER*/
    P_electric = outputs.at(0).designPower;//assumption: only output port is H2
    double i_l = 900.;//[mA/cm^2] limiting current density at the electrode that has the lowest limiting current density ("Fuel Cell Systems Explained" p. 57)
    double i_n = 2.;//[mA/cm^2] sum current density equivalent of fuel crossover and the internal current density ("Fuel Cell Systems Explained" p. 57)
    double p_0(100000.);//[Pa] standard pressure (1 bar = 100kPa is used in Dicks and Rand's "Fuel Cell Systems Explained" p. 36)
    double A_active = 250.;//[cm^2] active area (source: EKPO PEMFC stack module NM12)
    //###########################
    //Calculate reversible open-circuit voltage with values from Dicks and Rand's "Fuel Cell Systems Explained" table 2.2 page 34
    std::vector<double> T_vec({25., 80., 100., 200., 400., 600., 800., 1000.});
    std::vector<double> U_r_0_vec({1.23, 1.18, 1.17, 1.14, 1.09, 1.04, 0.98, 0.92});
    double U_r_0 = interpn(U_r_0_vec, T_vec, this->T_operating - 273.15);//[V] reversible open-circuit voltage
    double I_tmp = P_electric / this->U_stack_nominal;//[A] temporary current in the fuel cell stack
    double i_tmp = this->i_rel_design * i_l;
    double delta(this->calculateDelta(this->T_operating, this->p_operating));//Stoichiometric factor of water. Depends on operating temperature
    double U_r = U_r_0 + GAS_CONSTANT*this->T_operating/(2.*FARADAY_CONSTANT)*log(this->alpha/delta*pow(this->beta, 0.5)*pow(this->p_operating/p_0, 0.5));
    //Calculate temporary design variables
    this->desVars_tmp.clearDesignVariables();//first, clear all "old" design variables
    this->desVars_tmp.numberOfCellsInStack = RoundUp(I_tmp * 1000 / (i_tmp * A_active), 0);
    this->desVars_tmp.P_electric_max = P_electric / this->i_rel_design;
    P_electric_max_design = this->desVars_tmp.P_electric_max;
    //Recalculate i_tmp with new number of cells in stack
    i_tmp = I_tmp * 1000. / A_active / static_cast<double>(this->desVars_tmp.numberOfCellsInStack);
    //calculate cell voltage
    double U_cell = U_r - (i_tmp + i_n)*this->r_spec
                    - GAS_CONSTANT*this->T_operating/(2.*this->a_CT*FARADAY_CONSTANT)*log((i_tmp+i_n)/i_0)
                    + GAS_CONSTANT*this->T_operating/(2.*FARADAY_CONSTANT)*log(1.-(i_tmp+i_n)/i_l);//[V] temporary cell voltage
    this->eta = U_cell / U_heatingValue * this->H2_fuelEfficiency;
    this->P_thermal = P_electric / this->eta;
    double mDot_H2 = this->P_thermal / lower_heating_value_H2;
    inputs.at(0).designPower = mDot_H2;//assumption:only input port is H2
    /*MISSION POWER*/
    for (size_t i=0; i < data_->mission.Steps.size(); i++) {
        P_electric = outputs.at(0).missionPower.at(i).baseLoad;
        double I_tmp = this->P_electric / this->U_stack_nominal;//[A] temporary current in the fuel cell stack
        double i_tmp = std::min(I_tmp * 1000. / (A_active * static_cast<double>(this->desVars_tmp.numberOfCellsInStack)), (i_l - (i_n+1)));
        double delta(this->calculateDelta(this->T_operating, this->p_operating));//Stoichiometric factor of water. Depends on operating temperature
        double U_r = U_r_0 + GAS_CONSTANT*T_operating/(2.*FARADAY_CONSTANT)*log(this->alpha/delta*pow(this->beta, 0.5)*pow(this->p_operating/p_0, 0.5));
        double U_cell = U_r - (i_tmp + i_n)*this->r_spec
                        - GAS_CONSTANT*this->T_operating/(2.*this->a_CT*FARADAY_CONSTANT)*log((i_tmp+i_n)/i_0)
                        + GAS_CONSTANT*this->T_operating/(2.*FARADAY_CONSTANT)*log(1.-(i_tmp+i_n)/i_l);//[V] temporary cell voltage
        this->eta = U_cell * this->H2_fuelEfficiency / U_heatingValue;
        I_tmp = i_tmp / (1000. / (A_active * static_cast<double>(this->desVars_tmp.numberOfCellsInStack)));
        this->P_electric = I_tmp * this->U_stack_nominal;
        this->P_thermal = this->P_electric / this->eta;
        mDot_H2 = this->P_thermal / lower_heating_value_H2;
        inputs.at(0).missionPower.at(i).baseLoad = mDot_H2;//assumption:only input port is H2
        /*PEAK LOADS*/
        for (size_t j=0; j < outputs.at(0).missionPower.at(i).peakLoad.size(); j++) {
            P_electric = outputs.at(0).missionPower.at(i).peakLoad.at(j);
            double I_tmp = this->P_electric / this->U_stack_nominal;//[A] temporary current in the fuel cell stack
            double i_tmp = std::min(I_tmp * 1000. / (A_active * static_cast<double>(this->desVars_tmp.numberOfCellsInStack)), (i_l - (i_n+1)));
            double delta(this->calculateDelta(this->T_operating, this->p_operating));//Stoichiometric factor of water. Depends on operating temperature
            double U_r = U_r_0 + GAS_CONSTANT*T_operating/(2.*FARADAY_CONSTANT)*log(this->alpha/delta*pow(this->beta, 0.5)*pow(this->p_operating/p_0, 0.5));
            double U_cell = U_r - (i_tmp + i_n)*this->r_spec
                            - GAS_CONSTANT*this->T_operating/(2.*this->a_CT*FARADAY_CONSTANT)*log((i_tmp+i_n)/i_0)
                            + GAS_CONSTANT*this->T_operating/(2.*FARADAY_CONSTANT)*log(1.-(i_tmp+i_n)/i_l);//[V] temporary cell voltage
            this->eta = U_cell * this->H2_fuelEfficiency / U_heatingValue;
            I_tmp = i_tmp / (1000. / (A_active * static_cast<double>(this->desVars_tmp.numberOfCellsInStack)));
            this->P_electric = I_tmp * this->U_stack_nominal;
            this->P_thermal = this->P_electric / this->eta;
            mDot_H2 = this->P_thermal / lower_heating_value_H2;
            inputs.at(0).missionPower.at(i).peakLoad.push_back(mDot_H2);//assumption:only input port is H2
            inputs.at(0).missionPower.at(i).timePeak.push_back(outputs.at(0).missionPower.at(i).timePeak.at(j));
        }
    }
}

double fuelCellATA49::getMass() {
    double FCMass = convertUnit(WATT, KILO, WATT, P_electric_max_design) / P_spec_grav;
    return  FCMass;
}

Vec3 fuelCellATA49::getRefPoint() {
    Vec3 cgPoint;
    cgPoint[0] = data_->theFuselage.back().origin[0] + static_cast<double>(FCpositionInFuselageLength * geom2::measure::length(data_->theFuselage.back()));
    cgPoint[1] = data_->theFuselage.back().origin[1];
    cgPoint[2] = data_->theFuselage.back().origin[2];
    return cgPoint;
}

double fuelCellATA49::calculateDelta(double aTemperature, double aPressure) {
    //saturated vapour pressure estimation coefficients 2.5-5 Wagner equation (source: VDI Waermeatlas 2013 p.153)
    double H2O_Wagner_A = -7.86975;
    double H2O_Wagner_B = 1.90561;
    double H2O_Wagner_C = -2.30891;
    double H2O_Wagner_D = -2.06472;
    double H2O_T_crit   = 647.1;//[K] critical water temperature (source: VDI Waermeatlas 2013 p. 176)
    double H2O_P_crit   = 220.64e5;//Pa] critical water pressure (source: VDI Waermeatlas 2013 p. 176)

    if (aTemperature > H2O_T_crit) {
        throw("In \"systemsDesing\" -> \"fuelCellATA49.cpp\" -> \"fuelCellATA49::calculateDelta()\": operating temperature is restricted to critical temperature of T_krit = "
              + num2Str(H2O_T_crit) + " K. Current temperature used is " + num2Str(aTemperature) + " K!");
    }

    double sat_vap_press = H2O_T_crit/aTemperature * (H2O_Wagner_A*(1-aTemperature/H2O_T_crit)
                           + H2O_Wagner_B * pow(1.- aTemperature/H2O_T_crit, 1.5)
                           + H2O_Wagner_C * pow(1.- aTemperature/H2O_T_crit, 2.5)
                           + H2O_Wagner_D * pow(1.- aTemperature/H2O_T_crit, 5.));
    return H2O_P_crit*exp(sat_vap_press) / aPressure;
}

fuelCellATA49::fuelCellDesignVariables::fuelCellDesignVariables()
    :
    P_electric_max(NAN),
    numberOfCellsInStack(0) {
    //ctor
}

fuelCellATA49::fuelCellDesignVariables::~fuelCellDesignVariables() {
    //dtor
}

void fuelCellATA49::fuelCellDesignVariables::clearDesignVariables() {
    P_electric_max = NAN;
    numberOfCellsInStack = 0;
}
