/*  Copyright (C) 2011-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef SYSTEMSDESIGN_SRC_AIRCRAFTSYSTEMS_FUELCELLATA49_H_
#define SYSTEMSDESIGN_SRC_AIRCRAFTSYSTEMS_FUELCELLATA49_H_

/* The following statements import necessary libraries */
#include <string>                                                                  // Import c++ standard library for using string
#include <vector>                                                                  // Import c++ standard library for using vector

/* The following statements import necessary header files */
#include <svl/SVL.h>                              // Import of header file form src folder for SVL
#include <runtimeInfo/runtimeInfo.h>              // Import of header file from src folder for runtimeInfo
#include <memory>
#include "../aircraftSystem.h"                           // Import of header file from src folder for aircraftSystem
#include "../powerProfile.h"                             // Import of header file from src folder for powerProfile
#include "../systemsIOData.h"                            // Import of header file from systemsIOData for accessing data_. Can't be included in
                                                         // aircraftSystem.h because that would lead to circular dependencies

/** \brief contains all properties of a fuel cell auxiliary power unit (ATA85)
 */
class fuelCellATA49: public aircraftSystem {
 public:
    double APUefficiencyFactorBleedAir = 0.0;/**< [-] efficiency factor for bleed air extraction from the APU */
    double APUpercentageBleedAir = 0.0;/**< [-] percentage of bleed air provided by the APU */
    double APUpercentageElectricPower = 0.0;/**< [-] percentage of electric power provided by the APU */
    double APUpercentageHydraulicPower = 0.0;/**< [-] percentage of hydraulic power provided by the APU */
    double FCefficiencyBleedAir = 0.0;/**< [-] efficiency factor for bleed air from fuel cell */
    double FCpercentageBleedAir = 0.0;/**< [-] percentage of bleed air provided by the fuel cell */
    double FCpercentageElectricPower = 0.0;/**< [-] percentage of electric power provided by the fuel cell */
    double FCpercentageHydraulicPower = 0.0;/**< [-] percentage of hydraulic power provided by the fuel cell */
    double FCdesignBleedAir = 0.0;//unused?!
    double FCdesignElectricPower = 0.0;//unused?!
    double FCdesignHydraulicPower = 0.0;//unused?!
    double FCgravimetricPowerDensity = 0.0;/**< gravimetric power density of the fuel cell */ //unused?!
    double FCpositionInFuselageLength = 0.0;/**< [-] fuel cell position relative to total fuselage length */

    double P_electric_max_design = 0.0;/**< [W] maximum electrical output power in the design case */


    /** \brief reads system specific parameters from config
     *
     * \param config const node&
     * \return void
     *
     */
    void readParameters(const node& config) override;

    /** \brief updates system specific parameters in acxml
     * \return void
     */
    void updateXML() override;
    void calculateSourcePower() override;
    void calculateSourceMass() override;

    /** \brief calculates the power of the APU (ATA85)
     *
     * \return void
     *
     */
    void getPower();

    /** \brief calculates the mass of the APU (ATA85)
     * \return double: [kg]
     *
     */
    double getMass();

    /** \brief calculates the reference point of the APU (ATA85)
     *
     * \param fuselageRefPt Vec3: [m] fuselage reference point in global x, y, z-coordinates
     * \param lFuselage double: [m] fuselage length
     * \param relAPU double: [-] relative location of the APU with respect to the fuselage
     * \return Vec3: [m] APU reference point in global x, y, z-coordinates
     *
     */
    Vec3 getRefPoint();

    /** \brief Constructor
     *
     * \param systemName string
     * \param systemIsOperating bool: switch whether the system is operating (1) or not(0)
     * \param theSettings const systemsDesignSettings&
     * \param myAircraft aircraft
     *
     */
    fuelCellATA49(const std::string& systemName, bool systemIsOperating, std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO, size_t index);
    virtual ~fuelCellATA49();

 private:
    /* Constants */
    double U_stack_nominal;/**<[V] nominal fuel cell stack voltage*/
    double T_operating;/**<Component operating temperature */
    double p_operating;/**<Component operating pressure */
    double a_CT;/**<charge transfer coefficient */
    double r_spec;/**<[kOhm/cm^2] area-specific resistance */
    double alpha;/**<Stoichiometric index of hydrogen (1 for pure hydrogen) */
    double beta;/**<Stoichiometric index of oxygen (1 for pure oxygen, 0.21 for air) */
    double i_0;/**<Exchange current density (Dicks: Fuel Cell Systems Explained p. 49) */
    double H2_fuelEfficiency;/**< Efficiency of hydrogen usage*/
    double i_rel_design;/**<relative current density for design point to oversize the fuel cell and increase efficiency */
    double P_spec_grav;/**<[kW/kg] gravimetric power density of the fuel cell*/
    double P_spec_vol;/**<[kW/L] volumetric power density of the fuel cell*/

    /* State variables */
    double U_heatingValue;/**<[V] max. theoretical cell voltage (1.48 for liquid water (higher heating value), 1-.25 for gaseous water (lower heating value) */
    double eta;/**< fuel cell efficiency*/
    double P_electric;/**<[W] fuel cell electric output power*/
    double P_thermal;/**<[W] fuel cell input thermal power*/

    /* Design variables */
    class fuelCellDesignVariables {
     public:
        double P_electric_max;/**<[W] maximum electrical output power */
        int numberOfCellsInStack;/**number of individual cells in fuel cell stack */

        void clearDesignVariables();

        fuelCellDesignVariables();/**<Constructor*/
        virtual ~fuelCellDesignVariables();/**<Constructor*/
    };
    fuelCellDesignVariables theDesignVariables;
    fuelCellDesignVariables desVars_tmp;
    std::vector<fuelCellDesignVariables> desVars_vec;


    /** \brief Function that calculates the stoichiometric index of water in the exhaust gas
     *
     * \param aTemperature double: [K] fuel cell operating temperature
     * \param aPressure double: [Pa] fuel cell operating pressure
     * \return double delta: Stoichiometric index of water
     *
     */
    double calculateDelta(double aTemperature, double aPressure);
};

#endif // SYSTEMSDESIGN_SRC_AIRCRAFTSYSTEMS_FUELCELLATA49_H_
