/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "PowerControlUnit.h"

#include <cmath>

#include <runtimeInfo/runtimeInfo.h>

PowerControlUnit::PowerControlUnit(const std::string& Type)
    :
    Mass(0.0),
    installedPower(0.0) {
    // NOTE (aigner/peter#1#): Efficiency factors and rotation speed from [Gro09]] p.89 resp. [Cha10] p. 46
    if (Type == "Slat") {
        efficiencyFactorMin = 0.729525;
        efficiencyFactorMax = 0.953246;
        safetyFactor = 1.15;
        rotationSpeedPCU = 1398 * 1 / 60; //in Sekunde
        hydraulicPressureMin = 135; //in bar
        hydraulicPressureMax = 207; //in bar
    } else if (Type == "Fowler") {
        efficiencyFactorMin = 0.707625;
        efficiencyFactorMax = 0.92463;
        safetyFactor = 1.15;
        rotationSpeedPCU = 1080 * 1 / 60; //in Sekunde
        hydraulicPressureMin = 135; //in bar
        hydraulicPressureMax = 207; //in bar
    } else {
        myRuntimeInfo->warn << "The system type \"" << Type << "\" is unknown. No power control unit can be created!" << std::endl;
    }
    //ctor
}

PowerControlUnit::~PowerControlUnit() {
    //dtor
}

void PowerControlUnit::getInstalledPower(double theMTOM, double SrefWing) {
    // NOTE (aigner/peter#1#): Functions and parameters added to installedPower by Lammering in SVN rev2131 March 2013
    double theX     = theMTOM * SrefWing * 0.000001;
    installedPower  = 1000. * (0.665 * pow(theX, 0.8397) + 2.5);  //in Watt
}

void PowerControlUnit::getMass(const std::string& deviceType, double SrefWing) {
    // NOTE (aigner/peter#1#): methods changed by Lammering in SVN rev2131 March 2013
    if (deviceType == "Slat" || deviceType == "Fowler") {
        Mass  = 11.849 * pow((installedPower / 1000.), 0.3527);
    } else {
        myRuntimeInfo->warn << "The system type \"" << deviceType << "\" is unknown. Cannot calculate PCU mass!" << std::endl;
    }

// NOTE (aigner/peter#1#): Former method from [Cha10] p.69-70
//    if ( deviceType == "Slat" )
//    {
//        //Mass = 40.0;
//        Mass = 0.75 * (15.99 * log(SrefWing) - 49.43);
//        if(Mass <= 0.0)
//        {
//           Mass = 0.75 * 15.99 * log(SrefWing);
//        }
//
//
//    }
//    else if (deviceType == "Fowler" )
//    {
//        Mass = 0.75 * (15.99 * log(SrefWing) - 49.43);
//        if(Mass <= 0.0)
//        {
//           Mass = 0.75 * 15.99 * log(SrefWing);
//        }
//    }
//    else {
//        myRuntimeInfo->warn << "The system type \"" << deviceType << "\" is unknown. Cannot calculate PCU mass!" << std::endl;
//    }
}
