/**
 * \file interfaceSystem.cpp
 * \author Ellen Seabrooke
 * \brief Interface system class to represent the power supply between the A/C and the onboard systems
 * \version 0.1
 * \date 2024-04-17
 * 
 * @copyright Copyright (c) 2024
 * 
 */

#include "interfaceSystem.h"
#include "systemsIOData.h"

interfaceSystem::interfaceSystem(const std::string& systemName, bool systemIsOperating,
                                 std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO, size_t index)
    :
    aircraftSystem(systemName, systemIsOperating, data, rtIO, index, "interface") {
    ATA = "ATA-00";
    /* Define how many input and output ports the code can handle: */
    allowed_inputs = 0;
    allowed_outputs = 1;
    /* Define port types */
    expected_input_types = {""};
    //lists possible outputs of all interfaces. Actual interface type is matched in getInterface()
    expected_output_types = {"Bleed", "ShaftPower", "Electric", "Kerosene", "Hydrogen"};
    /* Runtime Info */
    myRuntimeInfo->out << "  Create: " << Name << "  ->  " << ATA << std::endl;
    //ctor
}

void interfaceSystem::getPower() {
    //ouput ports are written through connections. But if max and average power are required for acxml, these functions are necessary
    if (outputs.at(0).type == "ShaftPower") {
        getShaftPower();
    } else if (outputs.at(0).type == "Bleed") {
        getBleedAirPower();
    }
}

void interfaceSystem::getShaftPower() {
    /* Only source connected to bleed air interface is engine (APU: kerosene interface) */
    //find output of type bleed
    int i = findPort("ShaftPower", "ShaftPowerOut", "output");
    /* check if the source is the engine */
    if (outputs.at(i).type == "Engine") {//todo: check if this works!
        double maxShaftOfftakes_tmp = 0.;
        int numberOfCruiseSteps = 0;
        for (size_t j = 0; j < data_->mission.Steps.size(); j++) {
            if (data_->mission.Steps.at(j).missionPhase == "Cruise") {
                /* Accumulation of shaft power offtakes during cruise for the calculation of the average shaft power offtakes during cruise */
                averageShaftOfftakesEngine += outputs.at(i).missionPower.at(j).baseLoad;
                numberOfCruiseSteps++;
            }
            //Catching power peaks that are too high ...
            double remainingPower = data_->data.Propulsion.maxShaftPowerOffTakes - outputs.at(i).missionPower.at(j).baseLoad;
            for (size_t ID = 0; ID < outputs.at(i).missionPower.at(j).peakLoad.size(); ID++) {
                if (remainingPower > 0. && outputs.at(i).missionPower.at(j).peakLoad.at(ID) > remainingPower) {
                    double factor = outputs.at(i).missionPower.at(j).peakLoad.at(ID) / remainingPower;
                    outputs.at(i).missionPower.at(j).peakLoad.at(ID) = outputs.at(i).missionPower.at(j).peakLoad.at(ID) / factor;
                    outputs.at(i).missionPower.at(j).timePeak.at(ID) = outputs.at(i).missionPower.at(j).timePeak.at(ID) * factor;
                }
                //find highest peak
                maxShaftOfftakes_tmp = outputs.at(i).missionPower.at(j).peakLoad.at(ID);
                if (maxShaftOfftakes_tmp > maxShaftOfftakesEngine) {
                    maxShaftOfftakesEngine = maxShaftOfftakes_tmp;
                }
            }
        }
        /*To get average shaft power offtakes in cruise: devide the accumulated shaft power offtake during cruise by the number of cruise steps*/
        averageShaftOfftakesEngine /= numberOfCruiseSteps;
    }
}

void interfaceSystem::getBleedAirPower() {
    /* Only source connected to bleed air interface is engine (APU: kerosene interface) */
    //find output of type bleed
    int i = findPort("Bleed", "BleedOut", "output");
    double maxBleedAirOfftakes_tmp = 0.;
    int numberOfCruiseSteps = 0;
    for (size_t j = 0; j < data_->mission.Steps.size(); j++) {
        if (data_->mission.Steps.at(j).missionPhase == "Cruise") {
            /* Accumulation of shaft power offtakes during cruise for the calculation of the average shaft power offtakes during cruise */
            averageBleedAirOfftakesEngine += outputs.at(i).missionPower.at(j).baseLoad;
            numberOfCruiseSteps++;
        }
        //Catching power peaks that are too high ...
        //TODO(seabrooke): uncomment once max bleed air offtakes is available
        // double remainingPower = data_->data.Propulsion.maxBleedAirOfftakes - outputs.at(i).missionPower.at(j).baseLoad;
        for (size_t ID = 0; ID < outputs.at(i).missionPower.at(j).peakLoad.size(); ID++) {
        //     if (remainingPower > 0. && outputs.at(i).missionPower.at(j).peakLoad.at(ID) > remainingPower) {
        //         double factor = outputs.at(i).missionPower.at(j).peakLoad.at(ID) / remainingPower;
        //         outputs.at(i).missionPower.at(j).peakLoad.at(ID) = outputs.at(i).missionPower.at(j).peakLoad.at(ID) / factor;
        //         outputs.at(i).missionPower.at(j).timePeak.at(ID) = outputs.at(i).missionPower.at(j).timePeak.at(ID) * factor;
        //     }
            /* Find highest peak */
            maxBleedAirOfftakes_tmp = outputs.at(i).missionPower.at(j).peakLoad.at(ID);
            if (maxBleedAirOfftakes_tmp > maxBleedAirOfftakesEngine) {
                maxBleedAirOfftakesEngine = maxBleedAirOfftakes_tmp;
            }
        }
    }
    /*To get average bleed air offtakes in cruise: devide the accumulated bleed air offtake during cruise by the number of cruise steps*/
    averageBleedAirOfftakesEngine /= numberOfCruiseSteps;
}

void interfaceSystem::updateMissionXml(node& missionFile) {
    /* Choose node name, unit and description according to interface system type */
    std::string node_name;
    std::string unit;
    std::string description;
    if (outputs.at(0).type == "ShaftPower") {
        node_name = "shaft_power_takeoff_schedule";
        unit = "W";
        description = "shaft power offtake from the engine";
    } else if (outputs.at(0).type == "Bleed") {
        node_name = "bleed_air_takeoff_schedule";
        unit = "kg/s";
        description = "bleed air offtake from the engine";
    } else if (outputs.at(0).type == "Kerosene") {
        node_name = "specific_kerosene_consumption";
        unit = "kg/s";
        description = "fuel (kerosene) consumption of systems";
    } else if (outputs.at(0).type == "Electric") {
        //myRuntimeInfo->err << "updateMissionXml not yet defined for an interface system of the type " << type << std::endl;
        //exit(1);
        node_name = "electric_power";
        unit = "W";
        description = "electric power offtake from battery";
    } else if (outputs.at(0).type == "Hydrogen") {
        node_name = "specific_hydrogen_consumption";
        unit = "kg/s";
        description = "fuel (hydrogen) consumption of systems";
    } else {
        myRuntimeInfo->err << "No prameters defined in interfaceSystem::updateMissionXml to update the type " << outputs.at(0).type << std::endl;
        exit(1);
    }
    /* Offtakes at mission steps */
    std::string xmlString;
    int departureStep(0);
    int cruiseStep(0);
    int approachStep(0);
    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        /* Determine xml path according to current mission phase */
        std::stringstream temp1;
        temp1.str("");
        if (data_->mission.Steps.at(i).missionPhase == "Departure") {
            temp1 << departureStep;
            std::string str_id = temp1.str();
            xmlString = "mission/departure/departure_step@" + str_id;
            departureStep++;
        } else if (data_->mission.Steps.at(i).missionPhase == "Cruise") {
            temp1 << cruiseStep;
            std::string str_id = temp1.str();
            xmlString = "mission/cruise/cruise_phase@" + str_id;
            cruiseStep++;
        } else if (data_->mission.Steps.at(i).missionPhase == "Approach") {
            temp1 << approachStep;
            std::string str_id = temp1.str();
            xmlString = "mission/approach/approach_step@" + str_id;
            approachStep++;
        }
        /* Create node for the step if not existent yet*/
        std::string subPath = xmlString + "/" + node_name;
        if (!(missionFile.find(subPath))) {
            missionFile[subPath];
        }
        if (!outputs.empty()) { //if all sources are own then offtakes vector will be empty
            // Base Load
            // sum up base load of all outputs
            double tmp_base_load = 0.0;
            for (size_t n(0); n < outputs.size(); n++) {
                tmp_base_load += outputs.at(n).missionPower.at(i).baseLoad;
            }
            //write to xml
            Endnode<double> base_load = data_->data.create_endnode(tmp_base_load, unit, subPath + "/background_power_level", description);
            base_load.update(missionFile);
            // Peak Loads
            // write peak loads of each output
            for (size_t n(0); n < outputs.size(); n++) {
                for (size_t j = 0; j < outputs.at(n).missionPower.at(i).peakLoad.size(); j++) {
                    missionFile.at(subPath + "/").appendChild("takeoff_schedule_step@" + num2Str(j));
                    Endnode<double> power = data_->data.create_endnode(outputs.at(n).missionPower.at(i).peakLoad.at(j), unit,
                        subPath + "/takeoff_schedule_step@" + num2Str(j) + "/power", "peak power");
                    power.update(missionFile);
                    Endnode<double> time = data_->data.create_endnode(outputs.at(n).missionPower.at(i).timePeak.at(j), unit,
                        subPath + "/takeoff_schedule_step@" + num2Str(j) + "/time", "duration of the peak");
                    time.update(missionFile);
                }
            }
        }
    }
}

void interfaceSystem::updateACxml(node& acxml) {
    if (outputs.at(0).type == "ShaftPower") {
        if (!data_->configuration.reqMission) {
            std::string subPath = "aircraft_exchange_file/component_design/systems/specific";
            if (!(acxml.find(subPath))) {
                acxml[subPath];
            }
            /* Average offtake */
            Endnode<double> avgCurisePowerDemand = data_->data.create_endnode(averageShaftOfftakesEngine, "W", subPath + "/average_power_demand",
                                                                       "average power (hydraulic + electric) demand");
            avgCurisePowerDemand.set_boundaries(0, INFINITY);
            avgCurisePowerDemand.update(acxml);
            /* Maximum offtake */
            Endnode<double> maxPower = data_->data.create_endnode(maxShaftOfftakesEngine, "W", subPath + "/maximium_power_demand",
                                                      "maximum power (hydraulic + electric) demand");
            maxPower.set_boundaries(0, INFINITY);
            maxPower.update(acxml);
        }
    } else if (outputs.at(0).type == "Bleed") {
        if (!data_->configuration.reqMission) {
            std::string subPath = "aircraft_exchange_file/component_design/systems/specific";
            if (!(acxml.find(subPath))) {
                acxml[subPath];
            }
            /* Average offtake */
            Endnode<double> avgCuriseBleedAirDemand = data_->data.create_endnode(averageBleedAirOfftakesEngine, "W", subPath + "/average_bleed_air_demand",
                                                                       "average bleed air demand");
            avgCuriseBleedAirDemand.set_boundaries(0, INFINITY);
            avgCuriseBleedAirDemand.update(acxml);
            /* Maximum offtake */
            Endnode<double> maxBleed = data_->data.create_endnode(maxBleedAirOfftakesEngine, "W", subPath + "/maximium_bleed_air_demand",
                                                      "maximum bleed air demand");
            maxBleed.set_boundaries(0, INFINITY);
            maxBleed.update(acxml);
        }
    } else if (outputs.at(0).type == "Kerosene") {
        myRuntimeInfo->warn << "updateACxml not yet defined for an interface system of the type " << outputs.at(0).type
                            << " and will be skipped." << std::endl;
    } else if (outputs.at(0).type == "Electric") {
        myRuntimeInfo->err << "updateACxml not yet defined for an interface system of the type " << outputs.at(0).type
                            << " and will be skipped." << std::endl;
    }
}

