/**
 * \file interfaceSystem.h
 * \author Ellen Seabrooke
 * \brief Interface system class to represent the power supply between the A/C and the onboard systems
 * \version 0.1
 * \date 2024-04-17
 * 
 * @copyright Copyright (c) 2024
 * 
 */

#ifndef SYSTEMS_DESIGN_SRC_INTERFACESYSTEM_H_
#define SYSTEMS_DESIGN_SRC_INTERFACESYSTEM_H_

#include <string>
#include <vector>
#include <memory>
#include <aixml/endnode.h>
#include <moduleBasics/runtimeIO.h>
#include "powerProfile.h"
#include "aircraftSystem.h"

class systemsIOData;

class interfaceSystem: public aircraftSystem {
 protected:
    //std::shared_ptr<systemsIOData> data_;/*pointer used for exchange of power profiles between data/ATA70 and this*/
    //std::shared_ptr<RuntimeIO> rtIO_;/*pointer useful for acxml access for update functions*/

 public:
    /** \brief Constructor
    * \param systemName string
    * \param systemIsOperating bool: switch whether the system is operating (1) or not(0)
    * \param data pointer to IOData
    */
    interfaceSystem(const std::string& systemName, bool systemIsOperating,
                                 std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO, size_t index);

    /** \brief Calls specific getPower function according to the interface system type
     *  \return void
    */
    void getPower();
    void getShaftPower();
    void getBleedAirPower();
    void getKerosenePower();
    void getElectricPower();

    /** \brief update the acxml file
     * \param acxml aircraft xml
     * \return void
    */
    void updateACxml(node& acxml);

    /** \brief Function to update the mission.xml file
     * \param missionFile pointer to mission file node
     * \return void
    */
    void updateMissionXml(node& missionFile);

    /*Store offtakes from engine*/
    //std::vector<missionLoads> shaftOffTakesEngine;/*<[W] engine shaft offtakes*/
    double maxShaftOfftakesEngine = 0.;/*<[W] maximum shaft power offtake*/
    double averageShaftOfftakesEngine = 0.;/*<[W] average shaft power offtakes over the complete mission*/
    //std::vector<missionLoads> bleedAirOffTakesEngine;/*<[kg/s] engine bleed air offtakes*/
    double averageBleedAirOfftakesEngine = 0.;/*<[kg/s] average bleed air offtakes from the engine*/
    double maxBleedAirOfftakesEngine = 0.;/*<[kg/s] maximum bleed air offtakes from the engine*/
};
#endif //SYSTEMS_DESIGN_SRC_INTERFACESYSTEM_H_
