/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef SYSTEMS_DESIGN_SRC_STANDARDCONFIG_H_
#define SYSTEMS_DESIGN_SRC_STANDARDCONFIG_H_

#include <string>
#include <vector>
#include <svl/SVL.h>
#include <memory>

#include <moduleBasics/runtimeIO.h>
#include "aircraftSystem.h"
#include "toolinfo.h"

using std::vector;
using std::string;
using std::endl;

/* STANDARD SYSTEMS DESIGN CONFIG */
/**
 * \brief Storage class for non system specific values from configuration xml
 * 
 */
class standardConfig {
 public:
    //Typedefs, Classes, Enums
    explicit standardConfig(const node& config) {
        std::string mission_mode = EndnodeReadOnly<std::string>("/module_configuration_file/program_settings/mission_mode").read(config).value();
        if (mission_mode == "design") {
            designCase = true;
        } else if (mission_mode == "study") {
            //study mode is always the third option and does not have a variable to check for it
        } else if (mission_mode == "requirement") {
            reqMission = true;
        } else {
            myRuntimeInfo->err << "Unknown mission file type: " << mission_mode << std::endl;
            exit(1);
        }
    }

    /** \brief class that describes a port from config file
     */
    class port {
     public:
        class distribution {
         public:
            std::string values;
            std::string sourcePortUID;
            std::string sourceSystemUID;
        };
        std::string type;/**port type (e.g., Electric, Hydraulic,...)*/
        std::string ID;/**port ID string*/
        std::vector<distribution> distributions;/**Input distributions, relevant if more than one input connection is specified*/

        void writePort(node* theNode, string direction);  // function fills the attributes of port object with information from port node
        string checkPortType(string type);  // check if type is correct (Hydraulic, Electric, ...). If no, throw error. If yes, return type
    };

    /** \brief class that contains the description of a system from the config file (name,
     * operating status, if it uses fuel)
     */
    class systemDescription {
     public:
        std::string systemName;/**name of the system*/
        bool systemOperating;/**< switch whether system is operating (1) or not (0) */
        //bool systemUsesFuel;/**switch whether the system uses fuel*/
        // neu
        std::string ATA;/**ATA-chapter of system*/
        std::string systemClass;/**type of system/ class for this system*/
        std::vector<port> inputs;/**input ports*/
        std::vector<port> outputs;/**output ports*/

        void writeDescription(node* theNode);

        systemDescription()
        :
        systemName(""),
        systemOperating(false) {}
    };

    /* Energy sources for shaft power consumption are defined by the user for certain systems.*/
    /** \brief class for the power source systems
     */
    class PowerSource {
     public:
        std::string type;/**Type of Power Source */
        std::string sourceID;/**< Source_ID of Power source*/
        double operationFactor;/**< operation factor of the power source. Can be in the range [0,1] */

        PowerSource()
        :
        type(""), /**Type of Power Source */
        sourceID(""), /**< Source_ID of Power source in string format */
        operationFactor(0.) {} /**< operation factor of the power source. Can be in the range [0,1] */
    };

    /** \brief special system properties for virtual systems
     */
    class virtSystemProperties {
     public:
        std::string systemName;/**<System Name */
        bool systemOperating;/**< operational status*/

        double Mass;/**< system mass*/
        Vec3 refPoint;/**< center of gravity of the system*/

        std::vector<port> inputs;/**input ports*/
        Vec3 powerBleed;/**< bleed power required by the system*/
        Vec3 powerHydraulic;/**< hydraulic power required by the system*/
        Vec3 powerElectric;/**< electric power required by the system*/
        Vec3 heatLoad;/**< heat load of the system*/

        virtSystemProperties()
        :
        systemName(""), /**<System Name */
        systemOperating(false), /**< operational status*/
        Mass(0.0), /**< system mass*/
        refPoint({0.0, 0.0, 0.0}), /**< reference point of the system*/
        powerBleed({0.0, 0.0, 0.0}), /**< bleed power required by the system*/
        powerHydraulic({0.0, 0.0, 0.0}), /**< hydraulic power required by the system*/
        powerElectric({0.0, 0.0, 0.0}), /**< electric power required by the system*/
        heatLoad({0.0, 0.0, 0.0}) {} /**< heat load of the system*/
    };

    /** \brief Detailed description of the pumps
    */
    class hydraulicPump {
     public:
        std::string name;/**<Name of the pump*/
        std::string type;/**<Type of pump (e.g. EngineDriven, Electric)*/
        PowerSource powerSource;/*power source of the pump*/
        double efficiency;/**<Pump efficiency*/
        double operationFactor;/**<Percentage of total pump performance in normal operation*/

        hydraulicPump()
        :
        name(""),
        type(""),
        efficiency(1.),
        operationFactor(0.) {}
    };

    /** \brief Detailed description of the hydraulic circuits
     */
    class hydraulicCircuit {
     public:
        Vec3 compartmentRefPt;/*compartment reference point*/
        int numberOfPumps;/**<Number of hydraulic pumps*/

        hydraulicCircuit() {
            compartmentRefPt[0] = 0.;
            compartmentRefPt[1] = 0.;
            compartmentRefPt[2] = 0.;
            numberOfPumps = 0;
        }
        std::vector<hydraulicPump> pumps;/*vector containing the pumps of the hydraulic circuit*/
    };

    /** \brief contains the properties of the hydraulic system (ATA29)
     */
    class hydraulicSystem {
     public:
        int numberOfhydraulicCircuits;/**< Number of hydraulic circuits*/
        std::vector<hydraulicCircuit> circuits;/*vector containing the individual hyd. circuits*/
    };

    /** \brief contains the properties of the electric system (ATA24)
     */
    class electricSystem {
     public:
        int numberOfElectricCircuits;/**<Number of electric circuits*/

        /** \brief Detailed description of the electrical circuits
         */
        class electricCircuit {
         public:
            int numberOfGenerators;/**<Number of generators*/

            electricCircuit() {
                numberOfGenerators = 0;
            }

            /** \brief Detailed description of the generators
             */
            class generator {
             public:
                std::string name;/**<Name of Generator*/
                std::string type;/**<Type of Generator (e.g. IDG, APUG)*/
                PowerSource powerSource;/**<Power Source of the generator*/
                double efficiency;/**<efficiency of Generators*/
                double operationFactor;/**<Percentage of total performance in normal operation*/


                generator()
                :
                name(""),
                type(""),
                efficiency(1.),
                operationFactor(0.) {}
            };
            std::vector<generator> generators;/*vector containing the generators of the electric circuit*/
        };
        std::vector<electricCircuit> circuits;/*vector containing all electric circuits*/
    };

    /* Special program settings if necessary **/
    bool texReportWithoutMasses;/*if this is true, the TeX report is written only with approvals*/
    bool controlDeviceWarning;/*if this is true a warning is issued if no suitable method is available for ControlDevice*/

    bool designCase = false;/**< Determine whether system architecture is also sized **/
    bool reqMission = false;/**< Determining the offtakes for Requirement Mission **/

    double OpItemsFactor;/**< [-] Scaling Factor for Op. Items Masses **/

    int numberOfVirtualSystems;/*number of virtual systems*/
    int numberOfConductors;/* number of conductor systems*/
    int numberOfSources;/*number of source systems*/

    std::string theProgramName; /**< Name of systems design program depending on design or study case **/

    std::vector<systemDescription> Sinks;/*vector containing the system descriptions of the sinks*/
    systemDescription theECS;/*vector containing the system description of the environmental control systems*/
    std::vector<systemDescription> Sources;/*vector containing the system descriptions of the sources*/
    std::vector<systemDescription> Conductors;/*vector containing the system descriptions of the conductors*/
    std::vector<systemDescription> Interfaces; /*vector contianing the system description of the interface systems*/
    std::vector<virtSystemProperties> virtualSystem;/*vector containing the system descriptions of the virtual system*/
    hydraulicSystem Hydraulic;//Hydraulic System
    electricSystem Electric;//Electric System

    /** \brief function to read all general information from the config
     * \attention This function does not read system specific parameters from the config.
     * Reading those is implemented in the class of each specific system. Rather this function
     * reads all general settings and the system architecture.
     * \return void
     */
    void readAll(const node& config) {
        readProgramSettings(config);
        // here: new method
        readSystems(config);
        // readEnergySources(config);
        // readEnergySinks(config);
        // readECS(config);
        // readEnergyConductors(config);
        readVirtualSystems(config);
    }

    /** \brief function to check a value for validity within given borders
     * 
     * \param value value of the parameter to be checked
     * \param lowerBound lower boundary for validity of the value
     * \param upperBound upper boundary for validity of the value
     * \param Text name of the parameter for reference
     * 
     * \return void
    */
    void checkForValidity(double Value, double lowerBound, double upperBound, const std::string& Text);

 private:
    /** \brief function to read all systems that are energy sources from the config file
     *  \param config config file as const node&
     *  \return void
    */
    void readEnergySources(const node& config);

    /** \brief function to read all systems that are energy sinks from the config file
     *  \param config config file as const node&
     *  \return void
    */
    void readEnergySinks(const node& config);

    /** \brief function to read the environmental control system from the config file
     *  \param config config file as const node&
     *  \return void
    */
    void readECS(const node& config);

    /** \brief function to read all systems that are energy conductors from the config file
     *  \param config config file as const node&
     *  \return void
    */
    void readEnergyConductors(const node& config);

    /** \brief function to read all systems from the config file
     *  \param config config file as const node&
     *  \return void
    */
    void readSystems(const node& config);

    /** \brief function to read all virtual systems from the config file
     *  \param config config file as const node&
     *  \return void
    */
    void readVirtualSystems(const node& config);

    /** \brief function to read the program settings from the config file
     *  \param config config file as const node&
     *  \return void
    */
    void readProgramSettings(const node& config);
};

#endif // SYSTEMS_DESIGN_SRC_STANDARDCONFIG_H_
