/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 * 
 * Description:
 * This file is part of UNICADO.
 */

#include <string>
#include "standardSystemsDesign.h"
#include "toolinfo.h"
#include <moduleBasics/plot.h>
#include <matplot/matplot.h>
#include <cstdlib>

void standardSystemsDesign::generatePlotData() {
    if (data_->configuration.designCase) {
        this->generatePlotSystemsMass();
    }
    for (size_t n = 0; n < data_->Systems.energySink.size(); n++) {
        this->generatePlotMissionEnergy(*data_->Systems.energySink.at(n));
    }
    for (size_t n = 0; n < data_->Systems.virtualSystem.size(); n++) {
        this->generatePlotMissionEnergy(*data_->Systems.virtualSystem.at(n));
    }
    for (size_t n = 0; n < data_->Systems.energyConductor.size(); n++) {
        this->generatePlotMissionEnergy(*data_->Systems.energyConductor.at(n));
    }
    this->generatePlotDataMission();
    this->generatePlotMissionEnergyAccumulated();
    for (size_t n = 0; n < data_->Systems.energySource.size(); n++) {
        this->generatePlotMissionEnergy(*data_->Systems.energySource.at(n));
    }
    this->generatePlotEnergySinksMaxEnergy();
}

void standardSystemsDesign::generatePlotSystemsMass() {
    std::ofstream plot;
    std::string plotDataFile(rtIO_->getCsvFilesDir() + "/" + data_->configuration.theProgramName + "_systems_mass_plot.csv");
    plot.open(plotDataFile.c_str());
    if (!plot) {
        myRuntimeInfo->err << std::string(data_->configuration.theProgramName) + "_systems_mass_plot.csv konnte nicht geoeffnet werden!" << std::endl;
        exit(1);
    } else {
        plot << "# " + win2lin(rtIO_->getCsvFilesDir()) + "/" + data_->configuration.theProgramName + "_systems_mass_plot.csv" << std::endl;
        /* Adjust data structure: **/
        plot << "# Mass of the individual systems" << std::endl;
        plot << "# (1)ATA; (2) Name; (3) Mass [kg];" << std::endl;
    }
    // Define the systems and their names in corresponding vectors
    std::vector<std::string> potentialComponents = {
        "ATA-21", "ATA-24", "ATA-25", "ATA-26", "ATA-27", "ATA-28", "ATA-29",
        "ATA-30", "ATA-33", "ATA-35", "ATA-36", "ATA-85", "ATA-49", "ATA-XX"
    };

    std::vector<std::string> correspondingNames = {
        "Air Conditioning", "Electric System", "Furnishing", "Fire Protection",
        "Flight Control", "Fuel System", "Hydraulic System", "Anti-Icing", "Lighting",
        "Oxygen", "Bleed Air", "FC", "APU", "Remaining consumers including Avionics"
    };

    // Loop through systems and only plot if they exist
    for (size_t i = 0; i < potentialComponents.size(); ++i) {
        const std::string& system = potentialComponents[i];
        if (data_->Systems.containsSystem(system)) {
            plot << system << ";" << correspondingNames[i] << ";" << data_->Systems.allSystems.at(data_->Systems.getSpecificSystem(system))->Mass << std::endl;
        }
    }
    // Prepare data for plotting

    std::vector<std::string> components;
    std::vector<std::string> names;
    std::vector<double> masses;

    for (size_t i = 0; i < potentialComponents.size(); ++i) {
        const std::string& system = potentialComponents[i];
        if (data_->Systems.containsSystem(system)) {
            components.push_back(system);
            names.push_back(correspondingNames[i]);
            masses.push_back(data_->Systems.allSystems.at(data_->Systems.getSpecificSystem(system))->Mass);
        }
    }

    // Plot with matplotplusplus
    auto fig = matplot::figure(true);
    auto ax = fig->current_axes();
    ax->bar(masses);
    std::vector<double> ticks(masses.size());
    std::iota(ticks.begin(), ticks.end(), 1.0);

    ax->title("Systems Mass");
    ax->xlabel("Component");
    ax->ylabel("Mass [kg]");
    ax->x_axis().tickangle(45);
    int lim = masses.size() + 1;
    ax->x_axis().limits({0, static_cast<double>(lim)});
    ax->xticks(ticks);
    ax->x_axis().ticklabels(components);
    ax->ylim({0, 3000});
    ax->y_axis().tick_label_format("%4.0f");
    ax->grid(true);

    // Save the plot
    std::string plotFilePath = rtIO_->getPlotDir() + "/" + data_->configuration.theProgramName + "_systems_mass_plot.svg";
    matplot::save(plotFilePath);
}

void standardSystemsDesign::generatePlotDataMission() {
    std::ofstream plot;
    std::string plotDataFile(rtIO_->getCsvFilesDir() + "/" + data_->configuration.theProgramName + "_mission_steps_plot.csv");
    plot.open(plotDataFile.c_str());
    if (!plot) {
        myRuntimeInfo->err << std::string(data_->configuration.theProgramName) + "_mission_steps_plot.csv could not be opened!" << std::endl;
        exit(1);
    } else {
        plot << "# " + win2lin(rtIO_->getCsvFilesDir()) + "/" + data_->configuration.theProgramName + "_mission_steps_plot.csv" << std::endl;
        /* Adjust data structure: **/
        plot << "# Mission profile from the mission XML file" << std::endl;
        plot << "# (1)Step; (2)Missionsphase; (3)Missionsmode; (4)endAltitude; (5)endFL; (6)endCAS;"
             << " (7)endTAS; (8)endMach; (9)pointAltitude; (10)pointTemperature; (11)pointCAS; (12)pointTAS; (13)pointMach; (14)endTemperature" << std::endl;
        for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
            plot << i+1 << "; ";
            plot << data_->mission.Steps.at(i).missionPhase << "; ";
            plot << data_->mission.Steps.at(i).modeStep << "; ";
            plot << data_->mission.Steps.at(i).endAltitude << "; ";
            plot << convertUnit(METER, FOOT, data_->mission.Steps.at(i).endAltitude) / 100. << "; ";
            plot << data_->mission.Steps.at(i).endCAS << "; ";
            plot << data_->mission.Steps.at(i).endTAS << "; ";
            plot << data_->mission.Steps.at(i).endMach << "; ";
            plot << data_->mission.Steps.at(i).Point.Altitude << "; ";
            plot << data_->mission.Steps.at(i).Point.Temperature << "; ";
            plot << data_->mission.Steps.at(i).Point.CAS << "; ";
            plot << data_->mission.Steps.at(i).Point.TAS << "; ";
            plot << data_->mission.Steps.at(i).Point.Mach << "; ";
            plot << data_->mission.Steps.at(i).endTemperature << "; ";
            plot << std::endl;
        }
        plot.close();
    }
    // Prepare data for plotting
    std::vector<double> steps;
    std::vector<double> endFL;
    std::vector<double> endTemperature;
    std::vector<double> endMach;
    std::vector<double> endCAS;
    std::vector<double> endTAS;

    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        steps.push_back(i + 1);
        endFL.push_back(convertUnit(METER, FOOT, data_->mission.Steps.at(i).endAltitude) / 100.);
        endTemperature.push_back(data_->mission.Steps.at(i).endTemperature);
        endMach.push_back(data_->mission.Steps.at(i).endMach);
        endCAS.push_back(data_->mission.Steps.at(i).endCAS);
        endTAS.push_back(data_->mission.Steps.at(i).endTAS);
    }

    // Plot with matplotplusplus
    auto fig = matplot::figure(true);
    //Title
    fig->title("Mission Profile");
    //Create Subplots
    auto ax1 = fig->add_subplot(2, 1, 0);
    auto ax2 = fig->add_subplot(2, 1, 1);
    //grid on
    ax1->grid(true);
    ax2->grid(true);
    //write subplot 1
    ax1->plot(steps, endFL, "-b", steps, endTemperature, "--k");
    ax1->ylabel("Altitude [FL]");
    ax1->y_axis().tick_label_format("%4.0f");
    ax1->ylim({0, 500});
    ax1->y2_axis().visible(true);
    ax1->y2_axis().ticklabels_mode(true);
    ax1->y2_axis().label("Temperature [K]");
    ax1->y2_axis().limits({200, 300});
    ax1->y2_axis().tick_values({200, 220, 240, 260, 280, 300});
    ax1->y2_axis().tick_label_format("%4.0f");
    ax1->xticks(steps);
    matplot::legend(ax1, {"Flight level", "Temperature"});
    ax1->legend()->location(matplot::legend::general_alignment::topright);
    //write subplot 2
    ax2->hold(true);
    ax2->plot(steps, endMach, "-b");
    ax2->plot(steps, endCAS, "-k")->use_y2(true);
    ax2->plot(steps, endTAS, "--k")->use_y2(true);
    ax2->xlabel("Mission Step");
    ax2->ylabel("Mach Number [-]");
    ax2->y_axis().tick_label_format("%1.1f");
    ax2->ylim({0, 1});
    ax2->y2_axis().label("Flight Speed [m/s]");
    ax2->y2_axis().limits({0, 500});
    ax2->y2_axis().tick_label_format("%3.0f");
    ax2->y2_axis().visible(true);
    ax2->xticks(steps);

    matplot::legend(ax2, {"Mach number", "CAS", "TAS"});
    ax2->legend()->location(matplot::legend::general_alignment::topright);
    // Save the plot
    std::string plotFilePath = rtIO_->getPlotDir() + "/" + data_->configuration.theProgramName + "_mission_steps_plot.svg";
    matplot::save(plotFilePath);
}

void standardSystemsDesign::generatePlotMissionEnergy(const aircraftSystem& acftSystem) {
    std::ofstream plot;
    std::string plotDataFile(rtIO_->getCsvFilesDir() + "/" + TOOL_NAME + "_missionPower_" + acftSystem.ATA + "_plot.csv");
    // Check which input types the system has
    bool electric_exists = false, hydraulic_exists = false, bleed_exists = false;
    for (const auto& input : acftSystem.inputs) {
        if (input.type == "Electric") {
            electric_exists = true;
        } else if (input.type == "Hydraulic") {
            hydraulic_exists = true;
        } else if (input.type == "Bleed") {
            bleed_exists = true;
        }
        // Exit early if all are found
        if (electric_exists && hydraulic_exists && bleed_exists) {
            break;
        }
    }
    plot.open(plotDataFile.c_str());
    if (!plot) {
        myRuntimeInfo->err << plotDataFile + " could not be opened!" << std::endl;;
        exit(1);
    } else {
        plot << "# " + plotDataFile << std::endl;
        /* Adjust data structure: **/
        plot << "# (1) MissionsStep; (2) BleedAir [kg/s]; (3) Electric [W]; (4) Hydraulic [W]; (5) Altitude [ft]; (6) Mach [-]" << std::endl;
        for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
            plot << i+1 << "; ";
            if (bleed_exists) {
                plot << acftSystem.inputs.at(acftSystem.findPort("Bleed", "BleedIn", "input")).missionPower.at(i).baseLoad << ";";
            }
            if (electric_exists) {
                plot << acftSystem.inputs.at(acftSystem.findPort("Electric", "ElectricIn", "input")).missionPower.at(i).baseLoad << ";";
            }
            if (hydraulic_exists) {
                plot << acftSystem.inputs.at(acftSystem.findPort("Hydraulic", "HydraulicIn", "input")).missionPower.at(i).baseLoad << ";";
            }
            plot << data_->mission.Steps.at(i).Point.Altitude << ";";
            plot << data_->mission.Steps.at(i).Point.Mach << ";";
            plot << acftSystem.heatLoad.missionPower.at(i).baseLoad << ";";
            plot << std::endl;
        }
        plot.close();
    }
    // Prepare data for plotting
    std::vector<double> missionSteps, bleedAir, bleedAirPeak, electricPower, electricPeak, hydraulicPower, hydraulicPeak;
    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        //mission steps
        missionSteps.push_back(i + 1);
        if (bleed_exists) {
            bleedAir.push_back(acftSystem.inputs.at(acftSystem.findPort("Bleed", "BleedIn", "input")).missionPower.at(i).baseLoad);
            //Insert 0.0 peak load to fill each mission step (necessary so peak loads are aligned with the correct mission step)
            bleedAirPeak.push_back(0.0);
            //Actual Peak loads if they exist
            for (size_t k = 0; k < acftSystem.inputs.at(acftSystem.findPort("Bleed", "BleedIn", "input")).missionPower.at(i).peakLoad.size(); k++) {
                bleedAirPeak.at(i) = acftSystem.inputs.at(acftSystem.findPort("Bleed", "BleedIn", "input")).missionPower.at(i).peakLoad.at(k);
            }
        }
        if (electric_exists) {
            electricPower.push_back(acftSystem.inputs.at(acftSystem.findPort("Electric", "ElectricIn", "input")).missionPower.at(i).baseLoad / 1000.);
            //Insert 0.0 peak load to fill each mission step (necessary so peak loads are aligned with the correct mission step)
            electricPeak.push_back(0.0);
            for (size_t k = 0; k < acftSystem.inputs.at(acftSystem.findPort("Electric", "ElectricIn", "input")).missionPower.at(i).peakLoad.size(); k++) {
                electricPeak.at(i) = acftSystem.inputs.at(acftSystem.findPort("Electric", "ElectricIn", "input")).missionPower.at(i).peakLoad.at(k) / 1000.;
            }
        }
        if (hydraulic_exists) {
            hydraulicPower.push_back(acftSystem.inputs.at(acftSystem.findPort("Hydraulic", "HydraulicIn", "input")).missionPower.at(i).baseLoad / 1000.);
            //Insert 0.0 peak load to fill each mission step (necessary so peak loads are aligned with the correct mission step)
            hydraulicPeak.push_back(0.0);
            //Actual Peak loads if they exist
            for (size_t k = 0; k < acftSystem.inputs.at(acftSystem.findPort("Hydraulic", "HydraulicIn", "input")).missionPower.at(i).peakLoad.size(); k++) {
                hydraulicPeak.at(i) = acftSystem.inputs.at(acftSystem.findPort("Hydraulic", "HydraulicIn", "input")).missionPower.at(i).peakLoad.at(k) / 1000.;
            }
        }
    }

    // Plot with matplotplusplus
    auto fig = matplot::figure(true);
    fig->title("Mission Power of " + acftSystem.ATA);
    auto ax = fig->add_subplot(1, 1, 0);
    //clearing any previous plot data
    ax->clear();
    //turn on grid
    ax->grid(true);
    //turn on hold
    ax->hold(true);
    //scaling without values
    ax->ylim({0, 60});
    ax->y2_axis().limits({0, 3});
    //plot and adapt scaling
    if (electric_exists) {
        ax->plot(missionSteps, electricPower, "-b");//electric base load
        //use impulse for peak powers
        ax->plot(missionSteps, electricPeak, "-b")->impulse(true);//electric peak load
        double electric_max = getYAxisScalingBA(&acftSystem.inputs.at(acftSystem.findPort("Electric", "ElectricIn", "input")));
        ax->ylim({0, electric_max});
    }
    if (hydraulic_exists) {
        ax->plot(missionSteps, hydraulicPower, "-k");//hydraulic base load
        //use impulse for peak powers
        ax->plot(missionSteps, hydraulicPeak, "-k")->impulse(true);//hydraulic peak load
        double hydraulic_max = getYAxisScalingBA(&acftSystem.inputs.at(acftSystem.findPort("Hydraulic", "HydraulicIn", "input")));
        ax->ylim({0, hydraulic_max});
    }
    if (bleed_exists) {
        ax->plot(missionSteps, bleedAir, "--r")->use_y2(true);//bleed air base load
        auto bleed_peak = ax->plot(missionSteps, bleedAirPeak, "--r")->use_y2(true);//bleed air peak load
        bleed_peak.impulse(true);
        double BA_max = getYAxisScalingBA(&acftSystem.inputs.at(acftSystem.findPort("Bleed", "BleedIn", "input")));
        ax->y2_axis().limits({0, BA_max});
    }
    if (hydraulic_exists && electric_exists) {
        double EHC_max = getYAxisScalingEHC(&acftSystem.inputs.at(acftSystem.findPort("Hydraulic", "HydraulicIn", "input")),
                                            &acftSystem.inputs.at(acftSystem.findPort("Electric", "ElectricIn", "input")));
        // for power sources
        if (acftSystem.inputs.at(acftSystem.findPort("Hydraulic", "HydraulicIn", "input")).designPower < 0 &&
            acftSystem.inputs.at(acftSystem.findPort("Electric", "ElectricIn", "input")).designPower < 0) {
            EHC_max = -1.0 * EHC_max;
        }
        ax->ylim({0, EHC_max});
    }

    ax->ylabel("Power [kW]");
    ax->y_axis().tick_label_format("%.2f");
    ax->xticks(missionSteps);

    ax->y2_axis().visible(true);
    ax->y2_axis().ticklabels_mode(true);
    ax->y2_axis().label("Mass flow [kg/s]");
    ax->y2_axis().tick_label_format("%.2f");

    matplot::legend(ax, {"Electric Power", "Hydraulic Power", "Bleed Air", "", "", ""});
    ax->legend()->location(matplot::legend::general_alignment::topleft);
    ax->legend()->num_rows(1);
    ax->legend()->font_size(7);
    ax->legend()->box(false);

    // Save the plot
    std::string plotFilePath = rtIO_->getPlotDir() + "/" + data_->configuration.theProgramName + "_mission_power_" + acftSystem.ATA + "_plot.svg";
    matplot::save(fig, plotFilePath);

    fig->backend()->should_close();
    ax->clear();
}

void standardSystemsDesign::generatePlotMissionEnergyAccumulated() {
    //initialize pointer to conducting systems
    decltype(&data_->Systems.energyConductor.at(0)) electric_system = nullptr;
    decltype(&data_->Systems.energyConductor.at(0)) hydraulic_system = nullptr;
    decltype(&data_->Systems.energyConductor.at(0)) bleed_air_system = nullptr;
    //get conducting systems their output ports contain the total required power of that type
    for (size_t i(0); i < data_->Systems.energyConductor.size(); i++) {
        if (data_->Systems.energyConductor.at(i)->ATA == "ATA-24") {
            electric_system = &data_->Systems.energyConductor.at(i);
        } else if (data_->Systems.energyConductor.at(i)->ATA == "ATA-29") {
            hydraulic_system = &data_->Systems.energyConductor.at(i);
        } else if (data_->Systems.energyConductor.at(i)->ATA == "ATA-36") {
            bleed_air_system = &data_->Systems.energyConductor.at(i);
        }
    }
    //get output ports
    decltype(&data_->Systems.energyConductor.at(0)->outputs.at(0)) electric_port = nullptr;
    decltype(&data_->Systems.energyConductor.at(0)->outputs.at(0)) hydraulic_port = nullptr;
    decltype(&data_->Systems.energyConductor.at(0)->outputs.at(0)) bleed_air_port = nullptr;
    if (electric_system) {
        electric_port = &(*electric_system)->outputs.at((*electric_system)->findPort("Electric", "ElectricOut", "output"));
    }
    if (hydraulic_system) {
        hydraulic_port = &(*hydraulic_system)->outputs.at((*hydraulic_system)->findPort("Hydraulic", "HydraulicOut", "output"));
    }
    if (bleed_air_system) {
        bleed_air_port = &(*bleed_air_system)->outputs.at((*bleed_air_system)->findPort("Bleed", "BleedOut", "output"));
    }
    //write csv file with plot data
    std::ofstream plot;
    std::string plotDataFile(rtIO_->getCsvFilesDir() + "/" + data_->configuration.theProgramName + "_mission_power_" + "accumulated_power" + "_plot.csv");
    plot.open(plotDataFile.c_str());
    if (!plot) {
        myRuntimeInfo->err << plotDataFile + " could not be opened!" << std::endl;;
        exit(1);
    } else {
        plot << "# " + plotDataFile << std::endl;
        /* Adjust data structure: **/
        plot << "# (1) MissionsStep; (2) BleedAir [kg/s]; (3) Electric [W]; (4) Hydraulic [W] " << std::endl;
        for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
            plot << i+1 << "; ";
            plot << bleed_air_port->missionPower.at(i).baseLoad << ";";
            plot << electric_port->missionPower.at(i).baseLoad << ";";
            plot << hydraulic_port->missionPower.at(i).baseLoad << ";";
            plot << std::endl;
        }
        plot.close();
    }

    // Prepare data for plotting
    std::vector<double> missionSteps, bleedAirLoad, electricLoad, hydraulicLoad;
    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        missionSteps.push_back(i + 1);
        bleedAirLoad.push_back(bleed_air_port->missionPower.at(i).baseLoad);
        electricLoad.push_back(electric_port->missionPower.at(i).baseLoad / 1000.);
        hydraulicLoad.push_back(hydraulic_port->missionPower.at(i).baseLoad / 1000.);
    }

    // Plot with matplotplusplus
    auto fig = matplot::figure(true);//generate figure in quiet mode
    fig->title("Accumulated Power Requirements of the Systems");
    auto ax = fig->current_axes();
    //add grid
    ax->grid(true);
    //hold plots
    ax->hold(true);

    //Labels
    ax->ylabel("Power, kW");
    ax->xticks(missionSteps);
    ax->xlabel("Mission Step");
    ax->y2_axis().label("Mass Flow, kg/s");
    ax->ytickformat("%.2f");
    ax->y2tickformat("%.2f");

    //Electric
    ax->plot(missionSteps, electricLoad, "-b");

    //Hydraulic
    ax->plot(missionSteps, hydraulicLoad, "-k");

    //Bleed Air
    ax->plot(missionSteps, bleedAirLoad, "--r")->use_y2(true);

    //y-axis scaling
    double BA_max = getYAxisScalingBA(&(*bleed_air_port));
    ax->y2_axis().limits({0, 1.1 * BA_max}); //extra 10% room for legend
    double EHC_max = getYAxisScalingEHC(&(*hydraulic_port), &(*electric_port));
    ax->ylim({0, EHC_max});

    //legend
    matplot::legend(ax, {"Electric Power", "Hydraulic Power", "Bleed Air", "", "", ""});
    ax->legend()->location(matplot::legend::general_alignment::topleft);
    ax->legend()->num_rows(1);
    ax->legend()->font_size(7);
    ax->legend()->box(false);

    // Save the plot
    std::string plotFilePath = rtIO_->getPlotDir() + "/" + data_->configuration.theProgramName + "_accumulated_power_plot.svg";
    matplot::save(fig, plotFilePath);
}

void standardSystemsDesign::generatePlotEnergySinksMaxEnergy() {
    std::ofstream plot;
    std::string plotDataFile(rtIO_->getCsvFilesDir() + "/" + data_->configuration.theProgramName + "_energy_sinks_max_energy_plot.csv");
    plot.open(plotDataFile.c_str());
    if (!plot) {
        myRuntimeInfo->err << plotDataFile << " could not be opened!" << std::endl;
        exit(1);
    } else {
        powerProfile bleed_air;
        powerProfile hydraulic;
        powerProfile electric;
        plot << "# " + win2lin(rtIO_->getCsvFilesDir()) + "/" + data_->configuration.theProgramName + "_energy_sinks_max_energy_plot.csv" << std::endl;
        /* Adjust data structure: **/
        plot << "# Design Energy of the individual systems" << std::endl;
        plot << "# (1)System; (2) ATA; (3) Design Bleed Air [kg/s]; (4) Design Electric [W]; (5) Design Hydraulic [W]; ";
        plot <<   "(6) Max Mission Bleed Air [kg/s]; (7) Max Mission Electric [W]; (8) Max Mission Hydraulic [W]" << std::endl;
        for (size_t i = 0; i < data_->Systems.energySink.size(); i++) {
            //check which ports exist
            for (size_t j = 0; j < data_->Systems.energySink.at(i)->inputs.size(); j++) {
                if (data_->Systems.energySink.at(i)->inputs.at(j).type == "Bleed") {
                    bleed_air = data_->Systems.energySink.at(i)->inputs.at(j);
                }
                if (data_->Systems.energySink.at(i)->inputs.at(j).type == "Hydraulic") {
                    hydraulic = data_->Systems.energySink.at(i)->inputs.at(j);
                }
                if (data_->Systems.energySink.at(i)->inputs.at(j).type == "Electric") {
                    electric = data_->Systems.energySink.at(i)->inputs.at(j);
                }
            }
            plot << data_->Systems.energySink.at(i)->Name << "; ";
            plot << data_->Systems.energySink.at(i)->ATA << "; ";
            plot << bleed_air.designPower << "; ";
            plot << electric.designPower << "; ";
            plot << hydraulic.designPower << "; ";
            plot << bleed_air.maxMissionPower << "; ";
            plot << electric.maxMissionPower << "; ";
            plot << hydraulic.maxMissionPower << "; ";
            if (electric.designPower / 1000. >= this->maxShaftPower) {
                this->maxShaftPower = electric.designPower / 1000.;
            }
            if (hydraulic.designPower / 1000. >= this->maxShaftPower) {
                this->maxShaftPower = hydraulic.designPower / 1000.;
            }
            if (bleed_air.designPower >= this->maxBleedAir) {
                this->maxBleedAir = bleed_air.designPower;
            }
            plot << std::endl;
        }
        for (size_t i = 0; i < data_->Systems.virtualSystem.size(); i++) {
            //check which ports exist
            for (size_t j = 0; j < data_->Systems.virtualSystem.at(i)->inputs.size(); j++) {
                if (data_->Systems.virtualSystem.at(i)->inputs.at(j).type == "Bleed") {
                    bleed_air = data_->Systems.virtualSystem.at(i)->inputs.at(j);
                }
                if (data_->Systems.virtualSystem.at(i)->inputs.at(j).type == "Hydraulic") {
                    hydraulic = data_->Systems.virtualSystem.at(i)->inputs.at(j);
                }
                if (data_->Systems.virtualSystem.at(i)->inputs.at(j).type == "Electric") {
                    electric = data_->Systems.virtualSystem.at(i)->inputs.at(j);
                }
            }
            plot << data_->Systems.virtualSystem.at(i)->Name << "; ";
            plot << data_->Systems.virtualSystem.at(i)->ATA << "; ";
            plot << bleed_air.designPower << "; ";
            plot << electric.designPower << "; ";
            plot << hydraulic.designPower << "; ";
            plot << bleed_air.maxMissionPower << "; ";
            plot << electric.maxMissionPower << "; ";
            plot << hydraulic.maxMissionPower << "; ";
            plot << std::endl;
        }
        for (size_t i = 0; i < data_->Systems.energyConductor.size(); i++) {
            //check which ports exist
            for (size_t j = 0; j < data_->Systems.energyConductor.at(i)->inputs.size(); j++) {
                if (data_->Systems.energyConductor.at(i)->inputs.at(j).type == "Bleed") {
                    bleed_air = data_->Systems.energyConductor.at(i)->inputs.at(j);
                }
                if (data_->Systems.energyConductor.at(i)->inputs.at(j).type == "Hydraulic") {
                    hydraulic = data_->Systems.energyConductor.at(i)->inputs.at(j);
                }
                if (data_->Systems.energyConductor.at(i)->inputs.at(j).type == "Electric") {
                    electric = data_->Systems.energyConductor.at(i)->inputs.at(j);
                }
            }
            plot << data_->Systems.energyConductor.at(i)->Name << "; ";
            plot << data_->Systems.energyConductor.at(i)->ATA << "; ";
            plot << bleed_air.designPower << "; ";
            plot << electric.designPower << "; ";
            plot << hydraulic.designPower << "; ";
            plot << bleed_air.maxMissionPower << "; ";
            plot << electric.maxMissionPower << "; ";
            plot << hydraulic.maxMissionPower << "; ";
            plot << std::endl;
        }
        for (size_t i = 0; i < data_->Systems.energySource.size(); i++) {
            //check which ports exist
            for (size_t j = 0; j < data_->Systems.energySource.at(i)->inputs.size(); j++) {
                if (data_->Systems.energySource.at(i)->inputs.at(j).type == "Bleed") {
                    powerProfile bleed_air = data_->Systems.energySource.at(i)->inputs.at(j);
                }
                if (data_->Systems.energySource.at(i)->inputs.at(j).type == "Hydraulic") {
                    powerProfile hydraulic = data_->Systems.energySource.at(i)->inputs.at(j);
                }
                if (data_->Systems.energySource.at(i)->inputs.at(j).type == "Electric") {
                    powerProfile electric = data_->Systems.energySource.at(i)->inputs.at(j);
                }
            }
            plot << data_->Systems.energySource.at(i)->Name << "; ";
            plot << data_->Systems.energySource.at(i)->ATA << "; ";
            plot << bleed_air.designPower << "; ";
            plot << electric.designPower << "; ";
            plot << hydraulic.designPower << "; ";
            plot << bleed_air.maxMissionPower << "; ";
            plot << electric.maxMissionPower << "; ";
            plot << hydraulic.maxMissionPower << "; ";
            plot << std::endl;
        }
    }
    plot.close();

    //Prepare data for plotting
    std::vector<std::string> ATA;
    powerProfile bleed_air;
    powerProfile hydraulic;
    powerProfile electric;
    std::vector<double> hydraulic_design_power;
    std::vector<double> hydraulic_max_mission_power;
    std::vector<double> electric_design_power;
    std::vector<double> electric_max_mission_power;
    std::vector<double> bleed_air_design_power;
    std::vector<double> bleed_air_max_mission_power;
    for (size_t i = 0; i < data_->Systems.allSystems.size(); i++) {
        //check which ports exist
        for (size_t j = 0; j < data_->Systems.allSystems.at(i)->inputs.size(); j++) {
            if (data_->Systems.allSystems.at(i)->inputs.at(j).type == "Bleed") {
                bleed_air = data_->Systems.allSystems.at(i)->inputs.at(j);
            }
            if (data_->Systems.allSystems.at(i)->inputs.at(j).type == "Hydraulic") {
                hydraulic = data_->Systems.allSystems.at(i)->inputs.at(j);
            }
            if (data_->Systems.allSystems.at(i)->inputs.at(j).type == "Electric") {
                electric = data_->Systems.allSystems.at(i)->inputs.at(j);
            }
        }
        ATA.push_back(data_->Systems.allSystems.at(i)->ATA);
        bleed_air_design_power.push_back(bleed_air.designPower);
        electric_design_power.push_back(electric.designPower / 1000.);
        hydraulic_design_power.push_back(hydraulic.designPower / 1000.);
        bleed_air_max_mission_power.push_back(bleed_air.maxMissionPower);
        electric_max_mission_power.push_back(electric.maxMissionPower / 1000.);
        hydraulic_max_mission_power.push_back(hydraulic.maxMissionPower / 1000.);
    }
    //vector of empty strings for xticks
    std::vector<std::string> empty_xticks(16, "");

    //Plot
    auto fig = matplot::figure(true);
    fig->title("Max Energy");
    auto ax1 = fig->add_subplot(3, 1, 0);
    auto ax2 = fig->add_subplot(3, 1, 1);
    auto ax3 = fig->add_subplot(3, 1, 2);

    //Subplot 1: Electric Energy
    ax1->bar(electric_design_power);
    ax1->y_axis().limits({0, 50});
    ax1->ylabel("Electric Power in kW");
    ax1->grid(true);

    //Subplot 2: Hydraulic Energy
    ax2->bar(hydraulic_design_power);
    ax2->y_axis().limits({0, 100});
    ax2->ylabel("Hydraulic Power in kW");
    ax2->grid(true);

    //Subplot 3: Bleed Air
    ax3->bar(bleed_air_design_power);
    ax3->y_axis().limits({0.0, 2.0});
    ax3->y_axis().ticklabels({"0.0", "0.5", "1.0", "1.5", "2.0"});
    ax3->y_axis().tick_label_format("%0.2f");
    ax3->ylabel("Bleed Air in kg/s");
    ax3->grid(true);

    // Explicitly set the x-tick labels after plotting the bars
    std::vector<double> ticks(bleed_air_design_power.size());
    std::iota(ticks.begin(), ticks.end(), 1.0);
    ax1->x_axis().ticklabels(empty_xticks);
    ax1->xticks(ticks);
    ax2->x_axis().ticklabels(empty_xticks);
    ax2->xticks(ticks);
    ax3->x_axis().ticklabels(ATA);
    ax3->xticks(ticks);
    ax3->x_axis().tickangle(45);
    ax3->y_axis().tick_label_format("%4.0f");

    // Save the plot
    std::string plotFilePath = rtIO_->getPlotDir() + "/" + data_->configuration.theProgramName + "_energy_sinks_max_energy_plot.svg";
    matplot::save(plotFilePath);

    ax1->clear();
    ax2->clear();
    ax3->clear();
}

double standardSystemsDesign::getYAxisScalingBA(const powerProfile *bleedAir) {
    double maxBA_tmp = 0.0;
    double BA_max_intern = 0.0;
    int maxSteps = static_cast<int>(data_->mission.Steps.size()) - 1;
    for (int j = 0; j <= maxSteps; j++) {
        double BA_tmp = fabs(bleedAir->missionPower.at(j).baseLoad);
        if (BA_tmp > maxBA_tmp) {
            maxBA_tmp = BA_tmp;
        }
        //check peaks
        for (size_t k = 0; k < bleedAir->missionPower.at(j).peakLoad.size(); k++) {
            double BA_peak_tmp = fabs(bleedAir->missionPower.at(j).peakLoad.at(k));
            if (BA_peak_tmp > maxBA_tmp) {
                maxBA_tmp = BA_peak_tmp;
            }
        }
    }
    if (maxBA_tmp < 5) {
        BA_max_intern = roundUp(maxBA_tmp, 1);
    } else {
        BA_max_intern = roundUp(maxBA_tmp, 5);
    }
    return BA_max_intern;
}

double standardSystemsDesign::getYAxisScalingEHC(const powerProfile *hydraulicSum, const powerProfile *electricSum) {
    double maxE_tmp = 0.0;
    double maxH_tmp = 0.0;
    double maxEHC_tmp = 0.0;
    double EHC_max_intern = 0.0;

    int maxSteps = data_->mission.Steps.size() - 1;
    for (int j = 0; j <= maxSteps; j++) {
        double E_tmp = fabs(electricSum->missionPower.at(j).baseLoad) / 1000;
        double H_tmp = fabs(hydraulicSum->missionPower.at(j).baseLoad) / 1000;
        if (E_tmp > maxE_tmp) {
            maxE_tmp = E_tmp;
        }
        if (H_tmp > maxH_tmp) {
            maxH_tmp = H_tmp;
        }
        //check peaks
        for (size_t k = 0; k < electricSum->missionPower.at(j).peakLoad.size(); k++) {
            double E_peak_tmp = fabs(electricSum->missionPower.at(j).peakLoad.at(k)) / 1000;
            if (E_peak_tmp > maxE_tmp) {
                maxE_tmp = E_peak_tmp;
            }
        }
        for (size_t k = 0; k < hydraulicSum->missionPower.at(j).peakLoad.size(); k++) {
            double H_peak_tmp = fabs(hydraulicSum->missionPower.at(j).peakLoad.at(k)) / 1000;
            if (H_peak_tmp > maxH_tmp) {
                maxH_tmp = H_peak_tmp;
            }
        }
    }
    maxEHC_tmp = std::max(maxE_tmp, maxH_tmp);
    if (maxEHC_tmp < 5) {
        EHC_max_intern = roundUp(maxEHC_tmp, 1);
    } else if (maxEHC_tmp < 50) {
        EHC_max_intern = roundUp(maxEHC_tmp, 15);
    } else {
        EHC_max_intern = roundUp(maxEHC_tmp, 150);
    }
    return EHC_max_intern;
}

int standardSystemsDesign::roundUp(int numToRound, int multiple) {
    if (multiple == 0) {
        return numToRound;
    }
    int roundDown = (static_cast<int>(numToRound) / multiple) * multiple;
    int roundUp = roundDown + multiple;
    int roundCalc = roundUp;
    return (roundCalc);
}

bool standardSystemsDesign::all_zeros(const std::vector<double>& vec) {
    return std::all_of(vec.begin(), vec.end(), [](double v) { return v == 0.0; });
}
