/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include <iostream>
#include <memory>
#include <string>
#include <cstdlib>
#include <utility>
#include "standardSystemsDesign.h"
#include "systemsIOData.h"
#include "standardData.h"
#include "systemsDesignDefines.h"

// Models of the individual aircraft systems
#include "aircraftSystems/conventionalATA21.h"
#include "aircraftSystems/electricATA21.h"
#include "aircraftSystems/conventionalATA24.h"
#include "aircraftSystems/conventionalATA25.h"
#include "aircraftSystems/conventionalATA26.h"
#include "aircraftSystems/conventionalATA27.h"
#include "aircraftSystems/conventionalATA28.h"
#include "aircraftSystems/conventionalATA29.h"
#include "aircraftSystems/conventionalATA30.h"
#include "aircraftSystems/electricATA30.h"
#include "aircraftSystems/conventionalATA32.h"
#include "aircraftSystems/conventionalATA33.h"
#include "aircraftSystems/conventionalATA35.h"
#include "aircraftSystems/conventionalATA36.h"
#include "aircraftSystems/conventionalATA49.h"
#include "aircraftSystems/efficienciesATA70.h"
#include "aircraftSystems/ATAXX.h"
#include "aircraftSystems/fuelCellATA49.h"

void standardSystemsDesign::initialize() {
    const node& config = rtIO_->moduleConfig;
    /* Initializing IOData */
    data_ = std::make_shared<systemsIOData>(*rtIO_, config, rtIO_->getMissionDataDir());
    /* Read geometry components from aircraft xml */
    data_->readFuselages(rtIO_->acxml, rtIO_->getGeometryDir());
    data_->readNacelles(rtIO_->acxml, rtIO_->getGeometryDir());
    data_->readWing(rtIO_->acxml, rtIO_->getAirfoilDataDir());
    data_->readEmpennage(rtIO_->acxml, rtIO_->getAirfoilDataDir());

    /* Initializing values with data from the aircraft xml */
    data_->readPropulsion(rtIO_->acxml, rtIO_->getEngineDataDir());
    data_->readMassesFromXML(rtIO_->acxml);
    data_->readPerformance(rtIO_->acxml);
    data_->readAccomodation(rtIO_->acxml);

    /* Initializing individual systems */
    initializeSystems();

    /* Checking inputs from xml and config for consistency */
    checkUserInputForConsistency();
    checkGeometryForConsistency();
}

void standardSystemsDesign::run() {
    myRuntimeInfo->out << "Run systemsDesign standard strategy" << std::endl;
    /* Run modes */
    if (data_->configuration.designCase) {
        myRuntimeInfo->out << "Run - Sizing Mode" << std::endl;
        size();
    } else if (data_->configuration.reqMission) {
        myRuntimeInfo->out << "Run - Requirement Mission Mode" << std::endl;
        requirements_mission();
    } else {
        myRuntimeInfo->out << "Run - Study Mode" << std::endl;
        study();
    }
}

void standardSystemsDesign::update() {
    if (data_->configuration.designCase == true) {
        myRuntimeInfo->out << "Update systems" << std::endl;
        data_->updateMassProperties();
            /* Update system specific data in the acxml */
        for (size_t i = 0; i < data_->Systems.allSystems.size(); i++) {
            data_->Systems.allSystems.at(i)->updateXML();
        }
        for (size_t i = 0; i < data_->Systems.energyInterface.size(); i++) {
            data_->Systems.energyInterface.at(i)->updateACxml(rtIO_->acxml);
        }
    }
    /* Update mission file */
    data_->updateMissionXML(rtIO_->getMissionDataDir());//this updates all values that are calculated with the mission steps, including values in the acxml
}

void standardSystemsDesign::report() {
    if (rtIO_->plotOn) {
        myRuntimeInfo->out << "Generate plots ..." << std::endl;
        generatePlotData();
    }
    myRuntimeInfo->out << "Write reports ..." << std::endl;
    setHtmlBody();
    setTexBody();
    report_.generateReports(data_->configuration.theProgramName);
}

void standardSystemsDesign::save() {
    //not used on this level
}

/* METHODS */

void standardSystemsDesign::size() {
    /*Design Mission*/
    myRuntimeInfo->out << "Start calculation ..." << std::endl;
    /* Calculating performance profile */
    calculatePerformanceProfile();
    /* Calculating the masses for the design case */
    myRuntimeInfo->out << "Calculate system masses ..." << std::endl;
    getSystemsMass();
    myWeightsPt = new weightsAndCGs(data_);
    myWeightsPt->getWeights();
    myRuntimeInfo->out << "... design mission calculation finished!" << std::endl;
}

void standardSystemsDesign::study() {
    /*Study Mission*/
    myRuntimeInfo->out << "Start calculation ..." << std::endl;
    /* Calculating performance profile */
    calculatePerformanceProfile();//for study mission
    myRuntimeInfo->out << "No system masses are determined for the analysis calculation.!" << std::endl;
}

void standardSystemsDesign::requirements_mission() {
    /*Requirements Mission*/
    myRuntimeInfo->out << "Start calculation ..." << std::endl;
    /* Calculating performance profile */
    calculatePerformanceProfile();//for study mission
    myRuntimeInfo->out << "No system masses are determined for the analysis calculation.!" << std::endl;
}

void standardSystemsDesign::initializeSystems() {
    //Initialization of energy sinks
    for (size_t i = 0; i < data_->configuration.Sinks.size(); i++) {
        std::string tmp_name = data_->configuration.Sinks.at(i).systemName;
        bool tmp_operating = data_->configuration.Sinks.at(i).systemOperating;
        if (data_->configuration.Sinks.at(i).systemClass == "remainingConsumers") {
            data_->Systems.energySink.push_back(new ATAXX(tmp_name, tmp_operating, data_, *rtIO_, i));
        } else if (data_->configuration.Sinks.at(i).systemClass == "conventionalFurnishing") {  // ATA-25
            data_->Systems.energySink.push_back(new conventionalATA25(tmp_name, tmp_operating, data_, *rtIO_, i));
        } else if (data_->configuration.Sinks.at(i).systemClass == "conventionalFireProtection") {  // ATA-26
            data_->Systems.energySink.push_back(new conventionalATA26(tmp_name, tmp_operating, data_, *rtIO_, i));
        } else if (data_->configuration.Sinks.at(i).systemClass == "conventionalFlightControl") {  // ATA-27
            data_->Systems.energySink.push_back(new conventionalATA27(tmp_name, tmp_operating, data_, *rtIO_, i));
        } else if (data_->configuration.Sinks.at(i).systemClass == "conventionalFuel") {  // ATA-28
            data_->Systems.energySink.push_back(new conventionalATA28(tmp_name, tmp_operating, data_, *rtIO_, i));
        } else if (data_->configuration.Sinks.at(i).systemClass == "conventionalIceRainProtection") {  // ATA-30
            data_->Systems.energySink.push_back(new conventionalATA30(tmp_name, tmp_operating, data_, *rtIO_, i));
        } else if (data_->configuration.Sinks.at(i).systemClass == "conventionalGear") {  // ATA-32
            data_->Systems.energySink.push_back(new conventionalATA32(tmp_name, tmp_operating, data_, *rtIO_, i));
        } else if (data_->configuration.Sinks.at(i).systemClass == "conventionalLighting") {  // ATA-33
            data_->Systems.energySink.push_back(new conventionalATA33(tmp_name, tmp_operating, data_, *rtIO_, i));
        } else if (data_->configuration.Sinks.at(i).systemClass == "conventionalOxygenSystem") {  // ATA-35
            data_->Systems.energySink.push_back(new conventionalATA35(tmp_name, tmp_operating, data_, *rtIO_, i));
        } else if (data_->configuration.Sinks.at(i).systemClass == "remainingConsumers") {// ATA-XX
                data_->Systems.energySink.push_back(new ATAXX(tmp_name, tmp_operating, data_, *rtIO_, i));
        } else if (data_->configuration.Sinks.at(i).systemClass == "conventionalFurnishing") {  // ATA-25
            data_->Systems.energySink.push_back(new conventionalATA25(tmp_name, tmp_operating, data_, *rtIO_, i));
        } else if (data_->configuration.Sinks.at(i).systemClass == "conventionalFireProtection") {  // ATA-26
            data_->Systems.energySink.push_back(new conventionalATA26(tmp_name, tmp_operating, data_, *rtIO_, i));
        } else if (data_->configuration.Sinks.at(i).systemClass == "conventionalFlightControl") {  // ATA-27
            data_->Systems.energySink.push_back(new conventionalATA27(tmp_name, tmp_operating, data_, *rtIO_, i));
        } else if (data_->configuration.Sinks.at(i).systemClass == "conventionalFuel") {  // ATA-28
            data_->Systems.energySink.push_back(new conventionalATA28(tmp_name, tmp_operating, data_, *rtIO_, i));
        } else if (data_->configuration.Sinks.at(i).systemClass == "electricIceRainProtection") {   // ATA-30
            data_->Systems.energySink.push_back(new electricATA30(tmp_name, tmp_operating, data_, *rtIO_, i));
        } else {
            myRuntimeInfo->err << "The system is unknown!     " << data_->configuration.Sinks.at(i).systemClass << std::endl;
            exit(1);
        }
    }

    // Initialization of the ECS
    if (data_->configuration.theECS.systemClass == "electricECS") {
        //ECS without bleed air input
        data_->Systems.energyECS.push_back(new electricATA21(data_->configuration.theECS.systemName, data_->configuration.theECS.systemOperating, data_, *rtIO_, 0));
    } else {
        //conventional ECS with bleed air from engine
        data_->Systems.energyECS.push_back(new conventionalATA21(data_->configuration.theECS.systemName, data_->configuration.theECS.systemOperating, data_, *rtIO_, 0));
    }

    // Initializing the Virtual Systems
    for (int i = 0; i < data_->configuration.numberOfVirtualSystems; i++) {
        if (data_->configuration.virtualSystem.at(i).systemName == "virtualSystem") {
            std::stringstream nrATA;
            data_->Systems.virtualSystem.push_back(new aircraftSystem(data_->configuration.virtualSystem.at(i).systemName,
                                                                      data_->configuration.virtualSystem.at(i).systemOperating, data_, *rtIO_, 0, "virtual"));
            nrATA << "ATA-virtual" << i + 1;
            data_->Systems.virtualSystem.back()->ATA = nrATA.str();
        } else {
            myRuntimeInfo->err << "The system is unknown!     " << data_->configuration.virtualSystem.at(i).systemName << std::endl;
            exit(1);
        }
    }

    // Initialization of energy sources
    for (size_t i = 0; i < data_->configuration.Sources.size(); i++) {
        if (data_->configuration.Sources.at(i).systemClass == "conventionalAPU") {
            data_->Systems.energySource.push_back(new conventionalATA49(data_->configuration.Sources.at(i).systemName,
                                                                     data_->configuration.Sources.at(i).systemOperating, data_, *rtIO_, i));
        } else if (data_->configuration.Sources.at(i).systemClass == "fuelCellAPU") {
            data_->Systems.energySource.push_back(new fuelCellATA49(data_->configuration.Sources.at(i).systemName,
                                                                     data_->configuration.Sources.at(i).systemOperating, data_, *rtIO_, i));
        } else if (data_->configuration.Sources.at(i).systemClass == "conventionalPropulsion") {
            data_->Systems.energySource.push_back(new efficienciesATA70(data_->configuration.Sources.at(i).systemName,
                                                                     data_->configuration.Sources.at(i).systemOperating, data_, *rtIO_, i));
        } else {
            myRuntimeInfo->err << "The system is unknown!     " << data_->configuration.Sources.at(i).systemClass << std::endl;
            exit(1);
        }
    }

    // Initialization of the power distributors
    for (size_t i = 0; i < data_->configuration.Conductors.size(); i++) {
        if (data_->configuration.Conductors.at(i).systemClass == "ElectricSystem") {
            data_->Systems.energyConductor.push_back(new conventionalATA24(data_->configuration.Conductors.at(i).systemName,
                                                                        data_->configuration.Conductors.at(i).systemOperating, data_, *rtIO_, i));
        } else if (data_->configuration.Conductors.at(i).systemClass == "HydraulicSystem") {
            data_->Systems.energyConductor.push_back(new conventionalATA29(data_->configuration.Conductors.at(i).systemName,
                                                                        data_->configuration.Conductors.at(i).systemOperating, data_, *rtIO_, i));
        } else if (data_->configuration.Conductors.at(i).systemClass == "BleedAirSystem") {
            data_->Systems.energyConductor.push_back(new conventionalATA36(data_->configuration.Conductors.at(i).systemName,
                                                                        data_->configuration.Conductors.at(i).systemOperating, data_, *rtIO_, i));
        } else {
            myRuntimeInfo->err << "The system is unknown!     " << data_->configuration.Conductors.at(i).systemClass << std::endl;
            exit(1);
        }
    }
    /* Group all systems into one vector */
    data_->Systems.getAllSystems();
    /* Call read function of each system for reading system specific parameters from config */
    for (size_t i = 0; i < data_->Systems.allSystems.size(); i++) {
        data_->Systems.allSystems.at(i)->readParameters(data_->configFile);
    }
    /* since ECS is not yet in the allSystems vector: */
    data_->Systems.energyECS.back()->readParameters(data_->configFile);

    for (size_t i = 0; i < data_->configuration.Interfaces.size(); i++) {
        data_->Systems.energyInterface.push_back(new interfaceSystem(data_->configuration.Interfaces.at(i).systemName, data_->configuration.Interfaces.at(i).systemOperating,
                                                                     data_, *rtIO_, i));
    }
    // read connections
    readConnections();
}

void standardSystemsDesign::calculatePerformanceProfile() {
    /* Getting energy consumption */
    getSinkEnergyConsumption();
    // redistribute Power from Sink to conductors
    distributePower("conductor", "consumer");
    // power ECS und iterativ bis konvergenz!!!
    double avgECSBleedDeviation(1.);
    std::vector<double> avgECSBleedLoad;
    int Run(1);
    do {
        // Initialization of the accumulated heat load of the ECS
        if (data_->data.ECSHeatLoad.missionPower.empty()) {
            clearInitializeHeatLoad(&data_->data.ECSHeatLoad);
        }
        // Initialization of the accumulated heat load of the conductors
        if (data_->data.conductorHeatLoad.missionPower.empty()) {
            clearInitializeHeatLoad(&data_->data.conductorHeatLoad);
        }
        //connections of previous iteration
        vector<connection*> ECSconnectionsWithOldValues;
        for (size_t i=0; i < conductorToECS.size(); i++) {
            connection* tmp_connection = new connection();
            tmp_connection->connectionType = conductorToECS.at(i).connectionType;
            tmp_connection->providerPort = new powerProfile(*conductorToECS.at(i).providerPort);
            tmp_connection->receiverPort = new powerProfile(*conductorToECS.at(i).receiverPort);
            ECSconnectionsWithOldValues.push_back(tmp_connection);
        }
        //substract ECS power from previous iteration in output ports of connected conductors
        substractPreviousPower("conductor", "ecs", ECSconnectionsWithOldValues);
        // Calculating the energy consumption of the air conditioning system
        getECSEnergyConsumption();
        //distribute power
        distributePower("conductor", "ecs");
        // solve conductor power calculation and power distribution between different conductors iteratively
        double maxPowerDeviation = 1;
        int cnt = 0;
        while (maxPowerDeviation > ACCURACY_MEDIUM && cnt < 50) {
            cnt++;
            // copy old values
            //connections of previous iteration
            vector<connection*> conductorConnectionsWithOldValues;
            //conductor systems of previous iteration
            vector<aircraftSystem> conductorsWithOldValues;
            for (size_t i=0; i < conductorToConductor.size(); i++) {
                connection* tmp_connection = new connection();
                tmp_connection->connectionType = conductorToConductor.at(i).connectionType;
                tmp_connection->providerPort = new powerProfile(*conductorToConductor.at(i).providerPort);
                tmp_connection->receiverPort = new powerProfile(*conductorToConductor.at(i).receiverPort);
                conductorConnectionsWithOldValues.push_back(tmp_connection);
            }
            for (size_t i=0; i < data_->Systems.energyConductor.size(); i++) {
                conductorsWithOldValues.push_back(*data_->Systems.energyConductor.at(i));
            }
            // clear input ports of conductors before recalculating conducting power
            for (size_t i=0; i < data_->Systems.energyConductor.size(); i++) {
                for (size_t j=0; j < data_->Systems.energyConductor.at(i)->inputs.size(); j++) {
                    data_->Systems.energyConductor.at(i)->inputs.at(j).setPowerToZero();
                    //set design power of input also to zero
                    data_->Systems.energyConductor.at(i)->inputs.at(j).designPower = 0.;
                }
            }
            // calculate Conductor power
            getConductorEnergyConsumption();
            // calculate conductor transfer - substract conductor output port (connected to conductor input port) values first
            substractPreviousPower("conductor", "conductor", conductorConnectionsWithOldValues);
            //distribute power between conductors
            distributePower("conductor", "conductor");
            // recalculate maxPowerDeviation
            maxPowerDeviation = getMaxConductorPowerDeviation(conductorsWithOldValues);
            //delete old values
            for (connection* conn : conductorConnectionsWithOldValues) {
                delete conn->providerPort;
                delete conn->receiverPort;
                delete conn;
            }
        }
        // Cleaning up the peaks
        cleanPowerPeaks(&data_->data.totalHeatLoad);
        //cleanPowerPeaks(&data_->data.totalConsumption.Bleed);
        powerProfile bleedProfile;
        //for conventional ECS bleed air is written to output, for electric it is in the bleedAir power profile of the system
        if (data_->configuration.theECS.systemClass == "electricECS") {
            bleedProfile = data_->Systems.energyECS.at(0)->bleedAir;
        } else {
            bleedProfile = data_->Systems.energyECS.at(0)->inputs.at(data_->Systems.energyECS.at(0)->findPort("Bleed", "BleedIn", "input"));
        }
        avgECSBleedLoad.push_back(getAvgBleedECS(bleedProfile));
        Run = static_cast<int>(avgECSBleedLoad.size()) - 1;
        if (Run > 1) {
            avgECSBleedDeviation = fabs(1. - avgECSBleedLoad.at(Run) / avgECSBleedLoad.at(Run - 1));
        } else {
            avgECSBleedDeviation = 1.0;
        }
    } while (avgECSBleedDeviation > 0.0001 && Run <= 50);
    //output heat loads of systems
    for (size_t m = 0; m < data_->Systems.energySink.size(); m++) {
        myRuntimeInfo->info << "Heat load of " << data_->Systems.energySink.at(m)->ATA << " is "
        << data_->Systems.energySink.at(m)->heatLoad.getMaxPower() << std::endl;
    }
    for (size_t m = 0; m < data_->Systems.energyConductor.size(); m++) {
        myRuntimeInfo->info << "Heat load of " << data_->Systems.energyConductor.at(m)->ATA << " is "
        << data_->Systems.energyConductor.at(m)->heatLoad.getMaxPower() << std::endl;
    }
    // distribute Power Profiles from conductors to sources
    distributePower("source", "conductor");
    distributePower("interface", "conductor");
    // Power calculation of the sources
    getRequiredSourcePower();
    distributePower("interface", "source");
    /* Interface Systems. Get Power is used to calculate max and average power demand for updating acxml */
    for (size_t i = 0; i < data_->Systems.energyInterface.size(); i++) {
        data_->Systems.energyInterface.at(i)->getPower();
    }
    // Moving the ECS into the energy sink and adding it to the all systems vector
    data_->Systems.energySink.insert(data_->Systems.energySink.end(), data_->Systems.energyECS.back());
    data_->Systems.allSystems.insert(data_->Systems.allSystems.end(), data_->Systems.energyECS.back());
}

double standardSystemsDesign::getMaxConductorPowerDeviation(vector<aircraftSystem> oldConductors) {
    // loop over conductor power inputs, get difference between actual value and last value and find max element
    double maxDeviationValue = 0;
    for (size_t i=0; i < data_->Systems.energyConductor.size(); i++) {
        for (size_t n=0; n < data_->Systems.energyConductor.at(i)->outputs.size(); n++) {
            powerProfile& currOutput = data_->Systems.energyConductor.at(i)->outputs.at(n);
            powerProfile& currOldOutput = oldConductors.at(i).outputs.at(n);
            // loop over mission power and check deviation
            for (size_t m=0; m < currOutput.missionPower.size(); m++) {
                double baseLoadDeviation = std::abs((currOutput.missionPower.at(m).baseLoad - currOldOutput.missionPower.at(m).baseLoad)/currOutput.missionPower.at(m).baseLoad);
                if (baseLoadDeviation > maxDeviationValue) {
                    maxDeviationValue = baseLoadDeviation;
                }
                // check peak load Deviation
                // because peak loads might not exist in previous iteration only maximum peaks are compared
                double currMaxPeak = 0.;
                double oldMaxPeak = 0.;
                for (size_t k=0; k < currOldOutput.missionPower.at(m).peakLoad.size(); k++) {
                    if (currOldOutput.missionPower.at(m).peakLoad.at(k) > oldMaxPeak) {
                        oldMaxPeak = currOldOutput.missionPower.at(m).peakLoad.at(k);
                    }
                }
                for (size_t k=0; k < currOutput.missionPower.at(m).peakLoad.size(); k++) {
                    if (currOutput.missionPower.at(m).peakLoad.at(k) > currMaxPeak) {
                        currMaxPeak = currOutput.missionPower.at(m).peakLoad.at(k);
                    }
                }
                double peakLoadDeviation = std::abs((currMaxPeak - oldMaxPeak) / currMaxPeak);
                if (peakLoadDeviation > maxDeviationValue) {
                    maxDeviationValue = peakLoadDeviation;
                }
            }
        }
    }
    return maxDeviationValue;
}

void standardSystemsDesign::substractPreviousPower(std::string providerType, std::string receiverType, vector<connection*> oldConnections) {
    // identify the relevant connection group
    vector<connection> relevantConnections;
    if (providerType == "conductor" && receiverType == "consumer") {
        relevantConnections = conductorToConsumer;
    } else if (providerType == "conductor" && receiverType == "conductor") {
        relevantConnections = conductorToConductor;
    } else if (providerType == "source" && receiverType == "conductor") {
        relevantConnections = sourceToConductor;
    } else if (providerType == "interface" && receiverType == "conductor") {
        relevantConnections = interfaceToConductor;
    } else if (providerType == "interface" && receiverType == "source") {
        relevantConnections = interfaceToSource;
    } else if (providerType == "conductor" && receiverType == "ecs") {
        relevantConnections = conductorToECS;
    } else {
        myRuntimeInfo->err << "Unknown power distribution combination " << providerType << " and " << receiverType << endl;
        exit(1);
    }
    for (size_t i(0); i < relevantConnections.size(); i++) {
        //substract design load considering distribution factor
        myRuntimeInfo->debug << "Before substracting design power = " << relevantConnections.at(i).providerPort->designPower << std::endl;
        relevantConnections.at(i).providerPort->designPower -= oldConnections.at(i)->receiverPort->designPower * relevantConnections.at(i).designDistribution;
        myRuntimeInfo->debug << "Design power reciever = " << oldConnections.at(i)->receiverPort->designPower * relevantConnections.at(i).designDistribution << std::endl;
        myRuntimeInfo->debug << "After substracting design power = " << relevantConnections.at(i).providerPort->designPower << std::endl;
        //substract mission loads
        for (size_t j(0); j < data_->mission.Steps.size(); j++) {
            //base load
            relevantConnections.at(i).providerPort->missionPower.at(j).baseLoad -=
                oldConnections.at(i)->receiverPort->missionPower.at(j).baseLoad * relevantConnections.at(i).distribution.at(j);
            //peak loads (only if it's the same peak)
            for (size_t k(0); k < relevantConnections.at(i).providerPort->missionPower.at(j).peakLoad.size(); k++) {
                //check if old connection has a peak
                if (k < oldConnections.at(i)->receiverPort->missionPower.at(j).peakLoad.size()) {
                    //check if it is the same peak
                    if (accuracyCheck(relevantConnections.at(i).providerPort->missionPower.at(j).timePeak.at(k),
                                        oldConnections.at(i)->receiverPort->missionPower.at(j).timePeak.at(k),
                                        ACCURACY_HIGH)) {
                        relevantConnections.at(i).providerPort->missionPower.at(j).peakLoad.at(k) -=
                            oldConnections.at(i)->receiverPort->missionPower.at(j).peakLoad.at(k) * relevantConnections.at(i).distribution.at(j);
                    }
                }
            }
        }
    }
}

void standardSystemsDesign::distributePower(std::string providerType, std::string receiverType) {
    // identify the relevant connection group
    vector<connection> relevantConnections;
    if (providerType == "conductor" && receiverType == "consumer") {
        relevantConnections = conductorToConsumer;
    } else if (providerType == "conductor" && receiverType == "conductor") {
        relevantConnections = conductorToConductor;
    } else if (providerType == "source" && receiverType == "conductor") {
        relevantConnections = sourceToConductor;
    } else if (providerType == "interface" && receiverType == "conductor") {
        relevantConnections = interfaceToConductor;
    } else if (providerType == "interface" && receiverType == "source") {
        relevantConnections = interfaceToSource;
    } else if (providerType == "conductor" && receiverType == "ecs") {
        relevantConnections = conductorToECS;
    } else {
        myRuntimeInfo->err << "Unknown power distribution combination " << providerType << " and " << receiverType << endl;
        exit(1);
    }
    // loop over connections and add power from receiver to provider ports
    for (size_t i=0; i < relevantConnections.size(); i++) {
        relevantConnections.at(i).providerPort->addPowerProfile(relevantConnections.at(i).receiverPort,
            relevantConnections.at(i).distribution, relevantConnections.at(i).designDistribution);
    }
    // clean power peaks
    for (size_t i=0; i < relevantConnections.size(); i++) {
        cleanPowerPeaks(relevantConnections.at(i).providerPort);
    }
}

void standardSystemsDesign::eraseOutputValues(std::string providerType, std::string receiverType) {
    // identify the relevant connection group
    vector<connection> relevantConnections;
    if (providerType == "conductor" && receiverType == "consumer") {
        relevantConnections = conductorToConsumer;
    } else if (providerType == "conductor" && receiverType == "conductor") {
        relevantConnections = conductorToConductor;
    } else if (providerType == "source" && receiverType == "conductor") {
        relevantConnections = sourceToConductor;
    } else if (providerType == "interface" && receiverType == "conductor") {
        relevantConnections = interfaceToConductor;
    } else if (providerType == "interface" && receiverType == "source") {
        relevantConnections = interfaceToSource;
    } else if (providerType == "conductor" && receiverType == "ecs") {
        relevantConnections = conductorToECS;
    } else {
        myRuntimeInfo->err << "Unknown power distribution combination " << providerType << " and " << receiverType << endl;
        exit(1);
    }
    // loop over connections and set provider port values to zero
    for (size_t i=0; i < relevantConnections.size(); i++) {
        relevantConnections.at(i).providerPort->setPowerToZero();
    }
}

void standardSystemsDesign::readConnections() {
    // read Sink connections
    myRuntimeInfo-> out << "Read sink connections" << endl;
    for (size_t i=0; i < data_->configuration.Sinks.size(); i++) {
        for (size_t n=0; n < data_->configuration.Sinks.at(i).inputs.size(); n++) {
            standardConfig::port& currPort = data_->configuration.Sinks.at(i).inputs.at(n);
            // get pointer to current ports power profile
            string providerSystemType = "";
            powerProfile* currPowerProfilePt = getPowerProfilePt(data_->configuration.Sinks.at(i).systemName, currPort.ID, "input", &providerSystemType);
            // loop over distributions and create connections
            for (size_t m=0; m < currPort.distributions.size(); m++) {
                this->conductorToConsumer.push_back(createConnection(currPort.type, currPort.distributions.at(m), currPowerProfilePt, &providerSystemType));
            }
        }
    }
    /////// output section for debugging ///////////////
    for (size_t i=0; i < this->conductorToConsumer.size(); i++) {
        myRuntimeInfo->debug << "Connection " << this->conductorToConsumer.at(i).providerPort->ID << " to " << this->conductorToConsumer.at(i).receiverPort->ID
                           << " type " << this->conductorToConsumer.at(i).connectionType << endl;
        myRuntimeInfo->debug << "distribution " << this->conductorToConsumer.at(i).distribution.at(0);
        for (size_t n=1; n < this->conductorToConsumer.at(i).distribution.size(); n++) {
            myRuntimeInfo->debug << this->conductorToConsumer.at(i).distribution.at(n);
        }
        myRuntimeInfo->debug << endl;
    }
    // read Conductor connections
    myRuntimeInfo->out << "Read conductor connections" << endl;
    for (size_t i=0; i < data_->configuration.Conductors.size(); i++) {
        for (size_t n=0; n < data_->configuration.Conductors.at(i).inputs.size(); n++) {
            standardConfig::port& currPort = data_->configuration.Conductors.at(i).inputs.at(n);
            // get pointer to current ports power profile
            string dummy = "";
            powerProfile* currPowerProfilePt = getPowerProfilePt(data_->configuration.Conductors.at(i).systemName, currPort.ID, "input", &dummy);
            // loop over distributions and create connections
            for (size_t m=0; m < currPort.distributions.size(); m++) {
                string providerSystemType = "";
                connection newConnection = createConnection(currPort.type, currPort.distributions.at(m), currPowerProfilePt, &providerSystemType);
                if (providerSystemType == "conductor") {
                    this->conductorToConductor.push_back(newConnection);
                } else if (providerSystemType == "source") {
                    this->sourceToConductor.push_back(newConnection);
                } else if (providerSystemType == "interface") {
                    this->interfaceToConductor.push_back(newConnection);
                } else {
                    myRuntimeInfo->err << "Invalid provider system type " << providerSystemType << " for receiver type conductor" << endl;
                    exit(1);
                }
            }
        }
    }
    // read Source connections
    myRuntimeInfo->out << "Read source connections" << endl;
    for (size_t i=0; i < data_->configuration.Sources.size(); i++) {
        for (size_t n=0; n < data_->configuration.Sources.at(i).inputs.size(); n++) {
            standardConfig::port& currPort = data_->configuration.Sources.at(i).inputs.at(n);
            // get pointer to current ports power profile
            string dummy = "";
            powerProfile* currPowerProfilePt = getPowerProfilePt(data_->configuration.Sources.at(i).systemName, currPort.ID, "input", &dummy);
            // loop over distributions and create connections
            for (size_t m=0; m < currPort.distributions.size(); m++) {
                string providerSystemType = "";
                connection newConnection = createConnection(currPort.type, currPort.distributions.at(m), currPowerProfilePt, &providerSystemType);
                if (providerSystemType == "interface") {
                    this->interfaceToSource.push_back(newConnection);
                } else {
                    myRuntimeInfo->err << "Invalid provider system type " << providerSystemType << " for receiver type source" << endl;
                    exit(1);
                }
            }
        }
    }
    // read ECS connections
    myRuntimeInfo->out << "Read ECS connections" << endl;
    for (size_t n=0; n < data_->configuration.theECS.inputs.size(); n++) {
        standardConfig::port& currPort = data_->configuration.theECS.inputs.at(n);
        // get pointer to current ports power profile
        string dummy = "";
        powerProfile* currPowerProfilePt = getPowerProfilePt(data_->configuration.theECS.systemName, currPort.ID, "input", &dummy);
        // loop over distributions and create connections
        for (size_t m=0; m < currPort.distributions.size(); m++) {
            string providerSystemType = "";
            connection newConnection = createConnection(currPort.type, currPort.distributions.at(m), currPowerProfilePt, &providerSystemType);
            if (providerSystemType == "conductor") {
                this->conductorToECS.push_back(newConnection);
            } else {
                myRuntimeInfo->err << "Invalid provider system type " << providerSystemType << " for receiver type ecs" << endl;
                exit(1);
            }
        }
    }
}

standardSystemsDesign::connection standardSystemsDesign::createConnection(string portType, const standardConfig::port::distribution& thePortDistribution,
                                                                            powerProfile* currPowerProfilePt, string* providerSystemType) {
    standardSystemsDesign::connection newConnection;
    newConnection.connectionType = portType;
    // receiverport is this port - find corresponding powerprofile
    newConnection.receiverPort = currPowerProfilePt;
    newConnection.providerPort = getPowerProfilePt(thePortDistribution.sourceSystemUID, thePortDistribution.sourcePortUID, "output", providerSystemType);
    newConnection.designDistribution = readDesignDistribution(thePortDistribution.values, currPowerProfilePt->ID);
    newConnection.distribution = readMissionDistribution(thePortDistribution.values, currPowerProfilePt->ID);
    return newConnection;
}

double standardSystemsDesign::readDesignDistribution(string theDistribution, string receiverPortID) {
    // Separate the part before the "/", which is the design power distribution
    string delimiter = "/";
    size_t delimiterPos = theDistribution.find(delimiter);
    vector<string> splitted;
    if (delimiterPos != string::npos) {
        splitted.push_back(theDistribution.substr(0, delimiterPos));
    } else {
        myRuntimeInfo->err << "No design power distribution defined for port " << receiverPortID << std::endl;
        exit(1);
    }
    //convert to double
    double distribution = std::stod(splitted.at(0));
    if (distribution > 1.0 || distribution < 0.0) {
        myRuntimeInfo->err << "Port " << receiverPortID << " input distribution value is not between 0 and 1!" << endl;
        exit(1);
    }
    return distribution;
}

vector<double> standardSystemsDesign::readMissionDistribution(string theDistribution, string receiverPortID) {
    // Separate the part after the "/", which is the mission distribution
    string outer_delimiter = "/";
    size_t delimiterPos = theDistribution.find(outer_delimiter);
    if (delimiterPos != string::npos) {
        theDistribution = theDistribution.substr(delimiterPos + 1);
    }
    // split string into vector of string separated by ;, then convert entries to double vector
    vector<double> distributionDouble;
    vector<string> splitted;
    string delimiter = ";";
    size_t pos = 0;
    string token;
    size_t found = theDistribution.find(delimiter);
    if (found == string::npos) {
        splitted.push_back(theDistribution);
    } else {
        while ((pos = theDistribution.find(delimiter)) != string::npos) {
            token = theDistribution.substr(0, pos);
            splitted.push_back(token);
            theDistribution.erase(0, pos + delimiter.length());
        }
        splitted.push_back(theDistribution);
    }
    // check size. If is one, this value is valid for all mission segments. If not, check if size is equal to mission steps
    if (splitted.size() == 1) {
        // use value for all mission segments
        for (size_t i=0; i < data_->mission.Steps.size(); i++) {
            distributionDouble.push_back(stod(splitted.at(0)));
            if (distributionDouble.back() > 1 || distributionDouble.back() < 0) {
                myRuntimeInfo->err << "Port " << receiverPortID << " input distribution value is not between 0 and 1!" << endl;
                exit(1);
            }
        }
    } else if (splitted.size() == data_->mission.Steps.size()) {
        // convert to double
        for (size_t i=0; i < splitted.size(); i++) {
            distributionDouble.push_back(stod(splitted.at(i)));
            if (distributionDouble.back() > 1 || distributionDouble.back() < 0) {
                myRuntimeInfo->err << "Port " << receiverPortID << " input distribution value is not between 0 and 1!" << endl;
                exit(1);
            }
        }
    } else {
        myRuntimeInfo->err << "Distribution of inputs of port " << receiverPortID << " has wrong no of entries (" << splitted.size()
                           << "), must be equal to mission step no!" << endl;
        exit(1);
    }
    return distributionDouble;
}

powerProfile* standardSystemsDesign::getPowerProfilePt(string systemName, string portID, string direction, string* providerSystemType) {
    // search in different System vectors
    aircraftSystem* theSystem = NULL;
    for (size_t i=0; i < data_->Systems.energySink.size(); i++) {
        if (data_->Systems.energySink.at(i)->Name == systemName) {
            theSystem = data_->Systems.energySink.at(i);
            *providerSystemType = "sink";
            break;
        }
    }
    if (theSystem == NULL) {
        for (size_t i=0; i < data_->Systems.energySource.size(); i++) {
            if (data_->Systems.energySource.at(i)->Name == systemName) {
                theSystem = data_->Systems.energySource.at(i);
                *providerSystemType = "source";
                break;
            }
        }
    }
    if (theSystem == NULL) {
        for (size_t i=0; i < data_->Systems.energyConductor.size(); i++) {
            if (data_->Systems.energyConductor.at(i)->Name == systemName) {
                theSystem = data_->Systems.energyConductor.at(i);
                *providerSystemType = "conductor";
                break;
            }
        }
    }
    if (theSystem == NULL) {
        for (size_t i=0; i < data_->Systems.energyInterface.size(); i++) {
            myRuntimeInfo->out << "system name = " << systemName << std::endl;
            myRuntimeInfo->out << "interface system name = " << data_->Systems.energyInterface.at(i)->Name << std::endl;
            if (data_->Systems.energyInterface.at(i)->Name == systemName) {
                theSystem = data_->Systems.energyInterface.at(i);
                *providerSystemType = "interface";
                break;
            }
        }
    }
    if (theSystem == NULL) {
        for (size_t i=0; i < data_->Systems.energyECS.size(); i++) {
            if (data_->Systems.energyECS.at(i)->Name == systemName) {
                theSystem = data_->Systems.energyECS.at(i);
                *providerSystemType = "ecs";
                break;
            }
        }
    }
    if (theSystem == NULL) {
        myRuntimeInfo->err << "Could not find system with name " << systemName << endl;
        exit(1);
    }
    // identify port in theSystem
    if (direction == "input") {
        for (size_t i=0; i < theSystem->inputs.size(); i++) {
            if (theSystem->inputs.at(i).ID == portID) {
                return &theSystem->inputs.at(i);
            }
        }
    } else if (direction == "output") {
        for (size_t i=0; i < theSystem->outputs.size(); i++) {
            myRuntimeInfo->out << "Port " << theSystem->outputs.at(i).ID << endl;
            if (theSystem->outputs.at(i).ID == portID) {
                return &theSystem->outputs.at(i);
            }
        }
    } else {
        myRuntimeInfo->err << "Direction argument " << direction << " not allowed." << endl;
        exit(1);
    }
    myRuntimeInfo->err << "Could not find " << direction << " port with ID " << portID << " in System " << systemName << endl;
    exit(1);
}

/* Performance Profile Methods */
double standardSystemsDesign::getAvgBleedECS(powerProfile theBleed) {
    double avgBleedAir(0.0);
    for (size_t i = 0; i < theBleed.missionPower.size(); i++) {
        avgBleedAir += theBleed.missionPower.at(i).baseLoad;
    }
    return avgBleedAir / (static_cast<double>(theBleed.missionPower.size()) - 1.);
}

void standardSystemsDesign::getSinkEnergyConsumption() {
    //Initialization of the accumulated sink performances
    clearInitializeHeatLoad(&data_->data.sinkHeatLoad);
    /* Energy Sinks **/
    for (size_t i = 0; i < data_->configuration.Sinks.size(); i++) {
        myRuntimeInfo->debug << "Calculate system power for sink: " << data_->Systems.energySink.at(i)->Name << std::endl;
        data_->Systems.energySink.at(i)->calculateSystemsPower();
        getAccumulatedHeatLoad(*data_->Systems.energySink.at(i), &data_->data.sinkHeatLoad);
    }
    /* Virtual Systems are defined by Config **/
    for (size_t i = 0; i < data_->configuration.virtualSystem.size(); i++) {
        calculateVirtualSystemPower(i);
        getAccumulatedHeatLoad(*data_->Systems.virtualSystem.at(i), &data_->data.sinkHeatLoad);
    }
    cleanPowerPeaks(&data_->data.sinkHeatLoad);
}

void standardSystemsDesign::getECSEnergyConsumption() {
    data_->Systems.energyECS.back()->calculateECSPower(data_->data.sinkHeatLoad, data_->data.ECSHeatLoad, data_->data.conductorHeatLoad);
    clearInitializeHeatLoad(&data_->data.ECSHeatLoad);
    getAccumulatedHeatLoad(*data_->Systems.energyECS.back(), &data_->data.ECSHeatLoad);
    cleanPowerPeaks(&data_->data.ECSHeatLoad);
}

void standardSystemsDesign::getConductorEnergyConsumption() {
    // calculate conductor system power
    for (size_t i = 0; i < data_->configuration.Conductors.size(); i++) {
        data_->Systems.energyConductor.at(i)->calculateConductingPower();
    }
    // calculate total conductor heat load
    clearInitializeHeatLoad(&data_->data.conductorHeatLoad);
    for (size_t j = 0; j < data_->configuration.Conductors.size(); j++) {
        getAccumulatedHeatLoad(*data_->Systems.energyConductor.at(j), &data_->data.conductorHeatLoad);
    }
    cleanPowerPeaks(&data_->data.ECSHeatLoad);
}

// !!!rename to something like: getRequiredSourcePower
void standardSystemsDesign::getRequiredSourcePower() {
    // calling source power functions also calls the addEngineEfficiencies function
    for (size_t i = 0; i < data_->configuration.Sources.size(); i++) {
        data_->Systems.energySource.at(i)->calculateSourcePower();
    }
}

void standardSystemsDesign::clearInitializeHeatLoad(powerProfile *heatLoad) {
    heatLoad->missionPower.clear();
    heatLoad->designPower = 0.;
    heatLoad->maxMissionPower = 0.;
    //Mission Loads
    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        heatLoad->missionPower.push_back(missionLoads());
    }
}

void standardSystemsDesign::getAccumulatedHeatLoad(const aircraftSystem& theSystem, powerProfile *heatLoad) {
    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        heatLoad->missionPower.at(i).baseLoad += theSystem.heatLoad.missionPower.at(i).baseLoad;
        for (size_t j = 0; j < theSystem.heatLoad.missionPower.at(i).peakLoad.size(); j++) {
            heatLoad->missionPower.at(i).peakLoad.push_back(theSystem.heatLoad.missionPower.at(i).peakLoad.at(j));
            heatLoad->missionPower.at(i).timePeak.push_back(theSystem.heatLoad.missionPower.at(i).timePeak.at(j));
        }
    }
}

void standardSystemsDesign::calculateVirtualSystemPower(int ID) {
    auto& system = data_->Systems.virtualSystem.at(ID);
    auto& config = data_->configuration.virtualSystem.at(ID);
    decltype(&system->inputs.at(0)) bleedAirInput = nullptr;
    decltype(&system->inputs.at(0)) hydraulicInput = nullptr;
    decltype(&system->inputs.at(0)) electricInput = nullptr;
    // Check which input types the system has
    bool electric_exists = false, hydraulic_exists = false, bleed_exists = false;
    for (const auto& input : system->inputs) {
        if (input.type == "Electric") {
            electric_exists = true;
            electricInput = &system->inputs.at(system->findPort("Electric", "ElectricIn", "input"));
        } else if (input.type == "Hydraulic") {
            hydraulic_exists = true;
            hydraulicInput = &system->inputs.at(system->findPort("Hydraulic", "HydraulicIn", "input"));
        } else if (input.type == "Bleed") {
            bleed_exists = true;
            bleedAirInput = &system->inputs.at(system->findPort("Bleed", "BleedIn", "input"));
        }
        // Exit early if all are found
        if (electric_exists && hydraulic_exists && bleed_exists) {
            break;
        }
    }
    // Iterate through mission steps
    for (size_t n = 0; n < data_->mission.Steps.size(); n++) {
        int j = 0;
        if (data_->mission.Steps.at(n).missionPhase == "Cruise") {
            j = 1;
        } else if (data_->mission.Steps.at(n).missionPhase == "Approach") {
            j = 2;
        }

        /* Define baseLoad */
        if (bleedAirInput) {
            (*bleedAirInput).missionPower.at(n).baseLoad = config.powerBleed[j];
        }
        if (hydraulicInput) {
            (*hydraulicInput).missionPower.at(n).baseLoad = config.powerHydraulic[j];
        }
        if (electricInput) {
            (*electricInput).missionPower.at(n).baseLoad = config.powerElectric[j];
        }

        /* Assign heat load */
        system->heatLoad.missionPower.at(n).baseLoad = config.heatLoad[j];
    }

    /* Define maxMissionPower */
    if (bleedAirInput) {
        (*bleedAirInput).maxMissionPower = (*bleedAirInput).getMaxPower();
    }
    if (electricInput) {
        (*electricInput).maxMissionPower = (*electricInput).getMaxPower();
    }
    if (hydraulicInput) {
        (*hydraulicInput).maxMissionPower = (*hydraulicInput).getMaxPower();
    }
    system->heatLoad.maxMissionPower = system->heatLoad.getMaxPower();

    /* Define designPower */
    if (bleedAirInput) {
        (*bleedAirInput).designPower = (*bleedAirInput).maxMissionPower;
    }
    if (electricInput) {
        (*electricInput).designPower = (*electricInput).maxMissionPower;
    }
    if (hydraulicInput) {
        (*hydraulicInput).designPower = (*hydraulicInput).maxMissionPower;
    }
    system->heatLoad.designPower = system->heatLoad.maxMissionPower;
}

/* Methods for system masses */

void standardSystemsDesign::getSystemsMass() {
    for (size_t i = 0; i < data_->Systems.energySink.size(); i++) {
        data_->Systems.energySink.at(i)->calculateSinkMass();
        data_->data.accumulatedSystemsMass += data_->Systems.energySink.at(i)->Mass;
    }
    for (size_t i = 0; i < data_->configuration.virtualSystem.size(); i++) {
        data_->Systems.virtualSystem.at(i)->Mass = data_->configuration.virtualSystem.at(i).Mass;
        data_->data.accumulatedSystemsMass += data_->Systems.virtualSystem.at(i)->Mass;
        data_->Systems.virtualSystem.at(i)->refPoint[0] = data_->configuration.virtualSystem.at(i).refPoint[0];
        data_->Systems.virtualSystem.at(i)->refPoint[1] = data_->configuration.virtualSystem.at(i).refPoint[1];
        data_->Systems.virtualSystem.at(i)->refPoint[2] = data_->configuration.virtualSystem.at(i).refPoint[2];
    }
    for (size_t i = 0; i < data_->configuration.Sources.size(); i++) {
        data_->Systems.energySource.at(i)->calculateSourceMass();
        data_->data.accumulatedSystemsMass += data_->Systems.energySource.at(i)->Mass;
    }
    for (size_t i = 0; i < data_->configuration.Conductors.size(); i++) {
        data_->Systems.energyConductor.at(i)->calculateConductorMass();
        data_->data.accumulatedSystemsMass += data_->Systems.energyConductor.at(i)->Mass;
    }
}

/* CHECK METHODS */

bool standardSystemsDesign::checkEnergySource(std::string sourceName) {
    bool sourceFound;
    for (size_t i = 0; i < data_->configuration.Sources.size(); i++) {
        if (data_->configuration.Sources.at(i).systemName == sourceName) {
            sourceFound = true;
        }
    }
    return sourceFound;
}

void standardSystemsDesign::checkUserInputForConsistency() {
    /* Check flight controls */
    if (data_->Systems.systemExists("ATA-27")) {
        int ID = data_->Systems.getSpecificSystem("ATA-27");
        data_->Systems.allSystems.at(ID)->checkUserInput();
    }
}

void standardSystemsDesign::checkGeometryForConsistency() {
    bool aileronExists(false);
    for (size_t i(0); i < data_->theWing.at(0).controlSurface.size(); i++) {
        if (data_->theWing.at(0).controlSurface.at(i).name == "aileron" || data_->theWing.at(0).controlSurface.at(i).name == "Aileron"
            || data_->theWing.at(0).controlSurface.at(i).name == "droop_aileron" || data_->theWing.at(0).controlSurface.at(i).name == "DroopAileron")
            aileronExists = true;
    }
    if (!aileronExists) {
        myRuntimeInfo->warn << "There is no Aileron for the wing!" << std::endl;
    }
    /* SystemsDesign specific **/
    if (data_->theLandingGear.empty()) {
        myRuntimeInfo->err << "No landing gears integrated. Please integrate landing gears first" << std::endl;
        exit(1);
    } else {
        if (data_->theLandingGear.front().name != "nose_gear" && data_->theLandingGear.front().name != "nose_gear_assembly") {
            myRuntimeInfo->err << "First landing gear element is not nose landing gear. Abort!" << std::endl;
            exit(1);
        } else {
            if (data_->theLandingGear.at(0).strut_length <= 0.0 || data_->theLandingGear.at(0).strut_length >= 15.0) {
                myRuntimeInfo->err << "Gear.lengthStrut is not in the range of 0 and 15" << std::endl;
                exit(1);
            }
        }
    }
}
