/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 * 
 * Description:
 * This file is part of UNICADO.
 */

#ifndef SYSTEMS_DESIGN_SRC_STANDARDSYSTEMSDESIGN_H_
#define SYSTEMS_DESIGN_SRC_STANDARDSYSTEMSDESIGN_H_

#include <memory>
#include <vector>
#include <string>
#include <moduleBasics/strategySelector.h>
#include <moduleBasics/report.h>
#include <moduleBasics/plot.h>
#include "systemsDesign.h"
#include "systemsIOData.h"
#include "standardData.h"
#include "weightsAndCGs.h"
#include "powerProfile.h"

using std::endl;
using std::vector;

/**
 * \brief Class for standard systems design strategy
 * 
 */
class standardSystemsDesign : public Strategy {
 public:
    explicit standardSystemsDesign(std::shared_ptr<RuntimeIO> rtIO): rtIO_(rtIO), report_(rtIO) {}
    ~standardSystemsDesign() = default;
    /* Inherited functions from class Strategy */

    /** \brief Initializes the IOData pointer with the acxml and config file, reads the geometry information
     * for the fuselage, nacelles, wing, and empennage from the aircraft xml as well as all needed values for 
     * the propulsion system, system energy providers, masses, performance, and accomodation. All functions reading
     * from the acxml belong to the class systemsIOData. 
     * The individual systems are initialized based on the user definiton in the config document.
     * Finally the inputs from both files are checked for consistency.
     * \return void
    */
    void initialize();

    /** \brief Differentiates between the two modes (size and study) of the standard systemsDesign strategy.
     * \return void
    */
    void run();

    /** \brief For the sizing mode this function calls the update functions for the acxml and the mission xml file.
     * In the study mode only the mission xml file is updated.
     * \return void
    */
    void update();

    /** \brief Calls the functions for preparing and generating the plots and reports.
     * \return void
    */
    void report();

    void save();

    /** \brief Runs the sizing mode of systemsDesign. First the performance profile of all systems is determined for
     * the design mission. Then the systems are sized and there masses determined according to the design mission.
     * Finally, the weightsAndCGs class is used to calculated the accumulated masses and centers of gravity.
     * \return void
    */
    void size();

    /** \brief Runs the study mode of systemsDesign. The performance profile in the study mission for all systems is calculated.
     * \attention System and accumulated masses and CofG are not determined in the study mode.
    */
    void study();

    /** \brief Runs the requirements mission mode of systemsDesign. The performance profiles in the requirments mission for all
     * systems are calculated.
     * \attention System and accumulated masses and CofG are not determined in the requirments mission mode.
    */
    void requirements_mission();

    /** \brief Calls the specific implementation of the checkUserInput()-function of the aircraft systems ATA-29,
     * ATA-24, and ATA-27. These functions check user input from the configuration file.
     * \return void
    */
    void checkUserInputForConsistency();

    /** \brief Checks wether the acxml specifies ailerons for the wing and a landing gear. If not it throws an error
     * since these components are required.
     * \return void
    */
    void checkGeometryForConsistency();

    /** \brief Checks if the given energy source is specified in the config file.
     * \param sourceName name of the source needs to match the name specified in the config file.
     * \return bool is true when given source is defined in config file and false if it is not.
    */
    bool checkEnergySource(std::string sourceName);

    /** \brief Initializes every system based on the inputs from the config file. Based on the system name an
     * instance of the corresponding aircraftSystem is created.
     * \note The systems are grouped into sinks, environmental
     * control system, virtual systems, energy sources and energy conductors, which all have a corresponding vector
     * data_->Systems.energy... . Additionally, there is one vector containing all systems (data_->Systems.allSystems)
     * except the environmental control system.
     * 
     * The environmental control system, despite being a sink system, is kept seperate to allow for iteration between the 
     * power consumption and heat production of the different systems and conductors and the corresponding sizing of the 
     * environmental control system.
     * 
     * For each system the specific instance of the function "readParameters" is called to read the system specific parameter
     * from the config file.
     * \return void
    */
    void initializeSystems();

    /** \brief Calculates the total performance profile of all systems
     * \return void
    */
    void calculatePerformanceProfile();

    /** \brief Writes the power consumption of systems inputs to outputs of providing systems
     * \note 
     * \return void
    */
    void distributePower(std::string providerType, std::string receiverType);

    /** \brief Erases all power profile values except design power of output ports
     * \note 
     * \return void
    */
    void eraseOutputValues(std::string providerType, std::string receiverType);

    /** \brief Calculate the max value relative deviation between currrent conductor power profile values and oldConductors
     * \note 
     * \return double
    */
    double getMaxConductorPowerDeviation(vector<aircraftSystem> oldConductors);

    /** \brief Calculates the energy consumption (power profiles) of all systems that require energy (are energy sinks)
     * by calling the specific calculateSystemsPower-functions of each sink system and then @ref getAccumulatedPower.
     * After all power consumptions are accumulated the power peaks are cleaned.
     * \note power is split into electric power, hydraulic power, heat load, and bleed air
     * \return void
    */
    void getSinkEnergyConsumption();

    /** \brief Calculates the energy provided by the source systems by getting the total energy consumption of all systems
     * including the environmental control system and then calling the specific implementations of the 
     * calculateSourcePwer-function of each source system.
     * \note power is split into electric power, hydraulic power, heat load, and bleed air
     * \return void
    */
    void getRequiredSourcePower();

    /** \brief Calculates the energy consumption (power profiles) of the environmental control system by calling
     * the specific calculateSystemsPower-functions of ECS and then @ref getAccumulatedPower.
     * After all power consumptions are accumulated the power peaks are cleaned.
     * \note power is split into electric power, hydraulic power, heat load, and bleed air
     * \return void
    */
    void getECSEnergyConsumption();

    /** \brief Calculates the energy consumption (power profiles) of systems that conduct energy by calling the
     * specific implementation of the calculateConductingPower-function of each conducting system.
     * getAccumulatedPower.
     * \note power is split into electric power, hydraulic power, heat load, and bleed air
     * \return void
    */
    void getConductorEnergyConsumption();

    /** \brief sets all values to zero of a heat load power profiel
     *  \param heatLoad pointer to the heat load that shall be cleared
     * \return void
    */
    void clearInitializeHeatLoad(powerProfile *heatLoad);

    /** \brief Accumulates the heat load of a systems through the complete mission resulting in values for the max. mission
     * heat load, design heat load, and heat load in each mission segment.
     * \anchor getHeatLoad
     * \param theSystem the system of which the power demand/provision shall be calculated
     * \param heatLoad powerProfile to store the accumulated heat load
     * \return void
    */
    void getAccumulatedHeatLoad(const aircraftSystem& theSystem, powerProfile *heatLoad);

    /** \brief Writes the user defined (config) power requirement of a virtual system into the power profile of that system.
     * \param ID ID of the virtual system
     * \attention virtual systems are completely user defined. The power profile is NOT calculated.
     * \return void
    */
    void calculateVirtualSystemPower(int ID);

    /** \brief Calls the mass calculation function of each individual system.
     * \attention virtual systems are completely user defined and the mass is not calculated but copied from the config file
     * \note source and conductor systems require the total power consumption (bleed air, electric and hydraulic power).
     * Therefore, @ref getTotalConsumption is called before the masses of source and conductor systems is calculated.
     * \return void
     */
    void getSystemsMass();

    /** \brief Calculates the average bleed air consumption over the mission of the environmental control system
     * \param theBleed bleed air power profile of the ECS
     * \return double average bleed air
    */
    double getAvgBleedECS(powerProfile theBleed);

    /** \brief reads connections between ports defined in overall architecture from config data and writes to different connection vectors
     * \return void
    */
    void readConnections();

    class connection {
     public:
         string connectionType;/**<type of connection */
         powerProfile* receiverPort;/**< power profile of the reciever port */
         powerProfile* providerPort;/**< power profile of the provider port */
         double designDistribution;/**< distribution between provider ports for the design case */
         vector<double> distribution;/**< distribution between the provider ports for the mission steps */
    };

    /** \brief creates a new connection object
     * \return connection object
    */
    connection createConnection(string portType, const standardConfig::port::distribution& thePortDistribution, powerProfile* currPowerProfilePt, string* providerSystemType);

    /** \brief find back part after "/" of distribution (value/value;value;value) and convert value distribution of type value;value;value; into vector of double values.
     * \param theDistribution string of distribution
     * \param recieverPortID ID of the recieving port
     * \return double vector
    */
    vector<double> readMissionDistribution(string theDistribution, string receiverPortID);

    /** \brief find front part before "/" of distribution (value/value;value;value) and convert value distribution of type string into double.
     * \param theDistribution string of distribution
     * \param recieverPortID ID of the recieving port
     * \return double design power distribution
     */
    double readDesignDistribution(string theDistribution, string receiverPortID);

    /** \brief gets the pointer to the power profile in an aircraft systems' input or output
     * \return Pointer to powerProfile
    */
    powerProfile* getPowerProfilePt(string systemName, string portID, string direction, string* providerSystemType);

    /** \brief Substracts the power profile of the previous iteration.
     * \param providerType type of provider port
     * \param recieverType type of reciever port
     * \param oldConnections relevant connections of the previous iteration
     * \return void
     */
    void substractPreviousPower(std::string providerType, std::string receiverType, vector<connection*> oldConnections);

    /* Reporting elements */
    /**
    * \brief Defines the content of the HTML Report body. The content depends on wether the sizing or study mode
    * is used. Since that masses or only calculated in the sizing mode, they are also only written in this mode.
    * \return void
    */
    void setHtmlBody();

   /**
    * \brief Defines the content of the Tex Report body. The content depends on wether the sizing or study mode
    * is used. Since that masses or only calculated in the sizing mode, they are also only written in this mode.
    * \return void
    */
    void setTexBody();

    /* Plotting elements */
    /** \brief Calls all plot generation functions */
    void generatePlotData();

    /** \brief Generates bar plot of individual system masses */
    void generatePlotSystemsMass();

    /** \brief plots the mission data (altitude, temperature, mach, CAS, TAS) over the mission
     * steps. This data is read from the mission file and not altered in systemsDesign. (Plot can
     * serve as a reference when reading mission plots over mission steps of the individual systems)
     */
    void generatePlotDataMission();

    /** \brief Plots the power (hydraulic, electric, bleed air) of the system over the mission
     * steps. Base load and peak loads are plotted.
     * \param acftSystem aircraft system
     * \return void
     */
    void generatePlotMissionEnergy(const aircraftSystem& acftSystem);

    /** \brief Plots the accumulated power demand of all systems including peaks.
     * \return void
     */
    void generatePlotMissionEnergyAccumulated();

    /** \brief bar plot of the maximum power demand (electric, hydraulic and bleed air) of each sink system
     * \return void
     */
    void generatePlotEnergySinksMaxEnergy();

    /** \brief Calculates y-axis scaling for electric / hydraulic power based on the maximum value and including room for a legend
     * \param hydraulicSum pointer to the hydraulic power profile that will be plotted
     * \param electricSum pointer to the electric power profile that will be plotted
     * \return double: maximum value for power y-axis
     */
    double getYAxisScalingEHC(const powerProfile *hydraulicSum, const powerProfile *electricSum);

    /** \brief Calculates y-axis scaling for bleed air mass flow based on the maximum value and including room for a legen
     * \note Depending on values the extra room for the legend is sometimes smaller than that for electric/hydraulic power. It can be useful to add
     * an extra 10-20% of room by increasing the upper y-axis limit by this factor.
     * \param bleedAir pointer to the bleed air power profile
     * \return double: maximum value for mass flow y-axis
     */
    double getYAxisScalingBA(const powerProfile *bleedAir);

    /** \brief Helper function for getYAxisScalingBA and getYAxisScalingEHC
     * \param numToRound number to round (int)
     * \param multiple number to multiply (int)
     * \return int: rounded number
    */
    int roundUp(int numToRound, int multiple);

    /** \brief checks if all values of a vector are zero
     * \param vec vector to be checked
     * \return bool = true if all values of the vector are zero
    */
    bool all_zeros(const std::vector<double>& vec);

    double maxShaftPower = 0.;/**< [W] maximum shaft power offtake of all systems */
    double maxBleedAir = 0.;/**< [kg/s] maximum bleed air offtake of all system*/

    // informations about port connections
    vector<connection> conductorToConsumer;
    vector<connection> conductorToConductor;
    vector<connection> sourceToConductor;
    vector<connection> interfaceToConductor;
    vector<connection> interfaceToSource;
    vector<connection> conductorToECS;

 private:
    std::shared_ptr<RuntimeIO> rtIO_; /*<-- RuntimeIO object*/
    std::shared_ptr<systemsIOData> data_; /*<-- IOData object*/
    Report report_; /*Report object*/
    weightsAndCGs *myWeightsPt; /*Pointer to weightsAndCGs object*/
};

#endif //SYSTEMS_DESIGN_SRC_STANDARDSYSTEMSDESIGN_H_
