/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
*/

#include <string>
#include <memory>
#include <utility>
#include <unitConversion/unitConversion.h>
#include <algorithm>
#include <energyCarriers/energyCarriers.h>
#include "systemsIOData.h"
#include "aircraftSystem.h"

systemsIOData::systemsIOData(const RuntimeIO& rtIO, const node& config, std::string mission_data_dir)
    :
    rtIO_(rtIO),
    aircraftType(EndnodeReadOnly<std::string>("aircraft_exchange_file/general/type").read(rtIO_.acxml).value()),
    aircraftModel(EndnodeReadOnly<std::string>("aircraft_exchange_file/general/model").read(rtIO_.acxml).value()),
    configFile(config),
    configuration(standardConfig(config)),
    mission(missionData(mission_data_dir + "/" + getMissionFile(mission_data_dir))) {
        aircraftName = aircraftType + "-" + aircraftModel;
        //read delta ISA
        if (configuration.designCase) {
            delta_ISA = EndnodeReadOnly<double>("/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/design_mission/delta_ISA").
                read(rtIO_.acxml).value();
        } else {
            delta_ISA = EndnodeReadOnly<double>("/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/study_mission/delta_ISA").
                read(rtIO_.acxml).value();
        }
        //read mission
        mission.readMission(delta_ISA);
        //read config file
        configuration.readAll(config);
        std::string path_to_design_mission_req = "/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/design_mission";
        calibratedAirspeedUpToTransAltitude = EndnodeReadOnly<double>(path_to_design_mission_req + "/climb_speed_schedule/climb_speed_below_FL100").read(rtIO_.acxml).value();
        readLandingGears(rtIO_.acxml.at("aircraft_exchange_file/component_design/landing_gear/"));
        /* Initialize system vectors */
        Systems.energySink = std::vector<aircraftSystem*>();
        Systems.energyECS = std::vector<aircraftSystem*>();
        Systems.virtualSystem = std::vector<aircraftSystem*>();
        Systems.energySource = std::vector<aircraftSystem*>();
        Systems.energyConductor = std::vector<aircraftSystem*>();
        Systems.allSystems = std::vector<aircraftSystem*>();
}

std::string systemsIOData::getMissionFile(std::string mission_data_dir) {
    std::string xmlPath = "/aircraft_exchange_file/requirements_and_specifications/mission_files/";
    std::string missionType;
    std::string missionFile;
    if (configuration.reqMission) {
        missionType = "Requirement Mission";
        missionFile = EndnodeReadOnly<std::string>(xmlPath + "requirements_mission_file").read(rtIO_.acxml).value();
    } else {
        if (configuration.designCase) {
            missionType = "Design Mission";
            missionFile = EndnodeReadOnly<std::string>(xmlPath + "design_mission_file").read(rtIO_.acxml).value();
            myRuntimeInfo->debug << "Mission file: " << mission_data_dir + "/" + missionFile << std::endl;
        } else {
            missionType = "Study Mission";
            missionFile = EndnodeReadOnly<std::string>(xmlPath + "study_mission_file").read(rtIO_.acxml).value();
        }
    }
    if (!fileExists(mission_data_dir + "/" + missionFile)) {
        std::stringstream errorMsg;
        errorMsg << "Mission file \"" << mission_data_dir << "/" << missionFile << "\" for the " << missionType
                 << " does not exist. Please activate switch in createMissionXML. Abort program.";
        throwError(__FILE__, __func__, __LINE__, errorMsg.str());
    }
    return missionFile;
}

void systemsIOData::readWing(const node& acxml, const std::string& airfoilDir) {
    std::vector<node*> wingVector = acxml.getVector("aircraft_exchange_file/component_design/wing/specific/geometry/aerodynamic_surface");
    std::shared_ptr<node> ACXML = std::make_shared<node>(acxml); // wing factory requires this format
    node aircraftXML = acxml;
    std::shared_ptr<node> AcXML = std::make_shared<node>(aircraftXML);
    geom2::WingFactory theWingfactory{ACXML, airfoilDir}; // wing factory
    for (size_t i = 0; i < wingVector.size(); i++) {
        Wing newWing;
        newWing.aerodynamicSurface = theWingfactory.create("wing/specific/geometry/aerodynamic_surface@0");
        theWing.push_back(newWing);
        myRuntimeInfo->out << "Create: " << theWing.at(i).aerodynamicSurface.name << " ..." << std::endl;
        /* get control surfaces */
        /* Try-catch block to check wether control devices are present. With the current implementation no control devices on the wing lead to an error. */
        try {
            std::vector<node*> controlDeviceVector =
                acxml.getVector("aircraft_exchange_file/component_design/wing/specific/geometry/aerodynamic_surface/parameters/control_devices/control_device");
            geom2::ControlDeviceFactory theControlDevicefactory{ACXML, airfoilDir}; // control device factory
            for (size_t j = 0; j < controlDeviceVector.size(); j++) {
                theWing.at(i).controlSurface.push_back(theControlDevicefactory.create("wing/specific/geometry/aerodynamic_surface@" + num2Str(i) +
                                                                            "/control_devices/control_device@" + num2Str(j)));
                theWing.at(i).controlSurface.back().name = controlDeviceVector[j]->getAttribMap().at("description");
                theWing.at(i).controlDevicePosDeflection.push_back(controlDeviceVector[j]->at("/deflection/full_positive_deflection/value"));
                checkBoundaries("Wing control device positive deflection", theWing.at(i).controlDevicePosDeflection.back(), -0.01, 90.);
                theWing.at(i).controlDeviceNegDeflection.push_back(controlDeviceVector[j]->at("/deflection/full_negative_deflection/value"));
                checkBoundaries("Wing control device negative deflection", theWing.at(i).controlDeviceNegDeflection.back(), -90., 0.01);
            }
        } catch (const std::exception& e) {
            myRuntimeInfo->err << "There are no control devices defined in the acxml for the wing " << theWing.at(i).aerodynamicSurface.name << std::endl;
        }
        /* get spars */
        /* Try-catch block to check wether apars are defined in the acxml. With the current implementation no spars lead to an error. */
        try {
            std::vector<node*> sparVector =
                acxml.getVector("aircraft_exchange_file/component_design/wing/specific/geometry/aerodynamic_surface/parameters/spars");
            /* read relative chord position of front spar (needed for ATA30/wingAntiIcing)*/
            theWing.at(i).frontSparPos = sparVector[0]->at("/position/inner_position/chord/from/value");
        } catch (const std::exception& e) {
            myRuntimeInfo->err << "There are no spars defined in the acxml for the wing " << theWing.at(i).aerodynamicSurface.name << std::endl;
        }
    }
}

void systemsIOData::readEmpennage(const node& acxml, const std::string& airfoilDir) {
    std::vector<node*> empennageVector = acxml.getVector("aircraft_exchange_file/component_design/empennage/specific/geometry/aerodynamic_surface");
    std::shared_ptr<node> ACXML = std::make_shared<node>(acxml); // factory requires this format
    for (size_t i = 0; i < empennageVector.size(); i++) {
        std::string tmp_name = empennageVector[i]->at("/name/value");
        myRuntimeInfo->out << "Create: Empennage " << tmp_name << " ..." << std::endl;
        if (tmp_name == "vertical_stabiliser" || tmp_name == "horizontal_stabiliser" ||
            tmp_name == "vertical_stabilizer" || tmp_name == "horizontal_stabilizer") {
            geom2::WingFactory Finfactory{ACXML, airfoilDir};
            Empennage newEmpennage;
            newEmpennage.aerodynamicSurface = Finfactory.create("empennage/specific/geometry/aerodynamic_surface@" + num2Str(i));
            theEmpennage.push_back(newEmpennage);
        } else {
            myRuntimeInfo->err << "Empennage type " << tmp_name << " is unknown." << std::endl;
            exit(1);
        }
        std::vector<node*> EmpennageControlDeviceVector =
            acxml.getVector("aircraft_exchange_file/component_design/empennage/specific/geometry/aerodynamic_surface@" + num2Str(i) +
                            "/parameters/control_devices");
        for (size_t j = 0; j < EmpennageControlDeviceVector.size(); j++) {
            geom2::ControlDeviceFactory theControlDevicefactory{ACXML, airfoilDir}; // control device factory
            theEmpennage.at(i).controlSurface.push_back(theControlDevicefactory.create("empennage/specific/geometry/aerodynamic_surface@" + num2Str(i) +
                                                                           "/control_devices/control_device@" + num2Str(j)));
            theEmpennage.at(i).controlSurface.back().name = acxml.at("aircraft_exchange_file/component_design/empennage/specific/geometry/aerodynamic_surface@" + num2Str(i) +
                            "/parameters/control_devices/control_device@" + num2Str(j)).getAttribMap().at("description");
            theEmpennage.at(i).controlDevicePosDeflection.push_back(EmpennageControlDeviceVector[j]->at("/deflection/full_positive_deflection/value"));
            checkBoundaries("Empennage control device positive deflection", theEmpennage.at(i).controlDevicePosDeflection.back(), -0.01, 90.);
            theEmpennage.at(i).controlDeviceNegDeflection.push_back(EmpennageControlDeviceVector[j]->at("/deflection/full_negative_deflection/value"));
            checkBoundaries("Empennage control device negative deflection", theEmpennage.at(i).controlDeviceNegDeflection.back(), -90., 0.01);
        }
    }
}

void systemsIOData::readFuselages(const node& acxml, const std::string& geometryDir) {
    std::shared_ptr<node> ACXML = std::make_shared<node>(acxml); //fuselage factory requires this format
    std::vector<node*> fuselageVector = acxml.getVector("aircraft_exchange_file/component_design/fuselage/specific/geometry/fuselage");
    std::string path_to_fuselage = "aircraft_exchange_file/component_design/fuselage";
    geom2::FuselageFactory fuselageFactory{ACXML, geometryDir};
    for (size_t i(0); i < fuselageVector.size(); ++i) {
        theFuselage.push_back(fuselageFactory.create("fuselage/specific/geometry/fuselage@" + num2Str(i)));
        myRuntimeInfo->out << "Create: " << theFuselage.at(i).name << " ..." << std::endl;
        /* Point between cockpit and cabin is needed for cable length / avionics positioning */
        // current implementation only for TAW aircraft with one payload tube
        data.Accommodation.payload_tube_x_ref = fuselageVector[i]->at("/fuselage_accommodation/payload_tube@0/payload_tube_reference_points/front_reference_points/x/value");
    }
}

void systemsIOData::readNacelles(const node& acxml, const std::string& geometryDir) {
    std::shared_ptr<node> ACXML = std::make_shared<node>(acxml); // hull factory requires this format
    std::vector<node*> propulsorVector = acxml.getVector("aircraft_exchange_file/requirements_and_specifications/design_specification/propulsion/propulsor/");
    geom2::HullFactory nacelleFactory(ACXML, geometryDir);
    for (size_t j(0); j < propulsorVector.size(); j++) {
        std::vector<node*> nacelleVector = acxml.getVector("aircraft_exchange_file/component_design/propulsion/specific/propulsion@" + num2Str(j)
                                                            + "/nacelle");
        /* Engine node of propulsor contains parent component of engine/nacelle */
        for (size_t i(0); i < nacelleVector.size(); ++i) {
            myRuntimeInfo->out << "Create: Nacelle " << num2Str(j) << " ..." << std::endl;
            theNacelle.push_back(nacelleFactory.create("nacelle@" + num2Str(i)));
            std::string parentComponent = propulsorVector[j]->at("/position/parent_component/value");
            theNacelle.back().name = parentComponent;
        }
    }
}

void systemsIOData::readLandingGears(const node& geometry) {
    // for the landing gear only the center of gravity and the lengths of the main gear strut are required
    /* go through all legs of the landing gear and gather name and strut length */
    std::vector<node*> landing_gear_vector = geometry.getVector("specific/geometry/landing_gear_assembly");
    for (size_t i(0); i < landing_gear_vector.size(); ++i) {
        landing_gear current_gear;
        current_gear.name = std::string(geometry.at("specific/geometry/landing_gear_assembly@" + num2Str(i) + "/name/value"));
        current_gear.strut_length = landing_gear_vector[i]->at("/assembly_components/strut_length/value");
        theLandingGear.push_back(current_gear);
    }
    // center of gravity read in readcgComponents()
}

void systemsIOData::readEnergySources(const node& acxml) {
    std::vector<node*> providers = acxml.getVector("aircraft_exchange_file/requirements_and_specifications/design_specification/systems/energy_provider/");
    //get mission segments without knowing their names
    auto children = providers.at(0)->getChildren();
    //read powertrain and energy carrier for each mission segment defined in the acxml
    energyProvider provider;
    for (size_t i = 0; i < children.size(); i++) {
        provider.missionPhase = children[i]->getName();
        provider.powerTrain = std::string(children[i]->at("/powertrain/value"));
        provider.energyCarrier = children[i]->at("/energy_carrier_ID/value");
        //write powertrain and energy carrier to the vector of energy providers
        energyProviders.push_back(provider);
    }
}

void systemsIOData::updateMissionXML(std::string mission_data_dir) {
    myRuntimeInfo->out << "Write off-takes to mission file ..." << std::endl;
    std::string missionFile = getMissionFile(mission_data_dir);
    node& missionXML_tmp = aixml::openDocument(mission_data_dir + "/" + missionFile);
    for (size_t i = 0; i < Systems.energyInterface.size(); i++) {
        Systems.energyInterface.at(i)->updateMissionXml(missionXML_tmp);
    }
    aixml::saveDocument(missionXML_tmp, 1); //this should be moved: opening, updating, saving and closing should be same as for acxml
    aixml::closeDocument(missionXML_tmp);
}

void systemsIOData::readMassComponents(const node& acxml) {
    data.Mass.MTOM = EndnodeReadOnly<double>("aircraft_exchange_file/analysis/masses_cg_inertia/maximum_takeoff_mass/mass_properties/mass").read(acxml).value();
    data.Mass.WtoS = EndnodeReadOnly<double>("aircraft_exchange_file/sizing_point/wing_loading").read(acxml).value();
    data.Mass.OME = EndnodeReadOnly<double>("aircraft_exchange_file/analysis/masses_cg_inertia/operating_mass_empty/mass_properties/mass").read(acxml).value();
    std::string MLM_path = "aircraft_exchange_file/analysis/masses_cg_inertia/maximum_landing_mass/mass_properties/mass/value";
    if (!(rtIO_.acxml.find(MLM_path))) {
        //MLM is calculated by weightAndBalance and thus not avaialable in first iteration -> MTOM used instead to size landing gear system
        myRuntimeInfo->warn << "Acxml does not contain node for maximum landing mass. 0.85 * MTOM used instead to size ATA32 in this iteration." << std::endl;
        data.Mass.MLM = 0.85 * data.Mass.MTOM;
    } else {
    data.Mass.MLM              =
        EndnodeReadOnly<double>("aircraft_exchange_file/analysis/masses_cg_inertia/maximum_landing_mass/mass_properties/mass").read(acxml).value();
    }
    //energy capacity of tank
    double tank_energy = 0.;
    std::vector<node*> tank_vector = acxml.getVector("aircraft_exchange_file/component_design/tank/specific");
    for (size_t i = 0; i < tank_vector.at(0)->getChildren().size() - 1; i++) {
        double tank_energy_tmp =
            EndnodeReadOnly<double>("/aircraft_exchange_file/component_design/tank/specific/tank@" + num2Str(i) + "/maximum_energy_capacity").read(acxml).value();
        tank_energy += tank_energy_tmp;
    }
    int energy_carrier_id = 0;  // \todo: get energy carrier id from acxml
    std::string energy_carrier_type = rtIO_.get_fuel_type(energy_carrier_id);
    data.fuelDensity = rtIO_.get_fuel_density(energy_carrier_id);  // [kg/m^3]
    double volumetric_fuel_density = EnergyCarrier(energy_carrier_type, data.fuelDensity).volumetric_energy_density;  // [J/l]
    double gravimetric_fuel_density = EnergyCarrier(energy_carrier_type, data.fuelDensity).gravimetric_energy_density;  // [J/kg]
    data.Mass.totalFuelVolume    = tank_energy / volumetric_fuel_density;
    data.fuelDensity /= 1000; //in [kg/l]
    std::string max_fuel_mass_path = "aircraft_exchange_file/analysis/masses_cg_inertia/maximum_fuel_mass/mass_properties/mass";
    if (rtIO_.acxml.find(max_fuel_mass_path)) {
        data.Mass.maxUsableFuel = EndnodeReadOnly<double>(max_fuel_mass_path).read(acxml).value();
    } else {
        data.Mass.maxUsableFuel = tank_energy / gravimetric_fuel_density;
    }
    data.Mass.totalGroupAllEngines = EndnodeReadOnly<double>("aircraft_exchange_file/component_design/propulsion/mass_properties/mass").read(acxml).value();
    data.Mass.Wing = EndnodeReadOnly<double>("aircraft_exchange_file/component_design/wing/mass_properties/mass").read(acxml).value();
    data.Mass.Fuselage = EndnodeReadOnly<double>("aircraft_exchange_file/component_design/fuselage/mass_properties/mass").read(acxml).value();
    data.Mass.Empennage = EndnodeReadOnly<double>("aircraft_exchange_file/component_design/empennage/mass_properties/mass").read(acxml).value();
    data.Mass.Gear = EndnodeReadOnly<double>("aircraft_exchange_file/component_design/landing_gear/mass_properties/mass").read(acxml).value();
    std::vector<node*> propulsionVector = acxml.getVector("aircraft_exchange_file/component_design/propulsion/specific/propulsion");
    for (size_t i = 0; i < propulsionVector.size(); i++) {
        data.Mass.Pylon += EndnodeReadOnly<double>("aircraft_exchange_file/component_design/propulsion/specific/propulsion@" + num2Str(i) +
                                                    "/pylon/mass_properties/mass").read(acxml).value();
        data.Mass.Engines.push_back(EndnodeReadOnly<double>("aircraft_exchange_file/component_design/propulsion/specific/propulsion@" + num2Str(i) +
                                                    "/engine/mass_properties/mass").read(acxml).value());
    }
}

void systemsIOData::readCGComponents(const node& acxml) {
    // Wing
    data.CofG.Wing = readPosition(acxml, "aircraft_exchange_file/component_design/wing");
    // Fuselage
    data.CofG.Fuselage = readPosition(acxml, "aircraft_exchange_file/component_design/fuselage");
    // Empennage
    data.CofG.Empennage = readPosition(acxml, "aircraft_exchange_file/component_design/empennage");
    // Gear
    data.CofG.landingGear = readPosition(acxml, "aircraft_exchange_file/component_design/landing_gear");
    // Fuel
    data.CofG.Fuel[0] = data.CofG.Wing[0];
    data.CofG.Fuel[1] = data.CofG.Wing[1];
    data.CofG.Fuel[2] = data.CofG.Wing[2];
    // Payload
    data.CofG.Payload[0] = data.CofG.Fuselage[0];
    data.CofG.Payload[1] = data.CofG.Fuselage[1];
    data.CofG.Payload[2] = data.CofG.Fuselage[2];
    // Engines and Pylons
    std::vector<node*> propulsionVector = acxml.getVector("aircraft_exchange_file/component_design/propulsion/specific/propulsion");
    for (size_t i = 0; i < propulsionVector.size(); i++) {
        data.CofG.Engines.push_back(readPosition(acxml, "aircraft_exchange_file/component_design/propulsion/specific/propulsion@" + num2Str(i) + "/engine"));
        data.CofG.Pylon.push_back(readPosition(acxml, "aircraft_exchange_file/component_design/propulsion/specific/propulsion@" + num2Str(i) + "/pylon"));
    }
}

void systemsIOData::readPerformance(const node& acxml) {
    data.Performance.VMO = EndnodeReadOnly<double>(
        "/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/flight_envelope/maximum_operating_velocity").read(acxml).value();
    data.Performance.MMO = EndnodeReadOnly<double>(
        "/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/flight_envelope/maximum_operating_mach_number").read(acxml).value();
    data.Performance.MOA = EndnodeReadOnly<double>(
        "/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/flight_envelope/maximum_operating_altitude").read(acxml).value();
    data.Performance.designRange = EndnodeReadOnly<double>(
        "/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/design_mission/range").read(acxml).value();
}

void systemsIOData::readAccomodation(const node& acxml) {
    data.Accommodation.PAX = EndnodeReadOnly<double>(
        "/aircraft_exchange_file/requirements_and_specifications/design_specification/transport_task/passenger_definition/total_number_passengers").read(acxml).value();
    data.Accommodation.cabinCrew    =
        EndnodeReadOnly<double>("/aircraft_exchange_file/component_design/fuselage/specific/geometry/fuselage/number_of_required_cabin_crew").read(acxml).value();
    std::string flight_crew_path = "/aircraft_exchange_file/analysis/mission/design_mission/number_of_flight_crew";
    if (!(rtIO_.acxml.find(flight_crew_path))) {
        //flight crew only available after mission analysis, assuming 2 if no value in acxml
        myRuntimeInfo->warn << "Acxml does not contain node for number of flight crew. Assuming 2." << std::endl;
        data.Accommodation.flightCrew = 2;
    } else {
        data.Accommodation.flightCrew = EndnodeReadOnly<double>("/aircraft_exchange_file/analysis/mission/design_mission/number_of_flight_crew").read(acxml).value();
    }
    for (size_t i = 0; i < theFuselage.size(); i++) {
        //loop through all fuselages
        std::string path_to_fuselage = "aircraft_exchange_file/component_design/fuselage/specific/geometry/fuselage@" + num2Str(i);
        std::vector<node*> payload_tube_vector = acxml.getVector(path_to_fuselage + "/fuselage_accommodation/payload_tube");
        /*in flight entertainment*/
        std::vector<node*> fuselage_op_items = acxml.getVector(path_to_fuselage + "/mass_breakdown/fuselage_operator_items");
        data.Accommodation.numberIFE = data.Accommodation.PAX;
        double total_IFE_mass = 0.;
        for (size_t j = 0; j < fuselage_op_items.size(); j++) {
            //loop over operator items
            std::string tmp_name = fuselage_op_items[j]->at("name/value");
            if (tmp_name == "in_flight_entertainment") {
                total_IFE_mass = fuselage_op_items[j]->at("mass/value");
            }
        }
        data.Accommodation.massIFE = total_IFE_mass / data.Accommodation.numberIFE;//mass of single unit
        /*Accommodation*/
        /* ATA 25 (furnishing) mass is calculated by fuselage design (according to Airbus convention excluding lavatory and galley structure) */
        //add up all furnishing components
        std::vector<node*> furnishing_vector = acxml.getVector(path_to_fuselage + "/mass_breakdown/fuselage_furnishing");
        for (size_t k = 0; k < furnishing_vector.size(); k++) {
            data.Accommodation.furnishingMass += furnishing_vector[k]->at("mass/value");
        }
        for (size_t j = 0; j < payload_tube_vector.size(); j++) {
            //loop through all payload tubes
            std::string path_to_payload_tube = path_to_fuselage + "/fuselage_accommodation/payload_tube@" + num2Str(j);
            std::vector<node*> payload_deck_vector = acxml.getVector(path_to_payload_tube + "/payload_decks");
            /* cabin and cargo deck measurements */
            for (size_t k = 0; k < payload_deck_vector.size(); k++) {
                //loop through all payload decks of the tube
                std::string path_to_payload_deck = path_to_payload_tube + "/payload_deck@" + num2Str(k);
                std::string name_payload_deck = acxml.at(path_to_payload_deck + "/name/value");
                std::string first_word = name_payload_deck.substr(0, name_payload_deck.find('_'));
                if (first_word == "passenger") {
                    data.Accommodation.cabinVolume.push_back(EndnodeReadOnly<double>(path_to_payload_deck + "/payload_deck_water_volume").read(acxml).value());
                    data.Accommodation.cabinVolumeTotal += data.Accommodation.cabinVolume.back();
                    data.Accommodation.lengthDeck.push_back(EndnodeReadOnly<double>(path_to_payload_deck + "/payload_deck_length").read(acxml).value());
                    double numberWindows = 2. * 15. / 11.4 * data.Accommodation.lengthDeck.back();
                    data.Accommodation.numberWindows += round(numberWindows);
                    /* galleys */
                    data.Accommodation.galleyPower += EndnodeReadOnly<double>(path_to_payload_deck + "/payload_deck_required_galley_power").read(acxml).value();
                } else if (first_word == "cargo") {
                    data.Accommodation.cargoVolume.push_back(EndnodeReadOnly<double>(path_to_payload_deck + "/payload_deck_water_volume").read(acxml).value());
                    data.Accommodation.cargoVolumeTotal += data.Accommodation.cargoVolume.back();
                } else {
                    // deck unknown
                }
            }
        }
    }
}

void systemsIOData::readPropulsion(const node& acxml, std::filesystem::path engine_dir) {
    //TODO(aigner): Adapt for multiple EngineTypes
    std::string path_to_propulsion = "/aircraft_exchange_file/component_design/propulsion/specific/propulsion";
    //engine scale factor
    double engine_scale_factor = EndnodeReadOnly<double>(path_to_propulsion + "/engine/scale_factor").read(acxml).value();
    //engine model name
    std::string engine_model = EndnodeReadOnly<std::string>(path_to_propulsion + "/engine/model").read(acxml).value();
    //get engine
    Engine engine{engine_dir / engine_model, engine_scale_factor};
    double shaftPowerExtractionMax = engine.get_max_shaft_power_offtake_engine();
    double bleedAirExtractionMax = engine.get_max_bleed_offtake_at_current_operating_point();
    std::vector<node*> propulsorVector = acxml.getVector("aircraft_exchange_file/requirements_and_specifications/design_specification/propulsion/propulsor/");
    /* Loop through all propulsors and read propulsor powertrain, type, and energy carrier */
    for (size_t i = 0; i < propulsorVector.size(); i++) {
        /* read propulsor type */
        std::string propType = std::string(propulsorVector[i]->at("/type/value"));
        std::transform(propType.begin(), propType.end(), propType.begin(), ::toupper);
        // check if engine type is known (needed for determining methods in ATA25 and ATA26)
        if (propType != "FAN" && propType != "GTF" && propType != "PROP") {
            myRuntimeInfo->err << "Unknown engine type \"AcftExchangeFile/Propulsion/Engine@1/TypeOfEngine\" = \"" << propType << "\"!" << std::endl;
            myRuntimeInfo->err << "Abort program." << std::endl;
            exit(1);
        }
        /* read powertrain */
        std::string powertrain = propulsorVector[i]->at("/powertrain/value");
        /* read energy carrier ID */
        int energyCarrierID = propulsorVector[i]->at("/energy_carrier_ID/value");
        std::string parentComponent = propulsorVector[i]->at("/position/parent_component/value");
        data.Propulsion.Propulsor.push_back(standardData::propulsor(powertrain, propType, energyCarrierID, parentComponent));
    }
    data.Propulsion.numberEngines = propulsorVector.size();
    data.Propulsion.maxShaftPowerOffTakes = shaftPowerExtractionMax * static_cast<double>(data.Propulsion.numberEngines);
    data.Propulsion.maxBleedAirOffTakes = bleedAirExtractionMax * static_cast<double>(data.Propulsion.numberEngines);
}

void systemsIOData::updateMassProperties() {
    /* Empty inertia matrix */
    Eigen::Matrix<double, 3, 3> inertia = Eigen::Matrix<double, 3, 3>::Zero();
    /* Create path to systems if not existent yet*/
    std::string subPath = "aircraft_exchange_file/component_design/systems";
    if (!(rtIO_.acxml.find(subPath))) {
        rtIO_.acxml[subPath];
    }
    /* Create path to mass_properties if not existent yet*/
    subPath += "/mass_properties";
    if (!(rtIO_.acxml.find(subPath))) {
        rtIO_.acxml[subPath];
    }
    MassPropertiesIO system_mass_properties("component_design/systems", "systems");
    system_mass_properties.operator=(data.Mass.totalSystemMass);
    system_mass_properties.operator=(inertia);
    system_mass_properties.data["x"] = data.CofG.total[0];
    system_mass_properties.data["y"] = data.CofG.total[1];
    system_mass_properties.data["z"] = data.CofG.total[2];
    system_mass_properties.update(rtIO_.acxml);

    /* Operator items */
    // are listed seperately because parts of operator items are calculated in systemsDesign, fuselageDesign, and missionAnalysis
    // seperately listed for mass module to add them together
    /* Create path to operator items if not exitent yet */
    subPath = "aircraft_exchange_file/component_design/systems/operator_items";
    if (!(rtIO_.acxml.find(subPath))) {
        rtIO_.acxml[subPath];
    }
    // Potable water and toilet chemicals
    subPath += "/operator_item@0";
    if (!(rtIO_.acxml.find(subPath))) {
        rtIO_.acxml[subPath];
    }
    std::string massPath = subPath + "/mass_properties";
    if (!(rtIO_.acxml.find(massPath))) {
        rtIO_.acxml[subPath];
    }
    Endnode<std::string> waterName{subPath + "/name", "name of the operator item"};
    waterName.set_value("water_and_toilet_chemicals");
    waterName.update(rtIO_.acxml);
    MassPropertiesIO water_mass_properties(subPath, "water and toilet chemicals");
    water_mass_properties.operator=(data.Mass.waterAndChemicalsMass);
    water_mass_properties.operator=(inertia);
    water_mass_properties.data["x"] = data.CofG.waterAndChemicals[0];
    water_mass_properties.data["y"] = data.CofG.waterAndChemicals[1];
    water_mass_properties.data["z"] = data.CofG.waterAndChemicals[2];
    water_mass_properties.update(rtIO_.acxml);

    // residual fuel and oil
    subPath = "aircraft_exchange_file/component_design/systems/operator_items/operator_item@1";
    if (!(rtIO_.acxml.find(subPath))) {
        rtIO_.acxml[subPath];
    }
    massPath = subPath + "/mass_properties";
    if (!(rtIO_.acxml.find(massPath))) {
        rtIO_.acxml[subPath];
    }
    Endnode<std::string> resFuelName {subPath + "/name", "name of the operator item"};
    resFuelName.set_value("residual_oil_and_fuel");
    resFuelName.update(rtIO_.acxml);
    MassPropertiesIO resFuel_mass_properties(subPath, "water and toilet chemicals");
    resFuel_mass_properties.operator=(data.Mass.residualFuelMass);
    resFuel_mass_properties.operator=(inertia);
    resFuel_mass_properties.data["x"] = data.CofG.residualFuel[0];
    resFuel_mass_properties.data["y"] = data.CofG.residualFuel[1];
    resFuel_mass_properties.data["z"] = data.CofG.residualFuel[2];
    resFuel_mass_properties.update(rtIO_.acxml);
}

Vec3 systemsIOData::readPosition(const node& acxml, std::string component_path) {
    //values
    double x = EndnodeReadOnly<double>(component_path + "/mass_properties/center_of_gravity/x").read(acxml).value();
    double y = EndnodeReadOnly<double>(component_path + "/mass_properties/center_of_gravity/y").read(acxml).value();
    double z = EndnodeReadOnly<double>(component_path + "/mass_properties/center_of_gravity/z").read(acxml).value();
    Vec3 CofG(x, y, z);
    return CofG;
}

/* CHECK FUNCTIONS */
// own function to check boundaries, when the value is not of type node but double
void systemsIOData::checkBoundaries(std::string name, double value, const double& minValue, const double& maxValue) {
  if (value < minValue || value > maxValue) {
    std::string leftBoundary = "(";
    std::string rightBoundary = ")";
    throw std::string("parameter " + name + "=" + num2Str(value) + " not in range " + leftBoundary +
                        /*double*/ num2Str(minValue) + "," + /*double*/ num2Str(maxValue) + rightBoundary);
  }
}

/* CONVERSIONS */
std::string systemsIOData::camelToSnake(const std::string& str) {
    if (str.empty()) return str; // Handle empty string

    std::string result;
    result += tolower(str[0]); // Convert the first letter to lower case without adding '_'

    for (size_t i = 1; i < str.length(); ++i) {
        char ch = str[i];
        if (isupper(ch)) {
            result += '_';
            result += tolower(ch);
        } else {
            result += ch;
        }
    }
    return result;
}
