# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

"""Module containing functions for the preparation of geometrical data necessary for tank design."""
# Import standard libraries.
from math import pi as PI
from matplotlib import pyplot as plt
import sys
import numpy as np

# Import own modules.
# from pyunitconversion import constants
# import pyaircraftgeometry2 as geom2
# import pyaixml as aixml


# def read_aerodynamic_surface(paths, component, wing_id):
#     acxml = aixml.openDocument(paths["path_to_aircraft_exchange_file"])
#     airfoil_data_dir = f"{paths["project_directory"]}/geometryData/airfoilData"
#     aerodynamic_surface = geom2.factory.WingFactory(acxml, airfoil_data_dir).create(
#         f"{component}/specific/geometry/aerodynamic_surface@{wing_id}")

#     return aerodynamic_surface


# def read_spar(paths, component, wing_id, spar_id):
#     acxml = aixml.openDocument(paths["path_to_aircraft_exchange_file"])
#     airfoil_data_dir = f"{paths["project_directory"]}/geometryData/airfoilData"
#     spar = geom2.factory.SparFactory(acxml, airfoil_data_dir).create(
#         f"{component}/specific/geometry/aerodynamic_surface@{wing_id}/spars/spar@{spar_id}")

#     return spar

def prepare_geometrical_data(paths_and_names, dict_ac_data, dict_tank_design, runtime_output):
    """Prepare geometrical data for further use.

    This function calls the according preparation function depending on the energy carrier.
    In case of KEROSENE, the function adds the following data to the 'dict_tank_design':
        dict_tank_design = {
            'tank_0': {...,
                'geometry':{
                    'inner_surface': value,
                    'outer_surface': value}},
            ...,
            'geometry':{
                'fuselage': {
                    'fuselage_width': value},
                'wing': {
                    'section_0': {
                        'length_chord': value,
                        'relative_position_front_spar': value,
                        'relative_position_rear_spar': value,
                        'y_position_: value,
                        'thickness_ratio': value},
                    ...},
                'horizontal_stabilizer': {}
            }
        }

    If LIQUID HYDROGEN is used, the following parameters are added to the dict 'dict_tank_design':
        dict_tank_design = {}

    :param dict paths_and_names: Dictionary containing system paths and ElementTrees
    :param dict dict_ac_data: Dict containing parameter and values from aircraft exchange and module configuration file
    :param dict dict_tank_design: Dict containing results from tank design
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict dict_tank_design: Dict containing results from tank design
    """

    match dict_ac_data['energy_carrier']:
        case 'kerosene':
            # Kerosene: Prepare geometrical data for integral wing tanks and trim/additional center tanks if selected.
            dict_tank_design = prepare_for_kerosene(dict_ac_data, dict_tank_design, runtime_output)
        case 'liquid_hydrogen':
            # Liquid hydrogen: Prepare geometrical data for cryogenic liquid hydrogen tanks.
            dict_tank_design = prepare_for_liquid_hydrogen(dict_ac_data, dict_tank_design, runtime_output)

    return dict_tank_design


"""Prepare data for kerosene tanks."""
def prepare_for_kerosene(dict_ac_data, dict_tank_design, runtime_output):
    """Prepare geometrical data for kerosene tanks.

    :param dict dict_ac_data: Dict containing parameter and values from aircraft exchange and module configuration file
    :param dict dict_tank_design: Dictionary containing tank design parameter
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict dict_tank_design: Dictionary containing tank design parameter
    """
    # Initialize empty list in dictionary.
    dict_tank_design['geometry'] = {}

    """Prepare tank values according to tank locations."""
     # Wing tanks.
    if 'wing' in dict_tank_design['tank_configuration']:
        dict_tank_design = prepare_wing_tanks(dict_ac_data, dict_tank_design, runtime_output)
    # Trim tank.
    if 'trim' in dict_tank_design['tank_configuration']:
        dict_tank_design = prepare_trim_tank(dict_ac_data, dict_tank_design, runtime_output)
    # Additional center tank.
    if 'additional_center' in dict_tank_design['tank_configuration']:
        dict_tank_design = prepare_additional_center_tank(dict_ac_data, dict_tank_design, runtime_output)

    # Debug print.
    runtime_output.debug('Debug: Geometrical data for all kerosene tanks prepared.')

    return dict_tank_design


def prepare_wing_tanks(dict_ac_data, dict_tank_design, runtime_output):
    """Extract and process wing values.
    
    This function adds the 'wing' sub dictionary to the 'geometry' sub dictionary:
        dict_tank_design = {
            ...,
            'geometry':{
                'wing': {
                    'fuselage_center_line': {},
                    'wing_root': {},
                    'kink': {}, -> only if wing planform is compound tapered
                    'wing_tip': {},
                    'wing_planform': (str, information about wing planform),
                    'wing_symmetry': (bool, information on wing symmetry),
                    'constant_leading_edge_sweep': (bool, information on constant leading edge sweep),
                    'absolute_x_position': (float, x coordinate of complete wing in m),
                    (*) interpolated value lists
                }
            }
        }
    The three/four sub dictionaries contain the following values (exemplified by the 'wing_tip' dictionary):
        'wing_tip': {
            'length_chord': (float, chord length of segment in m),
            'x_position': (float, relative x coordinate of foremost wing section point in m),
            'y_position': (float, relative y coordinate of foremost wing section point in m),
            'z_position': (float, relative z coordinate of foremost wing section point in m),
            'profile' : (str, name of wing section profile),
            'relative_position_front_spar': (float, relative position of front spar in percent),
            'relative_position_rear_spar': (float, relative position of rear spar in percent),
            'thickness_ratio': (float, profile thickness ratio, derived from profile name in percent)
            }
        }
    (*) There are lists with interpolated values for the section tank area, chord lengths, relative front spar
    positions, maximum tank lengths and heights, and the leading edge sweep. Those lists each contain one value per
    mm span.

    :param dict dict_ac_data: Dict containing parameter and values from aircraft exchange and module configuration file
    :param dict dict_tank_design: Dictionary containing tank design parameter
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict dict_tank_design: Dictionary containing tank design parameter
    """
    # TODO: Divide dict into 'left' and 'right'?
    """AircraftGeometry2 Test."""
    # print(constants.AU)
    # wing = read_aerodynamic_surface(paths_and_names, 'wing', wing_id)
    # span = geom2.measure.span(wing)
    # chord_at_tip = geom2.measure.chord(wing,-span/2)
    # offset_leading_edge_x = geom2.measure.offset_LE(wing,-span/2).x()
    # thickness_maximum_at_root = geom2.measure.thickness_max(wing.sections[0])
    # spar = read_spar(paths_and_names, 'wing', wing_id, '0')
    # spar_area = geom2.measure.area(spar)
    # print(list(dir(geom2)))

    """Extract wing information for further operations."""
    # Extract wing ID and information on wing symmetry.
    wing_id = next(key for key, value in dict_ac_data['wing_name'].items()
                if value == 'main_wing').split('ID')[-1]
    wing_symmetry = dict_ac_data['wing_symmetry']['wing_symmetry_ID' + str(wing_id)]
    if not wing_symmetry:
        # Not implemented yet.
        runtime_output.critical('Error: Asymmetrical wing not implemented yet! '
                                + 'Program aborted.')
        sys.exit('Exit information: Tank design not possible for asymmetrical wing!')

    wing_sections_from_aircraft_xml = [key for key in dict_ac_data['wing_section_chord_origin_x'].keys()
                                 if key.split('_')[-2] == f'ID{str(wing_id)}']
    number_of_wing_sections = len(wing_sections_from_aircraft_xml)
    wing_section_ids = ['ID' + str(wing_id) + '_ID' + wing_sections_from_aircraft_xml[i][-1]
                        for i in range(number_of_wing_sections)]

    # Prepare data for further operation depending on number of wing sections.
    if number_of_wing_sections < 3:
        runtime_output.critical('Error: Not enough wing sections given. Program aborted.')
        sys.exit('Exit information: Not enough wing sections given!')
    elif number_of_wing_sections == 3:
        # Single tapered wing w/o kink.
        wing_planform = 'single_tapered'
        wing_section_list = ['fuselage_center_line', 'wing_root', 'wing_tip']
    elif number_of_wing_sections == 4:
        # Compound tapered wing w/ kink.
        wing_planform = 'compound_tapered'
        wing_section_list = ['fuselage_center_line', 'wing_root', 'kink', 'wing_tip']
    else:
        # Not implemented yet.
        runtime_output.critical('Error: Too many wing sections! Program aborted.')
        sys.exit('Exit information: Too many wing sections!')

    """Create wing geometry dictionary."""
    # Automatic creation of wing geometry dictionary.
    parameter_list = ['chord_length', 'profile']
    position_parameters = ['x', 'y', 'z']
    base_keys = {
        'x': 'wing_section_chord_origin_x',
        'y': 'wing_section_chord_origin_y',
        'z': 'wing_section_chord_origin_z',
        'chord_length': 'wing_section_chord_length',
        'profile': 'wing_section_profile'
    }

    # Create tmp_dict with 'position' sub dictionary for x, y, z positions.
    tmp_dict = {
        'sections': {
            wing_section_list[i]: {
                'position': {
                    pos_param: dict_ac_data[base_keys[pos_param]][f"{base_keys[pos_param]}_{wing_section_ids[i]}"]
                    for pos_param in position_parameters
                },
                **{
                    param: dict_ac_data[base_keys[param]][f"{base_keys[param]}_{wing_section_ids[i]}"]
                    for param in parameter_list
                }
            }
            for i in range(len(wing_section_list))
        }
    }

    # Add further values to 'tmp_dict'.
    for i, section in enumerate(wing_section_list):
        # Add span index to section.
        tmp_dict['sections'][section]['span_index'] = abs(int(round(
            tmp_dict['sections'][section]['position']['y']*100, 0)))
        # Add thickness_ratio.
        tmp_dict['sections'][section]['thickness_ratio'] = dict_ac_data['wing_thickness_to_length'][i]
        # Add 'thickness_ratio' (derived from the profile name).
        # profile_name = tmp_dict['sections'][section]['profile']
        # if profile_name.startswith(("naca", "n", "F15")):
        #     tmp_dict['sections'][section]['thickness_ratio'] = int(profile_name[-2:])/100
        # else:
        #     # Not implemented yet.
        #     runtime_output.critical('Error: Thickness ratio cannot be extracted! Program aborted.')
        #     sys.exit('Exit information: Thickness ratio cannot be extracted from profile name!')

    # Add 'tmp_dict' to 'dict_tank_design'.
    dict_tank_design['geometry']['wing'] = tmp_dict

    """Interpolate necessary values."""
    # Calculate number of points for value interpolation (one point per mm).
    number_of_points = abs(int(round(dict_tank_design['geometry']['wing']['sections']['wing_tip']['position']['y']
                                     *100, 0)))
    # Write wing section span indexes (equal distances in mm) to list.
    section_span_indexes = [tmp_dict['sections'][section]['span_index'] for section in wing_section_list]

    # Interpolate spar positions.
    wing_span_in_mm = int(round(dict_tank_design['geometry']['wing']['sections']['wing_tip']['position']['y']*100, 0))
    front_spar_span_indexes = [
        int(dict_ac_data['spar_position_inner_spanwise']['spar_position_inner_spanwise_ID0_ID0']*wing_span_in_mm),
        int(dict_ac_data['spar_position_outer_spanwise']['spar_position_outer_spanwise_ID0_ID0']*wing_span_in_mm)]
    front_spar_relative_position = [
        dict_ac_data['spar_position_inner_chord']['spar_position_inner_chord_ID0_ID0'],
        dict_ac_data['spar_position_outer_chord']['spar_position_outer_chord_ID0_ID0']]
    relative_position_front_spar = np.interp(range(number_of_points + 1), front_spar_span_indexes,
                                             front_spar_relative_position)
    rear_spar_span_indexes = [
        int(dict_ac_data['spar_position_inner_spanwise']['spar_position_inner_spanwise_ID0_ID1']*wing_span_in_mm),
        int(dict_ac_data['spar_position_outer_spanwise']['spar_position_outer_spanwise_ID0_ID1']*wing_span_in_mm)]
    rear_spar_relative_position = [
        dict_ac_data['spar_position_inner_chord']['spar_position_inner_chord_ID0_ID1'],
        dict_ac_data['spar_position_outer_chord']['spar_position_outer_chord_ID0_ID1']]
    relative_position_rear_spar = np.interp(range(number_of_points + 1), rear_spar_span_indexes,
                                             rear_spar_relative_position)
    x_coordinate_leading_edge = [dict_tank_design['geometry']['wing']['sections'][section]['position']['x']
                                for section in wing_section_list]
    interpolated_x_coordinate_leading_edge = np.interp(range(number_of_points + 1), section_span_indexes,
                                                 x_coordinate_leading_edge)
    # Calculate absolute front spar position.
    absolute_position_front_spar_list = [(x_coordinate_leading_edge[idx]
                                        + dict_tank_design['geometry']['wing']['sections'][name]['chord_length']
                                        *relative_position_front_spar[idx])
                                        for idx, name in enumerate(wing_section_list)]
    absolute_position_rear_spar_list = [(x_coordinate_leading_edge[idx]
                                          + dict_tank_design['geometry']['wing']['sections'][name]['chord_length']
                                          *relative_position_rear_spar[idx])
                                         for idx, name in enumerate(wing_section_list)]
    # Interpolate values for absolute front/rear spar positions and tank length.
    interpolated_absolute_front_spar = np.interp(range(number_of_points + 1), section_span_indexes,
                                                        absolute_position_front_spar_list)
    interpolated_absolute_rear_spar = np.interp(range(number_of_points + 1), section_span_indexes,
                                                        absolute_position_rear_spar_list)
    interpolated_tank_lengths = interpolated_absolute_rear_spar - interpolated_absolute_front_spar
    # Tank height interpolation.
    tank_height_list = [(dict_tank_design['geometry']['wing']['sections'][section]['chord_length']
                         *dict_tank_design['geometry']['wing']['sections'][section]['thickness_ratio']
                         *dict_ac_data['a_to_d_factor']) for section in wing_section_list]
    interpolated_tank_heights = np.interp(range(number_of_points + 1), section_span_indexes, tank_height_list)
    # z coordinate interpolation.
    z_coordinate_leading_edge = [dict_tank_design['geometry']['wing']['sections'][section]['position']['z']
                                for section in wing_section_list]
    interpolated_z_coordinate_leading_edge = np.interp(range(number_of_points + 1), section_span_indexes,
                                                 z_coordinate_leading_edge)

    # Write data to 'dict_tank_design'.
    dict_tank_design['geometry']['wing']['interpolated_values'] = {}
    dict_tank_design['geometry']['wing']['interpolated_values']['x_coordinate_leading_edge'] = (
        interpolated_x_coordinate_leading_edge)
    dict_tank_design['geometry']['wing']['interpolated_values']['z_coordinate_leading_edge'] = (
        interpolated_z_coordinate_leading_edge)
    dict_tank_design['geometry']['wing']['interpolated_values']['x_coordinate_front_spar'] = (
        interpolated_absolute_front_spar)
    dict_tank_design['geometry']['wing']['interpolated_values']['x_coordinate_rear_spar'] = (
        interpolated_absolute_rear_spar)
    dict_tank_design['geometry']['wing']['interpolated_values']['tank_lengths'] = interpolated_tank_lengths
    dict_tank_design['geometry']['wing']['interpolated_values']['tank_heights'] = interpolated_tank_heights

    # Write maximum tank height and length to dictionary.
    for section in wing_section_list:
        dict_tank_design['geometry']['wing']['sections'][section]['maximum_tank_height'] = (
            float(interpolated_tank_heights[dict_tank_design['geometry']['wing']['sections'][section]['span_index']]))
        dict_tank_design['geometry']['wing']['sections'][section]['maximum_tank_length'] = (
            float(interpolated_tank_lengths[dict_tank_design['geometry']['wing']['sections'][section]['span_index']]))

    """Calculate leading edge sweep angle."""
    number_of_wing_segments = number_of_wing_sections - 1
    x_tmp, y_tmp, sweep_leading_edge = [], [], [0]
    # Store x and y coordinates to temporary lists.
    for section in dict_tank_design['geometry']['wing']['sections'].keys():
        x_tmp.append(dict_tank_design['geometry']['wing']['sections'][section]['position']['x'])
        y_tmp.append(dict_tank_design['geometry']['wing']['sections'][section]['position']['y'])
    # Calculate sweep angles at wing sections.
    for i in range(number_of_wing_segments):
        opposite_leg = x_tmp[i+1] - x_tmp[i]
        adjacent_leg = y_tmp[i+1] - y_tmp[i]
        phi = (opposite_leg/adjacent_leg) * 180/PI
        sweep_leading_edge.append(phi)
    # Set 'constant_leading_edge_sweep' parameter.
    if wing_planform == 'single_tapered':
        constant_leading_edge_sweep = True
    elif wing_planform == 'compound_tapered':
        constant_leading_edge_sweep = all(round(x, 3) == round(sweep_leading_edge[2], 3)
                                           for x in sweep_leading_edge[2:])

    # Prepare output.
    dict_tank_design['geometry']['wing']['information'] = {}
    dict_tank_design['geometry']['wing']['information']['wing_planform'] = wing_planform
    dict_tank_design['geometry']['wing']['information']['wing_symmetry'] = wing_symmetry
    dict_tank_design['geometry']['wing']['information']['wing_section_list'] = wing_section_list
    dict_tank_design['geometry']['wing']['position'] = {}
    dict_tank_design['geometry']['wing']['position']['x'] = dict_ac_data['wing_position_x']
    dict_tank_design['geometry']['wing']['position']['y'] = dict_ac_data['wing_position_y']
    dict_tank_design['geometry']['wing']['position']['z'] = dict_ac_data['wing_position_z']
    dict_tank_design['geometry']['wing']['information']['constant_leading_edge_sweep'] = constant_leading_edge_sweep

    # Debug print.
    runtime_output.debug('Debug: Geometrical data for wing tanks prepared.')

    return dict_tank_design


def prepare_trim_tank(dict_ac_data, dict_tank_design, runtime_output):
    """Extract horizontal stabilizer geometry for trim tank.

    :param _type_ dict_ac_data: _description_
    :param _type_ dict_tank_design: _description_
    :param _type_ runtime_output: _description_
    :return _type_: _description_
    """
    horizontal_stabilizer_id = next(key for key, value in dict_ac_data['empennage_name'].items()
            if value in ('horizontal_stabilizer', 'horizontal_stabiliser')).split('ID')[-1]
    symmetry = dict_ac_data['empennage_symmetry']['empennage_symmetry_ID'
                                                                                + str(horizontal_stabilizer_id)]
    if not symmetry:
        # Not implemented yet.
        runtime_output.critical('Error: Asymmetrical horizontal stabilizer not implemented yet! '
                                + 'Program aborted.')
        sys.exit('Exit information: Tank design not possible for asymmetrical horizontal stabilizer!')

    horizontal_stabilizer_sections_from_aircraft_xml = [key for key in
                                                        dict_ac_data['empennage_section_chord_origin_x'].keys()
                                                        if key.split('_')[-2] == f'ID{str(horizontal_stabilizer_id)}']
    number_of_horizontal_stabilizer_sections = len(horizontal_stabilizer_sections_from_aircraft_xml)
    horizontal_stabilizer_section_ids = ['ID' + str(horizontal_stabilizer_id) + '_ID'
                                         + horizontal_stabilizer_sections_from_aircraft_xml[i][-1]
                                         for i in range(number_of_horizontal_stabilizer_sections)]
    # Prepare data for further operation depending on number of wing sections.
    if number_of_horizontal_stabilizer_sections < 2:
        runtime_output.critical('Error: Not enough horizontal stabilizer sections given. Program aborted.')
        sys.exit('Exit information: Not enough horizontal stabilizer sections given!')
    elif number_of_horizontal_stabilizer_sections == 2:
        # Single tapered wing w/o kink.
        planform = 'single_tapered'
        section_list = ['fuselage_center_line', 'tip']
    elif number_of_horizontal_stabilizer_sections == 3:
        # Single tapered wing w/o kink.
        planform = 'single_tapered'
        section_list = ['fuselage_center_line', 'root', 'tip']
    else:
        # Not implemented yet.
        runtime_output.critical('Error: Too many horizontal stabilizer sections! Program aborted.')
        sys.exit('Exit information: Too many horizontal stabilizer sections!')

    """Create horizontal stabilizer geometry dictionary."""
    # Automatic creation of horizontal stabilizer geometry dictionary.
    parameter_list = ['chord_length', 'profile']
    position_parameters = ['x', 'y', 'z']
    base_keys = {
        'x': 'empennage_section_chord_origin_x',
        'y': 'empennage_section_chord_origin_y',
        'z': 'empennage_section_chord_origin_z',
        'chord_length': 'empennage_section_chord_length',
        'profile': 'empennage_section_profile'
    }

    # Create tmp_dict with 'position' sub dictionary for x, y, z positions.
    tmp_dict = {
        'sections': {
            section_list[i]: {
                'position': {
                    pos_param: dict_ac_data[base_keys[pos_param]][
                        f"{base_keys[pos_param]}_{horizontal_stabilizer_section_ids[i]}"]
                    for pos_param in position_parameters
                },
                **{
                    param: dict_ac_data[base_keys[param]][f"{base_keys[param]}_{horizontal_stabilizer_section_ids[i]}"]
                    for param in parameter_list
                }
            }
            for i in range(len(section_list))
        }
    }

    # Add further values to 'tmp_dict'.
    for i, section in enumerate(section_list):
        # Add span index to section.
        tmp_dict['sections'][section]['span_index'] = abs(int(round(
            tmp_dict['sections'][section]['position']['y']*100, 0)))
        # Add thickness_ratio.
        tmp_dict['sections'][section]['thickness_ratio'] = dict_ac_data['horizontal_stabilizer_thickness_to_length'][i]
        # # Add 'thickness_ratio' (derived from the profile name).
        # profile_name = tmp_dict['sections'][section]['profile']
        # if profile_name.startswith(("naca", "n", "F15")):
        #     tmp_dict['sections'][section]['thickness_ratio'] = int(profile_name[-2:])/100
        # else:
        #     # Not implemented yet.
        #     runtime_output.critical('Error: Thickness ratio cannot be extracted! Program aborted.')
        #     sys.exit('Exit information: Thickness ratio cannot be extracted from profile name!')

    # Add 'tmp_dict' to 'dict_tank_design'.
    dict_tank_design['geometry']['horizontal_stabilizer'] = tmp_dict

    """Interpolate necessary values."""
    # Calculate number of points for value interpolation (one point per mm).
    number_of_points = abs(int(round(
        dict_tank_design['geometry']['horizontal_stabilizer']['sections']['tip']['position']['y']*100, 0)))
    # Write wing section span indexes (equal distances in mm) to list.
    section_span_indexes = [tmp_dict['sections'][section]['span_index']
                            for section in section_list]

    # Interpolate spar positions.
    horizontal_stabilizer_span_in_mm = number_of_points
    front_spar_span_indexes = [
        int(dict_ac_data['empennage_spar_position_inner_spanwise']
            ['empennage_spar_position_inner_spanwise_ID0_ID0']*horizontal_stabilizer_span_in_mm),
        int(dict_ac_data['empennage_spar_position_outer_spanwise']
            ['empennage_spar_position_outer_spanwise_ID0_ID0']*horizontal_stabilizer_span_in_mm)]
    front_spar_relative_position = [
        dict_ac_data['empennage_spar_position_inner_chord']['empennage_spar_position_inner_chord_ID0_ID0'],
        dict_ac_data['empennage_spar_position_outer_chord']['empennage_spar_position_outer_chord_ID0_ID0']]
    relative_position_front_spar = np.interp(range(number_of_points + 1), front_spar_span_indexes,
                                             front_spar_relative_position)
    rear_spar_span_indexes = [
        int(dict_ac_data['empennage_spar_position_inner_spanwise']
            ['empennage_spar_position_inner_spanwise_ID0_ID1']*horizontal_stabilizer_span_in_mm),
        int(dict_ac_data['empennage_spar_position_outer_spanwise']
            ['empennage_spar_position_outer_spanwise_ID0_ID1']*horizontal_stabilizer_span_in_mm)]
    rear_spar_relative_position = [
        dict_ac_data['empennage_spar_position_inner_chord']['empennage_spar_position_inner_chord_ID0_ID1'],
        dict_ac_data['empennage_spar_position_outer_chord']['empennage_spar_position_outer_chord_ID0_ID1']]
    relative_position_rear_spar = np.interp(range(number_of_points + 1), rear_spar_span_indexes,
                                             rear_spar_relative_position)
    x_coordinate_leading_edge = [
        dict_tank_design['geometry']['horizontal_stabilizer']['sections'][section]['position']['x']
        for section in section_list]
    interpolated_x_coordinate_leading_edge = np.interp(range(number_of_points + 1), section_span_indexes,
                                                 x_coordinate_leading_edge)
    # Calculate absolute front spar position.
    absolute_position_front_spar_list = [
        (x_coordinate_leading_edge[idx]
         + dict_tank_design['geometry']['horizontal_stabilizer']['sections'][name]['chord_length']
         *relative_position_front_spar[idx]) for idx, name in enumerate(section_list)]
    absolute_position_rear_spar_list = [
        (x_coordinate_leading_edge[idx]
         + dict_tank_design['geometry']['horizontal_stabilizer']['sections'][name]['chord_length']
         *relative_position_rear_spar[idx]) for idx, name in enumerate(section_list)]
    # Interpolate values for absolute front/rear spar positions and tank length.
    interpolated_absolute_front_spar = np.interp(range(number_of_points + 1), section_span_indexes,
                                                        absolute_position_front_spar_list)
    interpolated_absolute_rear_spar = np.interp(range(number_of_points + 1), section_span_indexes,
                                                        absolute_position_rear_spar_list)
    interpolated_tank_lengths = interpolated_absolute_rear_spar - interpolated_absolute_front_spar
    # Tank height interpolation.
    tank_height_list = [(dict_tank_design['geometry']['horizontal_stabilizer']['sections'][section]['chord_length']
                         *dict_tank_design['geometry']['horizontal_stabilizer']['sections'][section]['thickness_ratio']
                         *dict_ac_data['a_to_d_factor']) for section in section_list]
    interpolated_tank_heights = np.interp(range(number_of_points + 1), section_span_indexes, tank_height_list)
    # z coordinate interpolation.
    z_coordinate_leading_edge = [
        dict_tank_design['geometry']['horizontal_stabilizer']['sections'][section]['position']['z']
        for section in section_list]
    interpolated_z_coordinate_leading_edge = np.interp(range(number_of_points + 1), section_span_indexes,
                                                 z_coordinate_leading_edge)

    # Write data to 'dict_tank_design'.
    dict_tank_design['geometry']['horizontal_stabilizer']['interpolated_values'] = {}
    dict_tank_design['geometry']['horizontal_stabilizer']['interpolated_values']['x_coordinate_leading_edge'] = (
        interpolated_x_coordinate_leading_edge)
    dict_tank_design['geometry']['horizontal_stabilizer']['interpolated_values']['z_coordinate_leading_edge'] = (
        interpolated_z_coordinate_leading_edge)
    dict_tank_design['geometry']['horizontal_stabilizer']['interpolated_values']['x_coordinate_front_spar'] = (
        interpolated_absolute_front_spar)
    dict_tank_design['geometry']['horizontal_stabilizer']['interpolated_values']['x_coordinate_rear_spar'] = (
        interpolated_absolute_rear_spar)
    dict_tank_design['geometry']['horizontal_stabilizer']['interpolated_values']['tank_lengths'] = (
        interpolated_tank_lengths)
    dict_tank_design['geometry']['horizontal_stabilizer']['interpolated_values']['tank_heights'] = (
        interpolated_tank_heights)

    # Write maximum tank height and length to dictionary.
    for section in section_list:
        dict_tank_design['geometry']['horizontal_stabilizer']['sections'][section]['maximum_tank_height'] = (
            float(interpolated_tank_heights[
                dict_tank_design['geometry']['horizontal_stabilizer']['sections'][section]['span_index']]))
        dict_tank_design['geometry']['horizontal_stabilizer']['sections'][section]['maximum_tank_length'] = (
            float(interpolated_tank_lengths[
                dict_tank_design['geometry']['horizontal_stabilizer']['sections'][section]['span_index']]))

    """Calculate leading edge sweep angle."""
    number_of_horizontal_stabilizer_segments = number_of_horizontal_stabilizer_sections - 1
    x_tmp, y_tmp, sweep_leading_edge = [], [], [0]
    # Store x and y coordinates to temporary lists.
    for section in dict_tank_design['geometry']['horizontal_stabilizer']['sections'].keys():
        x_tmp.append(dict_tank_design['geometry']['horizontal_stabilizer']['sections'][section]['position']['x'])
        y_tmp.append(dict_tank_design['geometry']['horizontal_stabilizer']['sections'][section]['position']['y'])
    # Calculate sweep angles at wing sections.
    for i in range(number_of_horizontal_stabilizer_segments):
        opposite_leg = x_tmp[i+1] - x_tmp[i]
        adjacent_leg = y_tmp[i+1] - y_tmp[i]
        phi = (opposite_leg/adjacent_leg) * 180/PI
        sweep_leading_edge.append(phi)
    # Set 'constant_leading_edge_sweep' parameter.
    if planform == 'single_tapered':
        constant_leading_edge_sweep = True
    elif planform == 'compound_tapered':
        constant_leading_edge_sweep = all(round(x, 3) == round(sweep_leading_edge[2], 3)
                                           for x in sweep_leading_edge[2:])

    # Prepare output.
    dict_tank_design['geometry']['horizontal_stabilizer']['information'] = {}
    dict_tank_design['geometry']['horizontal_stabilizer']['information']['planform'] = planform
    dict_tank_design['geometry']['horizontal_stabilizer']['information']['symmetry'] = symmetry
    dict_tank_design['geometry']['horizontal_stabilizer']['information']['section_list'] = section_list
    dict_tank_design['geometry']['horizontal_stabilizer']['position'] = {}
    dict_tank_design['geometry']['horizontal_stabilizer']['position']['x'] = (
        dict_ac_data['empennage_position_x']
        + dict_ac_data['empennage_entity_position_x']['empennage_entity_position_x_ID'
                                                      + str(horizontal_stabilizer_id)])
    dict_tank_design['geometry']['horizontal_stabilizer']['position']['y'] = (
        dict_ac_data['empennage_position_y']
        + dict_ac_data['empennage_entity_position_y']['empennage_entity_position_y_ID'
                                                      + str(horizontal_stabilizer_id)])
    dict_tank_design['geometry']['horizontal_stabilizer']['position']['z'] = (
        dict_ac_data['empennage_position_z']
        + dict_ac_data['empennage_entity_position_z']['empennage_entity_position_z_ID'
                                                      + str(horizontal_stabilizer_id)])
    dict_tank_design['geometry']['horizontal_stabilizer']['information']['constant_leading_edge_sweep'] = (
        constant_leading_edge_sweep)

    # Debug print.
    runtime_output.debug('Debug: Geometrical data for trim tank prepared.')

    return dict_tank_design


def prepare_additional_center_tank(dict_ac_data, dict_tank_design, runtime_output):
    """Extract fuselage values for additional center tank.

    :param _type_ dict_ac_data: _description_
    :param _type_ dict_tank_design: _description_
    :param _type_ runtime_output: _description_
    :return _type_: _description_
    """
    # Extract wing position and index.
    # wing_id = next(key for key, value in dict_ac_data['wing_name'].items()
    #         if value == 'main_wing').split('ID')[-1]
    position_wing = dict_ac_data['wing_position_x']
    idx_wing = int(round(position_wing*100, 0))
    # Extract fuselage length.
    max_id_key = max(dict_ac_data['fuselage_section_origin_x'].keys(), key=lambda x: int(x.split('_')[-1][2:]))
    fuselage_length = dict_ac_data['fuselage_section_origin_x'][max_id_key]
    # Interpolate fuselage widths to enable extraction of fuselage width at wing position.
    number_of_points_fuselage_length = int(round(fuselage_length*100, 0))
    indexes = [0] + [int(round(value * 100, 0)) for value in dict_ac_data['fuselage_section_origin_x'].values()]
    widths = [0] + list(dict_ac_data['fuselage_section_width'].values())
    interpolated_fuselage_widths_list = np.interp(range(number_of_points_fuselage_length + 1), indexes, widths)
    fuselage_width_at_wing = interpolated_fuselage_widths_list[idx_wing]
    # Extract cabin floor z coordinate and structural floor thickness.
    ids = [0, 1]
    z_coordinates = [0.0, -1.84]
    structural_floor_thicknesses = [0.21, 0.21]
    # Main passenger deck is always the first payload deck and cargo compartment always the last payload deck defined.
    main_deck_id = ids[0]
    cargo_deck_id = ids[-1]
    z_coordinate_cabin_floor = z_coordinates[main_deck_id]
    cabin_floor_structural_thickness = structural_floor_thicknesses[main_deck_id]
    z_coordinate_cargo_floor = z_coordinates[cargo_deck_id]
    cargo_compartment_height = (abs(z_coordinate_cargo_floor) - abs(z_coordinate_cabin_floor)
                                - abs(cabin_floor_structural_thickness))

    # Write to geometry sub dictionary.
    dict_tank_design['geometry']['fuselage'] = {
        'width_at_wing_position': fuselage_width_at_wing,
        'cargo_compartment_height': cargo_compartment_height,
        'z_coordinate_cargo_floor': z_coordinate_cargo_floor,
        'interpolated_values': {
            'width': interpolated_fuselage_widths_list
        }
    }

    # Debug print.
    runtime_output.debug('Debug: Geometrical data for additional center tank prepared.')

    return dict_tank_design


"""Prepare data for liquid hydrogen tanks."""
def prepare_for_liquid_hydrogen(dict_ac_data, dict_tank_design, runtime_output):
    """Prepare geometrical data for liquid hydrogen tanks.

    This function combines the following geometrical values from the aircraft xml to one list each:
        * Length of individual segments (1)
        * Widths of individual segments
        * Heights of individual segments
    (1) The cumulated lengths result in the x-coordinates of the respective segments (x_coord).
    :param dict dict_ac_data: Dict containing parameter and values from aircraft exchange and module configuration file
    :param dict dict_tank_design: Dictionary containing tank design parameter
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict dict_tank_design: Dictionary containing tank design parameter
    """
    #TODO: Read position of bulkhead from aircraft exchange file. tail_length_before_bulkhead can then be deleted.
    # x_coord_pressure_bulkhead = 0
    tail_length_before_bulkhead = 8.63  # est. length of section from start of tail to bulkhead (from CSR-02)
    factor_usable_diameter = dict_ac_data['factor_usable_diameter']

    # Initialize empty lists in dictionary.
    dict_tank_design['geometry'] = {}
    dict_tank_design['geometry']['raw_data'] = {}
    dict_tank_design['geometry']['raw_data']['lengths'] = []
    dict_tank_design['geometry']['raw_data']['widths'] = [0]
    dict_tank_design['geometry']['raw_data']['heights'] = [0]
    # Iterate through the dictionary and add single values to lists.
    for key, value in dict_ac_data['geometrical_data'].items():
        if key.startswith('length_'):
            dict_tank_design['geometry']['raw_data']['lengths'].append(value)
        elif key.startswith('width_'):
            dict_tank_design['geometry']['raw_data']['widths'].append(value)
        elif key.startswith('height_'):
            dict_tank_design['geometry']['raw_data']['heights'].append(value)
    # Sum up the lengths to a list with x coordinates.
    x_coord = [0]
    for idx, value in enumerate(dict_tank_design['geometry']['raw_data']['lengths']):
        x_coord.append(x_coord[idx]+dict_tank_design['geometry']['raw_data']['lengths'][idx])
    # Extract fuselage length.
    length_fuselage = x_coord[-1]
    print("Total fuselage length: ", length_fuselage, ' m')

    n_points = round(x_coord[-1]*100)
    x_coord_array = np.linspace(0, x_coord[-1], n_points)
    width_fuselage_external_interpolated = np.interp(x_coord_array, x_coord,
                                                     dict_tank_design['geometry']['raw_data']['widths'])
    height_fuselage_external_interpolated = np.interp(x_coord_array, x_coord,
                                                      dict_tank_design['geometry']['raw_data']['heights'])
    # Calculate minimum diameter for every coordinate (every cm).
    diameter_outside_max = ([min(width_fuselage_external_interpolated[idx],
                                    height_fuselage_external_interpolated[idx]) for idx in range(0, n_points)])
    # Calculate maximum usable diameter for every coordinate (every cm).
    diameter_usable = [diameter_outside_max[idx]*factor_usable_diameter for idx in range(0, n_points)]
    dict_tank_design['geometry']['manipulated_data'] = {}
    dict_tank_design['geometry']['manipulated_data']['diameter_usable'] = diameter_usable
    # Find index and x-coordinate of begin of tail section.
    idx_begin_tail_section = determine_begin_tail_section(width_fuselage_external_interpolated,
                                                            height_fuselage_external_interpolated)
    x_coord_begin_tail_section = x_coord_array[idx_begin_tail_section]
    # Find x-coordinate of pressure bulkhead.
    x_coord_pressure_bulkhead = x_coord_begin_tail_section + tail_length_before_bulkhead
    dict_tank_design['geometry']['raw_data']['x_coord_pressure_bulkhead'] = x_coord_pressure_bulkhead

    # Iterate through tank entities.
    for tank_entity in dict_tank_design.keys():
        if tank_entity.startswith('tank_'):
            # Initialize empty dictionary and lists.
            dict_tank_design[tank_entity]['geometry'] = {}
            dict_tank_design[tank_entity]['geometry']['lengths'] = []
            dict_tank_design[tank_entity]['geometry']['widths_external'] = []
            dict_tank_design[tank_entity]['geometry']['heights_external'] = []

            # Extract tank configuration.
            tank_location = dict_tank_design[tank_entity]['tank_location']
            tank_position = dict_tank_design[tank_entity]['tank_position']

            if tank_location == 'fuselage' and tank_position == 'tailcone':
                #TODO: Figure out how to determine geometrical data of tailcone.
                idx_bulkhead = int(idx_begin_tail_section + tail_length_before_bulkhead*100 + 1)
                #TODO: Geometry data, except for box dimensions, necessary? Delete?
                print('Tank position is in tailcone of fuselage.')
                dict_tank_design[tank_entity]['tank_shape'] = 'conical'
                dict_tank_design[tank_entity]['geometry']['lengths'] = x_coord_array[
                    idx_begin_tail_section:idx_bulkhead]
                dict_tank_design[tank_entity]['geometry']['widths_external'] = (
                    width_fuselage_external_interpolated[idx_begin_tail_section:idx_bulkhead])
                dict_tank_design[tank_entity]['geometry']['heights_external'] = (
                    height_fuselage_external_interpolated[idx_begin_tail_section:idx_bulkhead])
                # Extract box dimensions (necessary for calculation).
                dict_tank_design[tank_entity]['geometry']['box_dimensions'] = {}
                dict_tank_design[tank_entity]['geometry']['box_dimensions']['maximum_length'] = (
                    dict_tank_design[tank_entity]['geometry']['lengths'][-1]
                    - dict_tank_design[tank_entity]['geometry']['lengths'][0]
                )
                # Assumption: Only the smaller height/width can be used.
                dict_tank_design[tank_entity]['geometry']['box_dimensions']['maximum_diameter_front'] = min(
                    dict_tank_design[tank_entity]['geometry']['widths_external'][0],
                    dict_tank_design[tank_entity]['geometry']['heights_external'][0]
                )
                dict_tank_design[tank_entity]['geometry']['box_dimensions']['maximum_diameter_rear'] = min(
                    dict_tank_design[tank_entity]['geometry']['widths_external'][-1],
                    dict_tank_design[tank_entity]['geometry']['heights_external'][-1]
                )

            elif tank_location == 'fuselage' and tank_position == 'rear':
                print('Tank position is in rear of fuselage.')
                dict_tank_design[tank_entity]['tank_shape'] = 'cylindrical'
                dict_tank_design[tank_entity]['geometry']['lengths'] = 0
                dict_tank_design[tank_entity]['geometry']['widths_external'] = (
                    width_fuselage_external_interpolated[idx_begin_tail_section-1])
                dict_tank_design[tank_entity]['geometry']['heights_external'] = (
                    height_fuselage_external_interpolated[idx_begin_tail_section-1])
                # Box dimensions.
                dict_tank_design[tank_entity]['geometry']['box_dimensions'] = {}
                dict_tank_design[tank_entity]['geometry']['box_dimensions']['maximum_length'] = 0
                dict_tank_design[tank_entity]['geometry']['box_dimensions']['maximum_diameter_front'] = min(
                    dict_tank_design[tank_entity]['geometry']['widths_external'],
                    dict_tank_design[tank_entity]['geometry']['heights_external']
                )
                dict_tank_design[tank_entity]['geometry']['box_dimensions']['maximum_diameter_rear'] = (
                    dict_tank_design[tank_entity]['geometry']['box_dimensions']['maximum_diameter_front']
                )

            elif tank_location == 'fuselage' and tank_position == 'front':
                print('Tank position is in front of fuselage.')
                dict_tank_design[tank_entity]['tank_shape'] = 'cylindrical'
                # TODO: Implement proper extraction of geometric parameters.
                dict_tank_design[tank_entity]['geometry']['lengths'] = 0
                dict_tank_design[tank_entity]['geometry']['widths_external'] = (
                    width_fuselage_external_interpolated[idx_begin_tail_section-1])
                dict_tank_design[tank_entity]['geometry']['heights_external'] = (
                    height_fuselage_external_interpolated[idx_begin_tail_section-1])
                # Box dimensions.
                dict_tank_design[tank_entity]['geometry']['box_dimensions'] = {}
                dict_tank_design[tank_entity]['geometry']['box_dimensions']['maximum_length'] = 0
                dict_tank_design[tank_entity]['geometry']['box_dimensions']['maximum_diameter_front'] = min(
                    dict_tank_design[tank_entity]['geometry']['widths_external'],
                    dict_tank_design[tank_entity]['geometry']['heights_external']
                )
                dict_tank_design[tank_entity]['geometry']['box_dimensions']['maximum_diameter_rear'] = (
                    dict_tank_design[tank_entity]['geometry']['box_dimensions']['maximum_diameter_front']
                )

            else:
                #TODO: Proper abortion of program.
                print('No valid tank configuration. Program aborted.')

    # Debug print.
    runtime_output.debug('Debug: Geometrical data for liquid hydrogen tanks prepared.')

    return dict_tank_design


# Determine start of tail section.
def determine_begin_tail_section(width, height):
    """_summary_

    :param _type_ width: _description_
    :param _type_ height: _description_
    :return _type_: _description_
    """
    idx = 0
    while ((width[idx + 1] >= width[idx]) and (height[idx + 1] >= height[idx])):
        idx += 1

    return idx
