# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

"""Module providing calculation functions provided by the user."""
# Import standard modules.
import sys

# Import own modules.
from src.tube_and_wing.empirical.tank_design_tu_berlin.general.tankdesigntuberlin import tank_design_tu_berlin
from src.tube_and_wing.empirical.tank_design_tu_berlin.kerosene.calculatetanks import calculate_kerosene_tanks


def method_kerosene(paths_and_names, routing_dict, dict_ac_exchange, dict_mod_config, runtime_output):
    """Tank sizing method for kerosene-powered aircraft.

    This function performs the tank sizing for kerosene-powered aircraft configurations.
    The following steps are performed:

    1. Data Merging: The `dict_ac_exchange` and `dict_mod_config` dictionaries are merged into a single dictionary,
    `dict_complete_data`, which holds all relevant design and configuration parameters.
    2. Data Extraction: General parameters are extracted from `dict_complete_data` and stored in `dict_general_data`. 
    Parameters related to the design mission are filtered and saved in `dict_design_data`.
    3. Validation of Design Data: The design data is checked for completeness. If any key parameters are missing (i.e.,
    `None`), the tank design process is deemed impossible, and the program is aborted.
    4. Fuselage Check: The number of fuselages is determined from the design data. If no fuselage or more than one 
    fuselage is found, the tank design cannot proceed, and the program is aborted.
    5. Tank Design Preparation: The general and design parameters are merged into `dict_design`, and a new tank design
    is prepared using the `tank_design_tu_berlin` function. If the tank definition is invalid, the program is
    terminated.
    6. Tank Calculation: Kerosene tank volumes and configurations are calculated using the `calculate_kerosene_tanks` 
    function, that takes the design parameters and performs necessary computations.
    7. Debugging and Final Output: A debug message is logged indicating successful execution, and a final message
    confirms that the tank design has been completed.

    The output dictionary 'kerosene_output_dict' contains the results of the tank design and is structured
    according to the following scheme:
        kerosene_output_dict = {
            'tank_0': {...},
            'tank_1': {...},
            ...,
            'number_of_tanks': (),
            'tank_configuration': ()}

    :param dict paths_and_names: Dictionary containing system paths and ElementTrees
    :param dict routing_dict: Dictionary containing routing parameters
    :param dict dict_ac_exchange: Dict containing parameters and according values from aircraft exchange file
    :param dict dict_mod_config: Dict containing parameters and according values from module configuration file
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict kerosene_output_dict: Dictionary containing results from calculation for kerosene-powered aircraft
    """

    """ Preparation. """
    # Check if any values from aircraft exchange file are missing.
    for key, value in dict_ac_exchange.items():
        # Check, if any value is 'None'.
        if value is None:
            # If any value is 'None', the calculation of the costs is impossible and the program aborted.
            runtime_output.critical('Error: Tank design not possible due to missing parameter in aircraft exchange '
                                    + 'file (' + key + ')! '
                                    + 'Program aborted.')
            sys.exit(1)
    # Check if any necessary values from module configuration file are missing.
    necessary_values_for_kerosene = ['obelisk_calculation_method', 'a_to_d_factor', 'factor_volume_usable',
                                     'temperature_expansion_allowance', 'buffer_inner_tank_segment',
                                     'buffer_outer_tank_segment', 'buffer_center_tank_segment',
                                     'factor_volume_allowance']
    for key, value in dict_mod_config.items():
        # Check, if any value is 'None'.
        if key in necessary_values_for_kerosene and value is None:
            # If any value is 'None', the calculation of the costs is impossible and the program aborted.
            runtime_output.critical('Error: Tank design not possible due to missing parameter in module configuration '
                                    + 'file (' + key + ')! '
                                    + 'Program aborted.')
            sys.exit(1)
    # Merge both data dictionaries.
    dict_design = dict_ac_exchange | dict_mod_config
    dict_design['energy_carrier'] = routing_dict['user_layer']

    """Calculate tanks."""
    # Tank design preparation.
    kerosene_output_dict = tank_design_tu_berlin(paths_and_names, routing_dict, dict_design, runtime_output)
    # Tank calculation.
    kerosene_output_dict = calculate_kerosene_tanks(dict_design, kerosene_output_dict, runtime_output)

    # Debug print.
    runtime_output.debug('Debug: The "method_kerosene" function was successfully executed.')
    runtime_output.print('Tank design completed.')

    return kerosene_output_dict
