# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

# Import standard libraries.
import math


def calculate_thickness_wall(diameter, internal_pressure=1.5, tensile_stress_allowable=172, method='asme'):
    """Calculate wall thickness for aluminum walls with internal pressure of 1.5 bar according to ASME Code.

    This function calculates the wall thickness of a pressure vessel made of aluminum Al-2219 operated at 1.5 bar
    internal pressure. The calculation is according to the ASME Code.

    TODO: Research joint efficiency factor, angle.
    TODO: Add source.
    TODO: Add AD2000 implementation.

    :param float diameter: Pressure vessel diameter in m
    :param float internal_pressure: Internal pressure in bar, defaults to 1.5
    :param float tensile_stress_allowable: Allowable tensile stress in N/mm, defaults to 172
    :param str method: Method used to calculate wall thickness, defaults to 'asme'
    :return float: Wall thickness in m
    """
    match method:
        case 'asme':
            # Conversion factors and conversions.
            conv_bar_to_PSI = 14.5037738
            conv_m_to_inch = 39.37
            conv_Nmm_to_PSI = 145.03773800722
            conv_inch_to_cm = 2.54
            conv_deg_to_rad = 0.01745329
            internal_pressure = internal_pressure*conv_bar_to_PSI
            tensile_stress_allowable = tensile_stress_allowable*conv_Nmm_to_PSI
            # Variables.
            angle = 1
            joint_efficiency_factor = 0.925
            # Calculation of wall thickness according to formula (3.59).
            diameter = diameter*conv_m_to_inch
            thickness_wall_inch = (internal_pressure*diameter)/(2*math.cos(angle*conv_deg_to_rad)
                                                                * tensile_stress_allowable*joint_efficiency_factor
                                                                - 0.6*internal_pressure)
            thickness_wall = thickness_wall_inch*conv_inch_to_cm/100
        case 'ad-2000':
            print('Note: AD-2000 not implemented yet. Wall thickness calculated according to ASME.')
            # ----------- TODO Change code to AD-2000 implementation -----------
            # Conversion factors and conversions.
            conv_bar_to_PSI = 14.5037738
            conv_m_to_inch = 39.37
            conv_Nmm_to_PSI = 145.03773800722
            conv_inch_to_cm = 2.54
            conv_deg_to_rad = 0.01745329
            internal_pressure = internal_pressure*conv_bar_to_PSI
            tensile_stress_allowable = tensile_stress_allowable*conv_Nmm_to_PSI
            # Variables.
            angle = 1
            joint_efficiency_factor = 0.925
            # Calculation of wall thickness according to formula (3.59).
            diameter = diameter*conv_m_to_inch
            thickness_wall_inch = (internal_pressure*diameter)/(2*math.cos(angle*conv_deg_to_rad)
                                                                * tensile_stress_allowable*joint_efficiency_factor
                                                                - 0.6*internal_pressure)
            thickness_wall = thickness_wall_inch*conv_inch_to_cm/100
            # ------------------------------------------------------------------
    return thickness_wall
