/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "dihedralMethods.h"

#include <aircraftGeometry2/processing/transform.h>
#include <unitConversion/constants.h>

#include <map>
#include <cmath>
#include <iostream>
#include <numbers>
#include <tuple>
#include <functional>


constexpr double to_degree = 180.0 / std::numbers::pi;
constexpr double to_radian = std::numbers::pi / 180.0;

double user_defined::dihedral(const double user_dihedral) { return user_dihedral; }

double by_wing_position_and_quarter_chord_sweep::dihedral(const std::string& wing_position,
                                                          const double quarter_chord_sweep, const double mach_design, const double raymer_range_value) {
  return raymer::dihedral_limitation(wing_position, quarter_chord_sweep, mach_design, raymer_range_value);
}

double raymer::dihedral_limitation(const std::string& wing_position,
                                                       const double quarter_chord_sweep, const double mach_design, const double raymer_range_value) {
  std::map<std::string, std::tuple<double, double>> unswept = {{"low", {5. * geom2::detail::to_radians, 7. * geom2::detail::to_radians}},
                                                               {"mid", {2. * geom2::detail::to_radians, 4. * geom2::detail::to_radians}},
                                                               {"high", {0. * geom2::detail::to_radians, 2. * geom2::detail::to_radians}}};

  std::map<std::string, std::tuple<double, double>> swept_subsonic = {{"low", {3. * geom2::detail::to_radians, 7. * geom2::detail::to_radians}},
                                                                      {"mid", {-2. * geom2::detail::to_radians, 2. * geom2::detail::to_radians}},
                                                                      {"high", {-5. * geom2::detail::to_radians, -2. * geom2::detail::to_radians}}};

  std::map<std::string, std::tuple<double, double>> swept_supersonic = {{"low", {0. * geom2::detail::to_radians, 5. * geom2::detail::to_radians}},
                                                                        {"mid", {-5. * geom2::detail::to_radians, 0. * geom2::detail::to_radians}},
                                                                        {"high", {-5. * geom2::detail::to_radians, 0. * geom2::detail::to_radians}}};

  double dihedral;
  if (mach_design < 1.) {
    if (fabs(quarter_chord_sweep) < ACCURACY_MEDIUM) {
      const auto [lo, up] = unswept[wing_position];
      dihedral = std::lerp(lo, up, raymer_range_value);
    } else {
      const auto [lo, up] =  swept_subsonic[wing_position];
      dihedral = std::lerp(lo, up, raymer_range_value);
    }
  } else {
    const auto [up, lo] =  swept_supersonic[wing_position];
    dihedral = std::lerp(lo, up, raymer_range_value);
  }
  return dihedral;
}

double get_design_dihedral(modes dihedral_mode, const std::string& wing_mounting, const double user_dihedral,
                                 const double quarter_chord_sweep, const double mach_design, const double raymer_range_value) {
  std::map<modes, std::tuple<std::string, std::function<double(void)>>> dihedral_selection = {
      {"mode_0", {"user_defined", [=]() { return user_defined::dihedral(user_dihedral); }}},
      {"mode_1", {"raymer", [=]() { return ::raymer::dihedral_limitation(wing_mounting, quarter_chord_sweep, mach_design, raymer_range_value); }}}};


  const auto [method, limits_method] = dihedral_selection[dihedral_mode];
  const double value = limits_method();
  std::cout << "Dihedral [method]: " << method << "\n";
  std::cout << "Dihedral  [value]: " << value * geom2::detail::to_degrees << " [deg]\n";
  return value;


}
